Package: plyranges
Type: Package
Title: A fluent interface for manipulating GenomicRanges
Version: 1.32.0
Authors@R: c(
      person("Stuart", "Lee", role = c("aut"), 
             comment = c(ORCID = "0000-0003-1179-8436")),
      person("Michael", "Lawrence", role = c("aut", "ctb")),
      person("Dianne", "Cook", role = c("aut", "ctb")),
      person("Spencer", "Nystrom", role = c("ctb"), 
             comment = c(ORCID = "0000-0003-1000-1579")),
      person("Pierre-Paul", "Axisa", role=c("ctb")),
      person("Michael", "Love", role = c("ctb","cre"),
             email = "michaelisaiahlove@gmail.com")
      )
Maintainer: Michael Love <michaelisaiahlove@gmail.com>
Description: A dplyr-like interface for interacting with the common Bioconductor 
    classes Ranges and GenomicRanges. By providing a grammatical
    and consistent way of manipulating these classes their accessiblity for new
    Bioconductor users is hopefully increased.
URL: https://tidyomics.github.io/plyranges
Depends: R (>= 3.5), BiocGenerics, IRanges (>= 2.12.0), GenomicRanges
        (>= 1.28.4), dplyr
Imports: methods, rlang (>= 0.2.0), magrittr, tidyselect (>= 1.0.0),
        rtracklayer, GenomicAlignments, Seqinfo, Rsamtools, S4Vectors
        (>= 0.23.10), utils
biocViews: Infrastructure, DataRepresentation, WorkflowStep, Coverage
BugReports: https://github.com/tidyomics/plyranges
License: Artistic-2.0
Encoding: UTF-8
ByteCompile: true
Suggests: knitr, BiocStyle, rmarkdown, testthat (>= 2.1.0),
        HelloRanges, HelloRangesData, BSgenome.Hsapiens.UCSC.hg19,
        pasillaBamSubset, covr, ggplot2
VignetteBuilder: knitr
Roxygen: list(markdown = TRUE)
RoxygenNote: 7.3.3
Collate: 'class-AnchoredRanges.R' 'class-Operator.R'
        'class-DeferredGenomicRanges.R' 'class-GroupedRanges.R'
        'dplyr-arrange.R' 'dplyr-filter.R' 'dplyr-groups.R'
        'dplyr-mutate.R' 'dplyr-pull.R' 'dplyr-select.R'
        'dplyr-slice.R' 'dplyr-summarize.R' 'endo-coverage.R'
        'endo-tile.R' 'io-bam.R' 'io-bed.R' 'io-bigwig.R' 'io-gff.R'
        'io-wig.R' 'methods-DeferredGenomicRanges.R'
        'methods-Operator.R' 'plyranges.R' 'ranges-add-distance.R'
        'ranges-anchors.R' 'ranges-arithmetic-flank.R'
        'ranges-arithmetic-setters.R' 'ranges-arithmetic-shift.R'
        'ranges-arithmetic-stretch.R' 'ranges-bind.R' 'ranges-chop.R'
        'ranges-colwise.R' 'ranges-construct.R' 'ranges-disjoin.R'
        'ranges-eval-quo.R' 'ranges-eval.R' 'ranges-expand.R'
        'ranges-genomeinfo.R' 'ranges-join-follow.R'
        'ranges-join-mcols.R' 'ranges-join-nearest.R'
        'ranges-join-precede.R' 'ranges-overlap-count.R'
        'ranges-overlap-filter.R' 'ranges-overlap-find.R'
        'ranges-overlap-groups.R' 'ranges-overlap-joins-intersect.R'
        'ranges-overlap-joins-outer.R' 'ranges-overlap-self-joins.R'
        'ranges-pairs.R' 'ranges-rangewise-setops.R' 'ranges-reduce.R'
        'ranges-setops.R' 'utils-pipe.R'
git_url: https://git.bioconductor.org/packages/plyranges
git_branch: RELEASE_3_23
git_last_commit: 1ace171
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-28 23:54:57 UTC; biocbuild
Author: Stuart Lee [aut] (ORCID: <https://orcid.org/0000-0003-1179-8436>),
  Michael Lawrence [aut, ctb],
  Dianne Cook [aut, ctb],
  Spencer Nystrom [ctb] (ORCID: <https://orcid.org/0000-0003-1000-1579>),
  Pierre-Paul Axisa [ctb],
  Michael Love [ctb, cre]

# Generated by roxygen2: do not edit by hand

S3method(anchor,AnchoredGenomicRanges)
S3method(anchor,AnchoredIntegerRanges)
S3method(anchor_3p,GenomicRanges)
S3method(anchor_5p,GenomicRanges)
S3method(anchor_center,GenomicRanges)
S3method(anchor_center,IntegerRanges)
S3method(anchor_end,GenomicRanges)
S3method(anchor_end,IntegerRanges)
S3method(anchor_start,GenomicRanges)
S3method(anchor_start,IntegerRanges)
S3method(arrange,DelegatingGenomicRanges)
S3method(arrange,DelegatingIntegerRanges)
S3method(arrange,Ranges)
S3method(as_granges,DataFrame)
S3method(as_granges,data.frame)
S3method(as_granges,default)
S3method(as_iranges,DataFrame)
S3method(as_iranges,data.frame)
S3method(as_iranges,default)
S3method(as_ranges,Rle)
S3method(as_ranges,RleList)
S3method(chop_by_gaps,DeferredGenomicRanges)
S3method(chop_by_gaps,GenomicRanges)
S3method(chop_by_introns,DeferredGenomicRanges)
S3method(chop_by_introns,GenomicRanges)
S3method(complement_ranges,GenomicRanges)
S3method(complement_ranges,IntegerRanges)
S3method(complement_ranges_directed,GenomicRanges)
S3method(compute_coverage,default)
S3method(count_overlaps,GenomicRanges)
S3method(count_overlaps,IntegerRanges)
S3method(count_overlaps_directed,GenomicRanges)
S3method(count_overlaps_within,GenomicRanges)
S3method(count_overlaps_within,IntegerRanges)
S3method(count_overlaps_within_directed,GenomicRanges)
S3method(disjoin_ranges,GenomicRanges)
S3method(disjoin_ranges,GroupedGenomicRanges)
S3method(disjoin_ranges,GroupedIntegerRanges)
S3method(disjoin_ranges,IntegerRanges)
S3method(disjoin_ranges_directed,GenomicRanges)
S3method(disjoin_ranges_directed,GroupedGenomicRanges)
S3method(dplyr::pull,Ranges)
S3method(filter,BamFileOperator)
S3method(filter,DeferredGenomicRanges)
S3method(filter,DelegatingGenomicRanges)
S3method(filter,DelegatingIntegerRanges)
S3method(filter,GroupedGenomicRanges)
S3method(filter,GroupedIntegerRanges)
S3method(filter,Ranges)
S3method(filter_by_non_overlaps,GenomicRanges)
S3method(filter_by_non_overlaps,IntegerRanges)
S3method(filter_by_non_overlaps_directed,GenomicRanges)
S3method(filter_by_overlaps,BamFileOperator)
S3method(filter_by_overlaps,DeferredGenomicRanges)
S3method(filter_by_overlaps,GenomicRanges)
S3method(filter_by_overlaps,IntegerRanges)
S3method(filter_by_overlaps_directed,GenomicRanges)
S3method(find_overlaps,GenomicRanges)
S3method(find_overlaps,IntegerRanges)
S3method(find_overlaps_directed,GenomicRanges)
S3method(find_overlaps_within,GenomicRanges)
S3method(find_overlaps_within,IntegerRanges)
S3method(find_overlaps_within_directed,GenomicRanges)
S3method(get_genome_info,BamFileOperator)
S3method(get_genome_info,DeferredGenomicRanges)
S3method(get_genome_info,Seqinfo)
S3method(get_genome_info,default)
S3method(group_by,DeferredGenomicRanges)
S3method(group_by,GenomicRanges)
S3method(group_by,GroupedGenomicRanges)
S3method(group_by,GroupedIntegerRanges)
S3method(group_by,IntegerRanges)
S3method(group_by_overlaps,GenomicRanges)
S3method(group_by_overlaps,IntegerRanges)
S3method(group_data,GroupedGenomicRanges)
S3method(group_data,GroupedIntegerRanges)
S3method(group_indices,GroupedGenomicRanges)
S3method(group_indices,GroupedIntegerRanges)
S3method(group_indices,Ranges)
S3method(group_keys,GroupedGenomicRanges)
S3method(group_keys,GroupedIntegerRanges)
S3method(group_keys,Ranges)
S3method(group_split,GroupedGenomicRanges)
S3method(group_split,GroupedIntegerRanges)
S3method(group_split,Ranges)
S3method(group_vars,GroupedGenomicRanges)
S3method(group_vars,GroupedIntegerRanges)
S3method(group_vars,Ranges)
S3method(groups,GroupedGenomicRanges)
S3method(groups,GroupedIntegerRanges)
S3method(groups,Ranges)
S3method(intersect_ranges,GenomicRanges)
S3method(intersect_ranges,IntegerRanges)
S3method(intersect_ranges_directed,GenomicRanges)
S3method(join_follow,GenomicRanges)
S3method(join_follow,IntegerRanges)
S3method(join_follow_left,GenomicRanges)
S3method(join_follow_left,IntegerRanges)
S3method(join_follow_upstream,GenomicRanges)
S3method(join_nearest_downstream,GenomicRanges)
S3method(join_nearest_upstream,GenomicRanges)
S3method(join_overlap_intersect,GenomicRanges)
S3method(join_overlap_intersect,IntegerRanges)
S3method(join_overlap_intersect_directed,GenomicRanges)
S3method(join_overlap_intersect_within,GenomicRanges)
S3method(join_overlap_intersect_within,IntegerRanges)
S3method(join_overlap_intersect_within_directed,GenomicRanges)
S3method(join_overlap_left,GenomicRanges)
S3method(join_overlap_left,IntegerRanges)
S3method(join_overlap_left_directed,GenomicRanges)
S3method(join_overlap_left_within,GenomicRanges)
S3method(join_overlap_left_within,IntegerRanges)
S3method(join_overlap_left_within_directed,GenomicRanges)
S3method(join_overlap_self,GenomicRanges)
S3method(join_overlap_self,IntegerRanges)
S3method(join_overlap_self_directed,GenomicRanges)
S3method(join_overlap_self_within,GenomicRanges)
S3method(join_overlap_self_within,IntegerRanges)
S3method(join_overlap_self_within_directed,GenomicRanges)
S3method(join_precede,GenomicRanges)
S3method(join_precede,IntegerRanges)
S3method(join_precede_downstream,GenomicRanges)
S3method(join_precede_right,GenomicRanges)
S3method(join_precede_right,IntegerRanges)
S3method(load_genomic_file,BamFileOperator)
S3method(mutate,AnchoredGenomicRanges)
S3method(mutate,AnchoredIntegerRanges)
S3method(mutate,DeferredGenomicRanges)
S3method(mutate,DelegatingGenomicRanges)
S3method(mutate,DelegatingIntegerRanges)
S3method(mutate,GroupedGenomicRanges)
S3method(mutate,GroupedIntegerRanges)
S3method(mutate,Ranges)
S3method(overscope_ranges,DelegatingGenomicRanges)
S3method(overscope_ranges,DelegatingIntegerRanges)
S3method(overscope_ranges,GroupedGenomicRanges)
S3method(overscope_ranges,GroupedIntegerRanges)
S3method(overscope_ranges,Ranges)
S3method(pair_follow,GenomicRanges)
S3method(pair_follow,IntegerRanges)
S3method(pair_nearest,GenomicRanges)
S3method(pair_nearest,IntegerRanges)
S3method(pair_overlaps,GenomicRanges)
S3method(pair_overlaps,IntegerRanges)
S3method(pair_precede,GenomicRanges)
S3method(pair_precede,IntegerRanges)
S3method(pull,DelegatingGenomicRanges)
S3method(pull,DelegatingIntegerRanges)
S3method(pull,GroupedGenomicRanges)
S3method(pull,GroupedIntegerRanges)
S3method(reduce_ranges,GenomicRanges)
S3method(reduce_ranges,GroupedGenomicRanges)
S3method(reduce_ranges,GroupedIntegerRanges)
S3method(reduce_ranges,IntegerRanges)
S3method(reduce_ranges_directed,GenomicRanges)
S3method(reduce_ranges_directed,GroupedGenomicRanges)
S3method(select,BamFileOperator)
S3method(select,DeferredGenomicRanges)
S3method(select,DelegatingGenomicRanges)
S3method(select,DelegatingIntegerRanges)
S3method(select,Ranges)
S3method(set_end,Ranges)
S3method(set_seqnames,GenomicRanges)
S3method(set_start,Ranges)
S3method(set_strand,GenomicRanges)
S3method(set_width,AnchoredGenomicRanges)
S3method(set_width,AnchoredIntegerRanges)
S3method(set_width,Ranges)
S3method(setdiff_ranges,GenomicRanges)
S3method(setdiff_ranges,IntegerRanges)
S3method(setdiff_ranges_directed,GenomicRanges)
S3method(slice,GroupedGenomicRanges)
S3method(slice,GroupedIntegerRanges)
S3method(slice,Ranges)
S3method(slice_head,GroupedGenomicRanges)
S3method(slice_head,GroupedIntegerRanges)
S3method(slice_head,Ranges)
S3method(slice_max,GroupedGenomicRanges)
S3method(slice_max,GroupedIntegerRanges)
S3method(slice_max,Ranges)
S3method(slice_min,GroupedGenomicRanges)
S3method(slice_min,GroupedIntegerRanges)
S3method(slice_min,Ranges)
S3method(slice_sample,GroupedGenomicRanges)
S3method(slice_sample,GroupedIntegerRanges)
S3method(slice_sample,Ranges)
S3method(slice_tail,GroupedGenomicRanges)
S3method(slice_tail,GroupedIntegerRanges)
S3method(slice_tail,Ranges)
S3method(slide_ranges,Ranges)
S3method(stretch,AnchoredGenomicRanges)
S3method(stretch,AnchoredIntegerRanges)
S3method(stretch,Ranges)
S3method(summarise,DeferredGenomicRanges)
S3method(summarise,DelegatingGenomicRanges)
S3method(summarise,GroupedGenomicRanges)
S3method(summarise,GroupedIntegerRanges)
S3method(summarise,Ranges)
S3method(tbl_vars,GenomicRanges)
S3method(tbl_vars,Ranges)
S3method(tile_ranges,Ranges)
S3method(unanchor,AnchoredGenomicRanges)
S3method(unanchor,AnchoredIntegerRanges)
S3method(ungroup,GroupedGenomicRanges)
S3method(ungroup,GroupedIntegerRanges)
S3method(ungroup,Ranges)
S3method(union_ranges,GenomicRanges)
S3method(union_ranges,IntegerRanges)
S3method(union_ranges_directed,GenomicRanges)
export("%>%")
export("%intersect%")
export("%setdiff%")
export("%union%")
export(add_nearest_distance)
export(add_nearest_distance_downstream)
export(add_nearest_distance_left)
export(add_nearest_distance_right)
export(add_nearest_distance_upstream)
export(anchor)
export(anchor_3p)
export(anchor_5p)
export(anchor_center)
export(anchor_centre)
export(anchor_end)
export(anchor_start)
export(as_granges)
export(as_iranges)
export(as_ranges)
export(between)
export(bind_ranges)
export(chop_by_gaps)
export(chop_by_introns)
export(complement_ranges)
export(complement_ranges_directed)
export(compute_coverage)
export(count_overlaps)
export(count_overlaps_directed)
export(count_overlaps_within)
export(count_overlaps_within_directed)
export(disjoin_ranges)
export(disjoin_ranges_directed)
export(expand_ranges)
export(filter_by_non_overlaps)
export(filter_by_non_overlaps_directed)
export(filter_by_overlaps)
export(filter_by_overlaps_directed)
export(find_overlaps)
export(find_overlaps_directed)
export(find_overlaps_within)
export(find_overlaps_within_directed)
export(flank_downstream)
export(flank_left)
export(flank_right)
export(flank_upstream)
export(genome_info)
export(get_genome_info)
export(group_by_overlaps)
export(id_to_column)
export(intersect_ranges)
export(intersect_ranges_directed)
export(interweave)
export(join_follow)
export(join_follow_left)
export(join_follow_upstream)
export(join_mcols_inner)
export(join_mcols_left)
export(join_nearest)
export(join_nearest_downstream)
export(join_nearest_left)
export(join_nearest_right)
export(join_nearest_upstream)
export(join_overlap_inner)
export(join_overlap_inner_directed)
export(join_overlap_inner_within)
export(join_overlap_inner_within_directed)
export(join_overlap_intersect)
export(join_overlap_intersect_directed)
export(join_overlap_intersect_within)
export(join_overlap_intersect_within_directed)
export(join_overlap_left)
export(join_overlap_left_directed)
export(join_overlap_left_within)
export(join_overlap_left_within_directed)
export(join_overlap_self)
export(join_overlap_self_directed)
export(join_overlap_self_within)
export(join_overlap_self_within_directed)
export(join_precede)
export(join_precede_downstream)
export(join_precede_right)
export(n)
export(n_distinct)
export(names_to_column)
export(overscope_ranges)
export(pair_follow)
export(pair_nearest)
export(pair_overlaps)
export(pair_precede)
export(read_bam)
export(read_bed)
export(read_bed_graph)
export(read_bigwig)
export(read_gff)
export(read_gff1)
export(read_gff2)
export(read_gff3)
export(read_narrowpeaks)
export(read_wig)
export(reduce_ranges)
export(reduce_ranges_directed)
export(remove_names)
export(set_end)
export(set_genome_info)
export(set_seqnames)
export(set_start)
export(set_strand)
export(set_width)
export(setdiff_ranges)
export(setdiff_ranges_directed)
export(shift_downstream)
export(shift_left)
export(shift_right)
export(shift_upstream)
export(slide_ranges)
export(span)
export(stretch)
export(tile_ranges)
export(unanchor)
export(union_ranges)
export(union_ranges_directed)
export(write_bed)
export(write_bed_graph)
export(write_bigwig)
export(write_gff)
export(write_gff1)
export(write_gff2)
export(write_gff3)
export(write_narrowpeaks)
export(write_wig)
exportClasses(BamFileOperator)
exportClasses(DeferredGenomicRanges)
exportClasses(FileOperator)
exportClasses(GroupedGenomicRanges)
exportClasses(GroupedIntegerRanges)
importFrom(BiocGenerics,"%in%")
importFrom(BiocGenerics,"end<-")
importFrom(BiocGenerics,"score<-")
importFrom(BiocGenerics,"start<-")
importFrom(BiocGenerics,"strand<-")
importFrom(BiocGenerics,"width<-")
importFrom(BiocGenerics,end)
importFrom(BiocGenerics,rep.int)
importFrom(BiocGenerics,score)
importFrom(BiocGenerics,start)
importFrom(BiocGenerics,strand)
importFrom(BiocGenerics,unique)
importFrom(BiocGenerics,unlist)
importFrom(BiocGenerics,which)
importFrom(BiocGenerics,width)
importFrom(GenomicAlignments,extractAlignmentRangesOnReference)
importFrom(GenomicAlignments,readGAlignmentPairs)
importFrom(GenomicAlignments,readGAlignments)
importFrom(GenomicRanges,GRanges)
importFrom(GenomicRanges,distanceToNearest)
importFrom(GenomicRanges,granges)
importFrom(IRanges,IRanges)
importFrom(IRanges,as.env)
importFrom(IRanges,countOverlaps)
importFrom(IRanges,coverage)
importFrom(IRanges,disjoin)
importFrom(IRanges,distanceToNearest)
importFrom(IRanges,extractList)
importFrom(IRanges,findOverlapPairs)
importFrom(IRanges,findOverlaps)
importFrom(IRanges,flank)
importFrom(IRanges,follow)
importFrom(IRanges,mid)
importFrom(IRanges,nearest)
importFrom(IRanges,pgap)
importFrom(IRanges,pintersect)
importFrom(IRanges,precede)
importFrom(IRanges,psetdiff)
importFrom(IRanges,punion)
importFrom(IRanges,ranges)
importFrom(IRanges,reduce)
importFrom(IRanges,resize)
importFrom(IRanges,shift)
importFrom(IRanges,slidingWindows)
importFrom(IRanges,subsetByOverlaps)
importFrom(IRanges,tile)
importFrom(Rsamtools,"bamFlag<-")
importFrom(Rsamtools,"bamMapqFilter<-")
importFrom(Rsamtools,"bamTag<-")
importFrom(Rsamtools,"bamWhat<-")
importFrom(Rsamtools,"bamWhich<-")
importFrom(Rsamtools,BamFile)
importFrom(Rsamtools,ScanBamParam)
importFrom(Rsamtools,bamFlag)
importFrom(Rsamtools,bamWhich)
importFrom(Rsamtools,scanBamFlag)
importFrom(Rsamtools,scanBamWhat)
importFrom(S4Vectors,"mcols<-")
importFrom(S4Vectors,"metadata<-")
importFrom(S4Vectors,DataFrame)
importFrom(S4Vectors,Pairs)
importFrom(S4Vectors,Rle)
importFrom(S4Vectors,cbind)
importFrom(S4Vectors,decode)
importFrom(S4Vectors,elementNROWS)
importFrom(S4Vectors,endoapply)
importFrom(S4Vectors,expand)
importFrom(S4Vectors,mcols)
importFrom(S4Vectors,merge)
importFrom(S4Vectors,metadata)
importFrom(S4Vectors,parallelVectorNames)
importFrom(S4Vectors,queryHits)
importFrom(S4Vectors,queryLength)
importFrom(S4Vectors,rbind)
importFrom(S4Vectors,runValue)
importFrom(S4Vectors,splitAsList)
importFrom(S4Vectors,subjectHits)
importFrom(S4Vectors,zipup)
importFrom(Seqinfo,"seqinfo<-")
importFrom(Seqinfo,"seqlevels<-")
importFrom(Seqinfo,"seqnames<-")
importFrom(Seqinfo,Seqinfo)
importFrom(Seqinfo,genome)
importFrom(Seqinfo,isCircular)
importFrom(Seqinfo,seqinfo)
importFrom(Seqinfo,seqlengths)
importFrom(Seqinfo,seqlevels)
importFrom(Seqinfo,seqnames)
importFrom(dplyr,arrange)
importFrom(dplyr,bind_cols)
importFrom(dplyr,bind_rows)
importFrom(dplyr,filter)
importFrom(dplyr,group_by)
importFrom(dplyr,group_data)
importFrom(dplyr,group_indices)
importFrom(dplyr,group_keys)
importFrom(dplyr,group_split)
importFrom(dplyr,group_vars)
importFrom(dplyr,groups)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,pull)
importFrom(dplyr,sample_n)
importFrom(dplyr,select)
importFrom(dplyr,slice)
importFrom(dplyr,slice_head)
importFrom(dplyr,slice_max)
importFrom(dplyr,slice_min)
importFrom(dplyr,slice_sample)
importFrom(dplyr,slice_tail)
importFrom(dplyr,summarise)
importFrom(dplyr,summarize)
importFrom(dplyr,tbl_vars)
importFrom(dplyr,ungroup)
importFrom(magrittr,"%>%")
importFrom(methods,as)
importFrom(methods,callNextMethod)
importFrom(methods,getGeneric)
importFrom(methods,getGenerics)
importFrom(methods,hasMethod)
importFrom(methods,initialize)
importFrom(methods,is)
importFrom(methods,new)
importFrom(methods,selectMethod)
importFrom(methods,setClass)
importFrom(methods,setMethod)
importFrom(methods,setValidity)
importFrom(methods,show)
importFrom(rlang,"!!!")
importFrom(rlang,":=")
importFrom(rlang,.data)
importFrom(rlang,dots_values)
importFrom(rlang,enquo)
importFrom(rlang,enquos)
importFrom(rlang,env_bind)
importFrom(rlang,env_get)
importFrom(rlang,env_parent)
importFrom(rlang,eval_tidy)
importFrom(rlang,have_name)
importFrom(rlang,new_data_mask)
importFrom(rlang,quo_is_missing)
importFrom(rlang,quo_name)
importFrom(rlang,quos)
importFrom(rlang,syms)
importFrom(rtracklayer,BigWigFile)
importFrom(rtracklayer,BigWigSelection)
importFrom(rtracklayer,export.bed)
importFrom(rtracklayer,export.bedGraph)
importFrom(rtracklayer,export.bw)
importFrom(rtracklayer,export.gff)
importFrom(rtracklayer,export.gff1)
importFrom(rtracklayer,export.gff2)
importFrom(rtracklayer,export.gff3)
importFrom(rtracklayer,export.wig)
importFrom(rtracklayer,import.bed)
importFrom(rtracklayer,import.bedGraph)
importFrom(rtracklayer,import.bw)
importFrom(rtracklayer,import.gff)
importFrom(rtracklayer,import.gff1)
importFrom(rtracklayer,import.gff2)
importFrom(rtracklayer,import.gff3)
importFrom(rtracklayer,import.wig)
importFrom(tidyselect,eval_select)
importFrom(utils,relist)

# plyranges: fluent genomic data analysis <img id="plyranges_logo" src="man/figures/logo.png" align="right" width = "125" />

<!-- badges: start -->

[![R-CMD-check-bioc](https://github.com/tidyomics/plyranges/workflows/R-CMD-check-bioc/badge.svg)](https://github.com/tidyomics/plyranges/actions?query=workflow%3AR-CMD-check-bioc)
[![BioC
status](http://www.bioconductor.org/shields/build/release/bioc/plyranges.svg)](https://bioconductor.org/checkResults/release/bioc-LATEST/plyranges)

<!-- badges: end -->

_plyranges_ provides a consistent interface for importing and wrangling genomics
data from a variety of sources. The package defines a grammar of genomic
data transformation based on _dplyr_ and the Bioconductor packages
_IRanges_, _GenomicRanges_, and _rtracklayer_. It does this by providing
a set of verbs for developing analysis pipelines based on _GRanges_
objects that represent genomic regions:

- Modify genomic regions with the `mutate()` and `stretch()` functions.
- Modify genomic regions while fixing the start/end/center coordinates
  with the `anchor_` family of functions.
- Sort genomic ranges with `arrange()`.
- Modify, subset, and aggregate genomic data with the `mutate()`,
  `filter()`, and `summarise()`functions.
- Any of the above operations can be performed on partitions of the data
  with `group_by()`.
- Find nearest neighbour genomic regions with the `join_nearest_` family
  of functions.
- Find overlaps between ranges with the `join_overlaps_` family of
  functions.
- Add additional metadata between ranges and a table with the `join_mcols_` 
  family of functions.
- Merge all overlapping and adjacent genomic regions with
  `reduce_ranges()`.
- Merge the end points of all genomic regions with `disjoin_ranges()`.
- Import and write common genomic data formats with the `read_/write_`
  family of functions.

# Documentation

For more details on the features of _plyranges_, read the 
[introductory vignette](https://tidyomics.github.io/plyranges/articles/an-introduction.html)
and the
[examples vignette](https://tidyomics.github.io/plyranges/articles/more-examples.html).

For a complete case-study on using _plyranges_ to combine ATAC-seq and
RNA-seq results read the [*fluentGenomics*
workflow](https://tidyomics.github.io/fluentGenomics).

_plyranges_ is part of the [tidyomics](https://github.com/tidyomics)
project, providing a _dplyr_-based interface for many types of
genomics datasets represented in Bioconductor.

# Installation

_plyranges_ can be installed from the latest Bioconductor release:

``` r
# install.packages("BiocManager")
BiocManager::install("plyranges")
```

To install the development version from GitHub:

``` r
BiocManager::install("tidyomics/plyranges")
```

# Learning more

In addition to the two package vignettes, see the following for more informtion:

- The [fluentGenomics workflow](https://sa-lee.github.io/fluentGenomics) 
  package shows how to combine differential gene expression and differential
  chromatin accessibility using _plyranges_.

- The [extended vignette in the plyrangesWorkshops
  package](https://github.com/sa-lee/plyrangesWorkshops) has a detailed
  walk through of using _plyranges_ for coverage analysis.

- The collection of genomic range applications including _plyranges_:
  [tidy ranges tutorial](https://tidyomics.github.io/tidy-ranges-tutorial).

# Citation

If you found _plyranges_ useful for your work please cite our
[paper](http://doi.org/10.1186/s13059-018-1597-8):

    @ARTICLE{Lee2019,
      title    = "plyranges: a grammar of genomic data transformation",
      author   = "Lee, Stuart and Cook, Dianne and Lawrence, Michael",
      journal  = "Genome Biol.",
      volume   =  20,
      number   =  1,
      pages    = "4",
      month    =  jan,
      year     =  2019,
      url      = "http://dx.doi.org/10.1186/s13059-018-1597-8",
      doi      = "10.1186/s13059-018-1597-8",
      pmc      = "PMC6320618"
    }

# Contributing

We welcome contributions from the R/Bioconductor community. We ask that
contributors follow the [code of conduct](.github/CODE_OF_CONDUCT.md)
and the guide outlined [here](.github/CONTRIBUTING.md).

summarize_rng <- function(.data, dots) {
  overscope <- overscope_ranges(.data)
  ans <- overscope_eval_update(overscope, dots, FALSE)
  
  # maintain list columns instead of collapsing them
  is_list <- vapply(ans, function(.) is(., "List") || is(., "list"), logical(1))
  
  # compress list columns
  if (any(is_list)) {
    nr <- check_n(.data)
    for (i in which(is_list)) {
      if (length(ans[[i]]) == 1) {
        ans[[i]] <- as(rep(ans[[i]], nr), "CompressedList")
      } else {
        
        if (all(lengths(ans[[i]]) == length(ans[[i]][[1]]))) {
          ans[[i]] <- as(BiocGenerics::Reduce(S4Vectors::pc, ans[[i]]), 
                         "CompressedList")
          # check length is equal to number of rows or records
          stopifnot(length(ans[[i]]) == nr)
        }
        

      }
    } 
  }
  
  results <- DataFrame(ans)
  rownames(results) <- NULL
  results
}

check_n <- function(.data) {
  if (is(.data, "GroupedGenomicRanges") || is(.data, "GroupedIntegerRanges")) {
    return(.data@n)
  }
  1L
}

#' Reduce multiple values in a Ranges down to a single value
#'
#' @param .data a Ranges object
#' @param ... Name-value pairs of summary functions. The name will be the 
#' name of the variable in the result. The value should be an expression that 
#' will return a value that has length one or length equal to the number of 
#' groups.
#' 
#' @details Creates one or more variables as a \code{S4Vectors::\link[S4Vectors:DataFrame-class]{DataFrame()}}
#' from the input Ranges object. If the ranges object is grouped, there will
#' be a row for each group. Because grouping may remove whether a Ranges object
#' is valid, a DataFrame is always returned.
#'
#' @return A \code{S4Vectors::\link[S4Vectors:DataFrame-class]{DataFrame()}}
#' @importFrom S4Vectors rbind cbind
#' @importFrom dplyr summarise summarize
#' @examples
#' df <- data.frame(start = 1:10, width = 5,  seqnames = "seq1",
#' strand = sample(c("+", "-", "*"), 10, replace = TRUE), gc = runif(10))
#' rng <- as_granges(df)
#' rng %>% summarise(gc = mean(gc))
#' rng %>% group_by(strand) %>% summarise(gc = mean(gc))
#' @method summarise Ranges
#' @rdname ranges-summarise
#' @export
summarise.Ranges <- function(.data, ...) {
  dots <- set_dots_named(...)
  summarize_rng(.data, dots)
}

#' @method summarise DelegatingGenomicRanges
#' @export
summarise.DelegatingGenomicRanges <- function(.data, ...) {
  dots <- set_dots_named(...)
  delegate <- .data@delegate
  summarize_rng(delegate, dots)
}

#' @method summarise DelegatingGenomicRanges
#' @export
summarise.DelegatingIntegerRanges <- function(.data, ...) {
  dots <- set_dots_named(...)
  delegate <- .data@delegate
  summarize_rng(delegate, dots)
}

#' @importFrom rlang !!! enquos
#' @importFrom dplyr bind_cols bind_rows
#' @method summarise GroupedGenomicRanges
#' @export
summarise.GroupedGenomicRanges <- function(.data, ...) {
  dots <- set_dots_named(...)
  md <- .data@group_keys
  res <- cbind(md, summarize_rng(.data, dots))
  res[order(res[, group_vars(.data)]), ]
}

#' @method summarise GroupedIntegerRanges
#' @export
summarise.GroupedIntegerRanges <- summarise.GroupedGenomicRanges

#' @include class-Operator.R

#' @title DeferredGenomiRanges objects
#'
#' @description  Enables deferred reading of files (currently only BAM files) by 
#' caching results after a plyranges verb is called.
#'
#' @slot delegate a GenomicRanges object to be cached
#' @slot ops  A FileOperator object
#' 
#' @seealso `read_bam()`
#' @export
#' @rdname ranges-deferred
setClass("DeferredGenomicRanges", 
         slots = c(ops = "FileOperator"),
         contains = "DelegatingGenomicRanges"
)

new_DeferredGenomicRanges <- function(delegate, ops) {
  new("DeferredGenomicRanges", delegate = delegate, ops = ops)
}

is_empty_delegate <- function(.data) length(.data@delegate) == 0L

# make a generic accessor for delegating Ranges?
load_delegate <-   function(.data) {
  if (is_empty_delegate(.data)) {
    load_genomic_file(.data@ops)
  } else {
    .data@delegate
  }
}

context("reading BAM files")

test_that("overlaps works as expected", {
  fl <- system.file("extdata", "ex1.bam", package="Rsamtools")
  olap <- GRanges(seqnames = "seq1", ranges = IRanges(1, width = 100))
  result <- read_bam(fl) %>% filter_by_overlaps(olap)
  expect_s4_class(result, "DeferredGenomicRanges")
  # now has cache filled
  expect_true(!is_empty_delegate(result))
  # cache has seqlengths methods
  expect_identical(c(seq1=1575L, seq2=1584L), seqlengths(result))
  # can we override cache by selecting a column not present
  result2 <- result %>% select(mapq)
  expect_true("mapq" %in% names(mcols(result2)))
  # but don't override cache if taking out a column
  result3 <- result %>% select(-cigar)
  expect_false("cigar" %in% names(mcols(result3)))
  # filter checks, removing minus strand should result in 1647 Ranges
  result4 <- read_bam(fl) %>% filter(!is_minus_strand)
  expect_true(length(result4) == 1647L)
  # override cache
  result5 <- read_bam(fl) %>% select(seq) %>% filter(!is_unmapped_query)
  expect_true(length(result5) == 3271L)

  # mapq filters
  result6 <- read_bam(fl) %>% filter(mapq > 30)
  expect_true(length(result6) == 3210L)
  
  # proper pairs 
  result7 <- read_bam(fl, paired = TRUE) %>%
    filter(is_proper_pair, !is_duplicate, !is_secondary_alignment, mapq > 30)
  expect_true(length(result7) == 3102L)
  
})

test_that("no index", {
  fl <- system.file("unitTests", "cases", "ex1_noindex.bam",
                    package="Rsamtools")
  result <- read_bam(fl, index = NULL)
  expect_s4_class(result, "DeferredGenomicRanges")
  expect_true(is_empty_delegate(result))
  exp <- DataFrame(n = 3271L)
  expect_identical(exp,  result %>% summarise(n = n()))
})

test_that("selection works with pipes", {
  fl <- system.file("extdata", "ex1.bam", package="Rsamtools")
  exp <- DataFrame(nm_sum = 924L)
  result <- read_bam(fl) %>% 
    select(NM) %>% 
    summarise(nm_sum = sum(NM))
  expect_identical(result, exp)
  
  result <- read_bam(fl) %>% 
    select(FO) %>% 
    summarise(fo_na = all(is.na(FO)))
  expect_true(result$fo_na)
})

#' Find overlaps within a Ranges object
#'
#' @param x A Ranges object
#' @param maxgap,minoverlap The maximimum gap between intervals as an integer
#' greater than or equal to zero. The minimum amount of overlap between intervals
#' as an integer greater than zero, accounting for the maximum gap.
#'
#' @details Self overlaps find any overlaps (or overlaps within or overlaps
#' directed) between a ranges object and itself.
#'
#' @return a Ranges object
#'
#' @examples
#' query <- data.frame(start = c(5,10, 15,20), width = 5, gc = runif(4)) %>%
#'              as_iranges()
#'
#' join_overlap_self(query)

#'
#' # -- GRanges objects, strand is ignored by default
#' query  <- data.frame(seqnames = "chr1",
#'                start = c(11,101),
#'                end = c(21, 200),
#'                name = c("a1", "a2"),
#'                strand = c("+", "-"),
#'                score = c(1,2)) %>%
#'            as_granges()
#'
#' # ignores strandedness
#' join_overlap_self(query)
#' join_overlap_self_within(query)
#' # adding directed prefix includes strand
#' join_overlap_self_directed(query)
#'
#'
#' @rdname ranges-overlaps-self
#' @seealso [find_overlaps()], [join_overlap_inner()]
#' @export
join_overlap_self <- function(x, maxgap, minoverlap) {
  UseMethod("join_overlap_self")
}

#' @export
join_overlap_self.GenomicRanges <- function(x, maxgap = -1L, minoverlap = 0L) {
  find_overlaps(x,x, maxgap, minoverlap, suffix = c("", ".overlap"))
}

#' @export
join_overlap_self.IntegerRanges <- join_overlap_self.GenomicRanges


#' @rdname ranges-overlaps-self
#' @export
join_overlap_self_within <- function(x, maxgap, minoverlap) {
  UseMethod("join_overlap_self_within")
}

#' @export
join_overlap_self_within.GenomicRanges <- function(x, maxgap = -1L, minoverlap = 0L) {
  find_overlaps_within(x,x, maxgap, minoverlap, suffix = c("", ".overlap"))
}

#' @export
join_overlap_self_within.IntegerRanges <- join_overlap_self_within.GenomicRanges

#' @rdname ranges-overlaps-self
#' @export
join_overlap_self_directed <- function(x, maxgap, minoverlap) {
  UseMethod("join_overlap_self_directed")
}

#' @export
join_overlap_self_directed.GenomicRanges <- function(x, maxgap = -1L, minoverlap = 0L) {
  find_overlaps_directed(x,x, maxgap, minoverlap, suffix = c("", ".overlap"))
}

#' @rdname ranges-overlaps-self
#' @export
join_overlap_self_within_directed <- function(x, maxgap, minoverlap) {
  UseMethod("join_overlap_self_within_directed")
}

#' @export
join_overlap_self_within_directed.GenomicRanges <- function(x, maxgap = -1L, minoverlap = 0L) {
  find_overlaps_within_directed(x,x, maxgap, minoverlap, suffix = c("", ".overlap"))
}

#' @method select DeferredGenomicRanges
#' @importFrom Rsamtools bamWhat<- bamTag<-
#' @export
select.DeferredGenomicRanges <- function(.data, ..., .drop_ranges = FALSE) {
 
  # check to see if we need to update cache, by selecting columns
  # that are not present in the delegate
  vars <- tbl_vars(.data@delegate)
  names(vars) <- vars 
  pos <- try(tidyselect::eval_select(rlang::expr(c(...)),
                                      vars,
                                      exclude = parallelVectorNames(.data@delegate)),
              silent = TRUE)
  
  
  if (is_empty_delegate(.data) || is(pos, "try-error")) {
    .data@ops <- select(.data@ops, ..., .drop_ranges = .drop_ranges)
    .data@delegate <- load_genomic_file(.data@ops)
  } else {
    .data@delegate <- select(.data@delegate, ..., .drop_ranges = .drop_ranges)
  }
  .data
}

#' @method filter DeferredGenomicRanges
#' @importFrom Rsamtools scanBamFlag bamFlag<-
#' @export
filter.DeferredGenomicRanges <- function(.data, ...) {
  dots <- quos(...)
  # check to see if we need to override cache 
  # this happens if any valid_flag_filters are available
  dot_names <- gsub("^!", "", unlist(lapply(dots, quo_name)))
  update_cache <- any(dot_names %in% unlist(names(valid_flag_filters())))
  # clear the cache
  if (update_cache & !is_empty_delegate(.data)) {
    .data@delegate <- GRanges()
  }
  
  if (is_empty_delegate(.data)) {
    .data@ops <- filter(.data@ops, ...)
    .data@delegate <- load_genomic_file(.data@ops)
  } else {
    .data@delegate <- filter(.data@delegate, ...)
  }
  .data
}



#' @importFrom Rsamtools bamWhich<-
#' @export
filter_by_overlaps.DeferredGenomicRanges <- function(x, y, maxgap = -1L, minoverlap = 0L) {
  if (is_empty_delegate(x)) {
    x@ops <- filter_by_overlaps(x@ops, y, maxgap, minoverlap)
    x@delegate <- load_genomic_file(x@ops)
  } else {
    x@delegate <- filter_by_overlaps(x@delegate, y, maxgap, minoverlap)
  }
  x
}

#' @method summarise DeferredGenomicRanges
#' @export
summarise.DeferredGenomicRanges <- function(.data, ...) {
  summarize(load_delegate(.data), ...)
}

#' @method mutate DeferredGenomicRanges
#' @export
mutate.DeferredGenomicRanges <- function(.data, ...) {
  mutate(load_delegate(.data), ...)
}

#' @method group_by DeferredGenomicRanges
#' @export
group_by.DeferredGenomicRanges <- function(.data, ...) {
  group_by(load_delegate(.data), ...)
}

# ranges-arithmetic-flank.R

#' Generate flanking regions
#' @description Find flanking regions  to the left or right or
#' upstream or downstream of a Ranges object.

#' @param x a Ranges object.
#' @param width the width of the flanking region relative to the ranges in
#' `x`. Either an integer vector of length 1 or an integer vector
#' the same length as x. The width can be negative in which case the
#' flanking region is reversed.
#'
#' @details The function
#' `flank_left` will create the flanking region to the left of starting
#' coordinates in `x`, while `flank_right` will create the flanking
#' region to the right of the starting coordinates in `x`. The function
#' `flank_upstream` will `flank_left` if the strand of rows in `x` is
#' not negative and will `flank_right` if the strand of rows in `x` is
#' negative. The function `flank_downstream` will `flank_right` if the strand of rows in `x` is
#' not negative and will `flank_leftt` if the strand of rows in `x` is
#' negative.
#'
#' By default `flank_left` and `flank_right` will
#' ignore strandedness of any ranges, while `flank_upstream` and
#' `flank_downstream` will take into account the strand of `x`.
#'
#' @return A Ranges object of same length as `x`.
#'
#' @seealso \code{IRanges::\link[IRanges:intra-range-methods]{flank()}},
#' \code{GenomicRanges::\link[GenomicRanges:intra-range-methods]{flank()}}
#' @importFrom IRanges flank
#' @examples
#' gr <- as_granges(data.frame(start = 10:15,
#'                             width = 5,
#'                             seqnames = "seq1",
#'                             strand = c("+", "+", "-", "-", "+", "*")))
#' flank_left(gr, width = 5L)
#' flank_right(gr, width = 5L)
#' flank_upstream(gr, width = 5L)
#' flank_downstream(gr, width = 5L)
#' @rdname ranges-flank
#' @export
flank_left <- function(x, width = 0L) {
  stopifnot(is.numeric(width))
  flank(x, width, start = TRUE, ignore.strand = TRUE)
}

#' @rdname ranges-flank
#' @export
flank_right <- function(x, width = 0L) {
  stopifnot(is.numeric(width))
  flank(x, width, start = FALSE, ignore.strand = TRUE)
}

#' @rdname ranges-flank
#' @export
flank_upstream <- function(x, width = 0L) {
  stopifnot(is.numeric(width))
  stopifnot(is(x, "GenomicRanges"))
  flank(x, width, start = TRUE, ignore.strand = FALSE)
}

#' @rdname ranges-flank
#' @export
flank_downstream <- function(x, width = 0L) {
  stopifnot(is.numeric(width))
  stopifnot(is(x, "GenomicRanges"))
  flank(x, width, start = FALSE, ignore.strand = FALSE)
}

context("flanks")

test_that("matches HelloRanges", {
  setwd(system.file("unitTests", "data", "flank", package="HelloRanges"))
  gr_a <- read_bed("a.bed")
  left <- flank(gr_a, 5L, ignore.strand = TRUE)
  right <- flank(gr_a, 5L, start = FALSE, ignore.strand = TRUE)

  expect_identical(flank_left(gr_a, 5L), left)
  expect_identical(flank_right(gr_a, 5L), right)

  up <- flank(gr_a, 5L, ignore.strand = FALSE)
  expect_identical(flank_upstream(gr_a, 5L), up)
  down <- flank(gr_a, 5L, start = FALSE, ignore.strand = FALSE)
  expect_identical(flank_downstream(gr_a, 5L), down)

})

test_that("matches GRanges tests", {
  expect_identical(flank_left(GRanges(), 10), GRanges())

  gr <- GRanges(seqnames = c("chr1", "chr2", "chr1", "chrM"),
                ranges = IRanges(21:24, width=10),
                strand = strand(c("+", "-", "*", "-")))

  target <- GRanges(seqnames = seqnames(gr),
                    ranges = IRanges(c(11, 32, 13, 34), width=10),
                    strand = strand(gr))
  current <- flank_upstream(gr, 10L)
  expect_identical(target, current)

  target <- GRanges(seqnames = seqnames(gr),
                    ranges = IRanges(c(31, 12, 33, 14), width=10),
                    strand = strand(gr))
  current <- flank_downstream(gr, 10L)
  expect_identical(target, current)

  target <- GRanges(seqnames = seqnames(gr),
                    ranges = IRanges(c(11, 12, 13, 14), width=10),
                    strand = strand(gr))
  current <- flank_left(gr, 10L)
  expect_identical(target, current)

  target <- GRanges(seqnames = seqnames(gr),
                    ranges = IRanges(c(31, 32, 33, 34), width=10),
                    strand = strand(gr))
  current <- flank_right(gr, 10L)
  expect_identical(target, current)
})

generate_slice <- function(.data, dots) {
  os <- overscope_ranges(.data)
  inx <- overscope_eval_update(os, dots, bind_envir = FALSE)
  check_num <- vapply(inx, is.numeric, logical(1))
  
  if (any(!check_num)) {
    stop("slice condition does not evaluate to an integer or numeric vector")
  }
  inx
}

#' Choose ranges by their position
#' 
#' @description
#' The `slice` family of functions let you select ranges by position.  They
#' follow the same conventions as their dplyr counterparts for data frames:
#'
#' * `slice()` selects ranges by explicit integer position.
#' * `slice_head()` / `slice_tail()` select the first or last ranges.
#' * `slice_sample()` randomly selects ranges.
#' * `slice_min()` / `slice_max()` select ranges with the smallest or largest
#'   values of a metadata column.
#'
#' All functions preserve grouping: when `.data` has been grouped with
#' [group_by()], selection is performed within each group.
#'
#' @param .data a `Ranges` object.
#' @param ... <[`data-masking`][rlang::args_data_masking]> Integer row values
#'   indicating positions to keep.  Only used by `slice()`; ignored elsewhere.
#' @param n Number of ranges to select.  Mutually exclusive with `prop`.
#'   For `slice_min()` and `slice_max()`, negative values are not supported.
#' @param prop Fraction of ranges to select, in `(0, 1]`.  Mutually exclusive
#'   with `n`.
#' @param order_by <[`data-masking`][rlang::args_data_masking]> Unquoted name
#'   of a numeric metadata column to order by.  Used by `slice_min()` and
#'   `slice_max()`.
#' @param with_ties If `TRUE` (the default), all ranges tied at the boundary
#'   rank are kept, which may return more than `n` ranges.  Set to `FALSE` to
#'   return exactly `n`.  Used by `slice_min()` and `slice_max()`.
#' @param weight_by <[`data-masking`][rlang::args_data_masking]> Unquoted name
#'   of a non-negative numeric metadata column used as sampling weights.  Used
#'   by `slice_sample()`.
#' @param replace Should sampling be with replacement?  Default `FALSE`.  Used
#'   by `slice_sample()`.
#' @param .preserve Ignored; retained for compatibility with the dplyr generic.
#'
#' @return A `GRanges` (or `IRanges`) object of the same class as `.data`, with
#'   grouping structure re-applied when `.data` was grouped.
#' 
#' @method slice Ranges
#' @importFrom dplyr slice
#' @rdname slice-ranges
#'
#' @examples
#' df <- data.frame(start = 1:10,
#'                  width = 5,
#'                  seqnames = "seq1",
#'                  strand = sample(c("+", "-", "*"), 10, replace = TRUE),
#'                  gc = runif(10))
#' rng <- as_granges(df)
#' dplyr::slice(rng, 1:2)
#' dplyr::slice(rng, -n())
#' dplyr::slice(rng, -5:-n())
#' 
#' by_strand <- group_by(rng, strand)
#' 
#' # slice with group by finds positions within each group
#' dplyr::slice(by_strand, n())
#' dplyr::slice(by_strand, which.max(gc))
#' 
#' # if the index is beyond the number of groups slice are ignored
#' dplyr::slice(by_strand, 1:3)
#' 
#' 
#' @export
slice.Ranges <- function(.data, ..., .preserve = FALSE) {
  dots <- set_dots_unnamed(...)
  inx <- generate_slice(.data, dots)
  inx <- Reduce(union, inx)
  .data[inx, ]
}

#' @rdname slice-ranges
#' @method slice GroupedGenomicRanges
#' @export
slice.GroupedGenomicRanges <- function(.data, ..., .preserve = FALSE) {
  dots <- set_dots_unnamed(...)
  inx <- .group_rows(.data)
  rng <- .data@delegate
  mch <- as(
    lapply(inx, 
           function(i) unlist(generate_slice(rng[i], dots))
    ),
    "IntegerList"
  )
  inx <- inx[mch]
  inx <- inx[!is.na(inx)]
  
  rng <- rng[sort(unlist(inx))]
  
  group_by(rng, !!!groups(.data))
  
}

#' @rdname slice-ranges
#' @method slice GroupedIntegerRanges
#' @export
slice.GroupedIntegerRanges <- slice.GroupedGenomicRanges

#' @importFrom dplyr sample_n
sample_n.Ranges <- function(tbl, size, replace = FALSE, weight = NULL, ...) 
{
  size <- rlang::enquo(size)
  weight <- rlang::enquo(weight)
  slice(tbl, sample.int(n(), check_length(!!size, n(), replace = replace), 
                        replace = replace, prob = !!weight))
}

check_length <- function(length, n, replace = FALSE) {
  if (length <= n || replace) {
    return(invisible(length))
  }
  
  stop(
  paste("`size`` must be less than or equal to", n, 
        "the length of the data. Set `replace` = TRUE to use sampling with replacement")
  )
}

# the following generated by Claude Sonnet 4.6 as part of adding new slice functionality
# 2026-03-18

# ── slice_head() methods ───────────────────────────────────────────────────────

#' @rdname slice-ranges
#' @importFrom dplyr slice_head
#' @method slice_head Ranges
#' @export
slice_head.Ranges <- function(.data, ..., n = NULL, prop = NULL,
                              .preserve = FALSE) {
  inx <- head_tail_index(length(.data), n = n, prop = prop, from_head = TRUE)
  .data[inx, ]
}

#' @rdname slice-ranges
#' @method slice_head GroupedGenomicRanges
#' @export
slice_head.GroupedGenomicRanges <- function(.data, ..., n = NULL, prop = NULL,
                                            .preserve = FALSE) {
  apply_grouped_slice(
    .data,
    function(grp_size) head_tail_index(grp_size, n = n, prop = prop,
                                       from_head = TRUE)
  )
}

#' @rdname slice-ranges
#' @method slice_head GroupedIntegerRanges
#' @export
slice_head.GroupedIntegerRanges <- slice_head.GroupedGenomicRanges


# ── slice_tail() methods ───────────────────────────────────────────────────────

#' @rdname slice-ranges
#' @importFrom dplyr slice_tail
#' @method slice_tail Ranges
#' @export
slice_tail.Ranges <- function(.data, ..., n = NULL, prop = NULL,
                              .preserve = FALSE) {
  inx <- head_tail_index(length(.data), n = n, prop = prop, from_head = FALSE)
  .data[inx, ]
}

#' @rdname slice-ranges
#' @method slice_tail GroupedGenomicRanges
#' @export
slice_tail.GroupedGenomicRanges <- function(.data, ..., n = NULL, prop = NULL,
                                            .preserve = FALSE) {
  apply_grouped_slice(
    .data,
    function(grp_size) head_tail_index(grp_size, n = n, prop = prop,
                                       from_head = FALSE)
  )
}

#' @rdname slice-ranges
#' @method slice_tail GroupedIntegerRanges
#' @export
slice_tail.GroupedIntegerRanges <- slice_tail.GroupedGenomicRanges


# ── slice_sample() methods ─────────────────────────────────────────────────────

#' @rdname slice-ranges
#' @importFrom dplyr slice_sample
#' @method slice_sample Ranges
#' @export
slice_sample.Ranges <- function(.data, ..., n = NULL, prop = NULL,
                                weight_by = NULL, replace = FALSE,
                                .preserve = FALSE) {
  total   <- length(.data)
  size    <- resolve_n_prop(total, n = n, prop = prop, replace = replace)
  weights <- eval_weight_by(.data, rlang::enquo(weight_by))
  inx     <- sample.int(total, size, replace = replace, prob = weights)
  .data[inx, ]
}

#' @rdname slice-ranges
#' @method slice_sample GroupedGenomicRanges
#' @export
slice_sample.GroupedGenomicRanges <- function(.data, ..., n = NULL,
                                              prop = NULL, weight_by = NULL,
                                              replace = FALSE,
                                              .preserve = FALSE) {
  weight_quo <- rlang::enquo(weight_by)
  rng        <- .data@delegate
  grp_rows   <- .group_rows(.data)

  selected <- lapply(grp_rows, function(i) {
    sub       <- rng[i]
    size      <- resolve_n_prop(length(i), n = n, prop = prop, replace = replace)
    w         <- eval_weight_by(sub, weight_quo)
    local_inx <- sample.int(length(i), size, replace = replace, prob = w)
    i[local_inx]
  })

  rng_out <- rng[sort(unlist(selected))]
  group_by(rng_out, !!!groups(.data))
}

#' @rdname slice-ranges
#' @method slice_sample GroupedIntegerRanges
#' @export
slice_sample.GroupedIntegerRanges <- slice_sample.GroupedGenomicRanges


# ── slice_min() methods ────────────────────────────────────────────────────────

#' @rdname slice-ranges
#' @importFrom dplyr slice_min
#' @method slice_min Ranges
#' @export
slice_min.Ranges <- function(.data, order_by, ..., n = NULL, prop = NULL,
                             with_ties = TRUE, .preserve = FALSE) {
  order_quo <- rlang::enquo(order_by)
  inx <- minmax_index(.data, order_quo, n = n, prop = prop,
                      with_ties = with_ties, decreasing = FALSE)
  .data[inx, ]
}

#' @rdname slice-ranges
#' @method slice_min GroupedGenomicRanges
#' @export
slice_min.GroupedGenomicRanges <- function(.data, order_by, ..., n = NULL,
                                           prop = NULL, with_ties = TRUE,
                                           .preserve = FALSE) {
  order_quo <- rlang::enquo(order_by)
  apply_grouped_minmax(.data, order_quo, n = n, prop = prop,
                       with_ties = with_ties, decreasing = FALSE)
}

#' @rdname slice-ranges
#' @method slice_min GroupedIntegerRanges
#' @export
slice_min.GroupedIntegerRanges <- slice_min.GroupedGenomicRanges


# ── slice_max() methods ────────────────────────────────────────────────────────

#' @rdname slice-ranges
#' @importFrom dplyr slice_max
#' @method slice_max Ranges
#' @export
slice_max.Ranges <- function(.data, order_by, ..., n = NULL, prop = NULL,
                             with_ties = TRUE, .preserve = FALSE) {
  order_quo <- rlang::enquo(order_by)
  inx <- minmax_index(.data, order_quo, n = n, prop = prop,
                      with_ties = with_ties, decreasing = TRUE)
  .data[inx, ]
}

#' @rdname slice-ranges
#' @method slice_max GroupedGenomicRanges
#' @export
slice_max.GroupedGenomicRanges <- function(.data, order_by, ..., n = NULL,
                                           prop = NULL, with_ties = TRUE,
                                           .preserve = FALSE) {
  order_quo <- rlang::enquo(order_by)
  apply_grouped_minmax(.data, order_quo, n = n, prop = prop,
                       with_ties = with_ties, decreasing = TRUE)
}

#' @rdname slice-ranges
#' @method slice_max GroupedIntegerRanges
#' @export
slice_max.GroupedIntegerRanges <- slice_max.GroupedGenomicRanges


# ── Internal helpers ───────────────────────────────────────────────────────────

generate_slice <- function(.data, dots) {
  os  <- overscope_ranges(.data)
  inx <- overscope_eval_update(os, dots, bind_envir = FALSE)
  check_num <- vapply(inx, is.numeric, logical(1))
  if (any(!check_num)) {
    stop("slice condition does not evaluate to an integer or numeric vector")
  }
  inx
}

# Resolve n / prop to a concrete integer size.
resolve_n_prop <- function(total, n = NULL, prop = NULL, replace = FALSE) {
  if (!is.null(n) && !is.null(prop)) {
    stop("Exactly one of `n` or `prop` must be supplied, not both.")
  }
  if (!is.null(prop)) {
    if (prop <= 0 || prop > 1) stop("`prop` must be in (0, 1].")
    n <- max(1L, floor(prop * total))
  }
  if (is.null(n)) n <- 1L  # default to n = 1
  n <- as.integer(n)
  if (!replace && n > total) {
    stop(paste("`n` (", n, ") must be <= length of .data (", total, ").",
               "Set `replace = TRUE` to use sampling with replacement."))
  }
  n
}

# Build a positional index for head / tail selection.
head_tail_index <- function(total, n = NULL, prop = NULL, from_head) {
  size <- resolve_n_prop(total, n = n, prop = prop)
  if (from_head) seq_len(size) else seq.int(total - size + 1L, total)
}

# Evaluate an optional weight_by quosure against .data's mcols.
eval_weight_by <- function(.data, weight_quo) {
  if (rlang::quo_is_null(weight_quo)) return(NULL)
  col_name <- rlang::as_name(weight_quo)
  w <- mcols(.data)[[col_name]]
  if (is.null(w)) stop(paste0("Column '", col_name, "' not found in .data."))
  if (any(w < 0, na.rm = TRUE)) stop("`weight_by` values must be non-negative.")
  w
}

# Shared guard used by slice_min / slice_max.
check_column_numeric <- function(x, col) {
  if (!is.numeric(x)) {
    stop(paste0("`", col, "` must be a numeric column for slice_min / slice_max"))
  }
}

# Build an ordering index for slice_min / slice_max.
minmax_index <- function(.data, order_quo, n = NULL, prop = NULL, with_ties,
                         decreasing) {
  col_name <- rlang::as_name(order_quo)
  vals     <- mcols(.data)[[col_name]]
  if (is.null(vals)) stop(paste0("Column '", col_name, "' not found in .data."))
  check_column_numeric(vals, col_name)

  total <- length(.data)
  size  <- resolve_n_prop(total, n = n, prop = prop)
  ord   <- order(vals, decreasing = decreasing, na.last = NA)

  if (with_ties) {
    threshold <- vals[ord[min(size, length(ord))]]
    keep <- if (decreasing) which(vals >= threshold) else which(vals <= threshold)
    sort(keep)
  } else {
    sort(ord[seq_len(min(size, length(ord)))])
  }
}

# Apply a grouped slice using a per-group index function.
# `index_fn` receives the group size and returns *local* indices.
apply_grouped_slice <- function(.data, index_fn) {
  rng      <- .data@delegate
  grp_rows <- .group_rows(.data)

  selected <- lapply(grp_rows, function(i) {
    local_inx <- index_fn(length(i))
    local_inx <- local_inx[local_inx >= 1L & local_inx <= length(i)]
    i[local_inx]
  })

  rng_out <- rng[sort(unlist(selected))]
  group_by(rng_out, !!!groups(.data))
}

# Grouped min / max helper.
apply_grouped_minmax <- function(.data, order_quo, n = NULL, prop = NULL,
                                 with_ties, decreasing) {
  col_name <- rlang::as_name(order_quo)
  rng      <- .data@delegate
  grp_rows <- .group_rows(.data)

  selected <- lapply(grp_rows, function(i) {
    sub  <- rng[i]
    vals <- mcols(sub)[[col_name]]
    if (is.null(vals)) stop(paste0("Column '", col_name, "' not found."))
    check_column_numeric(vals, col_name)

    size <- resolve_n_prop(length(i), n = n, prop = prop)
    ord  <- order(vals, decreasing = decreasing, na.last = NA)

    local_inx <- if (with_ties) {
      threshold <- vals[ord[min(size, length(ord))]]
      if (decreasing) which(vals >= threshold) else which(vals <= threshold)
    } else {
      ord[seq_len(min(size, length(ord)))]
    }

    i[sort(local_inx)]
  })

  rng_out <- rng[sort(unlist(selected))]
  group_by(rng_out, !!!groups(.data))
}

# ranges-rangewise-setops.R

#' Vector-wise Range set-operations
#'
#' @param x,y Two Ranges objects to compare.
#'
#' @details These are usual set-operations that act on the sets of the
#' ranges represented in x and y. By default these operations will ignore
#' any strand information. The directed versions of these functions will
#' take into account strand for GRanges objects.
#'
#' @return A Ranges object
#'
#' @examples
#' gr1 <- data.frame(seqnames = "chr1",
#'                   start = c(2,9),
#'                   end = c(7,9),
#'                   strand = c("+", "-")) %>%
#'                as_granges()
#' gr2 <- data.frame(seqnames = "chr1", start = 5, width = 5, strand = "-") %>%
#'          as_granges()
#'
#' union_ranges(gr1, gr2)
#' union_ranges_directed(gr1, gr2)
#'
#' intersect_ranges(gr1, gr2)
#' intersect_ranges_directed(gr1, gr2)
#'
#' setdiff_ranges(gr1, gr2)
#' setdiff_ranges_directed(gr1, gr2)
#' # taking the complement of a ranges requires annotation information
#' gr1 <- set_genome_info(gr1, seqlengths = 100)
#' complement_ranges(gr1)
#'
#' @rdname ranges-setops
#' @export
intersect_ranges <- function(x,y) { UseMethod("intersect_ranges") }

#' @export
intersect_ranges.IntegerRanges <- function(x,y) {
  IRanges::intersect(x,y)
}

#' @export
intersect_ranges.GenomicRanges <- function(x,y) {
  GenomicRanges::intersect(x,y,ignore.strand = TRUE)
}

#' @export
#' @rdname ranges-setops
intersect_ranges_directed <- function(x,y) { UseMethod("intersect_ranges_directed") }

#' @export
intersect_ranges_directed.GenomicRanges <- function(x,y) {
  GenomicRanges::intersect(x,y, ignore.strand = FALSE)
}

#' @export
#' @rdname ranges-setops
union_ranges <- function(x,y) { UseMethod("union_ranges") }

#' @export
union_ranges.IntegerRanges <- function(x,y) {
  IRanges::union(x,y)
}

#' @export
union_ranges.GenomicRanges <- function(x,y) {
  GenomicRanges::union(x,y,ignore.strand = TRUE)
}

#' @export
#' @rdname ranges-setops
union_ranges_directed <- function(x,y) { UseMethod("union_ranges_directed") }

#' @export
union_ranges_directed.GenomicRanges <- function(x,y) {
  GenomicRanges::union(x,y, ignore.strand = FALSE)
}

#' @export
#' @rdname ranges-setops
setdiff_ranges <- function(x,y) { UseMethod("setdiff_ranges") }

#' @export
setdiff_ranges.IntegerRanges <- function(x,y) {
  IRanges::setdiff(x,y)
}

#' @export
setdiff_ranges.GenomicRanges <- function(x,y) {
  GenomicRanges::setdiff(x,y, ignore.strand = TRUE)
}

#' @export
#' @rdname ranges-setops
setdiff_ranges_directed <- function(x,y) { UseMethod("setdiff_ranges_directed") }

#' @export
setdiff_ranges_directed.GenomicRanges <- function(x,y) {
  GenomicRanges::setdiff(x,y, ignore.strand = FALSE)
}

#' @rdname ranges-setops
#' @export
complement_ranges  <- function(x) { UseMethod("complement_ranges") }

#' @export
complement_ranges.IntegerRanges <- function(x) {
  IRanges::setdiff(IRanges(start = min(start(x)), end = max(end(x))), x)
}

#' @export
complement_ranges.GenomicRanges <- function(x) {
  y <- try(get_genome_info(x), silent = TRUE)
  if (is(y, "try-error")) {
    stop("complement_ranges requires sequence lengths information", call. = FALSE)
  }
  setdiff_ranges(y, x)
}

#' @rdname ranges-setops
#' @export
complement_ranges_directed <- function(x) {UseMethod("complement_ranges_directed")}

#' @export
complement_ranges_directed.GenomicRanges <- function(x) {
  y <- try(get_genome_info(x), silent = TRUE)
  if (is(y, "try-error")) {
    stop("complement_ranges_directed requires sequence lengths information", call. = FALSE)
  }
  setdiff_ranges_directed(y, x)
}

#' Pair together two ranges objects
#'
#' @param x,y Ranges objects to pair together.
#' @param maxgap,minoverlap The maximimum gap between intervals as an integer
#' greater than or equal to negative one. The minimum amount of overlap between intervals
#' as an integer greater than zero, accounting for the maximum gap.
#' @param suffix A character vector of length two used to identify metadata columns
#' coming from x and y.
#'
#' @details These functions return a DataFrame object, and is one way of
#' representing paired alignments with plyranges.
#'
#' @return a DataFrame with two ranges columns and the corresponding metadata
#' columns.
#'
#' @seealso [join_nearest()][join_overlap_inner()][join_precede()][join_follow()]
#' @examples
#' query <- data.frame(start = c(5,10, 15,20), width = 5, gc = runif(4)) %>%
#'              as_iranges()
#' subject <- data.frame(start = 2:6, width = 3:7, label = letters[1:5]) %>%
#'              as_iranges()
#'
#' pair_overlaps(query, subject)
#' pair_overlaps(query, subject, minoverlap = 5)
#' pair_nearest(query, subject)
#'
#'
#' query  <- data.frame(seqnames = "chr1",
#'                start = c(11,101),
#'                end = c(21, 200),
#'                name = c("a1", "a2"),
#'                strand = c("+", "-"),
#'                score = c(1,2)) %>%
#'            as_granges()
#' subject <- data.frame(seqnames = "chr1",
#'                       strand = c("+", "-", "+", "-"),
#'                       start = c(21,91,101,201),
#'                       end = c(30,101,110,210),
#'                       name = paste0("b", 1:4),
#'                       score = 1:4) %>%
#'                    as_granges()
#'
#' # ignores strandedness
#' pair_overlaps(query, subject, suffix = c(".query", ".subject"))
#' pair_follow(query, subject, suffix = c(".query", ".subject"))
#' pair_precede(query, subject, suffix = c(".query", ".subject"))
#' pair_precede(query, subject, suffix = c(".query", ".subject"))
#' @rdname ranges-pairs
#' @export
pair_overlaps <- function(x, y, maxgap, minoverlap, suffix) {
  UseMethod("pair_overlaps")
}

#' @export
pair_overlaps.IntegerRanges <- function(x, y, maxgap = -1L, minoverlap = 0L, suffix = c(".x", ".y")) {
  hits <- findOverlaps(x,y, maxgap, minoverlap, type = "any", select = "all")
  expand_by_hits(x,y, suffix, hits, return_data_frame = TRUE)
}

#' @export
pair_overlaps.GenomicRanges <- function(x, y, maxgap = -1L, minoverlap = 0L, suffix = c(".x", ".y")) {
  hits <- findOverlaps(x,y, maxgap, minoverlap,
                       type = "any", select = "all", ignore.strand = TRUE)
  expand_by_hits(x,y, suffix, hits, return_data_frame = TRUE)
}

#' @rdname ranges-pairs
#' @export
pair_nearest <- function(x, y, suffix) {
  UseMethod("pair_nearest")
}

#' @export
pair_nearest.IntegerRanges <- function(x,y, suffix = c(".x", ".y")) {
  hits <- nearest(x,y, select = "arbitrary")
  expand_by_hits(x,y, suffix, hits, return_data_frame = TRUE)
}

#' @export
pair_nearest.GenomicRanges <- function(x,y, suffix = c(".x", ".y")) {
  hits <- nearest(x,y, select = "arbitrary", ignore.strand = TRUE)
  expand_by_hits(x,y, suffix, hits, return_data_frame = TRUE)
}

#' @rdname ranges-pairs
#' @export
pair_precede <- function(x, y, suffix) {
  UseMethod("pair_precede")
}

#' @export
pair_precede.IntegerRanges <- function(x,y, suffix = c(".x", ".y")) {
  hits <- precede(x,y)
  expand_by_hits(x,y,suffix, hits, return_data_frame = TRUE)
}

#' @export
pair_precede.GenomicRanges <- function(x,y, suffix = c(".x", ".y")) {
  hits <- precede(x,y, ignore.strand = TRUE)
  expand_by_hits(x,y, suffix, hits, return_data_frame = TRUE)
}

#' @rdname ranges-pairs
#' @export
pair_follow <- function(x, y, suffix) {
  UseMethod("pair_follow")
}
#' @export
pair_follow.IntegerRanges <- function(x,y, suffix = c(".x", ".y")) {
  hits <- follow(x,y)
  expand_by_hits(x,y,suffix, hits, return_data_frame = TRUE)
}

#' @export
pair_follow.GenomicRanges <- function(x,y, suffix = c(".x", ".y")) {
  hits <- follow(x,y, ignore.strand = TRUE)
  expand_by_hits(x,y, suffix, hits, return_data_frame =  TRUE)
}

#' @param ranges ranges object from join_nearest expand_hits internal function
#' @param hits  hits from distanceToNearest
#' @param colname name of column to hold distance values
#'
#' @noRd
add_distance_col <- function(ranges, hits, colname) {
  
  if (colname %in% names(mcols(ranges))){
    stop(paste0(colname, " already exists in destination metadata"))
  }
  
  if (is.null(mcols(ranges))){
    # handle IRanges NULL adding X column of NA's
    meta <- DataFrame("distance" = NA_integer_)
    names(meta) <- colname
    mcols(ranges) <- meta
  } else {
    mcols(ranges)[[colname]] <-  NA_integer_
  }
  
  mcols(ranges[queryHits(hits)])[[colname]] <- mcols(hits)$distance
  
  return(ranges)
}

#' Macro for building add_nearest_distance_* family of functions
#' Change this to edit defaults for all functions
#' @param fun nearest hits helper function
#' @noRd
make_add_nearest_distance <- function(fun){
  f <- function(x, y = x, name = "distance"){
    hits <- fun(x, y)
    
    add_distance_col(x, hits, colname = name)
  }
}

#' Add distance to nearest neighbours between two Ranges objects
#' 
#' Appends distance to nearest subject range to query ranges similar to setting
#' `distance` in `join_nearest_`. Distance is set to `NA` for features with no
#' nearest feature by the selected nearest metric.
#'
#' @param x The query ranges
#' @param y the subject ranges within which the nearest ranges are found.
#'   If missing, query ranges are used as the subject.
#' @param name column name to create containing distance values
#'   
#' @details By default `add_nearest_distance` will find arbitrary nearest
#' neighbours in either direction and ignore any strand information.
#' The `add_nearest_distance_left` and `add_nearest_distance_right`  methods
#' will find arbitrary nearest neighbour ranges on x that are left/right of
#' those on y and ignore any strand information.
#'
#' The `add_nearest_distance_upstream` method will find arbitrary nearest
#' neighbour ranges on x that are upstream of those on y. This takes into
#' account strandedness of the ranges.
#' On the positive strand nearest upstream will be on the
#' left and on the negative strand nearest upstream will be on the right.
#'
#' The `add_nearest_distance_downstream` method will find arbitrary nearest
#' neighbour ranges on x that are upstream of those on y. This takes into
#' account strandedness of the ranges. On the positive strand nearest downstream
#' will be on the right and on the negative strand nearest upstream will be on
#' the left.
#'
#' @return ranges in `x` with additional column containing the distance to the
#'   nearest range in `y`.
#'  
#' @rdname add-nearest-distance
#' @seealso \code{\link{join_nearest}}
#' @export
#'
#' @examples
#' query <- data.frame(start = c(5,10, 15,20),
#'                    width = 5,
#'                    gc = runif(4)) %>%
#'              as_iranges()
#' subject <- data.frame(start = c(2:6, 24),
#'                       width = 3:8,
#'                       label = letters[1:6]) %>%
#'              as_iranges()
#'              
#' add_nearest_distance(query, subject)
#' add_nearest_distance_left(query, subject)
#' add_nearest_distance_left(query)
#' @export
add_nearest_distance <- make_add_nearest_distance(hits_nearest)

#' @rdname add-nearest-distance
#' @export
add_nearest_distance_left <- make_add_nearest_distance(hits_nearest_left)

#' @rdname add-nearest-distance
#' @export
add_nearest_distance_right <- make_add_nearest_distance(hits_nearest_right)

#' @rdname add-nearest-distance
#' @export
add_nearest_distance_upstream <- make_add_nearest_distance(hits_nearest_upstream)

#' @rdname add-nearest-distance
#' @export
add_nearest_distance_downstream <- make_add_nearest_distance(hits_nearest_downstream)

#' Stretch a genomic interval
#'
#' @param x a Ranges object, to fix by either the start, end or center
#' of an interval use `anchor_start(x)`, `anchor_end(x)`,
#' `anchor_center(x)`. To fix by strand use `anchor_3p(x)` or
#' `anchor_5p(x)`.
#' @param extend the amount to alter the width of a Ranges object by. Either an
#' integer vector of length 1 or an integer vector the same length as x.
#' @description By default, `stretch(x)` will anchor by the center of a Ranges
#' object. This means that half of the value of `extend` will be added to
#' the end of the range and the remaining half subtracted from the start of 
#' the Range. The other anchors will leave the start/end fixed and stretch
#' the end/start respectively. 
#' @export
#' @return a Ranges object with modified start or end (or both) coordinates
#' @seealso `anchor()`, `mutate()`
#' @examples
#' rng <- as_iranges(data.frame(start=c(2:-1, 13:15), width=c(0:3, 2:0)))
#' rng2 <- stretch(anchor_center(rng), 10)
#' stretch(anchor_start(rng2), 10)
#' stretch(anchor_end(rng2), 10)
#' grng <- as_granges(data.frame(seqnames = "chr1",
#'                          strand = c("+", "-", "-", "+", "+", "-", "+"),
#'                          start=c(2:-1, 13:15),
#'                          width=c(0:3, 2:0)))
#' stretch(anchor_3p(grng), 10)
#' stretch(anchor_5p(grng), 10)
stretch <- function(x, extend) { UseMethod("stretch") }

#' @export
stretch.Ranges <- function(x, extend = 0L) {
  stopifnot(length(extend) == 1L | length(extend) == length(x))
  stretch_center(x, extend)
}

#' @export
stretch.AnchoredIntegerRanges <- function(x, extend = 0L) {
  anchor <- anchor(x)
  rng <- x@delegate
  stopifnot(length(extend) == 1L | length(extend) == length(rng))
  switch(
    anchor,
    start = mutate(rng, end = end + extend),
    end = mutate(rng, start = start - extend),
    center = stretch_center(rng, extend)
  )
}


#' @export
stretch.AnchoredGenomicRanges <- function(x, extend = 0L) {
  anchor <- anchor(x)
  rng <- x@delegate
  stopifnot(length(extend) == 1L | length(extend) == length(rng))
  
  switch(
    anchor,
    start = mutate(rng, end = end + extend),
    end = mutate(rng, start = start - extend),
    center = stretch_center(rng, extend),
    "3p" = stretch_by_strand(rng, extend, "3p"),
    "5p" = stretch_by_strand(rng, extend, "5p")
  )
}


#' @importFrom IRanges mid
stretch_center <- function(x, extend) {
  x + extend / 2
}

stretch_by_strand <- function(x, extend, anchor) {
  # anchor by 3p start is fixed for negative strand
  # and end is fixed for positive strand
  if (anchor == "3p") {
    x_pos <- x[strand(x) == "+" | strand(x) == "*"]
    start(x_pos) <- start(x_pos) - extend
    x_neg <- x[strand(x) == "-"]
    end(x_neg) <- end(x_neg) + extend
  } else {
    x_pos <- x[strand(x) == "+" | strand(x) == "*"]
    end(x_pos) <- end(x_pos) + extend
    x_neg <- x[strand(x) == "-"]
    start(x_neg) <- start(x_neg) - extend
  }
  x[strand(x) == "+" | strand(x) == "*"] <- x_pos
  x[strand(x) == "-"] <- x_neg
  return(x)
}

#' Disjoin then aggregate a Ranges object
#'
#' @param .data a Ranges object to disjoin
#' @param ... Name-value pairs of summary functions.
#'
#' @return a Ranges object that is now disjoint (no bases overlap). 
#' @rdname ranges-disjoin
#' @importFrom IRanges disjoin
#' @examples
#' df <- data.frame(start = 1:10, width = 5,  seqnames = "seq1",
#' strand = sample(c("+", "-", "*"), 10, replace = TRUE), gc = runif(10))
#' rng <- as_granges(df)
#' rng %>% disjoin_ranges()
#' rng %>% disjoin_ranges(gc = mean(gc))
#' rng %>% disjoin_ranges_directed(gc = mean(gc))
#' @export
disjoin_ranges <- function(.data, ...) { UseMethod("disjoin_ranges") }

#' @method disjoin_ranges IntegerRanges
#' @export
disjoin_ranges.IntegerRanges <- function(.data, ...) {
  reduce_single(.data, ..., rfun = disjoin)
}

#' @method disjoin_ranges GroupedIntegerRanges
#' @export
disjoin_ranges.GroupedIntegerRanges <- function(.data, ...) {
  reduce_by_grp(.data, ..., rfun = disjoin)
}

#' @method disjoin_ranges GenomicRanges
#' @export
disjoin_ranges.GenomicRanges <- function(.data, ...) {
  reduce_single(.data, ...,
                rfun = function(x, ...) {
                  disjoin(x, ..., ignore.strand = TRUE)
                })
}

#' @method disjoin_ranges GroupedGenomicRanges
#' @export
disjoin_ranges.GroupedGenomicRanges <- function(.data, ...) {
  reduce_by_grp(.data, ...,
                rfun = function(x, ...) {
                  disjoin(x, ..., ignore.strand = TRUE)
                })
}

#' @rdname ranges-disjoin
#' @export
disjoin_ranges_directed <- function(.data, ...) {
  UseMethod("disjoin_ranges_directed")
}

#' @importFrom IRanges disjoin
#' @method disjoin_ranges_directed GenomicRanges
#' @export
disjoin_ranges_directed.GenomicRanges <- function(.data, ...) {
  reduce_single(.data, ...,
                rfun = function(x, ...) {
                  disjoin(x, ..., ignore.strand = FALSE)
                })
}

#' @method disjoin_ranges_directed GroupedGenomicRanges
#' @export
disjoin_ranges_directed.GroupedGenomicRanges <- function(.data, ...) {
  reduce_by_grp(.data, ...,
                rfun = function(x, ...) {
                  disjoin(x, ..., ignore.strand = FALSE)
                })
}

context("disjoin ranges")

test_that("matches IRanges/GRanges tests", {
  x <- IRanges()
  expect_identical(x, disjoin_ranges(x))
  ir <- IRanges(c(1, 21, 10, 1, 15, 5, 20, 20),
                c(6, 20, 9, 3, 14, 11, 20, 19))
  correct_ir <- IRanges(c(1, 4, 5, 7, 10, 20), c(3, 4, 6, 9, 11, 20))
  test_ir <- disjoin_ranges(ir)
  expect_identical(test_ir, correct_ir)

  # check revmap
  test_ir <- ir %>%
    mutate(i = 1:n()) %>%
    disjoin_ranges(revmap = IRanges::IntegerList(i))

  mcols(correct_ir)$revmap <- IRanges::IntegerList(c(1, 4), 1, c(1, 6), 6, 6, 7)
  expect_identical(test_ir, correct_ir)

  # -- granges
  gr <- GRanges(Rle(factor(c("chr1", "chr2", "chr1", "chr3")), c(1, 3, 2, 4)),
          IRanges(1:10, width=10:1, names=head(letters, 10)),
          Rle(c("-", "+", "*", "+", "-"), c(1, 2, 2, 3, 2)),
          score=1:10, GC=seq(1, 0, length=10),
          seqinfo=Seqinfo(paste("chr", 1:3, sep="")))

  correct_gr <- GRanges(Rle(c("chr1", "chr2", "chr3"), c(3, 3, 4)),
                        IRanges(start=c(6, 1, 5, 2, 3, 4, 7, 8, 9, 10),
                                end=c(10, 10, 10, 2, 10, 10, 7, 10, 9, 10)),
                        c("+", "-", "*", "+", "+", "*", "+", "+", "-", "-"))
  # matches directed
  expect_identical(disjoin_ranges_directed(gr), correct_gr)
  # this is the same as disjoin unstranded on correct_gr
  expect_identical(disjoin_ranges(gr), disjoin_ranges(correct_gr))

  gr <- GRanges(Rle(c("chr1", "chr3"), c(2, 2)),
                IRanges(c(8, 6, 8, 6), c(11, 15, 11, 15),
                        names=c("k", "l", "m", "n")),
                c("-", "-", "+", "*"),
                score=11:14, GC=c(.2, .3, .3, .1))

  correct_gr <- GRanges(Rle(c("chr1", "chr3"), c(3, 2)),
                        IRanges(c(6, 8, 12, 8, 6), c(7, 11, 15, 11, 15)),
                        Rle(c("-", "+", "*"), c(3, 1, 1)))
  mcols(correct_gr)$revmap <- IRanges::IntegerList(2, 1:2, 2, 3, 4)
  expect_identical(gr %>%
                     mutate(i = 1:n()) %>%
                     disjoin_ranges_directed(revmap = IRanges::IntegerList(i)),
                   correct_gr)

  # grouping works as expected
  grl <- GRangesList( GRanges(Rle(factor(c("chr1", "chr2", "chr1", "chr3")), c(1, 3, 2, 4)),
                              IRanges(1:10, width=10:1),
                              Rle(c("-", "+", "*", "+", "-"), c(1, 2, 2, 3, 2)),
                              score=1:10, GC=seq(1, 0, length=10),
                              seqinfo=Seqinfo(paste("chr", 1:3, sep=""))),
                      GRanges("1", IRanges(1, 10), score=21, GC=.21),
                      GRanges(),
                      GRanges(Rle(c("chr1", "chr3"), c(2, 2)),
                              IRanges(c(8, 6, 8, 6), c(11, 15, 11, 15)),
                              strand(c("-", "-","+","*")),
                              score=41:44, GC=c(.41, .42, .43, .44)))

  gr_by_group <- stack(grl, "name") %>% group_by(name)

  target <- stack(disjoin(grl, ignore.strand = TRUE), "name")
  current <- disjoin_ranges(gr_by_group) %>%
    mutate(name = Rle(name))
  expect_identical(target, current)
})



test_that("matches HelloRanges multinter", {
  oldwd <- getwd()
  setwd(system.file("unitTests", "data", "multiinter", package="HelloRanges"))
  bed_files <- list.files(pattern = ".bed$")

  correct_gr <- GRanges("chr1",
                        IRanges(c(7, 9, 13, 16, 21, 23, 31, 33),
                                c(8, 12, 15, 20, 22, 30, 32, 34)),
                        i=IRanges::IntegerList(1, c(1,3), 1:3, 1:2, 2, 1:2, 2, 3))

  gr_l <- S4Vectors::List(lapply(bed_files, function(x) {
    mutate(read_bed(x), grp = sub(".bed$", "", basename(x)))
    }))

  gr_by_group_r <- unlist(gr_l) %>%
    mutate(grp = factor(grp, levels = c("a", "b", "c"))) %>%
    group_by(grp) %>%
    reduce_ranges()
  test_gr <- gr_by_group_r %>%
    mutate(i = factor(as.integer(grp))) %>%
    disjoin_ranges(i = IRanges::IntegerList(i))

  expect_identical(correct_gr, test_gr)

  # with names in place of integer
  mcols(correct_gr)$i <- as(
    extractList(factor(c("a", "b", "c")), mcols(correct_gr)$i),
    "SimpleFactorList"
  )
  test_gr <- gr_by_group_r %>%
    disjoin_ranges(i = IRanges::FactorList(grp, compress = FALSE))
  expect_identical(correct_gr, test_gr)

  setwd(oldwd)

})

select_rng <- function(.data, .drop_ranges, ...) {
  
  var_names <- tbl_vars(.data)
  names(var_names) <- var_names
  if (.drop_ranges) {
    pos <- tidyselect::eval_select(rlang::expr(c(...)), var_names)
    
    .env <- overscope_ranges(.data)
    
    ans <- lapply(syms(names(pos)), eval_tidy, data = .env)
    names(ans) <- names(pos)
    return(as(ans, "DataFrame"))
  } else {
    
    core <- S4Vectors::parallelVectorNames(.data)
    var_names <- var_names[!(var_names %in% core)]

    
    pos <- try(tidyselect::eval_select(rlang::expr(c(...)), var_names),
               silent = TRUE)
    
    if (is(pos, "try-error") || !length(var_names)) {
      invalid_cols <- paste(core, collapse = ", ")
      stop(paste0("Cannot select/rename the following columns: ",invalid_cols))
    }
  }
  
  mcols(.data) <- mcols(.data)[ , pos, drop = FALSE]
  names(mcols(.data)) <- names(pos)
  
  .data
}


#' Select metadata columns of the Ranges object by name or position
#'
#' @param .data a `Ranges` object
#' @param ... One or more metadata column names.
#' @param .drop_ranges If TRUE select will always return a tibble. In this
#' case, you may select columns that form the core part of the Ranges object.
#' @details Note that by default select only acts on the metadata columns (and
#' will therefore return a Ranges object) if a core component of a Ranges is dropped or selected
#' without the other required components (this includes the seqnames, strand, start, end,
#' width names), then select will throw an error unless .drop_ranges is set to TRUE.
#' @return a Ranges object or a tibble
#' @seealso [dplyr::select()]
#' @importFrom dplyr select
#' @importFrom tidyselect eval_select
#' @importFrom S4Vectors parallelVectorNames
#' @examples
#' df <- data.frame(start = 1:10, width = 5,  seqnames = "seq1",
#' strand = sample(c("+", "-", "*"), 10, replace = TRUE), gc = runif(10), counts = rpois(10, 2))
#' rng <- as_granges(df)
#' select(rng, -gc)
#' select(rng, gc)
#' select(rng, counts, gc)
#' select(rng, 2:1)
#' select(rng, seqnames, strand, .drop_ranges = TRUE)
#' @rdname ranges-select
#' @method select Ranges
#' @export
select.Ranges <- function(.data, ..., .drop_ranges = FALSE) {
  dots <- rlang::enquos(...)
  # no selection? - return .data
  if (length(dots) == 0) {
    return(.data)
  }

  select_rng(.data, .drop_ranges, ...)
}

#' @method select DelegatingGenomicRanges
#' @export
select.DelegatingGenomicRanges <- function(.data, ..., .drop_ranges = FALSE) {
  dots <- rlang::enquos(...)
  # no selection? - return .data
  if (length(dots) == 0) {
    return(.data)
  }

  delegate <- .data@delegate
  delegate <- select_rng(delegate, .drop_ranges,  dots)
  if (.drop_ranges) {
    return(delegate)
  } else {
    .data@delegate <- delegate
  }
  .data
}

#' @method select DelegatingIntegerRanges
#' @export
select.DelegatingIntegerRanges <- select.DelegatingGenomicRanges

#' Find nearest neighbours between two Ranges objects
#'
#' @param x,y Ranges objects, add the nearest neighbours of ranges in x to
#' those in y.
#' @param suffix A character vector of length two used to identify metadata columns
#' @param distance logical vector whether to add a column named "distance"
#'   containing the distance to the nearest region. If set to a character vector
#'   of length 1, will use that as distance column name.
#'
#' @details By default `join_nearest` will find arbitrary nearest
#' neighbours in either direction and ignore any strand information.
#' The `join_nearest_left` and `join_nearest_right`  methods
#' will find arbitrary nearest neighbour ranges on x that are left/right of
#' those on y and ignore any strand information.
#'
#' The `join_nearest_upstream` method will find arbitrary nearest
#' neighbour ranges on x that are upstream of those on y. This takes into
#' account strandedness of the ranges.
#' On the positive strand nearest upstream will be on the
#' left and on the negative strand nearest upstream will be on the right.
#'
#' The `join_nearest_downstream` method will find arbitrary nearest
#' neighbour ranges on x that are upstream of those on y. This takes into
#' account strandedness of the ranges.On the positive strand nearest downstream
#' will be on the right and on the negative strand nearest upstream will be on
#' the left.
#'
#' @return A Ranges object corresponding to the nearest ranges, all metadata
#' is copied over from the right-hand side ranges `y`.
#'
#' @examples
#' query <- data.frame(start = c(5,10, 15,20),
#'                    width = 5,
#'                    gc = runif(4)) %>%
#'              as_iranges()
#' subject <- data.frame(start = c(2:6, 24),
#'                       width = 3:8,
#'                       label = letters[1:6]) %>%
#'              as_iranges()
#'
#' join_nearest(query, subject)
#' join_nearest_left(query, subject)
#' join_nearest_right(query, subject)
#'
#' subject  <- data.frame(seqnames = "chr1",
#'                start = c(11,101),
#'                end = c(21, 200),
#'                name = c("a1", "a2"),
#'                strand = c("+", "-"),
#'                score = c(1,2)) %>%
#'            as_granges()
#' query <- data.frame(seqnames = "chr1",
#'                       strand = c("+", "-", "+", "-"),
#'                       start = c(21,91,101,201),
#'                       end = c(30,101,110,210),
#'                       name = paste0("b", 1:4),
#'                       score = 1:4) %>%
#'                    as_granges()
#' join_nearest_upstream(query, subject)
#' join_nearest_downstream(query, subject)
#' @rdname ranges-nearest
#' @importFrom IRanges nearest
#' @export
join_nearest <- function(x,y, suffix = c(".x", ".y"), distance = FALSE) {
  hits <- hits_nearest(x, y)
  distance_col <- set_distance_col(distance)
  expand_by_hits(x, y, suffix, hits, hits_mcols_to_keep = distance_col)
}

#' @rdname ranges-nearest
#' @export
join_nearest_left <- function(x,y, suffix = c(".x", ".y"), distance = FALSE) {
  hits <- hits_nearest_left(x, y)
  distance_col <- set_distance_col(distance)
  expand_by_hits(x, y, suffix, hits, hits_mcols_to_keep = distance_col)
}

#' @rdname ranges-nearest
#' @export
join_nearest_right <- function(x, y, suffix = c(".x", ".y"), distance = FALSE) {
  hits <- hits_nearest_right(x, y)
  distance_col <- set_distance_col(distance)
  expand_by_hits(x, y, suffix, hits, hits_mcols_to_keep = distance_col)
}

#' @rdname ranges-nearest
#' @export
join_nearest_upstream <- function(x, y,  suffix = c(".x", ".y"), distance = FALSE) { UseMethod("join_nearest_upstream")}

#' @export
join_nearest_upstream.GenomicRanges <- function(x, y,  suffix = c(".x", ".y"), distance = FALSE) {
  hits <- hits_nearest_upstream(x, y)
  distance_col <- set_distance_col(distance)
  expand_by_hits(x, y, suffix, hits, hits_mcols_to_keep = distance_col)
}

#' @rdname ranges-nearest
#' @importFrom IRanges nearest
#' @export
join_nearest_downstream <- function(x, y,  suffix = c(".x", ".y"), distance = FALSE) { UseMethod("join_nearest_downstream")}

#' @export
join_nearest_downstream.GenomicRanges <- function(x, y, suffix = c(".x", ".y"), distance = FALSE) {
  hits <- hits_nearest_downstream(x, y)
  distance_col <- set_distance_col(distance)
  expand_by_hits(x, y, suffix, hits, hits_mcols_to_keep = distance_col)
}

# -----------------------
# Directed hits helpers

hits_nearest <- function(x, y){ 
  UseMethod("hits_nearest")
}
hits_nearest_left <- function(x, y){ 
  UseMethod("hits_nearest_left")
}
hits_nearest_right <- function(x, y){ 
  UseMethod("hits_nearest_right")
}
hits_nearest_upstream <- function(x, y){ 
  UseMethod("hits_nearest_upstream")
}
hits_nearest_downstream <- function(x, y){ 
  UseMethod("hits_nearest_downstream")
}

#' @importFrom IRanges distanceToNearest
hits_nearest.IRanges <- function(x, y){
  make_hits(x, y, distanceToNearest, select = "arbitrary")
}

#' @importFrom GenomicRanges distanceToNearest
hits_nearest.GenomicRanges <- function(x, y){
  make_hits(x, y, distanceToNearest, select = "arbitrary", ignore.strand = TRUE)
}

#' @importFrom IRanges distanceToNearest
hits_nearest_left.IRanges <- function(x, y){
  hits <- make_hits(x, y, distanceToNearest, select = "all")
  get_hits_left(x, y, hits)
}

#' @importFrom GenomicRanges distanceToNearest
hits_nearest_left.GenomicRanges <- function(x, y){
  hits <- make_hits(x, y, distanceToNearest, select = "all", ignore.strand = TRUE)
  get_hits_left(x,y, hits)
}

#' @importFrom IRanges distanceToNearest
hits_nearest_right.IRanges <- function(x, y){
  hits <- make_hits(x, y, distanceToNearest, select = "all")
  get_hits_right(x, y, hits)
}

#' @importFrom GenomicRanges distanceToNearest
hits_nearest_right.GenomicRanges <- function(x, y){
  hits <- make_hits(x, y, distanceToNearest, select = "all", ignore.strand = TRUE)
  get_hits_right(x, y, hits)
}

#' @importFrom GenomicRanges distanceToNearest
hits_nearest_upstream.GenomicRanges <- function(x, y){
  hits <- distanceToNearest(x,y, select = "all", ignore.strand = FALSE)
  get_hits_upstream(x, y, hits)
}

#' @importFrom GenomicRanges distanceToNearest
hits_nearest_downstream.GenomicRanges <- function(x, y){
  hits <- distanceToNearest(x,y, select = "all", ignore.strand = FALSE)
  get_hits_downstream(x, y, hits)
}


# -----------------------
# Hits logic helpers

get_hits_left <- function(x, y, hits){
  mcols(hits)$is_left <- start(y[subjectHits(hits)]) <= start(x[queryHits(hits)]) &
    end(y[subjectHits(hits)]) <= start(x[queryHits(hits)])
  hits[mcols(hits)$is_left]
}

get_hits_right <- function(x, y, hits){
  mcols(hits)$is_right <- end(x[queryHits(hits)]) <= start(y[subjectHits(hits)])
  hits[mcols(hits)$is_right]
}

get_hits_upstream <- function(x, y, hits){
  mcols(hits)$is_right <- end(x[queryHits(hits)]) <= start(y[subjectHits(hits)])
  mcols(hits)$is_left <- start(x[queryHits(hits)]) >= start(y[subjectHits(hits)])
  mcols(hits)$direction <- strand(x[queryHits(hits)])
  mcols(hits)$is_upstream <- ifelse(mcols(hits)$direction == "+",
                                    mcols(hits)$is_left,
                                    mcols(hits)$is_right)
  hits[mcols(hits)$is_upstream]
}

get_hits_downstream <- function(x, y, hits){
  mcols(hits)$is_right <- end(x[queryHits(hits)]) <= start(y[subjectHits(hits)])
  mcols(hits)$is_left <- start(x[queryHits(hits)]) >= start(y[subjectHits(hits)])
  mcols(hits)$direction <- strand(x[queryHits(hits)])
  # on positive strand nearest downtream will be on the right
  # on negative strand nearest downstream will be on the left
  mcols(hits)$is_downstream <- ifelse(mcols(hits)$direction == "-",
                                    mcols(hits)$is_left,
                                    mcols(hits)$is_right)

  hits[mcols(hits)$is_downstream]
}

set_distance_col <- function(distance){
  if (distance == TRUE){
    keep_distance <- "distance"
  } else if (is.character(distance) & length(distance) == 1) {
    keep_distance <- c("distance")
    names(keep_distance) <- distance
  } else if (distance == FALSE) {
    keep_distance = NULL
  } else {
    stop("Invalid argument passed to distance")
  }
  
  return(keep_distance)
}

# test-overlaps-distance.R
context("overlap joins with distance")

test_that("adding distance works as expected",{

  a <- GRanges(
    seqnames = "chr1",
    ranges = IRanges(c(11, 66, 101), c(15, 70, 110)),
    strand = c("+", "-", "+"),
    a_name = paste0("a", 1:3),
    a_score = 1:3
  )

  b <- GRanges(
    seqnames = "chr1",
    strand = c("-", "+", "-", "+"),
    ranges = IRanges(c(21, 31, 41, 51), c(30, 40, 50, 60)),
    b_name = paste0("b", 1:4),
    b_score = 3:6
  )

  join_overlap_left(a, b, maxgap=5)
  join_overlap_left_directed(a, b, maxgap=15)

  join_nearest(a, b, distance=TRUE)
  join_nearest_downstream(a, b, distance=TRUE)

  join_overlap_left(a, b, maxgap=15, distance=TRUE)
  join_overlap_left_directed(a, b, maxgap=15, distance=TRUE)

})

compare_dist_vals <- function(add, join){
  # Checks that joins and adds produce the same values (minus NA's)
  add_dist <- mcols(join)$distance
  add_dist <- add_dist[!is.na(add_dist)]
  join_dist <- mcols(join)$distance
  
  expect_true(identical(add_dist, join_dist))
}

test_that("add_nearest_distance works on GRanges", {
  subject  <- data.frame(seqnames = "chr1",
               start = c(11,101),
               end = c(21, 200),
               name = c("a1", "a2"),
               strand = c("+", "-"),
               score = c(1,2)) %>%
           as_granges()
  query <- data.frame(seqnames = "chr1",
                        strand = c("+", "-", "+", "-"),
                        start = c(21,91,101,201),
                        end = c(30,101,110,210),
                        name = paste0("b", 1:4),
                        score = 1:4) %>%
                     as_granges()
  
  near <- add_nearest_distance(query, subject)
  near_left <- add_nearest_distance_left(query, subject)
  near_right <- add_nearest_distance_right(query, subject)
  near_up <- add_nearest_distance_upstream(query, subject)
  near_down <- add_nearest_distance_downstream(query, subject)
  
  j_near <- join_nearest(query, subject, distance = TRUE)
  j_near_left <- join_nearest_left(query, subject, distance = TRUE)
  j_near_right <- join_nearest_right(query, subject, distance = TRUE)
  j_near_up <- join_nearest_upstream(query, subject, distance = TRUE)
  j_near_down <- join_nearest_downstream(query, subject, distance = TRUE)
  
  
  compare_dist_vals(near, j_near)
  compare_dist_vals(near_left, j_near_left)
  compare_dist_vals(near_right, j_near_right)
  compare_dist_vals(near_up, j_near_up)
  compare_dist_vals(near_down, j_near_down)
  
})

test_that("add_nearest_distance works on IRanges", {
  query_ir <- IRanges(c(6, 11, 1, 13, 18), width=c(2, 2, 2, 2, 2))
  subject_ir <- IRanges(c(1, 2, 9, 15, 15), width=c(4, 3, 2, 2, 3))
  
  near <- add_nearest_distance(query_ir, subject_ir)
  near_left <- add_nearest_distance_left(query_ir, subject_ir)
  near_right <- add_nearest_distance_right(query_ir, subject_ir)
  
  j_near <- join_nearest(query_ir, subject_ir, distance = TRUE)
  j_near_left <- join_nearest_left(query_ir, subject_ir, distance = TRUE)
  j_near_right <- join_nearest_right(query_ir, subject_ir, distance = TRUE)
  
  compare_dist_vals(near, j_near)
  compare_dist_vals(near_left, j_near_left)
  compare_dist_vals(near_right, j_near_right)
 
  # no up/down method for IRanges 
  expect_error(add_nearest_distance_upstream(query_ir, subject_ir), "no applicable method")
  expect_error(add_nearest_distance_downstream(query_ir, subject_ir), "no applicable method")
  
})

# test-nearest.R
context("nearest/follows/precedes")

test_that("nearest/follows/precedes matches IRanges methods", {

  ##      ..
  ##           ..
  ## ..
  ##             ..
  ##                  ..
  ## xxxx
  ##  xxx
  ##         xx
  ##               xx
  ##               xxx

  query <- IRanges(c(6, 11, 1, 13, 18), width=c(2, 2, 2, 2, 2))
  mcols(query) <- DataFrame(query = paste0("a", 1:5))
  subject <- IRanges(c(1, 2, 9, 15, 15), width=c(4, 3, 2, 2, 3))
  mcols(subject) <- DataFrame(subject = paste0("b", 1:5))

  result <- join_nearest(query, subject)
  expect_identical(mcols(result)$subject,
                  mcols(subject)[nearest(query, subject, select = "arbitrary"),])

  # left/right joins select all nearest hits
  left <- join_nearest_left(query, subject)
  expect_identical(mcols(left)$subject,
                   mcols(subject)$subject[c(1,2,3,5)])
  right <- join_nearest_right(query, subject)
  expect_identical(mcols(right)$subject, mcols(subject)$subject[c(3,2,4,5)])

  result_pairs <- pair_nearest(query, subject)
  expect_s4_class(result_pairs, "DataFrame")
  expect_identical(result_pairs$subject,
                   mcols(subject)[nearest(query, subject, select = "arbitrary"),])

  expect_identical(result_pairs$ranges.y,
                   ranges(subject)[nearest(query, subject, select = "arbitrary"),])


  # follows --
  fol1 <- join_follow(query, subject)
  indx <- follow(query, subject)
  indx <- indx[!is.na(indx)]
  expect_identical(mcols(fol1)$subject, mcols(subject)$subject[indx])
  # all ranges that follow on the left
  fol2 <- join_follow_left(query, subject)
  expect_identical(mcols(fol2)$subject, mcols(subject)$subject[c(1,2,3,3,5)])

  result_pairs <- pair_follow(query, subject)
  expect_identical(result_pairs$subject, mcols(subject)$subject[indx])
  expect_identical(result_pairs$ranges.y,
                   ranges(subject)[indx])
  # precedes --
  pre1 <- join_precede(query, subject)
  indx <- precede(query, subject)
  qindx <- !is.na(indx)
  sindx <- indx[qindx]
  expect_identical(mcols(pre1)$query, mcols(query)$query[qindx])
  expect_identical(mcols(pre1)$subject, mcols(subject)$subject[sindx])

  pre2 <- join_precede_right(query, subject)
  expect_identical(mcols(pre2)$query, mcols(query)$query[c(1,2,2,3,4,4)])

  result_pairs <- pair_precede(query, subject)
  expect_identical(result_pairs$subject, mcols(subject)$subject[sindx])
  expect_identical(result_pairs$ranges.y,
                   ranges(subject)[sindx])
  expect_identical(result_pairs$ranges.x,
                   ranges(query)[qindx])
})

test_that("nearest/follows/precedes matches GenomicRanges", {
  query_ir <- IRanges(c(6, 11, 1, 13, 18), width=c(2, 2, 2, 2, 2))
  subject_ir <- IRanges(c(1, 2, 9, 15, 15), width=c(4, 3, 2, 2, 3))

  #+#      ..
  #*#           ..
  #-# ..
  #+#             ..
  #-#                  ..
  #-# xxxx
  #+#  xxx
  #+#         xx
  #-#               xx
  #+#               xxx
  query_gr <- GRanges(seqnames = "chr1", ranges = query_ir,
                      strand = c("+", "*", "-", "+", "-"),
                      query = paste0("a", 1:5))
  subject_gr <- GRanges(seqnames = "chr1", ranges = subject_ir,
                        strand = c("-", "+", "+", "-", "-"),
                        subject = paste0("b", 1:5))
  # upstream/downstream
  up <- join_nearest_upstream(query_gr, subject_gr)
  expect_identical(mcols(up)$subject, subject_gr$subject[3])
  expect_identical(mcols(up)$query, query_gr$query[4])

  down <- join_nearest_downstream(query_gr, subject_gr)
  expect_identical(mcols(down)$subject, subject_gr$subject[c(3,1,5)])
  expect_identical(mcols(down)$query, query_gr$query[c(1,3,5)])
  # keeps no strand calls, plus gives all follows
  follow_up <- join_follow_upstream(query_gr, subject_gr)
  expect_identical(mcols(follow_up)$subject, subject_gr$subject[c(2,3,4,5,3)])

  precede_down <- join_precede_downstream(query_gr, subject_gr)
  expect_identical(mcols(precede_down)$subject, subject_gr$subject[c(3,1,5)])
  # nearest/follows/precedes ignores strandedness so should be identical to
  # ranges analogues
  nearest_gr <- join_nearest(query_gr, subject_gr)
  nearest_ir <- join_nearest(query_ir, subject_ir)
  expect_identical(ranges(nearest_ir), ranges(nearest_gr))

  follows_gr <- join_follow(query_gr, subject_gr)
  follows_ir <- join_follow(query_ir, subject_ir)
  expect_identical(ranges(follows_ir), ranges(follows_gr))

  precedes_gr <- join_precede(query_gr, subject_gr)
  precedes_ir <- join_precede(query_ir, subject_ir)
  expect_identical(ranges(precedes_ir), ranges(precedes_gr))

})

# filter.R
filter_rng <- function(.data, dots) {
  overscope <- overscope_ranges(.data)
  r <- overscope_eval_update(overscope, dots, bind_envir = FALSE)
  r <- Reduce(`&`, r)
  if (!(is.logical(r) || is(r, "LogicalList"))) {
    if (!(is(r, "Rle") || is(r, "RleList"))) {
      stop("Argument to filter condition must evaluate to a logical vector or logical-Rle")
    }
  }
  # drop missing values in r
  r & !is.na(r)
}

filter_grp <- function(.data, ...) {
    dots <- set_dots_unnamed(...)
    ii <- filter_rng(.data, dots)
    inx <- S4Vectors::split(
      seq_along(.data@delegate),
      .data@group_indices
    )
    inx_update <- inx[ii]
    rng <- .data@delegate[sort(unlist(inx_update))]
    group_by(rng, !!!groups(.data))
}

#' Subset a `Ranges` object
#'
#' @param .data A `Ranges` object
#' @param ...  valid logical predictates to subset .data by. These
#' are determined by variables in `.data`. If more than
#' one condition is supplied, the conditions are combined with `&`. Only
#' rows where the condition evaluates to `TRUE` are kept.
#' @param .preserve when FALSE (the default) grouping structure is recalculated, TRUE is currently not implemented.
#'
#'
#' @details  For any Ranges objects
#' `filter` can act on all core components of the class including start, end,
#' width (for IRanges) or seqnames and strand (for GRanges) in addition to
#' metadata columns. If the Ranges object is grouped, `filter` will act
#' seperately on each  parition of the data.
#'
#' @return a Ranges object
#'
#' @importFrom dplyr filter
#' @importFrom IRanges as.env
#' @importFrom S4Vectors runValue endoapply
#' @seealso [dplyr::filter()]
#' @name filter-ranges
#' @rdname filter-ranges
#'
#' @examples
#' set.seed(100)
#' df <- data.frame(start = 1:10,
#'                  width = 5,
#'                  seqnames = "seq1",
#'                  strand = sample(c("+", "-", "*"), 10, replace = TRUE),
#'                  gc = runif(10))
#'
#' rng <- as_granges(df)
#'
#' filter(rng, strand == "+")
#' filter(rng, gc > 0.5)
#'
#' # multiple criteria
#' filter(rng, strand == "+" | start > 5)
#' filter(rng, strand == "+" & start > 5)
#'
#' # multiple conditions are the same as and
#' filter(rng, strand == "+", start > 5)
#'
#' # grouping acts on each subset of the data
#' rng %>%
#'   group_by(strand) %>%
#'   filter(gc > 0.5)
#'
#' @method filter Ranges
#' @export
filter.Ranges <- function(.data, ..., .preserve = FALSE) {
  dots <- set_dots_unnamed(...)
  .data[filter_rng(.data, dots)]
}

#' @method filter DelegatingGenomicRanges
#' @export
filter.DelegatingGenomicRanges <- function(.data, ..., .preserve = FALSE) {
  dots <- set_dots_unnamed(...)
  delegate <- .data@delegate
  .data@delegate <- delegate[filter_rng(delegate, dots)]
  return(.data)
}

#' @method filter DelegatingIntegerRanges
#' @export
filter.DelegatingIntegerRanges <- filter.DelegatingGenomicRanges

#' @method filter GroupedGenomicRanges
#' @export
filter.GroupedGenomicRanges <- function(.data, ..., .preserve = FALSE) {
  filter_grp(.data, ...)
}

#' @method filter GroupedIntegerRanges
#' @export
filter.GroupedIntegerRanges <- filter.GroupedGenomicRanges

norm_args_reader_gff <- function(genome_info) {
  if (!is.null(genome_info)) {

    if (is(genome_info, "GRanges")) {
      return(seqinfo(genome_info))
    }
    return(genome_info)
  }
  NA
}
#' Read a GFF/GTF/GVT file
#'
#' @param file A path to a file or a connection.
#' @param col_names An optional character vector for parsing specific
#' columns in `file` that are part of the GFF specification. These should
#' name either fixed fields, like source or type, or, for GFF2 and GFF3,
#' any attribute.
#' @param genome_info An optional character string or a Ranges object
#' that contains information about the genome build. For example the UCSC identifier
#'"hg19" will add build information to the returned GRanges.
#' @param overlap_ranges An optional Ranges object. Only the intervals in the file
#' that overlap the Ranges will be returned.
#'
#' @description This is a lightweight wrapper to the import family
#' of functions defined in \pkg{rtracklayer}.
#'
#' @return A GRanges object
#'
#' @importFrom rtracklayer import.gff
#' @importFrom Seqinfo seqinfo
#' @seealso \code{rtracklayer::\link[rtracklayer:GFFFile-class]{GFFFile()}}
#' @export
#' @rdname io-gff-read
#' @examples
#' test_path <- system.file("tests", package = "rtracklayer")
#' # gff3
#' test_gff3 <- file.path(test_path, "genes.gff3")
#' gr <- read_gff3(test_gff3)
#' gr
#' # alternatively with read_gff
#' gr <- read_gff(test_gff3, genome_info = "hg19")
#' gr
#' @return a GRanges object
#' @importFrom rtracklayer import.gff
read_gff <- function(file, col_names = NULL, genome_info = NULL, overlap_ranges = NULL) {
  import.gff(file,
             colnames = col_names,
             genome = norm_args_reader_gff(genome_info),
             which = overlap_ranges)
}

#' @export
#' @rdname io-gff-read
#' @importFrom rtracklayer import.gff1
read_gff1 <- function(file, col_names = NULL, genome_info = NULL, overlap_ranges = NULL) {
  import.gff1(file,
              colnames = col_names,
              genome = norm_args_reader_gff(genome_info),
              which = overlap_ranges)
}

#' @export
#' @rdname io-gff-read
#' @importFrom rtracklayer import.gff2
read_gff2 <- function(file, col_names = NULL, genome_info = NULL, overlap_ranges = NULL) {
  import.gff2(file,
              colnames = col_names,
              genome = norm_args_reader_gff(genome_info),
              which = overlap_ranges)
}

#' @export
#' @rdname io-gff-read
#' @importFrom rtracklayer import.gff3
read_gff3 <- function(file, col_names = NULL, genome_info = NULL, overlap_ranges = NULL) {
  import.gff3(file,
              colnames = col_names,
              genome = norm_args_reader_gff(genome_info),
              which = overlap_ranges)
}



#' Write a GFF(123) file
#'
#' @param x A GRanges object
#' @param file Path or connection to write to
#' @param index If TRUE the output file will be compressed and indexed using bgzf and
#' tabix.
#'
#' @description This is a lightweight wrapper to the export family
#' of functions defined in \pkg{rtracklayer}.
#'
#' @importFrom rtracklayer export.gff export.gff1 export.gff2 export.gff3
#' @seealso \code{rtracklayer::\link[rtracklayer:GFFFile-class]{GFFFile()}}
#' @return The write function returns a GFFFile object invisibly
#' @export
#' @examples 
#' \dontrun{
#'  test_path <- system.file("tests", package = "rtracklayer")
#'  test_gff3 <- file.path(test_path, "genes.gff3")
#'  gr <- read_gff3(test_gff3)
#'  out_gff3 <- file.path(tempdir(), "test.gff3")
#'  write_gff3(gr, out_gff3)
#'  read_gff3(out_gff3)
#' }
#' @rdname io-gff-write
write_gff <- function(x, file, index = FALSE) {
  export.gff(x, file, index = index)
}

#' @export
#' @rdname io-gff-write
write_gff1 <- function(x, file, index = FALSE) {
  export.gff1(x, file, index = index)
}

#' @export
#' @rdname io-gff-write
write_gff2 <- function(x, file, index = FALSE) {
  export.gff2(x, file, index = index)
}

#' @export
#' @rdname io-gff-write
write_gff3 <- function(x, file, index = FALSE) {
  export.gff3(x, file, index = index)
}

context("test GFF files")

createCorrectGFF <- function(version) {
  space <- c(rep("chr10", 15), rep("chr12", 16))
  start <- c(rep(92828, 4), 92997, rep(94555, 4), rep(94744, 3), rep(95122, 2),
             95348, rep(87984, 4),
             rep(c(88257, 88570, 88860, 89675, 90587, 90796), each = 2))
  end <- c(95504, 95178, 95504, rep(94054, 2), rep(94665, 2), 94615, 94665,
           rep(94852, 3), rep(95178, 2), 95504,
           rep(c(91263, 88017, 88392, 88771, 89018, 89827, 90655, 91263),
               each = 2))
  type <- c("gene", "mRNA", "mRNA", "exon", "CDS", "exon", "exon",
            "CDS", "CDS", "exon", "exon", "CDS", "exon", "CDS", "exon",
            "gene", "mRNA", "exon", "CDS", "exon", "CDS", "exon", "CDS",
            "exon", "CDS", "exon", "CDS", "exon", "CDS", "exon", "CDS")
  type <- factor(type, unique(type))
  source <- factor("rtracklayer")
  phase <- NA_integer_
  score <- c(5, rep(NA, length(type) - 1L))
  strand <- strand(c(rep("-", 14), "*", rep("+", 15), "*"))
  Alias <- CharacterList(c(list(c("FLJ40100", "TUBB8")),
                           rep(list(character()), 14), "LOC100288778",
                           rep(list(character()), 15)))
  ID <- c("GeneID:347688", "873", "872", rep(NA, 12), "GeneID:100288778",
          "4644", rep(NA, 14))
  Name <- c(rep("TUBB8", 3), rep(NA, 12), rep("LOC100288778", 2),
            rep(NA, 14))
  Parent <- CharacterList(c(list(character()), rep("GeneID:347688", 2),
                            rep(list(c("872", "873")), 2),
                            rep(c("872", "873"), 3), rep("873", 3), "872",
                            list(character()), "GeneID:100288778",
                            rep("4644", 14)))
  geneName <- c("tubulin, beta 8", rep(NA, 14),
                "WAS protein family homolog 1; pseudogene", rep(NA, 15))
  genome <- c("hg19", rep(NA, length(geneName) - 1))

  correct_gff3 <- GRanges(space, IRanges(start, end), strand,
                          source, type, score, phase,
                          ID, Name, geneName, Alias, genome, Parent)
  seqinfo(correct_gff3) <- Seqinfo(c("chr10", "chr12"))

  if (version == 3) {
    return(correct_gff3)
  } else if (version == 2) {
    correct_gff2 <- correct_gff3
    toCSV <- function(x) {
      csv <- sapply(x, paste, collapse = ",")
      csv[nchar(csv) == 0] <- NA
      csv
    }
    correct_gff2$Alias <- toCSV(correct_gff2$Alias)
    correct_gff2$Parent <- toCSV(correct_gff2$Parent)
    return(correct_gff2)
  } else {
    correct_gff1 <- correct_gff3[,c("source", "type", "score", "phase")]
    correct_gff1$group <- as.factor(seqnames(correct_gff3))
    return(correct_gff1)
  }
}

test_path <- system.file("tests", package = "rtracklayer")

test_that("reading GFF files returns correct GRanges", {

  correct_gff3 <- createCorrectGFF(3)

  # basic import
  test_gff3 <- file.path(test_path, "genes.gff3")
  test_gr <- read_gff3(test_gff3)
  expect_identical(correct_gff3, test_gr)

  # as GFF(3)File
  test_gff_file <- rtracklayer::GFF3File(test_gff3)
  test_gr <- read_gff3(test_gff_file)
  expect_identical(correct_gff3, test_gr)
  test_gff_file <- rtracklayer::GFFFile(test_gff3, version = "3")
  test_gr <- read_gff3(test_gff_file)
  expect_identical(correct_gff3, test_gr)
  test_gff_file <- rtracklayer::GFF2File(test_gff3)
  expect_error(read_gff3(test_gff_file))
  expect_warning(read_gff2(test_gff3))


  # genome with character arg
  si_hg19 <- rtracklayer::SeqinfoForBSGenome("hg19")
  correct_hg19 <- correct_gff3
  seqlevels(correct_hg19) <- seqlevels(si_hg19)
  seqinfo(correct_hg19) <- si_hg19

  test_gr <- read_gff(test_gff3, genome_info = "hg19")
  expect_identical(test_gr, correct_hg19)

  # genome as genome_info
  # si_hg19_gr <- get_genome_info(si_hg19)
  # test_gr <- read_gff(test_gff3, genome_info = si_hg19_gr)
  # expect_identical(test_gr, correct_hg19)

  # overlaps
  which <- GRanges("chr10:90000-93000")
  which_target <- filter_by_overlaps(correct_gff3, which)
  test_gr <- read_gff3(test_gff3, overlap_ranges = which)
  expect_identical(which_target, test_gr)

  # colnames := "geneName"
  test_gr <- read_gff(test_gff3, col_names = "geneName")
  target <- select(correct_gff3, geneName)
  expect_identical(target, test_gr)

  # exporting gff3
  test_gff3_out <- file.path(tempdir(), "genes.gff3")
  on.exit(unlink(test_gff3_out))
  correct_genome_hg19 <- correct_gff3
  genome(correct_genome_hg19) <- "hg19"
  write_gff3(correct_genome_hg19, test_gff3_out)
  test_gr <- read_gff3(test_gff3_out)
  expect_identical(test_gr, correct_hg19)


  # indexing a gff file
  write_gff3(correct_gff3, test_gff3_out, index = TRUE)
  test_gff_bgz <- paste(test_gff3_out, ".bgz", sep = "")
  on.exit(unlink(test_gff_bgz))
  on.exit(unlink(paste(test_gff_bgz, ".tbi", sep = "")))
  test_gr <- read_gff3(test_gff_bgz, overlap_ranges = which)
  expect_identical(which_target, test_gr)

})

test_that("writing/reading other GFF files", {
  correct_gff3 <- createCorrectGFF(3)
  # 'gff' extension
  test_gff_out <- file.path(tempdir(), "genes.gff")
  on.exit(unlink(test_gff_out))

  write_gff1(correct_gff3, test_gff_out)
  test_gr <- read_gff1(test_gff_out)
  correct_gff1 <- createCorrectGFF(1)
  expect_identical(test_gr, correct_gff1)

  write_gff2(correct_gff3, test_gff_out)
  test_gr <- read_gff2(test_gff_out)
  correct_gff2 <- createCorrectGFF(2)
  expect_identical(test_gr, correct_gff2)

  write_gff3(correct_gff3, test_gff_out)
  test_gr <- read_gff3(test_gff_out)
  expect_identical(test_gr, correct_gff3)

  # 'gff2' extension
  test_gff2_out <- file.path(tempdir(), "genes.gff2")
  write_gff2(correct_gff3, test_gff2_out)
  test_gr <- read_gff2(test_gff2_out)
  expect_identical(test_gr, correct_gff2)

  #  'gff1' extension
  test_gff1_out <- file.path(tempdir(), "genes.gff1")
  write_gff1(correct_gff3, test_gff1_out)
  test_gr <- read_gff1(test_gff1_out)
  expect_identical(test_gr, correct_gff1)
})

#' Filter by overlapping/non-overlapping ranges
#'
#' @param x,y Objects representing ranges
#' @param maxgap The maximimum gap between intervals as a single
#' integer greater than or equal to -1. If you modify this argument,
#' `minoverlap` must be held fixed.
#' @param minoverlap  The minimum amount of overlap between intervals
#' as a single integer greater than 0. If you modify this argument,
#' `maxgap` must be held fixed.
#'
#' @details By default, `filter_by_overlaps` and
#' `filter_by_non_overlaps` ignore strandedness for [GenomicRanges::GRanges()]
#' objects. To perform stranded operations use `filter_by_overlaps_directed` and `filter_by_non_overlaps_directed`. The argument `maxgap` is the maximum number of positions
#' between two ranges for them to be considered overlapping. Here the default
#' is set to be -1 as that is the the gap between two ranges that
#' has its start or end strictly inside the other. The argugment
#' `minoverlap` refers to the minimum number of positions
#' overlapping between ranges, to consider there to be overlap.
#'
#' @return  a Ranges object
#' @importFrom IRanges subsetByOverlaps
#' @seealso \code{IRanges::\link[IRanges:findOverlaps-methods]{subsetByOverlaps()}}
#' @export
#' @examples
#' df <- data.frame(seqnames = c("chr1", rep("chr2", 2),
#'                               rep("chr3", 3), rep("chr4", 4)),
#'                  start = 1:10,
#'                  width = 10:1,
#'                  strand = c("-", "+", "+", "*", "*", "+", "+", "+", "-", "-"),
#'                  name = letters[1:10])
#' query <- as_granges(df)
#'
#' df2 <- data.frame(seqnames = c(rep("chr2", 2), rep("chr1", 3), "chr2"),
#'                   start = c(4,3,7,13,1,4),
#'                   width = c(6,6,3,3,3,9),
#'                   strand = c(rep("+", 3), rep("-", 3)))
#' subject <- as_granges(df2)
#'
#' filter_by_overlaps(query, subject)
#'
#' filter_by_overlaps_directed(query, subject)
#'
#' filter_by_non_overlaps(query, subject)
#'
#' filter_by_non_overlaps_directed(query, subject)
#'
#' @rdname ranges-filter-overlaps
filter_by_overlaps <- function(x,y, maxgap = -1L, minoverlap = 0L) {
  UseMethod("filter_by_overlaps")
}

#' @export
filter_by_overlaps.IntegerRanges <- function(x, y, maxgap = -1L, minoverlap = 0L) {
  subsetByOverlaps(x,y, maxgap, minoverlap)
}

#' @export
filter_by_overlaps.GenomicRanges <- function(x, y, maxgap = -1L, minoverlap =0L) {
  subsetByOverlaps(x,y, maxgap, minoverlap, ignore.strand = TRUE)
}

#' @export
#' @rdname ranges-filter-overlaps
filter_by_non_overlaps <- function(x,y, maxgap, minoverlap) {
  UseMethod("filter_by_non_overlaps")
}

#' @export
filter_by_non_overlaps.IntegerRanges <- function(x,y, maxgap = -1L, minoverlap = 0L) {
  subsetByOverlaps(x,y, maxgap, minoverlap, invert = TRUE)
}

#' @export
filter_by_non_overlaps.GenomicRanges <- function(x,y, maxgap = -1L, minoverlap = 0L) {
  subsetByOverlaps(x,y, maxgap, minoverlap,
                   invert = TRUE,
                   ignore.strand = TRUE)
}

#' @export
#' @rdname ranges-filter-overlaps
filter_by_overlaps_directed <- function(x,y, maxgap = -1L, minoverlap = 0L) {
  UseMethod("filter_by_overlaps_directed")
}

#' @export
filter_by_overlaps_directed.GenomicRanges <- function(x,y, maxgap = -1L, minoverlap = 0L) {
  subsetByOverlaps(x,y, maxgap, minoverlap, ignore.strand = FALSE)
}

#' @export
#' @rdname ranges-filter-overlaps
filter_by_non_overlaps_directed <- function(x,y, maxgap, minoverlap) {
  UseMethod("filter_by_non_overlaps_directed")
}

#' @export
filter_by_non_overlaps_directed.GenomicRanges <- function(x,y, maxgap = -1L, minoverlap = 0L) {
  subsetByOverlaps(x,y, maxgap, minoverlap,
                   invert = TRUE,
                   ignore.strand = FALSE)
}

# TODO -- add in more variants here?

# test-filter-by-overlap.R

test_that("filters by overlap & non-overlap work", {

  subject_gr <- as_granges(data.frame(seqnames = rep("chr1", 3),
                               start = c(1,1, 100),
                               width = rep(100, 3),
                               strand = c("+", "-", "+")))

  query_gr <- as_granges(data.frame(seqnames = c("chr1", "chr2", "chr1"),
                               start = c(20, 20, 150),
                               width = c(10, 10, 10),
                               strand = c("+", "-", "+")))
  
  filter_directed_expected_gr <- as_granges(data.frame(seqnames = rep("chr1", 2),
                               start = c(1, 100),
                               width = rep(100, 2),
                               strand = c("+", "+")))
  
  filter_non_overlap_expected_gr <- GRanges(seqinfo = Seqinfo(seqnames = "chr1"))
  
  filter_non_overlap_directed_expected_gr <- as_granges(
    data.frame(seqnames = "chr1",
                               start = c(1),
                               width = rep(100),
                               strand = c("-")
               )
    )
  
  expect_identical(subject_gr, filter_by_overlaps(subject_gr, query_gr))
  expect_identical(filter_directed_expected_gr, filter_by_overlaps_directed(subject_gr, query_gr))
  
  expect_identical(filter_non_overlap_expected_gr, filter_by_non_overlaps(subject_gr, query_gr))
  expect_identical(filter_non_overlap_directed_expected_gr, filter_by_non_overlaps_directed(subject_gr, query_gr))

})

#' Construct a I/GRanges object from a tibble or data.frame
#'
#' @param .data a [data.frame()] or [dplyr::tibble()] to
#' construct a Ranges object from
#' @param ... optional named arguments specifying which the columns in .data
#' containin the core components a Ranges object.
#' @param keep_mcols place the remaining columns into the metadata columns slot
#' (default=TRUE)
#'
#' @description The as_i(g)ranges function looks for column names in .data called start,
#' end, width, seqnames and strand in order to construct an IRanges or GRanges
#' object. By default other columns in .data are placed into the mcols (
#' metadata columns) slot of the returned object.
#'
#' @return a Ranges object.
#' @seealso \code{IRanges::\link[IRanges:IRanges-constructor]{IRanges()}},
#' \code{GenomicRanges::\link[GenomicRanges:GRanges-class]{GRanges()}}
#'
#' @importFrom rlang quos eval_tidy
#' @importFrom S4Vectors mcols metadata mcols<- metadata<-
#' @importFrom BiocGenerics start end width strand score start<- end<- width<- score<- strand<-
#' @importFrom Seqinfo seqnames seqnames<- seqinfo<-
#' @importFrom IRanges IRanges
#' @importFrom GenomicRanges GRanges granges
#' @rdname ranges-construct
#' @examples
#' df <- data.frame(start=c(2:-1, 13:15), width=c(0:3, 2:0))
#' as_iranges(df)
#'
#' df <- data.frame(start=c(2:-1, 13:15), width=c(0:3, 2:0), strand = "+")
#' # will return an IRanges object
#' as_iranges(df)
#'
#' df <- data.frame(start=c(2:-1, 13:15), width=c(0:3, 2:0),
#' strand = "+", seqnames = "chr1")
#' as_granges(df)
#'
#' # as_g/iranges understand alternate name specification
#' df <- data.frame(start=c(2:-1, 13:15), width=c(0:3, 2:0),
#' strand = "+", chr = "chr1")
#' as_granges(df, seqnames = chr)
#'
#' # can also handle DFrame input
#' df <- methods::as(df, "DFrame")
#' df$y <- IRanges::IntegerList(c(1,2,3), NA, 5, 6, 8, 9, 10:12)
#' as_iranges(df)
#' as_granges(df, seqnames = chr)
#'
#' @export
as_iranges <- function(.data, ..., keep_mcols = TRUE) UseMethod("as_iranges")

#' @export
as_iranges.default <- function(.data, ..., keep_mcols = TRUE) {
  as_iranges.data.frame(as.data.frame(.data), ...)
}

#' @export
as_iranges.data.frame <- function(.data, ..., keep_mcols = TRUE) {
  dots <- quos(...)

  col_names <- names(.data)

  check_names(dots, c("start", "end", "width"))

  if (length(dots) > 0) {
    rd <- lapply(dots, eval_tidy,  data = .data)
  } else {
    rd <- NULL
  }
  # IRanges constructor generate quos for core parts of class
  core_ir <- quos(start = .data$start, end = .data$end, width = .data$width)

  ir <- irng_construct(.data, rd, col_names, core_ir)

  if (keep_mcols) {
    return(make_mcols(.data, ir, col_names, dots, core_ir))
  }
  ir
}

#' @export
as_iranges.DataFrame <- function(.data, ..., keep_mcols = TRUE) {
  s4_cols <- vapply(.data, isS4, logical(1))
  tbl <- as.data.frame(.data[, !s4_cols, drop = FALSE])
  rng <- as_iranges(tbl, ..., keep_mcols = keep_mcols)
  if (keep_mcols) return(append_s4_cols(.data, rng, s4_cols))
  rng
}

#' @rdname ranges-construct
#' @export
as_granges <- function(.data, ..., keep_mcols = TRUE) UseMethod("as_granges")

#' @export
as_granges.default <- function(.data, ..., keep_mcols = TRUE) {
  as_granges.data.frame(as.data.frame(.data), ...)
}

#' @export
as_granges.data.frame <- function(.data, ..., keep_mcols = TRUE) {
  dots <- quos(...)

  col_names <- names(.data)

  valid_names <- c("start", "end", "width", "seqnames", "strand")
  check_names(dots, valid_names)

  if (length(dots) > 0) {
    rd <- lapply(dots, eval_tidy,  data = .data)
  } else {
    rd <- NULL
  }

  if (!(any(names(rd) %in% "seqnames") | any(names(.data) %in% "seqnames"))) {
    stop("seqnames column is required for GRanges.", call. = FALSE)
  }
  # IRanges constructor generate quos for core parts of class
  core_ir <- quos(start = .data$start, end = .data$end, width = .data$width)
  ir <- irng_construct(.data, rd, col_names, core_ir)

  # GRanges constructor generate quos for core parts of class
  core_gr <- quos(seqnames = .data$seqnames, strand = .data$strand)
  ir <- grng_construct(.data, rd, ir, col_names, core_gr)

  if (keep_mcols) {
    return(make_mcols(.data, ir, col_names, dots, c(core_ir, core_gr)))
  }

  ir
}

#' @export
as_granges.DataFrame <- function(.data, ..., keep_mcols = TRUE) {
  dots <- rlang::enquos(...)
  if (length(dots) == 0) return(as(.data, "GRanges"))
  s4_cols <- vapply(.data, isS4, logical(1))

  tbl <- as.data.frame(.data[, !s4_cols, drop = FALSE])
  rng <- as_granges(tbl, !!!dots, keep_mcols = keep_mcols)
  if (keep_mcols) return(append_s4_cols(.data, rng, s4_cols))
  rng
}

check_names <- function(dots, valid_names) {
  if (length(dots) > 0) {
    valid_args <- names(dots) %in% valid_names
    if (any(!valid_args)) {
      stop(paste("Named arguments must be",
                 paste(valid_names, collapse = ","), "."),
           .call = FALSE)
    }
  }
}

append_s4_cols <- function(.data, rng, s4_cols) {
  cols_update <- setdiff(names(.data)[s4_cols], names(mcols(rng)))
  for (col in cols_update) {
    mcols(rng)[[col]] <- .data[[col]]
  }
  rng
}


make_mcols <- function(.data, ir, col_names, dots, core) {
  # remaining columns in data
  old_cols <- unlist(lapply(dots, quo_name))
  remain_cols <- !(col_names %in% c(old_cols, names(core)))
  if (any(remain_cols)) {
    mcols(ir) <- .data[, remain_cols, drop = FALSE]
    names(mcols(ir)) <- col_names[remain_cols]
  }
  ir
}

irng_construct <- function(.data, rd, col_names, core_ir) {

  match_cols_i <- names(core_ir) %in% col_names
  match_dots_i <- names(core_ir) %in% names(rd)

  if (sum(c(match_cols_i, match_dots_i)) < 2) {
    stop("Unable to construct IRanges from .data must have at least two of
         start, end or width columns.",
         call. = FALSE)
  } else {
    remain_cols <- match_cols_i & !match_dots_i
    remain_core <- core_ir[remain_cols]
    if (length(remain_core) > 0) {
      ir <- lapply(core_ir[match_cols_i], eval_tidy, data = .data)

      ir <- c(ir, rd[names(rd) %in% names(core_ir)])
    } else {
      ir <- rd[names(rd) %in% names(core_ir)]
    }

    ir <- IRanges(start = ir[["start"]],
                  end = ir[["end"]],
                  width = ir[["width"]])
  }

  return(ir)

}

grng_construct <- function(.data, rd, ir, col_names, core_gr) {

  if (length(rd[["seqnames"]]) > 0) {
    ir <- GRanges(seqnames = rd[["seqnames"]],
                  ranges = ir)
  } else {
    ir <- GRanges(seqnames = unlist(eval_tidy(core_gr[[1]], .data)),
                  ranges = ir)
  }

  if (length(rd[["strand"]]) > 0) {
    strand(ir) <- rd[["strand"]]
  } else {
    if (any(col_names %in% "strand")) {
      strand(ir) <- unlist(eval_tidy(core_gr[[2]], .data))
    }
  }
  return(ir)

}

#' Coerce an Rle or RleList object to Ranges
#'
#' @param .data a [S4Vectors::Rle()] or an [IRanges::RleList()] object.
#'
#'
#' @return an [IRanges::IRanges()] object if the input is an
#' [S4Vectors::Rle()] object or a [GenomicRanges::GRanges()] object for
#' an [IRanges::RleList()] object.
#'
#' @details This function is behind [compute_coverage()].
#'
#' @seealso \code{S4Vectors::\link[S4Vectors:Rle-class]{Rle()}},
#' \code{IRanges::\link[IRanges:AtomicList]{RleList()}}
#' @examples
#' x <- S4Vectors::Rle(10:1, 1:10)
#' as_ranges(x)
#'
#' # must have names set
#' y <- IRanges::RleList(chr1 = x)
#' as_ranges(y)
#'
#' @export
as_ranges <- function(.data) UseMethod("as_ranges")

#' @importFrom IRanges ranges
#' @importFrom S4Vectors runValue
#' @export
as_ranges.Rle <- function(.data) {
  rng <- ranges(.data)
  mcols(rng)[["score"]] <- runValue(.data)
  rng
}

#' @importFrom methods as
#' @export
as_ranges.RleList <- function(.data) as(.data, "GRanges")

context("Combining Ranges objects")

gr <- as_granges(data.frame(start = 10:15,
                            width = 5,
                            seqnames = "seq1"))

gr2 <- as_granges(data.frame(start = 11:14,
                            width = 1:4,
                            seqnames = "seq1"))

gr3 <- gr
names(gr3) <- letters[1:5]

test_that("simple bind is same as `c`", {
  expect_identical(bind_ranges(gr), gr)
  expect_identical(bind_ranges(gr, gr2), c(gr, gr2))
  expect_identical(bind_ranges(gr, list(gr, gr2), gr2), c(gr,gr,gr2,gr2))
})

test_that("setting .id column", {
  gr4  <- bind_ranges(list(a = gr, b = gr2), .id = "origin")
  expect_s4_class(gr4$origin, "Rle")
  expect_equal(S4Vectors::runLength(gr4$origin), c(length(gr), length(gr2)))
  expect_identical(gr4, bind_ranges(a = gr, b = gr2, .id = "origin"))
  expect_identical(gr4, bind_ranges(a = gr, list(b = gr2), .id = "origin"))
})

test_that("bind preserves names", {
  expect_null(names(bind_ranges(gr, gr2)))
  expect_equal(names(bind_ranges(gr3)), names(gr3))
  gr5 <- bind_ranges(gr3, gr2)
  expect_equal(names(gr5), c(names(gr3), rep("", length(gr2))))
})

context("expanding GRanges objects")

test_that("expanding makes sense", {
  # expanding empty GRanges returns error
  gr <- GRanges()
  expect_error(expand_ranges(gr), "No list columns to unnest.", fixed = TRUE)
  # no metadata columns returns error
  gr <- as_granges(data.frame(seqnames = c("chr1", "chr2"),
                              start = c(10, 20),
                              width = 5))
  expect_error(expand_ranges(gr), "No list columns to unnest.", fixed = TRUE)

  # by default expand_ranges does cartesian product of list columns
  gr <- as_granges(data.frame(seqnames = "chr1", start = 20:22, width = 1000))
  gr <- mutate(gr,
                 col1 = IntegerList(a = 1, b = c(4,5), c = c(2,3)),
                 col2 = IntegerList(c(1,2), c(3,4), c(5)),
                 score = 1:3)

  correct_gr <- GRanges(seqnames = "chr1",
                        ranges = IRanges(start = c(20,20,21,21,21,21,22,22),
                                         width = 1000),
                        col1 = c(1L,1L,4L,4L,5L,5L,2L,3L),
                        col2 = c(1L,2L,3L,4L,3L,4L,5L,5L),
                        score = c(1L,1L,2L,2L,2L,2L,3L,3L))
  test_gr <- expand_ranges(gr, .recursive = TRUE)
  expect_identical(correct_gr, test_gr)
  test_gr <- expand_ranges(gr, col1, col2, .recursive = TRUE)
  expect_identical(correct_gr, test_gr)

  # unnesting on non-existent column errors
  expect_error(expand_ranges(gr, gc),
               "Input column(s): gc not found.",
               fixed = TRUE)
  # unnesting on non-list column returns an error
  expect_error(expand_ranges(gr, score),
               "Input column(s): score are not list columns.",
               fixed = TRUE)

  # not specifying .drop returns other list columns
  correct_gr <- GRanges(seqnames = "chr1",
                          ranges = IRanges(start = c(20,21,21,22,22),
                                           width = 1000),
                          col1 = as.integer(c(1,4,5,2,3)),
                          col2 = IntegerList(c(1,2), c(3,4), c(3,4), 5,5),
                          score = as.integer(c(1,2,2,3,3)))
  test_gr <- expand_ranges(gr, col1)
  expect_identical(correct_gr, test_gr)

  correct_gr <- select(correct_gr, -col2)
  test_gr <- expand_ranges(gr, col1, .drop = TRUE)
  expect_identical(correct_gr, test_gr)

  # .id works as expected
  expect_error(expand_ranges(gr, .id = "name", .recursive = TRUE),
               "`.id` does not have same length as number of list columns.",
               fixed = TRUE)
  expect_error(expand_ranges(gr, col2, .id = c("name", "x0")),
               "`.id` does not have same length as number of list columns.",
               fixed = TRUE)

  correct_gr <- GRanges(seqnames = "chr1",
                        ranges = IRanges(start = c(20,20,21,21,21,21,22,22),
                                         width = 1000),
                        col1 = as.integer(c(1,1,4,4,5,5,2,3)),
                        col2 = as.integer(c(1,2,3,4,3,4,5,5)),
                        score = as.integer(c(1,1,2,2,2,2,3,3)),
                        id1 = c(rep("a", 2), rep("b", 4), rep("c", 2)),
                        id2 = as.integer(c(1,1, 2,2,2,2,3,3)))

  test_gr <- expand_ranges(gr, .id = c("id1", "id2"), .recursive = TRUE)
  expect_identical(correct_gr, test_gr)

  correct_gr <- GRanges(seqnames = "chr1",
                        ranges = IRanges(start = c(20,21,21,22,22),
                                         width = 1000),
                        col1 = as.integer(c(1,4,5,2,3)),
                        col2 = IntegerList(c(1,2), c(3,4), c(3,4), 5,5),
                        score = as.integer(c(1,2,2,3,3)),
                        id1 = c("a", "b", "b", "c", "c"))
  test_gr <- expand_ranges(gr, col1, .id = "id1", .recursive = TRUE)
  expect_identical(correct_gr, test_gr)
  
  # one column test
  gr <- as_granges(data.frame(seqnames = "chr1", start = 20:22, width = 1000))
  gr <- mutate(gr, col1 = IntegerList(a = 1, b = c(4,5), c = c(2,3)))
  correct_gr <- S4Vectors::expand(gr)
  expect_equal(expand_ranges(gr), correct_gr)
  
  correct_gr$id1 <- c("a", "b", "b", "c", "c")
  expect_equal(expand_ranges(gr, .id = "id1"), correct_gr)
  
  # drop emptys
  gr <- as_granges(data.frame(seqnames = "chr1", start = 20:22, width = 1000))
  gr <- mutate(gr, col1 = IntegerList(a = integer(), b = c(4,5), c = c(2,3)))
  correct_gr <- S4Vectors::expand(gr, keepEmptyRows = TRUE)
  expect_equal(expand_ranges(gr, .keep_empty = TRUE), correct_gr)
  correct_gr$id1 <- c("a", "b", "b", "c", "c")
  expect_equal(expand_ranges(gr, .id = "id1", .keep_empty = TRUE), correct_gr)

})

# Setup test data
setup_test_data <- function() {
  # Create basic GRanges with various column types
  df <- data.frame(
    seqnames = rep(c("chr1", "chr2"), each = 5),
    start = 1:10,
    end = 5:14,
    strand = rep(c("+", "-"), 5),
    score = runif(10),
    count = rpois(10, 5),
    category = factor(rep(c("A", "B"), 5))
  )
  
  rng <- as_granges(df)
  
  # Add S4 columns to mcols for testing
  mcols(rng)$s4_rle <- Rle(c("x", "y", "z"), c(3, 4, 3))
  mcols(rng)$factor_col <- as(factor(rep(c("group1", "group2"), 5)), "Rle")
  
  return(rng)
}

test_that("pull works with core GRanges columns", {
  rng <- setup_test_data()
  
  # Test core genomic columns
  expect_equal(pull(rng, start), start(rng))
  expect_equal(pull(rng, end), end(rng))
  expect_equal(pull(rng, width), width(rng))
  expect_equal(pull(rng, seqnames), seqnames(rng))
  expect_equal(pull(rng, strand), strand(rng))
})

test_that("pull works with metadata columns", {
  rng <- setup_test_data()
  
  # Test regular mcols
  expect_equal(pull(rng, score), mcols(rng)[["score"]])
  expect_equal(pull(rng, count), mcols(rng)[["count"]])
  expect_equal(pull(rng, category), mcols(rng)[["category"]])
})

test_that("pull works with S4 metadata columns", {
  rng <- setup_test_data()
  
  # Test S4 columns in mcols
  expect_equal(pull(rng, s4_rle), mcols(rng)[["s4_rle"]])
  expect_equal(pull(rng, factor_col), mcols(rng)[["factor_col"]])
  
  # Verify they are indeed S4 objects
  expect_true(isS4(mcols(rng)[["s4_rle"]]))
  expect_true(isS4(mcols(rng)[["factor_col"]]))
})

test_that("pull works with positional indexing", {
  rng <- setup_test_data()
  
  # Test positive indexing (metadata columns only, excluding core columns)
  mcol_names <- setdiff(tbl_vars(rng), c("seqnames", "start", "end", "width", "strand"))
  
  expect_equal(pull(rng, 1), start(rng))
  expect_equal(pull(rng, 2), end(rng))
  
  # Test negative indexing (from the right)
  expect_equal(pull(rng, -1), mcols(rng)[[mcol_names[length(mcol_names)]]])
  expect_equal(pull(rng, -2), mcols(rng)[[mcol_names[length(mcol_names) - 1]]])
})

test_that("pull with name parameter works correctly", {
  rng <- setup_test_data()
  
  # Test naming with regular columns
  result <- pull(rng, score, name = category)
  expected <- mcols(rng)[["score"]]
  names(expected) <- mcols(rng)[["category"]]
  expect_equal(result, expected)
  
  # Test naming with core column
  result <- pull(rng, count, name = seqnames)
  expected <- mcols(rng)[["count"]]
  names(expected) <- as.character(seqnames(rng))
  expect_equal(result, expected)
})

test_that("pull with S4 name parameter converts to vector", {
  rng <- setup_test_data()
  
  # Test naming with S4 column - should convert to vector
  result <- pull(rng, score, name = s4_rle)
  expected <- mcols(rng)[["score"]]
  names(expected) <- as.vector(mcols(rng)[["s4_rle"]])
  expect_equal(result, expected)
  
  # Test naming with S4 factor
  result <- pull(rng, count, name = factor_col)
  expected <- mcols(rng)[["count"]]
  names(expected) <- as.vector(mcols(rng)[["factor_col"]])
  expect_equal(result, expected)
})

test_that("pull works with GroupedGenomicRanges", {
  rng <- setup_test_data()
  grouped_rng <- group_by(rng, strand)
  
  # Test that pull on grouped ranges returns same as ungrouped
  expect_equal(pull(grouped_rng, start), start(rng))
  expect_equal(pull(grouped_rng, score), mcols(rng)[["score"]])
  expect_equal(pull(grouped_rng, s4_rle), mcols(rng)[["s4_rle"]])
  
  # Test with names on grouped ranges
  result_grouped <- pull(grouped_rng, score, name = category)
  result_ungrouped <- pull(rng, score, name = category)
  expect_equal(result_grouped, result_ungrouped)
  
  # Test with S4 names on grouped ranges
  result_grouped <- pull(grouped_rng, count, name = s4_rle)
  result_ungrouped <- pull(rng, count, name = s4_rle)
  expect_equal(result_grouped, result_ungrouped)
})

test_that("pull works with GroupedIntegerRanges", {
  # Create IRanges version
  df <- data.frame(
    start = 1:10,
    width = 5,
    score = runif(10),
    group = rep(c("A", "B"), 5)
  )
  
  irng <- as_iranges(df)
  mcols(irng)$s4_rle <- Rle(c("x", "y"), c(5, 5))
  
  grouped_irng <- group_by(irng, group)
  
  # Test basic functionality
  expect_equal(pull(grouped_irng, start), start(irng))
  expect_equal(pull(grouped_irng, score), mcols(irng)[["score"]])
  expect_equal(pull(grouped_irng, s4_rle), mcols(irng)[["s4_rle"]])
  
  # Test with names
  result <- pull(grouped_irng, score, name = group)
  expected <- mcols(irng)[["score"]]
  names(expected) <- mcols(irng)[["group"]]
  expect_equal(result, expected)
})

test_that("pull handles edge cases appropriately", {
  rng <- setup_test_data()
  
  # Test quasiquotation support
  var_name <- "score"
  expect_equal(pull(rng, !!var_name), mcols(rng)[["score"]])
  
  # Test with symbols
  expect_equal(pull(rng, !!rlang::sym("count")), mcols(rng)[["count"]])
})

test_that("pull returns correct data types", {
  rng <- setup_test_data()
  
  # Verify return types match original column types
  expect_identical(class(pull(rng, score)), class(mcols(rng)[["score"]]))
  expect_identical(class(pull(rng, count)), class(mcols(rng)[["count"]]))
  expect_identical(class(pull(rng, category)), class(mcols(rng)[["category"]]))
  expect_identical(class(pull(rng, s4_rle)), class(mcols(rng)[["s4_rle"]]))
  expect_identical(class(pull(rng, start)), class(start(rng)))
  expect_identical(class(pull(rng, seqnames)), class(seqnames(rng)))
})

test_that("pull default argument (var = -1) works correctly", {
  rng <- setup_test_data()
  
  # Test default behavior - should pull last column
  all_vars <- tbl_vars(rng)
  last_var <- all_vars[length(all_vars)]
  
  # Default pull() should equal pull with last column explicitly
  expect_equal(pull(rng), pull(rng, !!last_var))
  
  # Test with grouped ranges
  grouped_rng <- group_by(rng, strand)
  expect_equal(pull(grouped_rng), pull(grouped_rng, !!last_var))
  
  # Test that default pulls from the rightmost position
  expect_equal(pull(rng), pull(rng, -1))
})

test_that("pull positional indexing works correctly", {
  rng <- setup_test_data()
  all_vars <- tbl_vars(rng)
  
  # Test positive indexing from left
  for (i in 1:min(5, length(all_vars))) {
    expect_equal(pull(rng, i), pull(rng, !!all_vars[i]),
                 info = paste("Failed for position", i))
  }
  
  # Test negative indexing from right
  for (i in 1:min(5, length(all_vars))) {
    expected_pos <- length(all_vars) - i + 1
    expect_equal(pull(rng, -i), pull(rng, !!all_vars[expected_pos]),
                 info = paste("Failed for negative position", -i))
  }
  
  # Test with grouped ranges
  grouped_rng <- group_by(rng, strand)
  expect_equal(pull(grouped_rng, 1), pull(rng, 1))
  expect_equal(pull(grouped_rng, -1), pull(rng, -1))
  expect_equal(pull(grouped_rng, 3), pull(rng, 3))
})

test_that("pull positional indexing with different Range types", {
  # Test with IRanges
  df_ir <- data.frame(
    start = 1:5,
    width = 3,
    score1 = runif(5),
    score2 = rpois(5, 3),
    group = letters[1:5]
  )
  irng <- as_iranges(df_ir)
  ir_vars <- tbl_vars(irng)
  
  # Test first and last positions
  expect_equal(pull(irng, 1), pull(irng, !!ir_vars[1]))
  expect_equal(pull(irng, -1), pull(irng, !!ir_vars[length(ir_vars)]))
  expect_equal(pull(irng), pull(irng, -1))  # Default behavior
  
  # Test with minimal IRanges (only core columns)
  minimal_irng <- IRanges(start = 1:3, width = 5)
  minimal_vars <- tbl_vars(minimal_irng)
  expect_equal(pull(minimal_irng, 1), pull(minimal_irng, !!minimal_vars[1]))
  expect_equal(pull(minimal_irng), pull(minimal_irng, -1))
})

test_that("pull default behavior with different column compositions", {
  # Test with ranges that have only core columns
  core_only <- GRanges(
    seqnames = "chr1",
    ranges = IRanges(1:3, width = 5),
    strand = "+"
  )
  core_vars <- tbl_vars(core_only)
  expect_equal(pull(core_only), pull(core_only, !!core_vars[length(core_vars)]))
  
  # Test with ranges that have mixed core + metadata columns
  rng <- setup_test_data()
  all_vars <- tbl_vars(rng)
  expect_equal(pull(rng), pull(rng, !!all_vars[length(all_vars)]))
  
  # Verify the last column is indeed what we expect
  last_col_name <- all_vars[length(all_vars)]
  if (last_col_name %in% names(mcols(rng))) {
    expect_equal(pull(rng), mcols(rng)[[last_col_name]])
  } else {
    # It's a core column
    if (last_col_name == "start") expect_equal(pull(rng), start(rng))
    if (last_col_name == "end") expect_equal(pull(rng), end(rng))
    if (last_col_name == "width") expect_equal(pull(rng), width(rng))
    if (last_col_name == "seqnames") expect_equal(pull(rng), seqnames(rng))
    if (last_col_name == "strand") expect_equal(pull(rng), strand(rng))
  }
})

test_that("pull with empty mcols still accesses core columns", {
  # Create ranges with no metadata columns
  rng <- GRanges(
    seqnames = "chr1",
    ranges = IRanges(start = 1:5, end = 5:9),
    strand = "+"
  )
  
  # Should still be able to pull core columns
  expect_equal(pull(rng, start), start(rng))
  expect_equal(pull(rng, seqnames), seqnames(rng))
  expect_equal(pull(rng, strand), strand(rng))
  
  # Test default behavior with no mcols
  core_vars <- tbl_vars(rng)
  expect_equal(pull(rng), pull(rng, !!core_vars[length(core_vars)]))
  expect_equal(pull(rng), pull(rng, -1))
})

test_that("pull maintains vector attributes", {
  rng <- setup_test_data()
  
  # Test that factor levels are preserved
  original_factor <- mcols(rng)[["category"]]
  pulled_factor <- pull(rng, category)
  expect_equal(levels(original_factor), levels(pulled_factor))
  
  # Test that S4 object structure is preserved
  original_rle <- mcols(rng)[["s4_rle"]]
  pulled_rle <- pull(rng, s4_rle)
  expect_identical(runLength(original_rle), runLength(pulled_rle))
  expect_identical(runValue(original_rle), runValue(pulled_rle))
})

# test-groups.R
context("Grouped Ranges and grouping")

gr0  <- GRanges()
ir0 <- IRanges()
gr1 <- GRanges(seqnames = "seq1",
               strand = c(rep("+", 4), rep("-", 3), rep("*", 3)),
               ranges = IRanges(start = 1:10, width = 5),
               score = as.integer(c(0,3,4,5,10,14,0, 0, 4,8)))

test_that("grouping works on empty ranges", {
  expect_s4_class(gr0 %>% group_by(seqnames), "GroupedGenomicRanges")
  expect_s4_class(ir0 %>% group_by(start), "GroupedIntegerRanges")
})

test_that("throws invalid object if group not found", {
  expect_error(gr0 %>% group_by(st))
  expect_error(ir0 %>% group_by(st))
})

test_that("grouping doesn't allow character names", {
  expect_error(gr0 %>% group_by("seqnames"))
  expect_error(ir0 %>% group_by("start"))
})

test_that("group_vars works as expected", {
  gr_by_strand_score <- gr1 %>% group_by(strand, score)
  expect_equal(group_vars(gr_by_strand_score),
                   c("strand", "score"))
  ir_by_start <- ir0 %>% group_by(start)
  expect_equal(group_vars(ir_by_start), "start")
})

test_that("ungrouping works as expected", {
  gr_by_strand_score <- gr1 %>% group_by(strand, score)
  gr_by_score <- gr1 %>% group_by(score)
  expect_identical(ungroup(gr_by_strand_score), gr1)
  expect_equivalent(gr_by_strand_score %>% ungroup(strand), gr_by_score)
})

test_that("grouping after filter or mutate leaves groups in tact", {
            by_strand <- gr1 %>% group_by(strand) %>% mutate(new_score = mean(score))
            expect_s4_class(by_strand, "GroupedGenomicRanges")
            expect_equal("strand", group_vars(by_strand))
            by_strand_filter <- gr1 %>% group_by(score) %>% filter(n() >= 2)
            expect_s4_class(by_strand_filter, "GroupedGenomicRanges")
            expect_equal("score", group_vars(by_strand_filter))
})

test_that("group by matches HelloRanges", {
  oldwd <- getwd()
  setwd(system.file("unitTests", "data", "groupby", package="HelloRanges"))

  a <- read_bed("values3.header.bed")
  exp <- GRanges(seqnames = c(rep("chr1", 5), rep("chr3",4)),
                 ranges = IRanges(start = c(1, 11, 12, 21, 121,1,11,21,121),
                                  end = c(10, 20, 21, 30, 130, 10, 20, 30, 130)),
                 strand = "*",
                 score.sum = c(10,5,5,45,1,1,2,3,8))

  result <- a %>%
    group_by(seqnames, start, end) %>%
    summarise(score.sum = sum(score)) %>%
    as_granges() %>%
    sort()

  expect_identical(exp, result)
  setwd(oldwd)
})

test_that("ungroup/groups/group_vars on ordinary Ranges",{
  expect_identical(ungroup(gr1), gr1)
  expect_identical(ungroup(ranges(gr1)), ranges(gr1))
  expect_null(groups(gr1))
  expect_null(groups(ranges(gr1)))
  expect_identical(group_vars(gr1), character(0))
  expect_identical(group_vars(ranges(gr1)), character(0))
})

# issue #42
test_that("group_by does not allow unknown columns", {
  gr <- GRanges("chr1", IRanges(start = 1:6, width = 10))
  foo <- 1:6
  expect_error(group_by(gr, foo), "Column `foo` is unknown")
  expect_error(group_by(gr, strand, foo), "Column `foo` is unknown")
})

# test-filter.R
context("filter methods")

gr0 <- GRanges(seqnames = "seq1",
               strand = c(rep("+", 4), rep("-", 3), rep("*", 3)),
               ranges = IRanges(start = 1:10, width = 5),
               score = as.integer(c(0,3,4,5,10,14,0, 0, 4,8)))

test_that("simple filter checks", {

  expect_identical(gr0 %>% filter(strand == "+"), gr0[strand(gr0) == "+"])
  expect_identical(gr0 %>% filter(score > 0), gr0[score(gr0) > 0])
  expect_identical(gr0 %>% filter(score > 0 & strand == "*"),
                   gr0[score(gr0) > 0 & strand(gr0) == "*"])
  expect_identical(gr0 %>% filter(score > 0, strand == "*"),
                   gr0[score(gr0) > 0 & strand(gr0) == "*"])
})

test_that("checking calls to n", {
  expect_identical(gr0 %>% filter(n() == 10), gr0)
  expect_identical(gr0 %>% filter(n() == 4), GRanges(score = integer(0),
                                                     seqinfo = seqinfo(gr0)))
  expect_error(gr0 %>% filter(1:n()))
})

gr_gfilter <- c(gr0[strand(gr0) == "+" & score(gr0) > 3],
                gr0[strand(gr0) == "-" & score(gr0) > 8],
                gr0[strand(gr0) == "*" & score(gr0) > 4])
test_that("grouped filter checks", {
  expect_identical(gr0 %>%
                     group_by(strand) %>%
                     filter(n() == 4) %>%
                     ungroup(),
                   gr0[1:4])
  expect_identical(gr0 %>% group_by(strand) %>%
                     filter(score > mean(score)) %>%
                     ungroup(),
                   gr_gfilter)
})

# see https://github.com/sa-lee/plyranges/issues/36
# note this should no longer be necessary as dplyr is part of Depends now
# may effect downstream packages
test_that("using `::` works", {
  expect_identical(gr0 %>% dplyr::group_by(strand) %>%
                     dplyr::filter(score > mean(score)) %>%
                     dplyr::ungroup(),
                   gr_gfilter)
  gr <- data.frame(seqnames = c("chr1", "chr2", "chr2", "chr1", "chr2"),
                   start = 1:5,
                   width = 5) %>%
    as_granges()
  expect_identical(dplyr::filter(gr, seqnames %in% c("chr1", "chr2")),
                   gr)
  expect_identical(gr %>% 
                     dplyr::group_by(seqnames) %>% 
                     dplyr::filter(seqnames %in% c("chr1", "chr2")) %>%
                     dplyr::ungroup(),
                   gr)
})

test_nearest_join_distance <- function(query, subject, join_fun, range_fun = granges){
  join <- join_fun(query, subject)
  
  join_dist <- join_fun(query, subject, distance = TRUE)
  join_test <- join_fun(query, subject, distance = "test")
  
  expect_equal(range_fun(join), range_fun(join_dist))
  expect_equal(mcols(join_dist)$distance, mcols(join_test)$test)
  
  meta <- DataFrame('test' = 1:length(query))
  mcols(query) <- meta
  expect_error(join_fun(query, subject, distance = "test"), "columns already exist")
  
}

test_that("join_nearest behaves correctly on GRanges", {

  subject  <- data.frame(seqnames = "chr1",
               start = c(11,101),
               end = c(21, 200),
               name = c("a1", "a2"),
               strand = c("+", "-"),
               score = c(1,2)) %>%
           as_granges()
  query <- data.frame(seqnames = "chr1",
                        strand = c("+", "-", "+", "-"),
                        start = c(21,91,101,201),
                        end = c(30,101,110,210),
                        name = paste0("b", 1:4),
                        score = 1:4) %>%
                     as_granges()
  
  test_nearest_join_distance(query, subject, join_nearest)
  test_nearest_join_distance(query, subject, join_nearest_left)
  test_nearest_join_distance(query, subject, join_nearest_right)
  test_nearest_join_distance(query, subject, join_nearest_upstream)
  test_nearest_join_distance(query, subject, join_nearest_downstream)
  
})

test_that("join_nearest works on IRanges", {
  query_ir <- IRanges(c(6, 11, 1, 13, 18), width=c(2, 2, 2, 2, 2))
  subject_ir <- IRanges(c(1, 2, 9, 15, 15), width=c(4, 3, 2, 2, 3))
  
  test_nearest_join_distance(query_ir, subject_ir, join_nearest, ranges)
  test_nearest_join_distance(query_ir, subject_ir, join_nearest_left, ranges)
  test_nearest_join_distance(query_ir, subject_ir, join_nearest_right, ranges)
})

test_add_nearest_distance <- function(query, subject){
  dist <- add_nearest_distance(query, subject)
  test <- add_nearest_distance(query, subject, name = "test")
  
  meta <- DataFrame('test' = 1:length(query))
  mcols(query) <- meta
  
  # test errors on existing column
  expect_error(add_nearest_distance(query, subject, name = "test"), "already exists in destination")
  
  expect_equal(mcols(dist)$distance, mcols(test)$test)
}

test_that("add_nearest_distance works", {
  subject  <- data.frame(seqnames = "chr1",
               start = c(11,101),
               end = c(21, 200),
               name = c("a1", "a2"),
               strand = c("+", "-"),
               score = c(1,2)) %>%
           as_granges()
  query <- data.frame(seqnames = "chr1",
                        strand = c("+", "-", "+", "-"),
                        start = c(21,91,101,201),
                        end = c(30,101,110,210),
                        name = paste0("b", 1:4),
                        score = 1:4) %>%
                     as_granges()
  
  query_ir <- IRanges(c(6, 11, 1, 13, 18), width=c(2, 2, 2, 2, 2))
  subject_ir <- IRanges(c(1, 2, 9, 15, 15), width=c(4, 3, 2, 2, 3))

  test_add_nearest_distance(query, subject)
  test_add_nearest_distance(query_ir, subject_ir)
})

set.seed(1999)
df <- data.frame(start = 1:10,
                 width = 5,
                 seqnames = "seq1",
                 strand = sample(c("+", "-", "*"), 10, replace = TRUE),
                 gc = runif(10))
rng <- as_granges(df)

test_that("rangewise slice produces expected results", {
  expect_identical(dplyr::slice(rng, 1:2), rng[1:2])
  expect_identical(dplyr::slice(rng, -n()), rng[1:9])
  expect_identical(dplyr::slice(rng, -5:-n()), rng[1:4])
  expect_error(dplyr::slice(rng, n()+1L))
  expect_error(dplyr::slice(rng, gc > 0.5))
})

test_that("new slice functionality produces expected results", {
  expect_identical(dplyr::slice_head(rng, n=2), rng[1:2])
  expect_identical(dplyr::slice_tail(rng, n=2), rng[9:10])
  expect_identical(dplyr::slice_max(rng, gc), rng[which.max(rng$gc)])
  expect_identical(dplyr::slice_min(rng, gc), rng[which.min(rng$gc)])

  expect_true(length(dplyr::slice_sample(rng, prop=.3)) == 3)
  
  # check that proportion weights look right:
  rng_wts <- rng
  rng_wts$weight <- c(1,1,1,1,1,2,2,2,2,2)
  rep_wts <- replicate(100,
    dplyr::slice_sample(rng_wts, prop=.5, weight_by=weight)$weight
  )
  tab <- table(as.vector(rep_wts))/500
  expect_true(tab[1] < tab[2])

})

by_strand <- group_by(rng, strand)

test_that("groupwise slice produces expected results", {

  target <- dplyr::slice(by_strand, n()) %>% ungroup()
  exp <- rng[c(7,9,10)]
  expect_identical(target, exp)
  
  target <- dplyr::slice(by_strand, which.max(gc))
  exp <- filter(by_strand, gc == max(gc))
  expect_identical(target, exp)
  
  target <- dplyr::slice(by_strand, 1:4)
  expect_identical(target, by_strand)

  expect_identical(dplyr::slice_head(by_strand) |> ungroup(), rng[c(1,2,5)])
  expect_identical(dplyr::slice_tail(by_strand) |> ungroup(), rng[c(7,9,10)])

  spl <- split(rng$gc, strand(rng))
  target <- dplyr::slice_max(by_strand, gc) %>% ungroup()
  idx <- as.character(strand(target))
  expect_identical(target$gc, sapply(spl[idx], max) %>% unname())
  target <- dplyr::slice_min(by_strand, gc) %>% ungroup()
  idx <- as.character(strand(target))
  expect_identical(target$gc, sapply(spl[idx], min) %>% unname())

  set.seed(123)
  target <- dplyr::slice_sample(by_strand, prop=.5)
  expect_equal(as.numeric(table(strand(target))), c(1,2,1))
  
})

# Representing seqinfo information as a Ranges

#' @name ranges-info
#' @rdname ranges-info
#' @export
genome_info <- function(genome = NULL, seqnames = NULL, seqlengths = NULL, is_circular = NULL) {

  if (length(seqlengths) == 0L) seqlengths <- NA
  if (length(is_circular) == 0L) is_circular <- NA
  if (length(genome) == 0L) genome <- NA

  seqinfo_res <- Seqinfo(seqnames, seqlengths, is_circular, genome)
  get_genome_info(seqinfo_res)

}

#' Construct annotation information
#'
#' @param .data A Ranges object to annotate or retrieve an annotation for.
#' @param genome A character vector of length one indicating the genome build.
#' @param seqnames A character vector containing the name of sequences.
#' @param seqlengths An optional integer vector containg the lengths of sequences.
#' @param is_circular An optional logical vector indicating whether a sequence is ciruclar.
#'
#' @return a GRanges object containing annotations. To retrieve the annotations
#' as a Ranges object use `get_genome_info`.
#'
#' @description To construct annotations by supplying annotation information
#' use `genome_info`. To add
#' annotations to an existing Ranges object use `set_genome_info`. To retrieve
#' an annotation as a Ranges object use `get_genome_info`.
#'
#' @importFrom Seqinfo Seqinfo seqnames seqlengths isCircular genome seqinfo
#' @seealso [Seqinfo::Seqinfo()]
#' @examples
#' x <- genome_info(genome = "toy",
#'                  seqnames = letters[1:4],
#'                  seqlengths = c(100, 300, 15, 600),
#'                  is_circular = c(NA, FALSE, FALSE, TRUE))
#'x
#'
#' rng <- as_granges(data.frame(seqnames = "a", start = 30:50, width = 10))
#' rng
#' rng <- set_genome_info(rng,
#'                        genome = "toy",
#'                        seqnames = letters[1:4],
#'                        seqlengths = c(100, 300, 15, 600),
#'                        is_circular = c(NA, FALSE, FALSE, TRUE))
#' get_genome_info(rng)
#'
#'\dontrun{
#' if (interactive()) {
#'  # requires internet connection
#'  genome_info(genome = "hg38")
#' }
#' }
#'
#' @rdname ranges-info
#' @export
set_genome_info <- function(.data, genome = NULL, seqnames = NULL,
                            seqlengths = NULL, is_circular = NULL) {


  info <- seqinfo(.data)

  if (is.null(genome)) {
     genome <- genome(info)
  }

  if (is.null(seqlengths)) {
     seqlengths <- seqlengths(info)
  }

  if (is.null(seqnames)) {
    seqnames <- seqnames(info)
  }

  if (is.null(is_circular)) {
    is_circular <- isCircular(info)
  }

  info <- Seqinfo(seqnames, seqlengths, is_circular, genome)

  seqinfo(.data) <- info
  return(.data)
}



#' @export
#' @importFrom methods hasMethod
#' @rdname ranges-info
get_genome_info <- function(.data)  UseMethod("get_genome_info")

#' @export
get_genome_info.default <- function(.data) {
  if (hasMethod("seqinfo", class(.data))) {
    return(get_genome_info.Seqinfo(seqinfo(.data)))
  } else {
    stop(paste("Unable to retrieve genome annotation from object of class:",
         class(.data)), call. = FALSE)
  }
}

#' @export
get_genome_info.Seqinfo <- function(.data) {

  if (any(is.na(seqlengths(.data)))) {
    stop("seqlengths must be non-missing to convert to GRanges")
  }
  as(.data, "GRanges")
}

#' @export
get_genome_info.BamFileOperator <- function(.data) {
  get_genome_info(seqinfo(.data@input))
}

#' @export
get_genome_info.DeferredGenomicRanges <- function(.data) {
  get_genome_info(.data@ops)
}

# test-coverage.R
# adapted from HelloRanges tests
context("coverage")

test_that("produces same result as IRanges",
          {
            ir <- IRanges(c(1, 8, 14, 15, 19, 34, 40),
                          width = c(12, 6, 6, 15, 6, 2, 7))
            cvg <- compute_coverage(ir)
            expect_identical(mcols(cvg)$score,
                             c(1L, 2L, 1L, 2L, 3L, 2L, 1L, 0L, 1L, 0L, 1L))
            expect_identical(width(cvg),
                             c(7L, 5L, 2L, 4L, 1L, 5L, 5L, 4L, 2L, 4L, 7L))

            ir <- IRanges(start=c(-2L, 6L, 9L, -4L, 1L, 0L, -6L, 10L),
                          width=c( 5L, 0L, 6L,  1L, 4L, 3L,  2L,  3L))
            cvg <- compute_coverage(ir)
            expect_identical(mcols(cvg)$score,
                             c(3L, 1L, 0L, 1L, 2L, 1L))
            expect_identical(width(cvg), c(2L, 2L, 4L, 1L, 3L, 2L))

            cvg <- ir %>% shift_right(7L) %>% compute_coverage()
            expect_identical(mcols(cvg)$score,
                             c(1L, 0L, 1L, 2L, 3L, 1L, 0L, 1L, 2L, 1L))
            expect_identical(width(cvg),
                             c(3L, 1L, 2L, 1L, 2L, 2L, 4L, 1L, 3L, 2L))

          })

test_that("produces same results as GRanges", {
  gr <- GRanges(seqnames = Rle(factor(c("chr1", "chr2", "chr1", "chr3")),
                               c(1, 3, 2, 4)),
                ranges = IRanges(1:10, width = 10:1))
  target <- IRanges::RleList(chr1=Rle(1:3, c(4, 1, 5)),
                    chr2=Rle(0:3, c(1, 1, 1, 7)),
                    chr3=Rle(0:4, c(6, 1, 1, 1, 1)),
                    compress=FALSE)
  cvg <- compute_coverage(gr)

  expect_identical(cvg$score, unlist(runValue(target), use.names = FALSE))
  expect_identical(width(cvg), unlist(runLength(target), use.names = FALSE))

  target <- RleList(chr1=Rle(1:3, c(4, 1, 5)),
                    chr2=Rle(c(0:3, 0L), c(1, 1, 1, 7, 10)),
                    chr3=Rle(c(0:4, 0L), c(6, 1, 1, 1, 1, 20)),
                    compress=FALSE)

  cvg <- compute_coverage(gr, width = c(chr1 = 10, chr2 = 20, chr3 = 30))

  expect_identical(cvg$score, unlist(runValue(target), use.names = FALSE))
  expect_identical(width(cvg), unlist(runLength(target), use.names = FALSE))

})
test_that("Compatible with GenomicRanges", {
  skip_if_not(
    requireNamespace("HelloRanges", quietly = TRUE),
    message = "'HelloRanges' must be installed to run coverage unit tests."
  )
  oldwd <- getwd()
  setwd(system.file("unitTests", "data", "genomecov", package="HelloRanges"))

  genome <- read.delim("test.genome", header = FALSE)

  genome <- genome_info(genome = "test",
                        seqnames = as.character(genome$V1),
                        seqlengths = genome$V2)
  y <- read_bed("y.bed", genome_info = genome)
  cov_gr <- as(coverage(granges(y)), "GRanges")
  exp <- cov_gr[score(cov_gr) > 0]
  cov_plyr <- y %>% compute_coverage() %>% filter(score > 0)
  expect_identical(exp, cov_plyr)
  setwd(oldwd)

})

# # ranges-anchors.R
#' Anchored Ranges objects
#'
#' @description The `GRangesAnchored` class and the `IRangesAnchored`
#' class allow components of a `GRanges` or `IRanges` (start, end, center)
#' to be held fixed.
#'
#' @details Anchoring will fix a Ranges start, end, or center positions,
#' so these positions will remain the same when performing arithimetic.
#' For `GRanges` objects, the function
#' (`anchor_3p()`) will fix the start for the negative strand,
#' while `anchor_5p()` will fix the end for the
#' positive strand. Anchoring modifies how arithmetic is performed, for example
#' modifying the width of a range with `set_width()` or stretching a
#' range with `stretch()`. To remove anchoring use `unanchor()`.
#'
#' @param x a Ranges object
#'
#' @section Constructors:
#' Depending on how you want to fix the components of a Ranges, there are
#' five ways to construct a RangesAnchored class. Here `x` is either
#' an `IRanges` or `GRanges` object.
#' \describe{
#'    \item{`anchor_start(x)`}{Fix the start coordinates}
#'    \item{`anchor_end(x)`}{Fix the end coordinates}
#'    \item{`anchor_center(x)`}{Fix the center coordinates}
#'    \item{`anchor_3p(x)`}{On the negative strand fix the start coordinates,
#'    and for positive or unstranded ranges fix the end coordinates.}
#'    \item{`anchor_5p(x)`}{On the positive or unstranded ranges fix the start coordinates,
#'    coordinates and for negative stranded ranges fix the end coordinates.}
#' }
#'
#' @section Accessors:
#' To see what has been anchored use the function `anchor`.
#' This will return a character vector containing a valid anchor.
#' It will be set to one of `c("start", "end", "center")` for an
#' `IRanges` object or one of
#' `c("start", "end", "center", "3p", "5p")` for a `GRanges` object.
#'
#' @seealso \link{mutate.Ranges}, \link{stretch}
#'
#' @return a RangesAnchored object which has the same appearance as a regular
#' Ranges object but with an additional slot displaying an anchor.
#' @examples
#' df <- data.frame(start = 1:10, width = 5)
#' rng <- as_iranges(df)
#' rng_by_start <- anchor_start(rng)
#' rng_by_start
#' anchor(rng_by_start)
#' mutate(rng_by_start, width = 3L)
#' grng <- as_granges(df,
#'                    seqnames = "chr1",
#'                    strand = c(rep("-", 5), rep("+", 5)))
#' rng_by_5p <- anchor_5p(grng)
#' rng_by_5p
#' mutate(rng_by_5p, width = 3L)
#'
#' @importFrom methods setClass setValidity setMethod show
#' @rdname ranges-anchor
#' @export
anchor <- function(x) { UseMethod("anchor") }

#' @export
anchor.AnchoredGenomicRanges <- function(x) {
  x@anchor
}

#' @export
anchor.AnchoredIntegerRanges <- anchor.AnchoredGenomicRanges

#' @rdname ranges-anchor
#' @export
unanchor <- function(x) { UseMethod("unanchor") }

#' @export
unanchor.AnchoredGenomicRanges <- function(x) {
  x@delegate
}

#' @export
unanchor.AnchoredIntegerRanges <- unanchor.AnchoredGenomicRanges


#' @rdname ranges-anchor
#' @export
anchor_start <- function(x) { UseMethod("anchor_start") }

#' @export
anchor_start.IntegerRanges <- function(x) {
  new_anchored_ir(x, "start")
}

#' @export
anchor_start.GenomicRanges <- function(x) {
  new_anchored_gr(x, "start")
}


#' @rdname ranges-anchor
#' @export
anchor_end <- function(x) { UseMethod("anchor_end") }

#' @export
anchor_end.IntegerRanges <- function(x) {
  new_anchored_ir(x, "end")
}

#' @export
anchor_end.GenomicRanges <- function(x) {
  new_anchored_gr(x, anchor = "end")
}

#' @rdname ranges-anchor
#' @export
anchor_center <- function(x) { UseMethod("anchor_center") }

#' @export
anchor_center.IntegerRanges <- function(x) {
  new_anchored_ir(x, "center")
}

#' @export
anchor_center.GenomicRanges <- function(x) {
  new_anchored_gr(x, "center")
}

#' @rdname ranges-anchor
#' @export
anchor_centre <- function(x) { UseMethod("anchor_center") }

#' @rdname ranges-anchor
#' @export
anchor_3p <- function(x) { UseMethod("anchor_3p") }

#' @export
anchor_3p.GenomicRanges <- function(x) {
  new_anchored_gr(x, "3p")
}

#' @rdname ranges-anchor
#' @export
anchor_5p <- function(x) { UseMethod("anchor_5p") }

#' @export
anchor_5p.GenomicRanges <- function(x) {
  new_anchored_gr(x, "5p")
}

context("stretch ranges")

test_that("matches IRanges", {
  ir1 <- IRanges(c(2,5,1), c(3,7,3))
  # stretch 0 is a no-op
  expect_identical(stretch(ir1), ir1)
  ir2 <- ir1 + 5
  expect_identical(stretch(ir1, 10), ir2)
  ir3 <- ir1
  end(ir3) <- end(ir3) + 10
  expect_identical(stretch(anchor_start(ir1), 10), ir3)
  # will cause negative width
  expect_error(stretch(anchor_end(ir1), -10))
})

test_that("matches GenomicRanges", {
  # default is to anchor by center
  gr1 <- GRanges(seqnames = Rle(factor(c("chr1", "chr2", "chr1", "chr3")),
                               c(1, 3, 2, 4)),
                ranges = IRanges(1:10, width = 10:1),
                strand = Rle(strand(c("-", "+", "*", "+", "-")),
                             c(1, 2, 2, 3, 2)))
  gr2 <- stretch(gr1, 10)
  expect_equal(mid(gr1), mid(gr2))
  expect_identical(gr2, gr1 + 5L )
  expect_identical(stretch(anchor_center(gr1), 10L), gr1 + 5L)
  
  # anchoring by 3p fixes start of - and end of +
  gr3 <- gr1
  pos <- strand(gr3) == "+" | strand(gr3) == "*"
  start(gr3[pos]) <- start(gr3[pos]) - 10
  end(gr3[strand(gr3) == "-"]) <- end(gr3[strand(gr3) == "-"]) + 10
  expect_identical(stretch(anchor_3p(gr1), 10), gr3)
  
  # anchoring by 3p fixes start of + and end of -
  gr4 <- gr1
  pos <- strand(gr4) == "+" | strand(gr4) == "*"
  start(gr4[strand(gr4) == "-"]) <- start(gr4[strand(gr4) == "-"]) - 10
  end(gr4[pos]) <- end(gr4[pos]) + 10
  expect_identical(stretch(anchor_5p(gr1), 10), gr4)
  
  # check ends 
  gr3 <- gr1
  start(gr3) <- start(gr3) - 10L
  expect_identical(stretch(anchor_end(gr1), 10L), gr3)
  
  gr4 <- gr1 
  end(gr4) <- end(gr4) + 10L
  expect_identical(stretch(anchor_start(gr1), 10L), gr4)
  
})

context("Anchored ranges")

test_that("AnchoredIntegerRanges and friends", {
  ir <- IRanges()
  astart <- anchor_start(ir)
  # check class and accessors
  expect_true(is(astart, "AnchoredIntegerRanges"))
  expect_s4_class(astart, "AnchoredIntegerRanges")
  expect_equal(anchor(astart), "start")
  # adding a new anchored overrides existing one
  expect_equal(anchor(anchor_center(astart)), "center")
  # can't anchor by strand for IRanges
  expect_error(anchor_5p(ir))
  expect_error(anchor_3p(ir))
  # stretching empty ranges returns identical range if anchored
  expect_identical(ir, stretch(astart, 5L))

  # anchoring coordinates leaves respective positions fixed
  ir <- IRanges(start = c(1, 20, 25, 25, 33),
                width = c(19, 5, 0, 8, 5))
  expect_equal(start(ir), start(anchor_center(ir)))
  correct_start <- start(ir)
  test_start <- ir %>% anchor_start() %>% stretch(10L) %>% start()
  expect_equal(test_start, correct_start)
  correct_end <- end(ir)
  test_end <- ir %>% anchor_end() %>% stretch(10L) %>% end()
  expect_equal(test_end, correct_end)
  correct_center <- resize(ir, fix = "center", width = 11L)
  test_center <- set_width(anchor_center(ir), 11L)
  expect_identical(correct_center, test_center)
})

test_that("AnchoredGenomicRanges and friends", {

  gr <- GRanges()
  # check class and acessors
  expect_true(is(anchor_5p(gr), "AnchoredGenomicRanges"))
  expect_s4_class(anchor_start(gr), "AnchoredGenomicRanges")
  by_3p <- anchor_3p(gr)
  expect_equal(anchor(by_3p), "3p")
  # unanchored returns error
  expect_error(anchor(gr))

  # anchoring coordinates works as expected
  gr <- GRanges(
    seqnames=Rle(paste("chr", c(1, 2, 1, 3), sep=""), c(1, 3, 2, 4)),
    ranges=IRanges(1:10, width=10:1, names=letters[1:10]),
    strand=Rle(strand(c("-", "+", "*", "+", "-")), c(1, 2, 2, 3, 2)),
    score=1:10,
    GC=seq(1, 0, length=10)
  )

  # check anchoring does not remove any Range info
  by_end <- anchor_end(gr)

  expect_equal(names(by_end), names(gr))
  # modifying width anchors by start
  correct_gr <- gr
  width(correct_gr) <- 5L
  expect_identical(set_width(anchor_start(gr), 5L), correct_gr)

  # anchoring by 5p fixes start for negative strand
  expect_identical(set_width(anchor_5p(gr), 5L),
                   resize(gr, fix = "start", width = 5L))
  # anchoring by 3p fixes end for neg strand
  expect_identical(set_width(anchor_3p(gr), 5L),
                   resize(gr, fix = "end", width = 5L))
  # TODO ranges derivatives work as expected

})

context("resizing ranges")

test_that("resizing with anchoring", {
  ir1 <- IRanges(c(2,5,1), c(3,7,3))

  expect_identical(mutate(ir1, width = 10),
                 IRanges(c(2, 5, 1), width=10))
  expect_identical(mutate(anchor_end(ir1), width = 10),
                 IRanges(c(-6, -2, -6), width=10))
  expect_identical(mutate(anchor_center(ir1), width = 10),
                 IRanges(c(-2, 1, -3), width=10))


  gr <- GRanges(seqnames = Rle(factor(c("chr1", "chr2", "chr1", "chr3")),
                               c(1, 3, 2, 4)),
                ranges = IRanges(1:10, width = 10:1),
                strand = Rle(strand(c("-", "+", "*", "+", "-")),
                             c(1, 2, 2, 3, 2)))
  # ignores strand by defualt
  gr_w <- mutate(gr, width = 10)
  expect_identical(rep(10L, length(gr)), width(gr_w))
  expect_identical(c(1L:10L), start(gr_w))
  ir_c <- ranges(mutate(anchor_center(gr), width = 10))
  expect_identical(ir_c, IRanges(rep(1:5, each=2), width = 10))
  # anchoring at 5p will fix start on + and fix end on negative
  ir_5p <- ranges(mutate(anchor_5p(gr), width = 10))
  expect_identical(ir_5p,
                 IRanges(c(1L, 2L, 3L, 4L, 5L, 6L, 7L, 8L, 1L, 1L),
                         width = 10L))
  # anchoring at 3p will fix start on - and fix end on +
  ir_3p <- ranges(mutate(anchor_3p(gr), width = 10))
  expect_identical(ir_3p,
                   IRanges(c(rep(1L, 8), 9, 10), width = 10L))

})

#' Interweave a pair of Ranges objects together
#'
#' @param left,right Ranges objects.
#' @param .id When supplied a new column that represents the origin column
#' and is linked to each row of the resulting Ranges object.
#'
#' @details The output of `interweave()` takes pairs of Ranges
#' objects and combines them into a single  Ranges object. If an .id
#' argument is supplied, an origin column with name .id is created indicated which side
#' the resulting Range comes from (eit)
#'
#' @examples
#'gr <- as_granges(data.frame(start = 10:15,
#'                             width = 5,
#'                             seqnames = "seq1",
#'                             strand = c("+", "+", "-", "-", "+", "*")))
#' interweave(flank_left(gr, width = 5L), flank_right(gr, width = 5L))
#' interweave(flank_left(gr, width = 5L), flank_right(gr, width = 5L), .id = "origin")
#' @importFrom S4Vectors zipup Pairs
#' @return a Ranges object
#' @export
#' @rdname ranges-interweave
interweave <- function(left, right, .id = NULL) {
  if (!is.null(.id)) {
    stopifnot(is.character(.id) & length(.id) == 1L)
    mcols(left)[[.id]] <- "left"
    mcols(right)[[.id]] <- "right"
  }
  unlist(zipup(Pairs(left, right)))
}


#' Combine Ranges by concatentating them together
#'
#' @param ... Ranges objects to combine. Each argument can be a Ranges object,
#' or a list of Ranges objects.
#' @param .id Ranges object identifier. When .id is supplied a new column
#' is created that links each row to the original Range object. The contents
#' of the column correspond to the named arguments or the names of the list
#' supplied.
#'
#' @note Currently GRangesList or IRangesList objects are not supported.
#' @return a concatenated Ranges object
#' @importFrom rlang dots_values have_name
#' @rdname ranges-bind
#' @examples
#'gr <- as_granges(data.frame(start = 10:15,
#'                             width = 5,
#'                             seqnames = "seq1"))
#'gr2 <- as_granges(data.frame(start = 11:14,
#'                             width = 1:4,
#'                             seqnames = "seq2"))
#'bind_ranges(gr, gr2)
#'
#'bind_ranges(a = gr, b = gr2, .id = "origin")
#'
#'bind_ranges(gr, list(gr, gr2), gr2)
#'
#'bind_ranges(list(a = gr, b = gr2), c = gr, .id = "origin")
#'
#' @export
#' @importFrom methods as
bind_ranges <- function(..., .id = NULL) {
  x <- unlist(rlang::dots_values(...), recursive = TRUE)
  .bind_rng(x, .id)
}

all_genomic_ranges <- function(x) {
  all(vapply(x, function(x) is(x, "GenomicRanges"), logical(1)))
}

all_integer_ranges <- function(x) {
  all(vapply(x, function(x) is(x, "IntegerRanges"), logical(1)))
}

.bind_rng <- function(x, .id = NULL) {
  if (!(all_genomic_ranges(x) | all_integer_ranges(x))) {
    stop("Cannot bind objects of different classes together, ... must be
         all Ranges objects.", 
         call. = FALSE)
  }
  x <- as(x, "List")
  x <- .check_id_var(x, .id)
  x <- .set_id_var(x, .id)
  x
}

# .check_ranges_list <- function(x, .id = NULL) {
#   if (is(x, "RangesList")) return(unlist(x, recursive = TRUE), .id))
#   x
# }

.check_id_var <- function(x, .id = NULL) {
  if (!is.null(.id)) {
    stopifnot(is.character(.id) & length(.id) == 1L)
    any_empty <-  vapply(x, function(x) length(x) == 0, logical(1))
    any_noname <- rlang::have_name(x)
    if (!all(any_empty | any_noname)) {
      x <- Filter(length, x)
      names(x) <- seq_along(x)
    }
  }
  x
}

.set_id_var <- function(x, .id = NULL) {
  res <- suppressWarnings(unlist(x, use.names = FALSE))
  if (!is.null(.id)) {
    mcols(res)[[.id]] <-  Rle(names(x), lengths(x))
  }
  res
}

# ported from HelloRanges
na_cols <- function(x) {
  if (is(x, "IRanges"))
    return(IRanges(0L, -1L))
  if (is(x, "GRanges"))
    return(GRanges(".", IRanges(0L, -1L)))
  if (is(x, "Rle"))
    x <- S4Vectors::decode(x)
  ans <- x[NA_integer_]
  if (is(x, "Rle"))
    ans <- S4Vectors::Rle(ans)
  ans
}

zero_cols <- function(x) {
  empty <- new(class(S4Vectors::decode(x)))
  if (is(x, "Rle")) return(Rle(values = empty))
  empty
}

na_ranges <- function(x) {
  na <- na_cols(x)
  if (length(mcols(x)) > 0) {
    mcols(na) <- DataFrame(lapply(mcols(x), na_cols))
  }
  na
}

na_dframe <- function(dframe, nrows) {
  
  empty <- new("DFrame", nrows = as.integer(nrows))
  
  # IRanges special case for mcols is null
  if (is.null(ncol(dframe))) return(NULL)
  if (ncol(dframe) == 0) return(empty)
  
  tform <- na_cols
  if (nrows == 0) tform <- zero_cols
  
  for (i in colnames(dframe)) {
      empty[[i]] <- tform(dframe[[i]])
  }

  empty
  
}


add_na_seqlevels <- function(x) {
  na <- na_cols(granges(x))
  seqlevels(x) <- union(seqlevels(na), seqlevels(x))
  x
}

.join_overlap_left <- function(x, y, suffix, f_in, distance = FALSE, ...) {
  # generate hits
  hits <- make_hits(x, y, f_in, ...)
  # overlaps found
  left <- x[queryHits(hits), ]
  right <- y[subjectHits(hits), ]
  mcols(left) <- mcols_overlaps_update(left, right, suffix)

  # add distance if requested
  if (distance) {
    # ignore.stand will be passed in the ...
    mcols(left)$distance <- GenomicRanges::distance(left, right, ...)
  }

  # overlaps not found
  only_left <- rep(TRUE, queryLength(hits))
  only_left[queryHits(hits)] <- FALSE

  # ranges object
  rng_only_left <- x[only_left]
  if (distance) {
    # add NA distances for the ranges only on the left
    mcols(rng_only_left)$distance <- rep(NA, length(rng_only_left))
  }
  
  # if there are no right mcols create an empty DataFrame
  # otherwise propagate missing values to the outer frame
  mcols_outer <- na_dframe(mcols(right), sum(only_left))
  
  if (!is.null(mcols(rng_only_left))) {
    mcols(rng_only_left) <- cbind(mcols(rng_only_left), mcols_outer)

  } else {
    mcols(rng_only_left) <- mcols_outer
  }

  names(mcols(rng_only_left)) <- names(mcols(left))

  left_outer <- c(left, rng_only_left)
  left_outer <- left_outer[order(c(queryHits(hits), which(only_left)))]
  left_outer
}

#' @importFrom S4Vectors queryLength decode
#' @importFrom Seqinfo seqlevels seqlevels<-
#' @rdname overlap-joins
#' @export
join_overlap_left <- function(x, y, maxgap, minoverlap, suffix = c(".x", ".y"), distance) {
  UseMethod("join_overlap_left")
}

#' @export
join_overlap_left.IntegerRanges <- function(x,y, maxgap = -1L, minoverlap = 0L, 
  suffix = c(".x", ".y"), distance = FALSE) {
  .join_overlap_left(x,y, suffix, findOverlaps, 
                 maxgap = maxgap, 
                 minoverlap = minoverlap, 
                 type = "any",
                 select = "all")
}

#' @export
join_overlap_left.GenomicRanges <- function(x, y, maxgap = -1L, minoverlap = 0L,
                                            suffix = c(".x", ".y"), distance = FALSE) {
  .join_overlap_left(x,
                     y, 
                     suffix, 
                     findOverlaps, 
                     maxgap = maxgap, 
                     minoverlap = minoverlap, 
                     type = "any",
                     select = "all",
                     ignore.strand = TRUE,
                     distance = distance)
}

#' @rdname overlap-joins
#' @export
join_overlap_left_within <- function(x, y, maxgap, minoverlap, suffix = c(".x", ".y")) {
  UseMethod("join_overlap_left_within")
}

#' @export
join_overlap_left_within.IntegerRanges <- function(x, y, maxgap = -1L, minoverlap = 0L, suffix = c(".x", ".y")) {
  .join_overlap_left(x,y, suffix, findOverlaps, 
                     maxgap = maxgap, 
                     minoverlap = minoverlap, 
                     type = "within",
                     select = "all")
}

#' @export
join_overlap_left_within.GenomicRanges <- function(x, y, maxgap = -1L, minoverlap = 0L, suffix = c(".x", ".y")) {
  .join_overlap_left((x),
                     (y), 
                     suffix, 
                     findOverlaps, 
                     maxgap = maxgap, 
                     minoverlap = minoverlap, 
                     type = "within",
                     select = "all",
                     ignore.strand = TRUE)
}

#' @rdname overlap-joins
#' @export
join_overlap_left_directed <- function(x, y, maxgap, minoverlap, suffix = c(".x", ".y"), distance) {
  UseMethod("join_overlap_left_directed")
}


#' @export
join_overlap_left_directed.GenomicRanges <- function(x, y, maxgap = -1L, minoverlap = 0L,
                                                     suffix = c(".x", ".y"), distance = FALSE) {
  .join_overlap_left((x),
                     (y), 
                     suffix, 
                     findOverlaps, 
                     maxgap = maxgap, 
                     minoverlap = minoverlap, 
                     type = "any",
                     select = "all",
                     ignore.strand = FALSE,
                     distance = distance)
}

#' @rdname overlap-joins
#' @export
join_overlap_left_within_directed <- function(x, y, maxgap, minoverlap, suffix = c(".x", ".y")) {
  UseMethod("join_overlap_left_within_directed")
}


#' @export
join_overlap_left_within_directed.GenomicRanges <- function(x, y, maxgap = -1L, minoverlap = 0L, suffix = c(".x", ".y")) {
  .join_overlap_left(x,
                     y, 
                     suffix, 
                     findOverlaps, 
                     maxgap = maxgap, 
                     minoverlap = minoverlap, 
                     type = "within",
                     select = "all",
                     ignore.strand = FALSE)
}

#' Read a BAM file
#'
#' @param file A connection or path to a BAM file
#' @param index The path to the BAM index file
#' @param paired Whether to treat alignments as paired end (TRUE)
#' or single end (FALSE). Default is FALSE.
#'
#' @details Reading a BAM file is deferred until an action
#' such as using `summarise()` or `mutate()` occurs. If paired is set to
#' TRUE, when alignments are loaded, the GRanges has two additional
#' columns called read_pair_id and read_pair_group corresponding
#' to paired reads and is grouped by the read_pair_group.
#'
#' Certain verbs have different behaviour, after using `read_bam()`.
#' 
#' For `select()` valid columns are the fields available in the
#' BAM file. Valid entries are qname (QNAME), flag (FLAG),
#' rname (RNAME), strand, pos (POS), qwidth (width of query),
#' mapq (MAPQ), cigar (CIGAR), mrnm (RNEXT), mpos (PNEXT), isize
#' (TLEN), seq (SEQ), and qual (QUAL). Any two character tags in the BAM file 
#' are also valid.
#'
#' For `filter()` the following fields are valid, to select the FALSE option
#' place `!` in front of the field:
#' 
#' * `is_paired` Select either unpaired (FALSE) or paired (TRUE) reads.
#' * `is_proper_pair` Select either improperly paired (FALSE) or properly
#' paired (TRUE) reads. This is dependent on the alignment software used.
#' * `is_unmapped_query``	Select unmapped (TRUE) or mapped (FALSE) reads.
#' * `has_unmapped_mate` Select reads with mapped (FALSE) or unmapped (TRUE) mates.
#' * `is_minus_strand` 	Select reads aligned to plus (FALSE) or minus (TRUE) strand.
#' * `is_mate_minus_strand`	Select reads where mate is aligned to plus (FALSE) or
#' minus (TRUE) strand.
#' * `is_first_mate_read`	Select reads if they are the first mate (TRUE) or
#' not (FALSE).
#' * `is_second_mate_read` Select reads if they are the second mate (TRUE) or
#' not (FALSE).
#' * `is_secondary_alignment` Select reads if their alignment status is
#' secondary (TRUE) or not (FALSE). This might be relevant if there are
#' multimapping reads.
#' * `is_not_passing_quality_controls` Select reads that either pass
#' quality controls (FALSE) or that do not (TRUE).
#' * `is_duplicate`	Select reads that are unduplicated (FALSE) or
#' duplicated (TRUE). This may represent reads that are PCR or
#' optical duplicates.
#'
#' @importFrom GenomicAlignments readGAlignments readGAlignmentPairs
#' @importFrom Rsamtools BamFile ScanBamParam
#' @export
#' @examples
#' if (require(pasillaBamSubset)) {
#'    bamfile <- untreated1_chr4()
#'    # nothing is read until an action has been performed
#'    print(read_bam(bamfile))
#'    # define a region of interest
#'    roi <- data.frame(seqnames = "chr4", start = 5e5, end = 7e5) %>%
#'             as_granges()
#'    rng <- read_bam(bamfile) %>% 
#'             select(mapq) %>%
#'             filter_by_overlaps(roi)
#' }
#'
#' @return A DeferredGenomicRanges object
#' @rdname io-bam-read
#' @seealso \code{Rsamtools::\link[Rsamtools:BamFile]{BamFile()}},[GenomicAlignments::readGAlignments()]
read_bam <- function(file, index = file, paired = FALSE) {
  ops <- new_bam_ops(file, index, paired)
  new_DeferredGenomicRanges(GRanges(), ops)
}

#' Shift all coordinates in a genomic interval left or right, upstream or downstream
#'
#' @param x a Ranges object .
#' @param shift the amount to move the genomic interval in the Ranges object by.
#' Either a non-negative integer vector of length 1 or an integer vector
#' the same length as x.
#'
#' @details Shifting left or right will ignore any strand information
#' in the Ranges object, while shifting upstream/downstream will shift coordinates
#' on the positive strand left/right and the negative strand right/left. By
#' default, unstranded features are treated as positive. When
#' using [shift_upstream()] or [shift_downstream()] when the `shift` argument is
#' indexed by the strandedness of the input ranges.
#'  
#'
#' @return a Ranges object with start and end coordinates shifted.
#'
#' @seealso \code{IRanges::\link[IRanges:intra-range-methods]{shift()}},
#' \code{GenomicRanges::\link[GenomicRanges:intra-range-methods]{shift()}}
#' @importFrom IRanges shift
#' @importFrom BiocGenerics "%in%" which
#' @rdname ranges-shift
#' @examples
#' ir <- as_iranges(data.frame(start = 10:15, width = 5))
#' shift_left(ir, 5L)
#' shift_right(ir, 5L)
#' gr <- as_granges(data.frame(start = 10:15,
#'                             width = 5,
#'                             seqnames = "seq1",
#'                             strand = c("+", "+", "-", "-", "+", "*")))
#' shift_upstream(gr, 5L)
#' shift_downstream(gr, 5L)
#' @export
shift_left <- function(x, shift = 0L) {
  .norm_args_shift(x, shift)
  shift_l <- -1L * shift
  shift(x, shift_l)
}

#' @rdname ranges-shift
#' @export
shift_right <- function(x, shift = 0L) {
  .norm_args_shift(x, shift)
  shift(x, shift)
}

#' @rdname ranges-shift
#' @export
shift_upstream <- function(x, shift = 0L) {
  stopifnot(is(x, "GenomicRanges"))
  .norm_args_shift(x, shift)
  neg <- strand(x) == "-"
  pos <- strand(x) %in% c("+", "*")
  if (length(x) == length(shift)) {
    shift_neg <- shift[which(neg)]
    shift_pos <- shift[which(pos)]
    x[neg] <- shift_right(x[neg], shift_neg)
    x[pos] <- shift_left(x[pos], shift_pos)
  } else {
    x[neg] <- shift_right(x[neg], shift)
    x[pos] <- shift_left(x[pos], shift)
  }
  x
}

#' @rdname ranges-shift
#' @export
shift_downstream <- function(x, shift = 0L) {
  stopifnot(is(x, "GenomicRanges"))
  .norm_args_shift(x, shift)
  neg <- strand(x) == "-"
  pos <- strand(x) %in% c("+", "*")
  if (length(x) == length(shift)) {
    shift_neg <- shift[which(neg)]
    shift_pos <- shift[which(pos)]
    x[neg] <- shift_left(x[neg], shift_neg)
    x[pos] <- shift_right(x[pos], shift_pos)
  } else {
    x[neg] <- shift_left(x[neg], shift)
    x[pos] <- shift_right(x[pos], shift)
  }
  x
}

.norm_args_shift <- function(x, shift) {
  stopifnot(all(shift >= 0) && is.numeric(shift))
  if (length(shift) != 1 && length(shift) != length(x)) {
    stop("shift must be of length equal to 1 or x")
  }
}

# test-shift.R
# Adapted from HelloRanges tests.
context("shift")
skip_if_not(
  requireNamespace("HelloRanges", quietly = TRUE),
  message = "'HelloRanges' must be installed to run shift unit tests."
)
oldwd <- getwd()
setwd(system.file("unitTests", "data", "shift", package="HelloRanges"))

genome <- read.delim("tiny.genome", header = FALSE)
genome <- genome_info(genome = "test",
                      seqnames = as.character(genome$V1),
                      seqlengths = genome$V2)

gr <- read_bed("a.bed", genome_info = genome)

test_that("compatible with GenomicRanges shift:", {
  exp <- shift(gr, 5L)
  expect_identical(exp, shift_right(gr, 5L))
  expect_identical(shift(gr, -5L), shift_left(gr, 5L))
  exp <- gr
  exp[strand(exp) == "-"] <- shift(exp[strand(exp) == "-"], 5L)
  exp[strand(exp) == "+"] <- shift(exp[strand(exp) == "+"], -5L)
  expect_identical(exp, shift_upstream(gr, 5L))
  exp <- gr
  exp[strand(exp) == "+"] <- shift(exp[strand(exp) == "+"], 5L)
  exp[strand(exp) == "-"] <- shift(exp[strand(exp) == "-"], -5L)
  expect_identical(exp, shift_downstream(gr, 5L))

})


test_that("shift inputs", {
  # shift must be postive
  expect_error(shift_left(gr, shift = -5))
  # shift must have same length as input ranges
  expect_error(shift_downstream(gr, 1:3))
  # see issue https://github.com/sa-lee/plyranges/issues/73
  # vector input
  exp <- shift(gr, c(5L, 3L))
  expect_identical(shift_right(gr, c(5L, 3L)), exp)
  exp <- shift(gr, c(-5L, -3L))
  expect_identical(shift_left(gr, c(5L, 3L)), exp)
  expect_identical(shift_upstream(gr, c(5,5)), shift_upstream(gr, 5))

  # shift vector is parallel to ranges
  exp <- gr
  exp[strand(exp) == "-"] <- shift(exp[strand(exp) == "-"], 3L)
  exp[strand(exp) == "+"] <- shift(exp[strand(exp) == "+"], -5L)
  expect_identical(exp, shift_upstream(gr, c(5L, 3L)))
  exp <- gr
  exp[strand(exp) == "+"] <- shift(exp[strand(exp) == "+"], 5L)
  exp[strand(exp) == "-"] <- shift(exp[strand(exp) == "-"], -3L)
  expect_identical(exp, shift_downstream(gr, c(5L, 3L)))

  # when there's unstranded ranges
  gr <- GRanges(seqnames="chr1",
                strand=c("+", "-", "+", "*"),
                ranges=IRanges(start=c(3, 8, 11, 5), width=2))
  expect_identical(shift_downstream(gr, c(3, 3, 3, 3)),
                   shift_downstream(gr, 3))

  exp <- gr
  exp[strand(exp) == "-"] <- shift(exp[strand(exp) == "-"], 3)
  exp[strand(exp) %in% c("*", "+")] <- shift(exp[strand(exp) %in% c("*", "+")],
                                             c(-0,-1,-2))
  expect_identical(exp, shift_upstream(gr, c(0,3,1,2)))

  exp <- gr
  exp[strand(exp) == "-"] <- shift(exp[strand(exp) == "-"], -3)
  exp[strand(exp) %in% c("*", "+")] <- shift(exp[strand(exp) %in% c("*", "+")],
                                             c(0,1,2))
  expect_identical(exp, shift_downstream(gr, c(0,3,1,2)))
})

setwd(oldwd)

#' Expand list-columns in a Ranges object
#'
#' @param data A Ranges object
#' 
#' @param ... list-column names to expand then unlist
#' @param .drop Should additional list columns be dropped (default = FALSE)?
#' By default `expand_ranges()` will keep other list columns even if they are nested.
#' @param .id A character vector of length equal to number of list columns.
#' If supplied will create new column(s) with name `.id`
#' identifying the index of the list column (default = NULL).
#' @param .keep_empty If a list-like column contains empty elements, should
#' those elements be kept? (default = FALSE)
#' @param .recursive If there are multiple list-columns, should the columns be 
#' treated as parallel? If FALSE each column will be unnested recursively,
#' otherwise they are treated as parallel, that is each list column has
#' identical lengths. (deafualt = FALSE)
#'
#' @importFrom S4Vectors expand
#'
#' @return a GRanges object with expanded list columns
#' @examples
#' grng <- as_granges(data.frame(seqnames = "chr1", start = 20:23, width = 1000))
#' grng <- mutate(grng, 
#'                exon_id = IntegerList(a = 1, b = c(4,5), c = 3, d = c(2,5))
#'                )
#' expand_ranges(grng)
#' expand_ranges(grng, .id = "name")
#' 
#' # empty list elements are not preserved by default
#' grng <- mutate(grng, 
#'                exon_id = IntegerList(a = NULL, b = c(4,5), c= 3, d = c(2,5))
#'                )
#' expand_ranges(grng)
#' expand_ranges(grng, .keep_empty = TRUE)
#' expand_ranges(grng, .id = "name", .keep_empty = TRUE)
#' 
#' @rdname ranges-expand
#' @export
expand_ranges <- function(data, ..., .drop = FALSE, .id = NULL, .keep_empty = FALSE, .recursive = FALSE) {
  
  list_cols <- get_list_cols(data)
  which_unnest <- unnest_cols(data, list_cols, ...)
  
  if (.drop) {
    list_cols_to_drop <- setdiff(names(list_cols), which_unnest)
    if (length(list_cols_to_drop) > 0) {
      mcols(data) <- mcols(data)[!(names(mcols(data)) %in% list_cols_to_drop)]
    }
  }
  
  if (ncol(mcols(data)) == length(which_unnest)) {
    expand_rng <- S4Vectors::expand(data, 
                                    keepEmptyRows = .keep_empty,
                                    recursive = .recursive)
  } else {
    expand_rng <- S4Vectors::expand(data, 
                                    which_unnest, 
                                    keepEmptyRows = .keep_empty,
                                    recursive = .recursive)
  }
  
  
  
  if (!is.null(.id)) {
    if (length(.id) != length(which_unnest)) {
      stop("`.id` does not have same length as number of list columns.",
           call. = FALSE)
    }
    
    values <- lapply(mcols(data)[, which_unnest, drop = FALSE],
                     extract_index,
                     .keep_empty = .keep_empty)
    
    names(values) <- .id
    values <- do.call("DataFrame", values)
    # for some reason if values is single column DF, expand
    # returns a List, with expanded values in last element
    if (length(.id) == 1) {
      expand_values <- expand(values, 
                              .id, 
                              keepEmptyRows = .keep_empty,
                              recursive = .recursive)
      expand_values <- expand_values[[.id]]
      expand_values <- DataFrame(expand_values)
      names(expand_values) <- .id
    } else {
      expand_values <- S4Vectors::expand(values, 
                                         keepEmptyRows = .keep_empty,
                                         recursive = .recursive )
    }
    
    mcols(expand_rng) <- cbind(mcols(expand_rng), expand_values)
  }
  
  return(expand_rng)
  
}


get_list_cols <- function(data) {
  list_cols_pos <- unlist(Map(function(.) is(., "List"), mcols(data)))
  
  list_cols <- Filter(isTRUE, list_cols_pos)
  
  if (length(list_cols) == 0L) {
    stop("No list columns to unnest.", call. = FALSE)
  }
  
  list_cols
}

unnest_cols <- function(data, list_cols, ...) {
  
  dots <- set_dots_unnamed(...)
  
  dot_names <- unlist(Map(function(.) quo_name(.), dots))
  
  if (any(!(dot_names %in% colnames(mcols(data))))) {
    stop(paste("Input column(s):",
               paste0(dot_names, collapse = ","), "not found."),
         call. = FALSE)
  }
  
  if (length(dot_names) == 0L) {
    which_unnest <- names(list_cols)
  } else {
    which_unnest <- intersect(dot_names, names(list_cols))
    
  }
  if (length(which_unnest) == 0L) {
    stop(paste("Input column(s):",
               paste0(dot_names, collapse = ","), "are not list columns."),
         call. = FALSE)
  }
  
  return(which_unnest)
}

extract_index <- function(col, .keep_empty) {

  v <- names(col)
  if (length(v) == 0) v <- seq_along(col)
  l <- lengths(col)
  if (.keep_empty) l[l == 0] <- 1
  inx <- IRanges::IntegerList(Map(function(i) rep(i, l[i]), seq_along(l)))
  IRanges::extractList(v, inx)
  
}

validGroupedRanges <- function(object) {
  group_names <- colnames(object@group_keys)
  check_valid_groups <- !(group_names %in% tbl_vars(object))
  
  if (any(check_valid_groups)) {
    paste("Invalid groups slot:",
          paste(group_names[check_valid_groups], collapse = ","),
          "not found in data.")
  }
}

#' @rdname group_by-ranges
#' @importFrom IRanges extractList 
#' @importFrom S4Vectors splitAsList
#' @importFrom BiocGenerics unlist
#' @export
setClass("GroupedGenomicRanges",
         slot = c(group_keys = "DFrame", 
                  group_indices = "Rle",
                  n = "integer"),
         contains = c("DelegatingGenomicRanges"),
         validity = 
)

initialize_GroupedRanges <- function(.Object, delegate, group_keys, group_indices, n ) {
  .Object@delegate <- delegate
  .Object@group_keys <- group_keys
  .Object@group_indices <- group_indices
  .Object@n <- n
  .Object
}

#' @importFrom methods setMethod initialize
setMethod("initialize", "GroupedGenomicRanges",
          function(.Object, delegate = GRanges(), group_keys = DataFrame(), group_indices = Rle(), n = integer()) {
            initialize_GroupedRanges(.Object, delegate, group_keys, group_indices, n)
          }
)

show_GroupedRanges <- function(object) {
  groups <- colnames(object@group_keys)
  groups <- paste(groups, collapse = ", ")
  output <- c("", utils::capture.output(show(object@delegate)))
  output[1] <- output[2]
  output[2] <- paste("Groups:", groups, paste0("[", object@n, "]"))
  cat(output, sep = "\n")
}
setMethod("show", "GroupedGenomicRanges", function(object) { 
  show_GroupedRanges(object)
})

# --- group-by backend ---
# generates index for grouping variables
new_grouping <- function(rng,  ..., target = "GroupedGenomicRanges") {
  new_groups <- rlang::enquos(...)
  if (length(new_groups) == 0) return(rng)
  
  new_groups <- rlang::quos_auto_name(new_groups)
  # check if we need to mutate, i.e. if quosure is a call
  update_groups <- Filter(rlang::quo_is_call, new_groups)
  
  if (length(update_groups) > 0) {
    rng <- mutate(rng, !!!update_groups)
  }
  
  
  check_names <- !(names(new_groups) %in% tbl_vars(rng))
  
  if (any(check_names)) {
    stop(paste0("Column `", 
               names(new_groups)[check_names],
               "` is unknown"))
  }
  
  group_df <- select(rng, !!!rlang::syms(names(new_groups)), .drop_ranges = TRUE)
  
  unique <- BiocGenerics::unique(group_df)
  inx <- Rle(BiocGenerics::match(group_df, unique))
  n <- nrow(unique)
  new(target, rng, unique, inx, n)
}

# --- GroupedIntegerRanges ---
#' @rdname group_by-ranges
#' @export
setClass("GroupedIntegerRanges",
         slot = c(group_keys = "DFrame", 
                  group_indices = "Rle",
                  n = "integer"),
         contains = "DelegatingIntegerRanges",
         validity = validGroupedRanges)


setMethod("initialize", "GroupedIntegerRanges", 
          function(.Object, delegate = IRanges(), group_keys = DataFrame(), group_indices = Rle(), n = integer()) {
            initialize_GroupedRanges(.Object, delegate, group_keys,group_indices, n)
            
          })

setMethod("show", "GroupedIntegerRanges", function(object) {
  show_GroupedRanges(object)
})



#' @importFrom dplyr tbl_vars
#' @export
tbl_vars.Ranges <- function(x) {
  c("start", "end", "width", names(mcols(x)))
}

#' @export
tbl_vars.GenomicRanges <- function(x) {
  c("start", "end", "width", "strand", "seqnames", names(mcols(x)))
}

# this enables the use of `::` so we can dispatch on the
# right generics

#' @importFrom methods getGeneric getGenerics
bioc_generics <- function(pkgs = c("BiocGenerics", "IRanges", "S4Vectors")) {
  pkgs <- lapply(pkgs, asNamespace)
  generics <-  getGenerics(pkgs)
  fn <- mapply(getGeneric,
               f = generics@.Data,
               package = generics@package,
               SIMPLIFY = FALSE)
  Filter(function(x) {
    fun = try(x@default, silent = TRUE)
    if (is(fun, "try-error")) FALSE
    !is.primitive(fun)
  },
  fn)
}

# bind the generics to a child env, so a quosure has access to them

scope_plyranges <- function(env, generics) {
  tail <- env
  nms <- character(0)
  # recurse through parent environments until we get to the empty env
  while (!identical(tail, rlang::empty_env())) {
    env_nms <- rlang::env_names(tail)
    nms <- unique(c(nms, intersect(names(generics), env_nms)))
    tail <- rlang::env_parent(tail)
  }
  nms <- setdiff(nms, rlang::env_names(rlang::global_env()))
  generics <- generics[nms]

  child <- rlang::child_env(env, !!!generics)
  child
}

# Given a set of quosures bind the generics as a child environment
quos_set_env <- function(quos, generics = bioc_generics()) {
  for (i in seq_along(quos)) {
    quos[[i]] <- rlang::quo_set_env(
      quos[[i]],
      env = scope_plyranges(rlang::quo_get_env(quos[[i]]), generics)
    )
  }
  quos
}


set_dots_unnamed <- function(...) {
  dots <- rlang::enquos(...)
  quos_set_env(dots)
}

set_dots_named <- function(...) {
  dots <- rlang::enquos(..., .named = TRUE)
  quos_set_env(dots)
}

#' plyranges: a grammar of genomic data manipulation
#'
#' plyranges is a dplyr like API to the Ranges/GenomicRanges infrastructure
#' in Bioconductor.
#'
#' plryanges provides a consistent interface for importing and
#' wrangling genomics data from a variety of sources. The package defines a
#' grammar of genomic data manipulation through a set of verbs. These verbs
#' can be used to construct human readable analysis pipelines based on Ranges
#' objects.
#'
#'   * Modify genomic regions with the `set_width()` and `stretch()` functions.
#'   * Modify genomic regions while fixing the start/end/center coordinates
#'     with the `anchors()` family of functions.
#'   * Sort genomic ranges with `arrange()`.
#'   * Modify, subset, and aggregate genomic data with the `mutate()`,
#'    `filter()`, and `summarise()`functions.
#'   * Any of the above operations can be performed on partitions of the
#'     data with `group_by()`.
#'   * Find nearest neighbour genomic regions with the `join_nearest()` family
#'   of functions.
#'   * Find overlaps between ranges with the `join_overlap_inner()` family of functions.
#'   * Merge all overlapping and adjacent genomic regions with `reduce_ranges()`.
#'   * Merge the end points of all genomic regions with `disjoin_ranges()`.
#'   * Import and write common genomic data formats with the `read_/write_` family
#'    of functions.
#'
#'    For more details on the features of plryanges, read the vignette:
#'    `browseVignettes(package = "plyranges")`
"_PACKAGE"

#' Read a BigWig file
#' @param file A path to a file or URL.
#' @param genome_info An optional character string or a Ranges object
#' that contains information about the genome build. For example the identifier
#'"hg19" will add build information to the returned GRanges.
#' @param overlap_ranges An optional Ranges object. Only the intervals in the file
#' that overlap the Ranges will be loaded.
#'
#' @examples
#' if (.Platform$OS.type != "windows") {
#'   test_path <- system.file("tests", package = "rtracklayer")
#'   bw_file <- file.path(test_path, "test.bw")
#'   gr <- read_bigwig(bw_file)
#'   gr
#' }
#' @return a GRanges object
#' @importFrom rtracklayer import.bw BigWigSelection BigWigFile
#' @seealso \code{rtracklayer::\link[rtracklayer:BigWigFile-class]{BigWigFile()}}
#' @export
#' @rdname io-bigwig-read
read_bigwig <- function(file, genome_info = NULL, overlap_ranges = NULL) {

  stopifnot(is(overlap_ranges, "GRanges") || is.null(overlap_ranges))

  if (!is.null(genome_info)) {
    if (is(genome_info, "GRanges")) {
      seq_info <- seqinfo(genome_info)
    } else if (is.character(genome_info)) {
      seq_info <- Seqinfo::Seqinfo(genome = genome_info)
    }
  } else {
      seq_info <- NULL
  }

  if (!is.null(overlap_ranges)) {
    selection <- BigWigSelection(overlap_ranges)
    rng <- import.bw(file, selection = selection, as = "GRanges")
  } else {
    rng <- import.bw(file, format = "bw", as = "GRanges")
  }

  if (!is.null(seq_info)) {
    seqinfo(rng) <- seq_info
  }
  rng
}

#' Write a BigWig file
#'
#' @param x A GRanges object
#' @param file File name, URL or connection specifying a file to write x to.
#'             Compressed files with extensions such as '.gz' are handled
#'             automatically.
#'
#' @description This is a lightweight wrapper to the export family
#' of functions defined in \pkg{rtracklayer}.
#'
#' @importFrom rtracklayer export.bw
#' @seealso \code{rtracklayer::\link[rtracklayer:BigWigFile-class]{BigWigFile()}}
#' @export
#' @return The write functions return a BigWigFile invisibly
#' @examples 
#' \dontrun{
#'  if (.Platform$OS.type != "windows") {
#'   test_path <- system.file("tests", package = "rtracklayer")
#'   bw_file <- file.path(test_path, "test.bw")
#'   gr <- read_bigwig(bw_file)
#'   gr
#'   bw_out <- file.path(tempdir(), "test_out.bw")
#'   write_bigwig(gr ,bw_out)
#'   read_bigwig(bw_out)
#'  }
#' }
#' @rdname io-bigwig-write
write_bigwig <- function(x, file) {
  export.bw(x, file)
}

context("BigWig files")

# tests adapted from rtracklayer
createCorrectGR <- function() {
  ir <- as(IRanges::PartitioningByWidth(rep(300, 9)), "IRanges")
  space <- factor(c(rep("chr2", 5), rep("chr19", 4)), c("chr2", "chr19"))
  score <- seq(-1, 1, length = 9)
  correct_fixed <- GRanges(space, ir, score = score)
  si <- rtracklayer::SeqinfoForBSGenome("hg19")
  seqlengths(correct_fixed) <- seqlengths(si)[levels(space)]
  correct_fixed
}

test_that("reading/ writing bigwig files returns correct GRanges", {
  skip_on_os(os  = "windows")
  
  # check that import by genome name works
  skip_if_not(
    requireNamespace("BSgenome.Hsapiens.UCSC.hg19", quietly = TRUE),
    message = "'BSgenome.Hsapiens.UCSC.hg19' must be installed to run tests"
  )
  
  test_path <- system.file("tests", package = "rtracklayer")
  test_bw <- file.path(test_path, "test.bw")

  correct_gr <- createCorrectGR()
  test_gr <- read_bigwig(test_bw)

  expect_identical(correct_gr, test_gr)

  # basic writing
  test_bw_out <- file.path(tempdir(), "test_out.bw")
  write_bigwig(correct_gr, test_bw_out)
  on.exit(unlink(test_bw_out))
  test_gr <- read_bigwig(test_bw_out)
  expect_identical(test_gr, correct_gr)

  # bedGraph
  correct_bedgraph <- correct_gr
  width(correct_bedgraph) <- seq(1, 300, length = 9)

  write_bigwig(correct_bedgraph, test_bw_out)
  test_gr <- read_bigwig(test_bw_out)
  expect_identical(test_gr, correct_bedgraph)

  ## overlap ranges
  which <- GRanges(c("chr2", "chr2"), IRanges(c(1, 300), c(400, 1000)))
  correct_which <- filter_by_overlaps(correct_bedgraph, which)
  ranges(correct_which) <- IRanges::ranges(IRanges::intersect(correct_which, which))
  test_gr <- read_bigwig(test_bw_out, overlap_ranges = correct_which)
  expect_identical(test_gr, correct_which)

  ## empty overlap
  which <- GRanges()
  correct_which <- filter_by_overlaps(correct_bedgraph, which)
  test_gr <- read_bigwig(test_bw_out, overlap_ranges = which)
  expect_identical(test_gr, correct_which)

})

#' Group a Ranges by one or more variables
#'
#' @param .data a Ranges object.
#' @param ... Variable names to group by. These can be either metadata columns
#' or the core variables of a Ranges.
#' @param add if `.data` is already a GroupedRanges object, when add = FALSE 
#' the (default), `group_by()` will override existing groups. If add = TRUE, 
#' additional groups will be added.
#' @param x a GroupedRanges object.
#'
#' @description The function `group_by` takes a Ranges object and defines
#' groups by one or more variables. Operations are then performed on the Ranges
#' by their "group". `ungroup()` removes grouping.
#'
#' @details
#' `group_by()` creates a new object of class `GroupedGenomicRanges` if
#' the input is a `GRanges` object or an object of class `GroupedIntegerRanges`
#' if the input is a `IRanges` object. Both of these classes contain a slot
#' called `groups` corresponding to the names of grouping variables. They
#' also inherit from their parent classes, `Ranges` and `GenomicRanges`
#' respectively. `ungroup()` removes the grouping and will return
#' either a `GRanges` or `IRanges` object.
#'
#' @section Accessors:
#' To return grouping variables on a grouped Ranges use either
#' \describe{
#'   \item{`groups(x)`}{Returns a list of symbols}
#'   \item{`group_vars(x)`}{Returns a character vector}
#' }
#'
#' @return The `group_by()` function will return a GroupedRanges object.
#' These have the same appearance as a regular Ranges object but with an
#' additional groups slot.
#'
#'
#' @importFrom dplyr group_by
#' @importFrom rlang quo_name quos syms
#' @importFrom methods new
#' @method group_by GenomicRanges
#' @name group_by-ranges
#' @rdname group_by-ranges
#' @export
#' @examples
#' set.seed(100)
#' df <- data.frame(start = 1:10,
#'                  width = 5,
#'                  gc = runif(10),
#'                  cat = sample(letters[1:2], 10, replace = TRUE))
#' rng <- as_iranges(df)
#' rng_by_cat <- rng %>% group_by(cat)
#' # grouping does not change appearance or shape of Ranges
#' rng_by_cat
#' # a list of symbols
#' groups(rng_by_cat)
#' # ungroup removes any grouping
#' ungroup(rng_by_cat)
#' # group_by works best with other verbs
#' grng <- as_granges(df,
#'                    seqnames = "chr1",
#'                    strand = sample(c("+", "-"), size = 10, replace = TRUE))
#'
#' grng_by_strand <- grng %>% group_by(strand)
#' grng_by_strand
#' # grouping with other verbs
#' grng_by_strand %>% summarise(gc = mean(gc))
#' grng_by_strand %>% filter(gc == min(gc))

#' grng_by_strand %>%
#'   ungroup() %>%
#'   summarise(gc = mean(gc))
#'
#'
group_by.GenomicRanges <- function(.data, ..., add = FALSE) {
  new_grouping(.data, ...)
}

#' @method group_by IntegerRanges
#' @export
group_by.IntegerRanges <- function(.data, ..., add = FALSE) {
  new_grouping(.data, ..., target = "GroupedIntegerRanges")
}

#' @method group_by GroupedGenomicRanges
#' @export
group_by.GroupedGenomicRanges <- function(.data, ..., add = FALSE) {
  dots <- enquos(...)
  
  if (add) {
    gvars <- groups(.data)
    dots <- c(gvars, dots)
  }
  
  new_grouping(.data@delegate, !!!dots, target = class(.data))
}

#' @method group_by GroupedIntegerRanges
#' @export
group_by.GroupedIntegerRanges <- group_by.GroupedGenomicRanges

#' @rdname group_by-ranges
#' @importFrom dplyr ungroup
#' @method ungroup GroupedGenomicRanges
#' @export
ungroup.GroupedGenomicRanges <- function(x, ...) {
  ungroups <- enquos(...)
  ungroups <- rlang::quos_auto_name(ungroups)
  if (length(ungroups) == 0L) {
    return(x@delegate)
  } else {
    gvars <- group_vars(x)
    names(gvars) <- gvars
    groups_update <- tidyselect::eval_select(rlang::expr(-c(...)), gvars)
    if (length(groups_update) == 0) {
      return(x@delegate)
    }
    
    groups_update <- syms(names(groups_update))
    new_grouping(x@delegate, !!!groups_update, target = class(x))
  }
}

#' @method ungroup GroupedIntegerRanges
#' @export
ungroup.GroupedIntegerRanges <- ungroup.GroupedGenomicRanges

#' @method ungroup Ranges
#' @export
ungroup.Ranges <- function(x, ...) x

#' @importFrom dplyr groups
#' @method groups GroupedGenomicRanges
#' @rdname group_by-ranges
#' @export
groups.GroupedGenomicRanges <- function(x)  syms(colnames(x@group_keys))

#' @importFrom dplyr group_vars
#' @method group_vars GroupedGenomicRanges
#' @export
group_vars.GroupedGenomicRanges <- function(x) colnames(x@group_keys) 

#' @method groups GroupedIntegerRanges
#' @rdname group_by-ranges
#' @export
groups.GroupedIntegerRanges <- groups.GroupedGenomicRanges

#' @method group_vars GroupedIntegerRanges
#' @export
group_vars.GroupedIntegerRanges <- group_vars.GroupedGenomicRanges

#' @method groups Ranges
#' @export
groups.Ranges <- function(x)  NULL 

#' @method group_vars Ranges
#' @export
group_vars.Ranges <- function(x) character(0)

#' @method group_keys GroupedGenomicRanges
#' @export
#' @importFrom dplyr group_keys
group_keys.GroupedGenomicRanges <- function(.tbl, ...) {
  .tbl@group_keys
}

#' @method group_keys GroupedIntegerRanges
#' @export
group_keys.GroupedIntegerRanges <- group_keys.GroupedGenomicRanges

#' @method group_keys Ranges
#' @export
group_keys.Ranges <- function(.tbl, ...) {
  if (length(enquos(...)) == 0) {
    return(new("DFrame", nrows = 1L))
  }
  NextMethod(group_by(.tbl, ...))
}

#' @method group_indices GroupedGenomicRanges
#' @export
#' @importFrom dplyr group_indices
group_indices.GroupedGenomicRanges <- function(.data, ...) {
  .data@group_indices
} 

#' @method group_indices GroupedIntegerRanges
#' @export
group_indices.GroupedIntegerRanges <- group_indices.GroupedGenomicRanges


#' @method group_indices Ranges
#' @export
group_indices.Ranges <- function(.data, ...) {
  if (length(enquos(...)) == 0) {
    return(rep.int(1, length(.data)))
  }
  NextMethod(group_by(.data, ...))
}


.group_rows <- function(.data) {
  as(unname(S4Vectors::split(
    seq_len(length(.data@delegate)),
    .data@group_indices
  )),
  "IntegerList"
  )
}

#' @method group_data GroupedGenomicRanges
#' @export
#' @importFrom dplyr group_data  
group_data.GroupedGenomicRanges <- function(.data) {
  S4Vectors::DataFrame(
    .data@group_keys,
    .rows = .group_rows(.data))
}

#' @method group_data GroupedIntegerRanges
#' @export
group_data.GroupedIntegerRanges <- group_data.GroupedGenomicRanges

#' @method group_data Ranges
group_data.Ranges <- function(.data) {
  .rows <- as(seq_len(length(.data)), "IntegerList")
  S4Vectors::DataFrame(.rows = .rows)
}

#' @method group_split GroupedGenomicRanges
#' @export
#' @importFrom dplyr group_split
group_split.GroupedGenomicRanges <- function(.tbl, ..., keep = TRUE) {
  if (length(enquos(...)) > 0) {
    warning("Ignoring arguments to `...` 
            and using existing group structure")
  }
  
  rng <- .tbl@delegate 
  
  if (!keep) {
    vars_drop <- lapply(group_vars(.tbl), function(.) rlang::quo(-!!.))
    rng <- select(rng, !!!vars_drop)
  } 
  
  unname(S4Vectors::split(rng, .tbl@group_indices))
}

#' @method group_split GroupedIntegerRanges
#' @export
group_split.GroupedIntegerRanges <- group_split.GroupedGenomicRanges

#' @method group_split Ranges
#' @export
group_split.Ranges <- function(.tbl, ..., keep = TRUE) {
  if (length(enquos(...)) == 0) {
    return(as(.tbl, "List"))
  }
  NextMethod(group_by(.tbl, ...))
}

#' Extract a single column from a Ranges object as a vector
#'
#' @param .data a `Ranges` object
#' @param var A variable specified as:
#'   * a literal variable name
#'   * a positive integer, giving the position counting from the left. In this
#'   case order is start, end, width, (strand, seqnames), gc and score. 
#'   * a negative integer, giving the position counting from the right.
#'   The default returns the last column (on the assumption that's the column 
#'   you've created most recently). This argument is taken by expression and 
#'   supports quasiquotation (you can unquote column names and column locations).
#' @param name An optional parameter that specifies the column to be used as 
#'   names for a named vector. Specified in a similar manner as `var`.
#' @param ... For use by methods.
#'
#'
#' @importFrom dplyr pull
#' @importFrom rlang enquo quo_is_missing
#' @importFrom tidyselect eval_select
#' @seealso [dplyr::pull()]
#' @name pull-ranges
#' @rdname pull-ranges
#'
#' @examples
#' df <- data.frame(start = 1:10,
#'                  width = 5,
#'                  seqnames = "seq1",
#'                  strand = sample(c("+", "-", "*"), 10, replace = TRUE),
#'                  gc = runif(10),
#'                  score = rpois(10, 2))
#' rng <- as_granges(df)
#'
#' # Pull parts of the range
#' pull(rng, start)
#' # equivalent to start(rng)
#'
#' # Pull by column name
#' pull(rng, gc)
#' pull(rng, score)
#'
#' # Pull by position (positive from left, negative from right)
#' pull(rng, 1)    # First metadata column
#' pull(rng, -1)   # Last metadata column (default)
#' pull(rng, -2)   # Second to last metadata column
#'
#' # Pull with names
#' pull(rng, score, name = gc)
#'
#' @method pull Ranges
#' @exportS3Method dplyr::pull
pull.Ranges <- function(.data, var = -1, name = NULL, ...) {
  
  var <- tidyselect::vars_pull(tbl_vars(.data), !!rlang::enquo(var))
  name <- rlang::enquo(name)
  env <- overscope_ranges(.data) 
  
  pulled <- eval_tidy(rlang::ensym(var), env)
  if (rlang::quo_is_null(name)) {
    return(pulled)
  }
  
  name <- tidyselect::vars_pull(tbl_vars(.data), !!name)
  named <- eval_tidy(rlang::ensym(name), env)
  
  rlang::set_names(pulled, nm = named)
  
}
  

#' @method pull DelegatingGenomicRanges
#' @export
pull.DelegatingGenomicRanges <- function(.data, var = -1, name = NULL, ...) {
  pull(.data@delegate, var = {{var}}, name = {{name}}, ...)
}

#' @method pull DelegatingIntegerRanges
#' @export
pull.DelegatingIntegerRanges <- pull.DelegatingGenomicRanges

#' @method pull GroupedGenomicRanges
#' @export
pull.GroupedGenomicRanges <- function(.data, var = -1, name = NULL, ...) {
  pull(.data@delegate, var = {{var}}, name = {{name}}, ...)
}

#' @method pull GroupedIntegerRanges
#' @export
pull.GroupedIntegerRanges <- pull.GroupedGenomicRanges

# A virtual class that delegates anchoring, so we can decorate
# a Ranges derivative object with an anchor.
# --- GenomicRanges and friends ---
validAnchoredRanges <- function(object, valid_anchors =  c("start", "end", "center", "centre", "5p", "3p")) {  
  if (length(object@anchor) > 1L) {
    paste("anchor must be character vector of length 1.")
  }
  
  if (!(object@anchor %in% valid_anchors)) {
    paste(object@anchor, "is not a valid anchor.")
  }
}

setClass("AnchoredGenomicRanges",
         representation = representation(anchor = "character"),
         contains = "DelegatingGenomicRanges")

setValidity("AnchoredGenomicRanges", function(object) {
  validAnchoredRanges(object)
})

# constructor
initialize_AnchoredRanges <- function(.Object, delegate, anchor) {
  .Object@delegate <- delegate
  .Object@anchor <- anchor
  .Object
}

setMethod("initialize", "AnchoredGenomicRanges", 
          function(.Object, delegate, anchor, ...) {
            initialize_AnchoredRanges(.Object, delegate, anchor)
})

new_anchored_gr <- function(rng, anchor) {
  new("AnchoredGenomicRanges", delegate = rng, anchor = anchor)
}

# mcols method for DelegatingGenomicRanges
setMethod("mcols", "DelegatingGenomicRanges", function(x, ...) {
  mcols(x@delegate, ...)
})

show_AnchoredRanges <- function(object) {
  output <- c("", utils::capture.output(show(object@delegate)))
  output[1] <- output[2]
  anchor <- object@anchor
  output[2] <- paste("Anchored by:", anchor)
  cat(output, sep = "\n")
}
setMethod("show", "AnchoredGenomicRanges", show_AnchoredRanges)

# --- IntegerRanges and friends ---
# Equivalent of DelegatingGenomicRanges for IntegerRanges class
setClass("DelegatingIntegerRanges",
          representation = representation(delegate = "IntegerRanges"),
          contains = c("VIRTUAL", "IntegerRanges"))

# methods for DelegatingIntegerRanges
setMethod("start", "DelegatingIntegerRanges", function(x, ...) start(x@delegate))
setMethod("end", "DelegatingIntegerRanges", function(x, ...) end(x@delegate))
setMethod("width", "DelegatingIntegerRanges", function(x) width(x@delegate))
setMethod("mcols", "DelegatingIntegerRanges", function(x, ...) {
    mcols(x@delegate, ...)
})

#' @importFrom methods callNextMethod
setMethod("mcols<-", "DelegatingGenomicRanges", function(x, ..., value) {
  x@delegate <- callNextMethod(x = x@delegate, ..., value = value)
  x
})

setMethod("mcols<-", "DelegatingIntegerRanges", function(x, ..., value) {
  x@delegate <- callNextMethod(x = x@delegate, ..., value = value)
  x
})

# --- ready for our AnchoredIntegerRanges ---
setClass("AnchoredIntegerRanges",
         slot = c(anchor = "character"),
         contains = "DelegatingIntegerRanges")

setValidity("AnchoredIntegerRanges", function(object) {
  validAnchoredRanges(object, c("start", "end", "center", "centre"))
})

setMethod("initialize", "AnchoredIntegerRanges", 
          function(.Object, delegate, anchor, ...) {
            initialize_AnchoredRanges(.Object, delegate, anchor)
})

setMethod("show", "AnchoredIntegerRanges", show_AnchoredRanges)

new_anchored_ir <- function(rng, anchor) {
  new("AnchoredIntegerRanges", delegate = rng, anchor = anchor)
}

.join_intersect <- function(x, y, suffix, f_in = findOverlapPairs, ...) {

  pairs <- f_in(x, y, ...)
  strand_arg <- quos(...)
  strand_arg <- strand_arg[names(strand_arg) == "ignore.strand"]
  if (length(strand_arg) == 1L) {
      inner_rng <- pintersect(pairs, ignore.strand = eval_tidy(strand_arg[[1]]))
  } else {
    inner_rng <- pintersect(pairs)
  }
  left_rng <- S4Vectors::first(pairs)
  right_rng <- S4Vectors::second(pairs)
  mcols(inner_rng) <- mcols_overlaps_update(left_rng, right_rng, suffix)
  inner_rng
}

#' Join by overlapping Ranges
#'
#' @param x,y Objects representing ranges
#' @param maxgap,minoverlap The maximimum gap between intervals as an integer
#' greater than or equal to zero. The minimum amount of overlap between intervals
#' as an integer greater than zero, accounting for the maximum gap.
#' @param suffix Character to vectors to append to common columns in x and y
#' (default = `c(".x", ".y")`).
#' @param distance logical, whether to add a column named "distance"
#'   containing the distance between the ranges, relevant with non-zero
#' `maxgap` values. Available for [join_overlap_left()] 
#' and the `_directed()` version.
#'
#' @details The function [join_overlap_intersect()] finds
#' the genomic intervals that are the overlapping ranges between x and y and
#' returns a new ranges object with metadata columns from x and y.
#'
#' The function [join_overlap_inner()] is equivalent to [find_overlaps()].
#'
#' The function [join_overlap_left()] performs a left outer join between x
#' and y. It returns all ranges in x that overlap or do not overlap ranges in y
#' plus metadata columns common to both. If there is no overlapping range
#' the metadata column will contain a missing value.
#'
#' The function [join_overlap_self()] find all overlaps between a ranges
#' object x and itself.
#'
#' All of these functions have two suffixes that modify their behavior.
#' The `within` suffix, returns only ranges in x that are completely
#' overlapped within in y. The `directed` suffix accounts for the strandedness
#' of the ranges when performing overlaps.
#'
#' @return a GRanges object
#'
#' @importFrom S4Vectors DataFrame
#' @importFrom IRanges findOverlapPairs
#' @examples
#' x <- as_iranges(data.frame(start = c(11, 101), end = c(21, 201)))
#' y <- as_iranges(data.frame(start = c(10, 20, 50, 100, 1),
#'                            end = c(19, 21, 105, 202, 5)))
#' 
#' # self
#' join_overlap_self(y)
#' 
#' # intersect takes common interval
#' join_overlap_intersect(x,y)
#'
#' # within
#' join_overlap_intersect_within(x,y)
#' 
#' # left, and inner join, it's often useful having an id column here
#' y <- y %>% mutate(id = 1:n()) 
#' x <- x %>% mutate(id = 1:n()) 
#' join_overlap_inner(x,y)
#' join_overlap_left(y,x, suffix = c(".left", ".right"))
#' 
#' @rdname overlap-joins
#' @export
#' @seealso [join_overlap_self()], [join_overlap_left()], [find_overlaps()]
join_overlap_intersect <- function(x, y, maxgap, minoverlap, suffix = c(".x", ".y")) {
  UseMethod("join_overlap_intersect")
}

#' @export
join_overlap_intersect.IntegerRanges <- function(x, y, maxgap = -1L, minoverlap = 0L,
                                          suffix = c(".x", ".y")) {
  .join_intersect(x, y, 
                  suffix, 
                  type = "any", 
                  maxgap = maxgap, 
                  minoverlap = minoverlap)
}

#' @export
join_overlap_intersect.GenomicRanges <- function(x, y, maxgap = -1L, minoverlap = 0L,
                                                 suffix = c(".x", ".y")) {
  .join_intersect(x, y, 
                  suffix, 
                  type = "any", 
                  maxgap = maxgap, 
                  minoverlap = minoverlap,
                  ignore.strand = TRUE)

}
#' @rdname overlap-joins
#' @export
join_overlap_intersect_within <- function(x, y, maxgap, minoverlap,
                                      suffix = c(".x", ".y")) {
  UseMethod("join_overlap_intersect_within")
}

#' @export
join_overlap_intersect_within.IntegerRanges <- function(x, y, maxgap = -1L, minoverlap = 0L,
                                             suffix = c(".x", ".y")) {
  .join_intersect(x, y, 
                  suffix, 
                  type = "within", 
                  maxgap = maxgap, 
                  minoverlap = minoverlap)
}

#' @export
join_overlap_intersect_within.GenomicRanges <- function(x, y, maxgap = -1L, minoverlap = 0L,
                                                    suffix = c(".x", ".y")) {
  .join_intersect(x, y, 
                  suffix, 
                  type = "within", 
                  maxgap = maxgap, 
                  minoverlap = minoverlap,
                  ignore.strand = TRUE)
}


#' @rdname overlap-joins
#' @export
join_overlap_intersect_directed <- function(x, y, maxgap, minoverlap,
                                            suffix = c(".x", ".y")) {
  UseMethod("join_overlap_intersect_directed")
}

#' @export
join_overlap_intersect_directed.GenomicRanges <- function(x, y,
                                                          maxgap = -1L,
                                                          minoverlap = 0L,
                                                          suffix = c(".x", ".y")) {
  .join_intersect(x, y, suffix,
                  type = "any",
                  maxgap = maxgap,
                  minoverlap = minoverlap,
                  ignore.strand = FALSE)
}

#' @rdname overlap-joins
#' @export
join_overlap_intersect_within_directed <- function(x, y, maxgap, minoverlap,
                                            suffix = c(".x", ".y")) {
  UseMethod("join_overlap_intersect_within_directed")
}

#' @export
join_overlap_intersect_within_directed.GenomicRanges <- function(x, y,
                                                          maxgap = -1L,
                                                          minoverlap = 0L,
                                                          suffix = c(".x", ".y")) {
  .join_intersect(x, y, suffix,
                  type = "within",
                  maxgap = maxgap,
                  minoverlap = minoverlap,
                  ignore.strand = FALSE)
}


#' @rdname overlap-joins
#' @export
join_overlap_inner <- function(x, y, maxgap = -1L, minoverlap = 0L, suffix = c(".x", ".y")) {
  UseMethod("find_overlaps")
}

#' @rdname overlap-joins
#' @export
join_overlap_inner_within <- function(x, y, maxgap = -1L, minoverlap = 0L, suffix = c(".x", ".y")) {
  UseMethod("find_overlaps_within")
}

#' @rdname overlap-joins
#' @export
join_overlap_inner_directed <- function(x, y, maxgap = -1L, minoverlap = 0L, suffix = c(".x", ".y")) {
  UseMethod("find_overlaps_directed")
}

#' @rdname overlap-joins
#' @export
join_overlap_inner_within_directed <- function(x, y, maxgap = -1L, minoverlap = 0L, suffix = c(".x", ".y")) {
  UseMethod("find_overlaps_within_directed")
}

# test-overlaps.R
context("overlap joins")

test_that("intersect join returns correct results",{

  # empty intersect returns empty GRanges()
  x <- GRanges()
  expect_identical(join_overlap_intersect(x, x), x)

  x <- IRanges()
  expect_identical(join_overlap_intersect(x, x), x)

  a <- GRanges(seqnames = "chr1",
               ranges = IRanges(c(11,101), c(21, 200)),
               name = c("a1", "a2"),
               strand = c("+", "-"),
               score = c(1,2))
  b <- GRanges(seqnames = "chr1",
               strand = c("+", "-", "+", "-"),
               ranges = IRanges(c(21,91,101,201), c(30,101,110,210)),
               name = paste0("b", 1:4),
               score = 1:4)

  exp <- GRanges("chr1", IRanges(c(21, 101,101), c(21, 101, 110)),
                 strand=c("+", "-", "-"),
                 name.x=c("a1", "a2", "a2"),
                 score.x=c(1, 2, 2),
                 name.y = c("b1", "b2", "b3"),
                 score.y = 1:3)

  target <- join_overlap_intersect(a,b)
  expect_identical(target, exp)

  # with IRanges
  target <- join_overlap_intersect(ranges(a), ranges(b))
  expect_identical(target, ranges(exp))

  # modifying minoverlap and maxgap
  exp <- GRanges("chr1", IRanges(101, 110), strand = "-",
                 name.x = "a2", score.x = 2, name.y = "b3", score.y = 3L)

  target <- join_overlap_intersect(a,b, minoverlap = 2L)
  expect_identical(target, exp)

  target <- join_overlap_intersect(ranges(a), ranges(b), minoverlap = 2L)
  expect_identical(target, ranges(exp))

  exp <- GRanges("chr1", IRanges(c(21, 101, 101, 201), c(21, 101, 110, 200)),
                 strand = c("+", "-", "-", "-"),
                 name.x = c("a1", "a2", "a2", "a2"),
                 score.x = c(1,2,2,2),
                 name.y = c("b1", "b2", "b3", "b4"),
                 score.y = 1:4)
  target <- join_overlap_intersect(a, b, maxgap = 10)
  expect_identical(target, exp)

  # within
  b_with <- b[3]
  exp <- GRanges("chr1", IRanges(101, 110), strand = c("+"),
                 name.x = "b3", score.x = 3L, name.y = "a2", score.y = 2)

  target <- join_overlap_intersect_within(b_with, a)
  expect_identical(target, exp)

  exp <- GRanges() %>%
    mutate(name.x = character(),
           score.x = numeric(),
           name.y = character(),
           score.y = integer())
  seqinfo(exp) <- Seqinfo(seqnames = "chr1")

  target <- join_overlap_intersect_within(a, b)
  expect_identical(target, exp)

  # directed
  exp <- GRanges(seqnames = "chr1",
                 strand = c("+", "-"),
                 IRanges(c(21, 101), c(21, 101)),
                 name.x = c("a1", "a2"),
                 score.x = c(1,2),
                 name.y = c("b1", "b2"),
                 score.y = c(1L,2L))
  target <- join_overlap_intersect_directed(a,b)
  expect_identical(target, exp)

})

test_that("inner join/find overlaps returns correct results", {

  #a# .....
  #b#    ....
  #c#         ..
  #A#  x
  #B#  xx
  #C#          xxx
  a <- IRanges(c(1, 4, 9), c(5, 7, 10)) %>%
    mutate(key = letters[1:3])
  b <- IRanges(c(2, 2, 10), c(2, 3, 12)) %>%
    mutate(key = LETTERS[1:3])

  exp <- IRanges(c(1,1,9), c(5,5,10)) %>%
    mutate(key.a = c("a", "a", "c"),
           key.b = c("B", "A", "C"))

  target <- join_overlap_inner(a,b, suffix = c(".a", ".b"))
  expect_identical(target, exp)
  target <- find_overlaps(a, b, suffix = c(".a", ".b"))
  expect_identical(target, exp)

  # with maxgap = 1
  exp <- IRanges(c(1,1,4,4,9), c(5,5,7,7,10)) %>%
    mutate(key.a = c("a", "a", "b", "b", "c"),
           key.b = c("B", "A", "B", "A","C"))
  target <- join_overlap_inner(a,b, maxgap = 1, suffix = c(".a", ".b"))
  expect_identical(target, exp)
  target <- find_overlaps(a,b, maxgap = 1, suffix = c(".a", ".b"))
  expect_identical(target, exp)

  # minoverlap
  exp <- IRanges() %>%
    mutate(key.a = character(), key.b = character())

  target <- join_overlap_inner(a,b, minoverlap = 3L, suffix = c(".a", ".b"))
  expect_identical(target, exp)
  target <- find_overlaps(a,b, minoverlap = 3L, suffix = c(".a", ".b"))
  expect_identical(target, exp)

  exp <- IRanges(1,5) %>%
    mutate(key.a = "a", key.b = "B")
  target <- join_overlap_inner(a,b, minoverlap = 2L, suffix = c(".a", ".b"))
  expect_identical(target, exp)
  target <- find_overlaps(a,b, minoverlap = 2L, suffix = c(".a", ".b"))

  #a# ..
  #b#     ..
  #c#   ....
  #d#    ......
  #A# xxxx
  #B#   xxxx
  #C#     xxxxx
  #D#      xxxx

  a <- IRanges(c(1, 5, 3, 4), width=c(2, 2, 4, 6)) %>%
    mutate(key.a = letters[1:4])
  b <- IRanges(c(1, 3, 5, 6), width=c(4, 4, 5, 4)) %>%
    mutate(key.b = LETTERS[1:4])

  exp <- IRanges(c(1,5,5,3,3,3,4,4,4), c(2, rep(6, 5), 9,9,9)) %>%
    mutate(key.a = c("a", rep("b", 2), rep("c", 3), rep("d", 3)),
           key.b = c("A", "B", "C", "A", "B", "C", "B", "C", "D"))

  target <- join_overlap_inner(a,b, minoverlap = 2L, suffix = c(".a", ".b"))
  expect_identical(target, exp)
  target <- find_overlaps(a,b, minoverlap = 2L, suffix = c(".a", ".b"))
  expect_identical(target, exp)

  # within
  exp <- IRanges(c(1,5,5,3), c(2,6,6,6)) %>%
    mutate(key.a = c("a", "b", "b", "c"), key.b = c("A", "B", "C", "B"))
  target <- join_overlap_inner_within(a,b, suffix = c(".a", ".b"))
  expect_identical(target, exp)
  target <- find_overlaps_within(a,b, suffix = c(".a", ".b"))
  expect_identical(target, exp)

  # GRanges

  #+# .....
  #-#    ....
  #+#         ..
  #-#  xxxx
  #+#  xxx
  #-#     xx
  #+#      xxxx
  a <- GRanges(seqnames = "chr1",
               strand = c("+", "-", "+"),
               ranges = IRanges(c(1, 4, 9), c(5, 7, 10)),
               key.a = letters[1:3])
  b <- GRanges(seqnames = "chr1",
               strand = c("-", "+", "-", "+"),
               ranges = IRanges(c(2, 2,5, 6), c(5, 4,6, 9)),
               key.b = LETTERS[1:4])
  exp <- GRanges(seqnames = "chr1",
                 strand = c(rep("+", 3), rep("-", 4), "+"),
                 ranges = IRanges(c(rep(1, 3), rep(4, 4), 9),
                                  c(rep(5,3), rep(7, 4), 10)),
                 key.a = c(rep("a", 3), rep("b", 4), "c"),
                 key.b = c(LETTERS[1:3], LETTERS[1:4], "D"))



  target <- join_overlap_inner(a,b, suffix = c(".a", ".b"))
  expect_identical(target, exp)
  target <- find_overlaps(a,b, suffix = c(".a", ".b"))
  expect_identical(target, exp)

  expect_identical(exp, join_overlap_left(a,b,suffix = c(".a", ".b")))

  # directed
  exp <- GRanges(seqnames = "chr1",
                 strand = c("+", "-", "-", "+"),
                 ranges = IRanges(c(1,4,4,9), c(5, 7,7,10)),
                 key.a = c("a", "b", "b", "c"),
                 key.b = c("B", "A", "C", "D"))
  target <- join_overlap_inner_directed(a,b, suffix = c(".a", ".b"))
  expect_identical(target, exp)

  target <- find_overlaps_directed(a,b, suffix = c(".a", ".b"))
  expect_identical(target, exp)
  
  expect_identical(exp, 
                   join_overlap_left_directed(a,b,suffix = c(".a", ".b")))

})

test_that("outer join returns correct results", {
  # left overlap join, returns expected result
  a <- GRanges(seqnames = "chr1",
               strand = c("+", "-"),
               ranges = IRanges(c(1, 9), c(7, 10)),
               key = letters[1:2])
  b <- GRanges(seqnames = "chr1",
               strand = c("-", "+"),
               ranges = IRanges(c(2, 6), c(4, 8)),
               key = LETTERS[1:2])
  exp <- a[c(1,1,2)]
  exp <- select(exp, key.a = key)
  mcols(exp)$key.b <- c("A", "B", NA)
  target <- join_overlap_left(a,b, suffix = c(".a", ".b"))
  
  expect_identical(exp, target)
  
  # if IRanges
  ir0 <- ranges(a)
  
  ir1 <- ranges(b)
  mcols(ir1) <- DataFrame(key = c("A", "B"))
  
  exp <- ir0[c(1,1,2)]
  mcols(exp) <- DataFrame(key = c("A", "B", NA))
  target <- join_overlap_left(ir0, ir1)
  expect_identical(exp, target)
  
  mcols(exp) <- NULL
  mcols(ir1) <- NULL
  expect_identical(join_overlap_left(ir0, ir1), exp)
  
  
  # no overlaps
  expect_identical(join_overlap_left(a, GRanges()), a)
  expect_identical(join_overlap_left(ir0, IRanges()), ir0)
  
  b <- GRanges(seqnames = "chr1",
               strand = c("-", "+"),
               ranges = IRanges(c(20, 11), width = 5),
               key = LETTERS[1:2])
  exp <- select(a, key.a = key)
  mcols(exp)$key.b <- NA_character_
  
  target <- join_overlap_left(a,b, suffix = c(".a", ".b"))
  expect_identical(rep(NA_character_, 2), mcols(target)$key.b)
  expect_identical(exp, target)
  
  # issue 70, no mcols returns result
  # https://github.com/sa-lee/plyranges/issues/70
  gr0 <- GRanges(Rle(c("chr2", "chr5", "chr1", "chr3"), c(1, 3, 2, 4)),
                 IRanges(1:10, width=10:1))
  
  gr1 <- GRanges(Rle(c("chr2", "chr3", "chr1", "chr3"), c(1, 3, 2, 4)),
                 IRanges(1:10, width=10:1))
  
  gr2 <- gr1
  gr2$AAA <- 1L
  
  exp <- join_overlap_left(gr0, gr1)
  target <- granges(join_overlap_left(gr0, gr1))
  
  expect_identical(exp, target)
  

  
  
})

.join_mcols <- function(x, y, by = NULL, all) {
  x_id <- x
  x_id$.id <- factor(seq_along(x))
  if (!is(y, "DataFrame")) {
    y <- DataFrame(y)
  }
  if (is.null(by)) {
    by <- intersect(names(mcols(x_id)), names(y))
    if (length(by) == 0) {
      rlang::abort(
        "`by` must be supplied when `x` and `y` have no common variables."
      )
    }
    message("Joining with `by = ", deparse(by), "`")
  }
  new_mcols <- S4Vectors::merge(mcols(x_id), y, by = by, sort = FALSE, all = all)
  new_mcols <- new_mcols[order(new_mcols$.id), ]
  new_x <- x[as.integer(new_mcols$.id)]
  new_mcols$.id <- NULL
  mcols(new_x) <- NULL
  mcols(new_x) <- new_mcols
  return(new_x)
}


#' Join data by metadata columns
#'
#' These functions enable joins based on metadata
#' of `x` and data in `y`, utilizing
#' [S4Vectors::merge()].
#'
#' @param x Object representing ranges, with metadata
#' columns containing variables for matching
#' @param y A table of data, DataFrame, data.frame or tibble
#' @param by Specifications of the columns used for merging.
#' Passed to `S4Vectors::merge()`
#' @param ... Additional arguments passed to `S4Vectors::merge()`
#' 
#' @details The function [join_mcols_inner()] returns ranges in 
#' `x` only when a match is found in the columns specified with
#' `by` in the table `y`.
#' 
#' The function [join_mcols_left()] returns all ranges in `x`
#' regardless of a match in `y`, with duplications possible
#' from multiple matches.
#' 
#' @return Object representing ranges, with new metadata columns
#'
#' @examples
#'
#' x <- as_granges(data.frame(
#'   seqnames = "chr1", start = 1:5, width=10,
#'   id=c(1:4,4), foo=letters[1:5]
#' ))
#' y <- DataFrame(id=c(1:2,2,4), bar=letters[6:9])
#'
#' # metadata joins:
#' join_mcols_inner(x, y, by="id")
#' join_mcols_left(x, y, by="id")
#'
#' @importFrom dplyr left_join
#' @importFrom rlang .data
#' @importFrom S4Vectors merge
#'
#' @rdname mcols-joins
#' @aliases join_mcols_inner join_mcols_left
#' @export
join_mcols_inner <- function(x, y, by = NULL, ...) {
  .join_mcols(x = x, y = y, by = by, all = FALSE, ...)
}

#' @rdname mcols-joins
#' @export
join_mcols_left <- function(x, y, by = NULL, ...) {
  .join_mcols(x = x, y = y, by = by, all = TRUE, ...)
}

test_that("join_mcols works with different types of y", {

  x <- as_granges(data.frame(seqnames = "chr1", start = 1:5, width=10, id=c(1:4,4), foo=letters[1:5]))
  y <- DataFrame(id=c(1:2,2,4), bar=letters[6:9])
  res1 <- x |> join_mcols_left(y, by="id")

  ydf <- data.frame(id=c(1:2,2,4), bar=letters[6:9])
  res2 <- x |> join_mcols_left(ydf, by="id")

  ytib <- tibble(id=c(1:2,2,4), bar=letters[6:9])
  res3 <- x |> join_mcols_left(ytib, by="id")

  expect_equal(res1,res2)
  expect_equal(res1,res3)

  # check inner vs left
  res_inner <- x |> join_mcols_inner(y, by="id")

  expect_equal(res_inner, res1[!is.na(res1$bar)])

})

# ranges-reduce
group_by_revmap <- function(.data, indexes, groups) {
  group_vars <- syms(c(group_vars(.data), "revmap"))
  .data <- ungroup(.data)
  .data <- .data[unlist(indexes)]
  mcols(.data)[["revmap"]] <- groups
  return(group_by(.data, !!!group_vars))
} 

make_key_rle <- function(x) {
  Rle(as.integer(S4Vectors::runValue(x)), S4Vectors::runLength(x))
}

make_revmap_rle <- function(x) {
  Rle(seq_along(x), elementNROWS(x))
}

unsplice <- function(x, y) x[unlist(y)]



add_revmap_grouping <- function(.data, key, revmap) {
  # lookup indexes for each group
  lookup <- S4Vectors::splitAsList(revmap, key)
  inx <- .group_rows(.data)
  # indexes we will expand by
  indexes <- S4Vectors::mendoapply(unsplice, inx, lookup)
  # groupings from revmap
  groups <- make_revmap_rle(revmap)
  group_by_revmap(.data, indexes, groups)
}

reduce_single <- function(.data, ..., rfun = reduce) {
  dots <- set_dots_named(...)
  if (length(dots) == 0L) {
    return(rfun(.data))
  }

  reduced <- rfun(.data, with.revmap = TRUE)
  
  .data <- group_by_revmap(.data,
                           mcols(reduced)[["revmap"]],
                           make_revmap_rle(mcols(reduced)[["revmap"]]))
  
  sd <- summarise(.data, !!!dots)
  
  sd <- sd[order(sd[["revmap"]]), -which(names(sd) == "revmap")[1], drop = FALSE]
  
  mcols(reduced) <- sd
  reduced
}

reduce_by_grp <- function(.data, ..., rfun = IRanges::reduce) {
  dots <- set_dots_named(...)

  by_groups <- dplyr::group_split(.data)
  
  if (length(dots) == 0L) {
    rng <- IRanges::stack(rfun(by_groups))
    sd <- dplyr::group_keys(.data)
    key <- make_key_rle(mcols(rng)[["name"]])
    mcols(rng) <- sd[key, , drop = FALSE]
    return(rng)
  }
  
  rng <- IRanges::stack(rfun(by_groups, with.revmap = TRUE))

  .data <- add_revmap_grouping(.data, 
                               mcols(rng)[["name"]], 
                               mcols(rng)[["revmap"]])
  
  sd <- summarise(.data, !!!dots)
  sd <- sd[order(sd[["revmap"]]), -which(names(sd) == "revmap"), drop = FALSE]
  
  mcols(rng) <- sd
  rng
}

#' Reduce then aggregate a Ranges object
#'
#' @param .data a Ranges object to reduce
#' @param min.gapwidth Ranges separated by a gap of at least min.gapwidth positions are not merged. 
#' @param ... Name-value pairs of summary functions.
#'
#' @return a Ranges object with the
#' @rdname ranges-reduce
#' @importFrom IRanges reduce
#' @importFrom utils relist
#' @examples
#' set.seed(10)
#' df <- data.frame(start = sample(1:10), 
#'                  width = 5,  
#'                  seqnames = "seq1",
#'                  strand = sample(c("+", "-", "*"), 10, replace = TRUE), 
#'                  gc = runif(10))
#'                  
#' rng <- as_granges(df)
#' rng %>% reduce_ranges()
#' rng %>% reduce_ranges(gc = mean(gc))
#' rng %>% reduce_ranges_directed(gc = mean(gc))
#' rng %>% reduce_ranges_directed(gc = mean(gc), min.gapwidth = 10)
#' 
#' x <- data.frame(start = c(11:13, 2, 7:6), 
#'                width=3, 
#'                id=sample(letters[1:3], 6, replace = TRUE),
#'                score= sample(1:6))
#' x <- as_iranges(x)
#' x %>% reduce_ranges()
#' x %>% reduce_ranges(score = sum(score))
#' x %>% group_by(id) %>% reduce_ranges(score = sum(score))
#' @export
reduce_ranges <- function(.data, min.gapwidth = 1L, ...) { UseMethod("reduce_ranges") }

#' @method reduce_ranges IntegerRanges
#' @export
reduce_ranges.IntegerRanges <- function(.data, min.gapwidth = 1L, ...) {
  reduce_single(.data, ..., 
                rfun = function(x, ...) {
                  reduce(x, ..., min.gapwidth = min.gapwidth)
                })
}

#' @method reduce_ranges GroupedIntegerRanges
#' @export
reduce_ranges.GroupedIntegerRanges <- function(.data, min.gapwidth = 1L, ...) {
  reduce_by_grp(.data, ..., 
                rfun = function(x, ...) {
                  reduce(x, ..., min.gapwidth = min.gapwidth)
                })
}

#' @method reduce_ranges GroupedGenomicRanges
#' @export
reduce_ranges.GroupedGenomicRanges <- function(.data, min.gapwidth = 1L, ...) {
  reduce_by_grp(.data, ..., 
                rfun = function(x, ...) {
                  reduce(x, ..., min.gapwidth = min.gapwidth, ignore.strand = TRUE)
                  })
}

#' @method reduce_ranges GenomicRanges
#' @export
reduce_ranges.GenomicRanges <- function(.data, min.gapwidth = 1L, ...) {
  reduce_single(.data, ..., 
                rfun = function(x, ...) {
                  reduce(x, ..., min.gapwidth = min.gapwidth, ignore.strand = TRUE)
                })
}

#' @rdname ranges-reduce
#' @export
reduce_ranges_directed <- function(.data, min.gapwidth = 1L, ...) {
  UseMethod("reduce_ranges_directed")
}

#' @importFrom IRanges reduce
#' @method reduce_ranges_directed GenomicRanges
#' @export
reduce_ranges_directed.GenomicRanges <- function(.data, min.gapwidth = 1L, ...) {
  reduce_single(.data, ..., 
                rfun = function(x, ...) {
                  reduce(x, ..., min.gapwidth = min.gapwidth, ignore.strand = FALSE)
                })
}

#' @method reduce_ranges_directed GroupedGenomicRanges
#' @export
reduce_ranges_directed.GroupedGenomicRanges <- function(.data, min.gapwidth = 1L, ...) {
  reduce_by_grp(.data, ..., 
                rfun = function(x, ...) {
                  reduce(x, ..., min.gapwidth = min.gapwidth, ignore.strand = FALSE)
                })
}

context("reduce_ranges method")

test_that("matches IRanges/GenomicRanges", {
  x <- IRanges()
  expect_identical(x, reduce_ranges(x))

  x <- IRanges(c(1:4, 10:11, 11), width=c(0,1,1,0,0,0,1))
  mcols(x) <- DataFrame(mapping = paste0("a", seq_along(x)))
  target <- IRanges(c(1:2, 10:11), width=c(0,2,0,1))
  mcols(target) <- DataFrame(mapping=CharacterList("a1", c("a2","a3","a4"), "a5", c("a6","a7")))
  expect_identical(reduce_ranges(x, mapping = mapping),
                   target)

  # drop.empty.ranges is just a filter
  current <- x %>%
    filter(width > 0) %>%
    reduce_ranges(mapping = mapping)
  target <- reduce(x, drop.empty.ranges=TRUE)
  mcols(target) <- DataFrame(mapping = CharacterList(c("a2","a3"), "a7"))

  expect_identical(current, target)

  # -- GRanges
  gr <- GRanges(Rle(factor(c("chr1", "chr2", "chr1", "chr3")), c(1, 3, 2, 4)),
          IRanges(1:10, width=10:1, names=head(letters, 10)),
          Rle(c("-", "+", "*", "+", "-"), c(1, 2, 2, 3, 2)),
          name = paste0("a", 1:10))

  target <- GRanges(Rle(c("chr1", "chr2", "chr3"), c(3, 2, 2)),
                    IRanges(start=c(6, 1, 5, 2, 4, 7, 9), end=10),
                    c("+", "-", "*", "+", "*", "+", "-"))
  expect_identical(reduce_ranges_directed(gr), target)

  mcols(target)$mapping <- CharacterList("a6", "a1", "a5", c("a2","a3"), 
                                         "a4", c("a7","a8"), c("a9","a10"))

  expect_identical(reduce_ranges_directed(gr, mapping = name),
                   target)
  target <- GRanges(Rle(c("chr1", "chr2", "chr3"), c(1,1,1)),
                    IRanges(start = c(1,2,7), end = c(10,10,10)))
  expect_identical(reduce_ranges(gr), target)

  mcols(target)$mapping <- CharacterList(
    c("a1","a5","a6"), 
    c("a2","a3","a4"), 
    c("a7","a8","a9","a10"))

  expect_identical(reduce_ranges(gr, mapping = name),
                   target)

  # reduce with min.gapwidth
  gr <- GRanges(Rle(factor(c("chr1", "chr2")), c(3, 3)),
          IRanges(c(1,6,12,1,6,12), width=4),
          name = paste0("a", 1:6))

  target <- GRanges(Rle(c("chr1", "chr2"), c(2, 2)),
                    IRanges(start=c(1, 12, 1, 12), end=c(9, 15, 9, 15)))

  expect_identical(reduce_ranges(gr, min.gapwidth = 2), target)
  
})


test_that("non-standard evaluation works as expected",{
  oldwd <- getwd()
  setwd(system.file("unitTests", "data", "merge", package="HelloRanges"))

  a <- read_bed("a.bed")

  exp <- reduce(a, ignore.strand = TRUE)

  expect_identical(exp, reduce_ranges(a))

  exp <- reduce(a, with.revmap=TRUE, ignore.strand = TRUE)
  mcols(exp)[["seqnames.count"]] <- lengths(exp$revmap)
  exp <- exp %>% select(-revmap)

  expect_identical(exp, a %>% reduce_ranges(seqnames.count = n()))

  mcols(a)$name <- paste0("a", 1:4)
  exp <- reduce(a, with.revmap=TRUE)
  mcols(exp)[["name.collapse"]] <- CharacterList("a1", c("a2","a3","a4"))
  exp <- exp %>% select(-revmap)

  expect_identical(exp, a %>%
                     reduce_ranges(name.collapse = name)
                   )
  a <- read_bed("a.full.bed")
  exp <- reduce(a, with.revmap=TRUE, ignore.strand=TRUE)
  mcols(exp)[["name.collapse"]] <- CharacterList("a1", c("a2", "a3","a4"), "a1", "a2", c("a3","a4"))
  mcols(exp)[["score.sum"]] <- c(1,9,5,6,15)

  exp <- exp %>% select(-revmap)

  expect_identical(exp, reduce_ranges(a,
                                      name.collapse = name,
                                      score.sum = sum(score)))

  exp <- reduce(a, with.revmap=TRUE, ignore.strand=TRUE)
  mcols(exp)[["score.count"]] <- c(1L,3L,1L,1L,2L)
  mcols(exp)[["score.sum"]] <- c(1,9,5,6,15)
  exp <- exp %>% select(-revmap)

  expect_identical(exp, reduce_ranges(a, score.count = n(),
                                      score.sum = sum(score)))
  setwd(oldwd)
})

test_that("grouping then reducing works as expected", {
  oldwd <- getwd()
  setwd(system.file("unitTests", "data", "multiinter", package="HelloRanges"))
  bed_files <- list.files(pattern = ".bed$")
  # GRangesList
  gr_l <- as(lapply(bed_files, function(x) {
    mutate(read_bed(x), grp = sub(".bed$", "", basename(x)))
  }), "GRangesList")
  names(gr_l) <- sub(".bed$", "", basename(bed_files))
  gr_l_reduced <- reduce(gr_l)
  correct_gr <- IRanges::stack(gr_l_reduced, "grp") %>%
    mutate(grp = as.character(grp))

  # GroupedGRanges
  gr_by_group <- unlist(gr_l, use.names = FALSE) %>% group_by(grp)
  test_gr <- reduce_ranges(gr_by_group)

  expect_identical(correct_gr, test_gr)

  # with an operation matches  revmap length
  gr_l_reduced <- reduce(gr_l, with.revmap = TRUE)
  correct_n <- lengths(IRanges::stack(gr_l_reduced)$revmap)
  test_n <- reduce_ranges(gr_by_group, n = n())$n
  expect_identical(correct_n, test_n)
  setwd(oldwd)
  
})


test_that("expected behaviour for grouped filter w reduce #37",
          {
            # see https://github.com/sa-lee/plyranges/issues/37
            set.seed(2019)
            n <- 10
            r <- GRanges(seqnames = rep("chr1", n),
                         ranges = IRanges(start = sample(20, n, replace = TRUE),
                                          width = sample(6,  n, replace = TRUE))
            )
            mcols(r) <- data.frame(score = runif(n, 0, 100), 
                                   condition = rep_len(c("One","Two"), n),
                                   stringsAsFactors = TRUE)
            red1 <- r %>% group_by(condition) %>% reduce_ranges()
            
            exp <- S4Vectors::split(r, r$condition) %>% 
              reduce() %>%
              unlist()
            exp <- exp %>%
              mutate(condition = as.factor(names(.)))
            names(exp) <- NULL
            
            expect_identical(red1, exp)
            
            red2 <- r %>%  
              group_by(condition) %>% 
              filter(score > 2) %>% 
              reduce_ranges()
            
            exp <- S4Vectors::split(r, r$condition) 
            is_gt2 <- as(lapply(exp, function(x) x$score > 2), "List")
            exp <- exp[is_gt2] %>% reduce() %>% unlist()
            exp <- exp %>%
              mutate(condition = as.factor(names(.)))
            names(exp) <- NULL
            
            expect_identical(red2, exp)
          }
)

#' Functional setters for Ranges objects
#' @details These methods are used internally in `mutate()` to modify 
#' core columns in Ranges objects.
#' @param x a Ranges object
#' @param width integer amount to modify width by
#' @param start integer amount to modify start by
#' @param end integer amount to modify end by
#' @param seqnames update seqnames column
#' @param strand update strand column
#' 
#' @return a Ranges object
#' 
#' @rdname ranges-setters
#' @importFrom IRanges resize
#' @export
set_width <- function(x, width) UseMethod("set_width")

#' @export
set_width.Ranges <- function(x, width = 0L) {
  width(x) <- width
  x
}

#' @export
set_width.AnchoredIntegerRanges <- function(x, width = 0L) {
  anchor <- anchor(x)
  rng <- x@delegate
  switch(anchor,
         start = resize(rng, width, fix = "start"),
         end = resize(rng, width, fix = "end"),
         center = resize(rng, width, fix = "center")
  )
}


#' @export
set_width.AnchoredGenomicRanges <- function(x, width = 0L) {
  anchor <- anchor(x)
  rng <- x@delegate
  switch(anchor,
         start = resize(rng, width, fix = "start", ignore.strand = TRUE),
         end = resize(rng, width, fix = "end", ignore.strand = TRUE),
         center = resize(rng, width, fix = "center", ignore.strand = TRUE),
         "3p" = resize_by_strand(rng, width, "3p"),
         "5p" = resize_by_strand(rng, width, "5p")
  )
}

# anchor by strand version of resize
resize_by_strand <- function(x, width, anchor) {
  # anchor_3p == fix by start for negative strand
  if (anchor == "3p") {
    return(resize(x, width, fix = "end", ignore.strand = FALSE))
  } else {
    return(resize(x, width, fix = "start", ignore.strand = FALSE))
  }
}

#' @rdname ranges-setters
#' @export
set_start <- function(x, start = 0L) { UseMethod("set_start") }

#' @export
set_start.Ranges <- function(x, start = 0L) {
  start(x) <- start
  x
}

#' @rdname ranges-setters
#' @export
set_end <- function(x, end = 0L) { UseMethod("set_end") }

#' @export
set_end.Ranges <- function(x, end = 0L) {
  end(x) <- end
  x
}

#' @rdname ranges-setters
#' @export
set_seqnames <- function(x, seqnames) { UseMethod("set_seqnames") }

#' @export
set_seqnames.GenomicRanges <- function(x, seqnames) {
  seqnames(x) <- seqnames
  x
}

#' @rdname ranges-setters
#' @export
set_strand <- function(x, strand) { UseMethod("set_strand") }

#' @export
set_strand.GenomicRanges <- function(x, strand = "*") {
  strand(x) <- strand
  x
}

# workhorse function for copying metadata columns in y
mcols_overlaps_update <- function(left, right, suffix, return_data_frame = FALSE) {

  left_names <- names(mcols(left))
  right_names <- names(mcols(right))
  common_name <- intersect(left_names, right_names)
  lname_inx <- left_names %in% common_name
  rname_inx <- right_names %in% common_name
  if (any(lname_inx)) {
    names(mcols(left))[lname_inx] <- paste0(left_names[lname_inx], suffix[1])
  }

  if (any(rname_inx)) {
    names(mcols(right))[rname_inx] <- paste0(right_names[rname_inx], suffix[2])
  }

  if (!is.null(mcols(left))) {
    additional_mcols <- mcols(left)
  } else {
    additional_mcols <- NULL
  }

  if (!is.null(mcols(right))) {
    if (is.null(additional_mcols)) {
      additional_mcols <- mcols(right)
    } else {
      additional_mcols <- cbind(additional_mcols, mcols(right))
    }
  }

  if (return_data_frame) {
    if (is(left, "GenomicRanges")) {
      ranges_df <- DataFrame(granges.x = granges(left),
                             granges.y = granges(right))
    } else {
      ranges_df <- DataFrame(ranges.x = ranges(left),
                             ranges.y = ranges(right))
    }
    names(ranges_df) <- paste0(gsub("\\..*", "" , names(ranges_df)), suffix)
    if (!is.null(additional_mcols)) {
      additional_mcols <- cbind(ranges_df, additional_mcols)
    } else {
      return(ranges_df)
    }
  }

  additional_mcols
}

# backend function for olaps and related methods
# takes the pair of ranges, a suffix, optional arguments
# and f_in an anyonymous function that generates a Hits object
make_hits <- function(x, y, f_in, ...) {
  f_in(x, y, ...)
}

# expand x,y, based on Hits object
# hits_mcols_to_keep is a named vector 
# where names correspond to colname in left, and value corresponds to colname in hits mcols
# values from hits will be added to colname in left, error if already filled
expand_by_hits <- function(x, y, suffix, hits, return_data_frame = FALSE, hits_mcols_to_keep = NULL) {
  if (is(hits, "Hits")) {
    left <- x[queryHits(hits), ]
    right <- y[subjectHits(hits), ]
  } else {
    # for vector case
    no_hits_id <- !is.na(hits)
    left <- x[no_hits_id, ]
    right <- y[hits[no_hits_id], ]
  }

  if (return_data_frame) {
    left <- add_hits_mcols(left, hits, hits_mcols_to_keep)
    return(mcols_overlaps_update(left, right, suffix, return_data_frame))
  }
  
  mcols(left) <-  mcols_overlaps_update(left, right, suffix)
  
  add_hits_mcols(left, hits, hits_mcols_to_keep)
}

# For adding hits metadata to mcols in `left` hits_mcols_to_keep if set to a
# vector will take those named columns from hits and append them to `left`. If
# hits_mcols_to_keep is a named vector, the names will correspond to the columns
# in `left` that will hold the values of the columns in `hits`.
add_hits_mcols <- function(left, hits, hits_mcols_to_keep = NULL){
  if (!is.null(hits_mcols_to_keep)){
    
    # If unnamed, use hits mcols names as new column names
    if (is.null(names(hits_mcols_to_keep))) {
      names(hits_mcols_to_keep) <- hits_mcols_to_keep
    }
    
    hit_meta <- mcols(hits)[hits_mcols_to_keep]
    names(hit_meta) <- names(hits_mcols_to_keep)
    
    if (names(hits_mcols_to_keep) %in% names(mcols(left))) {
      bad_names <- names(hits_mcols_to_keep)[names(hits_mcols_to_keep) %in% names(mcols(left))]
      bad_names <- paste(bad_names, collapse = ", ")
      error <- paste0("Cannot join metadata. The following columns already exist in input: ", bad_names)
      stop(error)
    }
    
    if (is.null(mcols(left))) {
      # handles IRanges NULL mcols
      mcols(left) <- hit_meta
    } else {
      mcols(left) <- cbind(mcols(left), hit_meta)
    }
    
  }
  return(left)
}

.find_overlaps <- function(x,y, suffix, f_in, ...) {
  hits <- make_hits(x, y, f_in, ...)
  expand_by_hits(x,y, suffix, hits)
}

#' Find overlap between two Ranges
#'
#' @rdname ranges-overlaps
#'
#' @param x,y Objects representing ranges
#' @param maxgap,minoverlap The maximimum gap between intervals as an integer
#' greater than or equal to negative one. The minimum amount of overlap between intervals
#' as an integer greater than zero, accounting for the maximum gap.
#' @param suffix A character vector of length two used to identify metadata columns
#' coming from x and y.
#'
#' @details `find_overlaps()` will search for any overlaps between ranges
#' x and y and return a Ranges object of length equal to the number of times x
#' overlaps y. This  Ranges object will have additional metadata columns 
#' corresponding to the metadata columns in y. `find_overlaps_within()` is
#' the same but will only search for overlaps within y. For GRanges objects strand is
#' ignored, unless `find_overlaps_directed()` is used. If the Ranges objects have no
#' metadata, one could use `group_by_overlaps()` to be able to
#' identify the index of the input Range x that overlaps a Range in y. 
#' Alternatively,
#' `pair_overlaps()` could be used to place the x ranges next to the range
#' in y they overlap.
#'
#' @return A Ranges object with rows corresponding to the
#' ranges in x that overlap y.  In the case of `group_by_overlaps()`, returns
#' a GroupedRanges object, grouped by the number of overlaps
#' of ranges in x that overlap y (stored in a column called query).
#'
#' @examples
#' query <- data.frame(start = c(5,10, 15,20), width = 5, gc = runif(4)) %>%
#'              as_iranges()
#' subject <- data.frame(start = 2:6, width = 3:7, label = letters[1:5]) %>%
#'              as_iranges()
#'
#' find_overlaps(query, subject)
#' find_overlaps(query, subject, minoverlap = 5)
#' find_overlaps_within(query, subject) # same result as minoverlap
#' find_overlaps(query, subject, maxgap = 1)
#'
#' # -- GRanges objects, strand is ignored by default
#' query  <- data.frame(seqnames = "chr1",
#'                start = c(11,101),
#'                end = c(21, 200),
#'                name = c("a1", "a2"),
#'                strand = c("+", "-"),
#'                score = c(1,2)) %>%
#'            as_granges()
#' subject <- data.frame(seqnames = "chr1",
#'                       strand = c("+", "-", "+", "-"),
#'                       start = c(21,91,101,201),
#'                       end = c(30,101,110,210),
#'                       name = paste0("b", 1:4),
#'                       score = 1:4) %>%
#'                    as_granges()
#'
#' # ignores strandedness
#' find_overlaps(query, subject, suffix = c(".query", ".subject"))
#' find_overlaps(query, subject, suffix = c(".query", ".subject"), minoverlap = 2)
#' # adding directed prefix includes strand
#' find_overlaps_directed(query, subject, suffix = c(".query", ".subject"))
#'
#' @seealso \code{IRanges::\link[IRanges:findOverlaps-methods]{findOverlaps()}},
#' \code{GenomicRanges::\link[GenomicRanges:findOverlaps-methods]{findOverlaps()}}
#' @importFrom IRanges findOverlaps
#' @importFrom S4Vectors queryHits subjectHits
#' @export
find_overlaps <- function(x, y, maxgap, minoverlap, suffix = c(".x", ".y")) {
  UseMethod("find_overlaps")
}

#' @rdname ranges-overlaps
#' @export
find_overlaps.IntegerRanges <- function(x, y, maxgap = -1L, minoverlap = 0L, suffix = c(".x", ".y")) {
  .find_overlaps(x,y, suffix, findOverlaps, 
                 maxgap = maxgap, 
                 minoverlap = minoverlap, 
                 type = "any",
                 select = "all")
}

#' @rdname ranges-overlaps
#' @export
find_overlaps.GenomicRanges <- function(x, y, maxgap = -1L, minoverlap = 0L, suffix = c(".x", ".y")) {
  .find_overlaps(x,y, suffix, 
                 findOverlaps, 
                 maxgap = maxgap, 
                 minoverlap = minoverlap, 
                 type = "any",
                 select = "all",
                 ignore.strand = TRUE)
}

#' @rdname ranges-overlaps
#' @export
find_overlaps_within <- function(x, y, maxgap, minoverlap, suffix = c(".x", ".y")) {
  UseMethod("find_overlaps_within")
}

#' @rdname ranges-overlaps
#' @export
find_overlaps_within.IntegerRanges <- function(x,y, maxgap = -1L, minoverlap = 0L, suffix = c(".x", ".y")) {
  .find_overlaps(x,y, suffix, findOverlaps, 
                 maxgap = maxgap, 
                 minoverlap = minoverlap, 
                 type = "within",
                 select = "all")
}

#' @rdname ranges-overlaps
#' @export
find_overlaps_within.GenomicRanges <- function(x,y, maxgap = -1L, minoverlap = 0L, suffix = c(".x", ".y")) {
  .find_overlaps(x,y, suffix, findOverlaps, 
                 maxgap = maxgap, 
                 minoverlap = minoverlap, 
                 type = "within",
                 select = "all",
                 ignore.strand = TRUE)
}

#' @rdname ranges-overlaps
#' @export
find_overlaps_directed <- function(x, y, maxgap, minoverlap, suffix = c(".x", ".y")) {
  UseMethod("find_overlaps_directed")
}

#' @rdname ranges-overlaps
#' @export
find_overlaps_directed.GenomicRanges <- function(x,y, maxgap = -1L, minoverlap = 0L, suffix = c(".x", ".y")) {
  .find_overlaps(x,y, suffix, findOverlaps, 
                 maxgap = maxgap, 
                 minoverlap = minoverlap, 
                 type = "any",
                 select = "all",
                 ignore.strand = FALSE)
}

#' @rdname ranges-overlaps
#' @export
find_overlaps_within_directed <- function(x, y, maxgap, minoverlap, suffix = c(".x", ".y")) {
  UseMethod("find_overlaps_within_directed")
}

#' @rdname ranges-overlaps
#' @export
find_overlaps_within_directed.GenomicRanges <- function(x, y, maxgap, minoverlap, suffix = c(".x", ".y")) {
  .find_overlaps(x,y, suffix, findOverlaps, 
                 maxgap = maxgap, 
                 minoverlap = minoverlap, 
                 type = "within",
                 select = "all",
                 ignore.strand = FALSE)
}

#' @method select BamFileOperator
#' @importFrom Rsamtools bamWhat<- bamTag<- scanBamWhat
#' @export
select.BamFileOperator <- function(.data, ..., .drop_ranges = FALSE) {
  dots <- set_dots_unnamed(...)
  
  # populate bam params
  all_fields_tags <- unlist(lapply(dots, quo_name))
  
  tags_inx <- vapply(all_fields_tags,
                     FUN = function(x) { nchar(x) == 2L},
                     FUN.VALUE = logical(1))
  
  tags <- all_fields_tags[tags_inx]
  if (length(tags) > 0 ) bamTag(.data@param) <- tags
  
  fields <- all_fields_tags[!tags_inx]
  
  if (length(fields) > 0) {
    if (any(!(fields %in% Rsamtools::scanBamWhat()))) {
      stop("Invalid field identifier.", call. = FALSE)
    }
    bamWhat(.data@param) <- fields
  }
  return(.data)
}

# -- filter method for BamFileOperator
valid_flag_filters <- function() {
  list(
    is_paired = "isPaired",
    is_proper_pair = "isProperPair",
    is_unmapped_query = "isUnmappedQuery",
    has_unmapped_mate = "hasUnmappedMate",
    is_minus_strand = "isMinusStrand",
    is_mate_minus_strand = "isMateMinusStrand",
    is_first_mate_read = "isFirstMateRead",
    is_second_mate_read = "isSecondMateRead",
    is_secondary_alignment = "isSecondaryAlignment",
    is_not_passing_quality_controls = "isNotPassingQualityControls",
    is_duplicate = "isDuplicate"
  )
}

#' @importFrom Rsamtools bamMapqFilter<-
#' @export
filter.BamFileOperator <- function(.data, ...) {
  dots <- set_dots_unnamed(...)
  filters <- unlist(valid_flag_filters())
  
  flags <- unlist(lapply(dots, quo_name))
  
  # mapq filter 
  mapq_filter <- grepl("mapq", flags)
  mapq_args <- flags[mapq_filter]
  if (length(mapq_args) > 0) {
    mapq_vals <- as.integer(sub(".*>", "", mapq_args))
    if (any(is.na(mapq_vals))) {
      message("mapq filter only accepts greater than statements")
    }
    mapq_vals <- mapq_vals[!is.na(mapq_vals)]
    if (length(mapq_vals) > 1) {
      message("mapq filter only accepts singular mapq filters; taking first
              argument")
    } 
    bamMapqFilter(.data@param) <- mapq_vals
  }
  
  # flag filters
  args <- !grepl("^!", flags[!mapq_filter])
  names(args) <- gsub("^!", "", flags[!mapq_filter])
  
  valid_flags <- intersect(names(filters), names(args))
  
  if (length(valid_flags) == 0L & length(mapq_args) == 0L) {
    stop("no valid flags found in filter", call. = FALSE)
  }
  
  if (any(!(names(args) %in% valid_flags))) {
    invalid_flag <- names(args)[!(names(args) %in% valid_flags)]
    stop(paste(paste(invalid_flag, collapse = ","), "are not valid flags"),
         call. = FALSE)
  }
  names(args) <- filters[valid_flags]
  args <- as.list(args)
  sam_flags <- do.call("scanBamFlag", 
                       args, 
                       envir = getNamespace("Rsamtools"))
  bamFlag(.data@param) <- sam_flags
  .data
}

#' @importFrom Rsamtools bamWhich<-
#' @export
filter_by_overlaps.BamFileOperator <- function(x, y, maxgap = -1L, minoverlap = 0L) {
  stopifnot(is(y, "GenomicRanges"))
  bamWhich(x@param) <- y
  x
}

#' @rdname ranges-overlaps
#' @importFrom rlang syms
#' @export
group_by_overlaps <- function(x, y, maxgap, minoverlap) { UseMethod("group_by_overlaps") }

#' @rdname ranges-overlaps
#' @export
group_by_overlaps.IntegerRanges <- function(x, y, maxgap = -1L, minoverlap = 0L) {
  hits <- make_hits(x, y, findOverlaps, maxgap = maxgap, minoverlap = minoverlap)
  left <- expand_by_hits(x, y, c(".query", ".subject"), hits)
  mcols(left)$query <- queryHits(hits)
  group_by(left, !!!rlang::syms("query"))
}

#' @rdname ranges-overlaps
#' @export
group_by_overlaps.GenomicRanges <- function(x, y, maxgap = -1L, minoverlap = 0L) {
  hits <- make_hits(x, y, findOverlaps, 
                    maxgap = maxgap, 
                    minoverlap = minoverlap,
                    ignore.strand = TRUE)
  left <- expand_by_hits(x, y, c(".query", ".subject"), hits)
  mcols(left)$query <- queryHits(hits)
  group_by(left, !!!rlang::syms("query"))
}

# TODO add in more variants here?

# test-overlaps.R
# adapted from IRanges/GRanges tests
context("overlaps")

make_subject <- function(class = c("GRanges", "IRanges")) {
  class <- match.arg(class)

  switch(class,
         GRanges = GRanges(seqnames = Rle(factor(c("chr1", "chr2", "chr1", "chr3")), c(1, 3, 2, 4)),
                          ranges = IRanges(1:10, width = 10:1),
                          strand = Rle(strand(c("-", "+", "+", "-", "-", "-")), c(1, 2, 1, 1, 3, 2)),
                          seqinfo = Seqinfo(seqnames = paste("chr", 1:3, sep="")),
                          score = 1:10,
                          GC = seq(1, 0, length=10)),
         IRanges = IRanges(c(2, 2, 10), c(2, 3, 12))
  )
}


make_query <- function(class = c("GRanges", "IRanges")) {
  class <- match.arg(class)

  switch(class,
         GRanges = GRanges(seqnames = c("chr1", "chr3", "chr1"),
                           ranges = IRanges(start=c(5,2,1), end=c(10, 7, 5)),
                           strand = c("+", "-", "-"),
                           match = c("no", "one", "two")),
         IRanges = IRanges(c(1, 4, 9), c(5, 7, 10))
  )
}


test_that("group_by_overlaps matches queryHits", {

  subject_gr <- make_subject("GRanges")
  query_gr <- make_query("GRanges")

  # strand is ignored unless method is directed
  hits <- findOverlaps(query_gr, subject_gr, ignore.strand = TRUE)

  expect_equal(group_by_overlaps(query_gr, subject_gr)$query,
               queryHits(hits))

  subject_ir <- make_subject("IRanges")
  query_ir <- make_query("IRanges")

  hits <- findOverlaps(query_ir, subject_ir)

  expect_equal(mcols(group_by_overlaps(query_ir, subject_ir))$query,
               queryHits(hits))

})

# io-bed.R
norm_args_reader <- function(genome_info) {

  if (is(genome_info, "GenomicRanges")) {
    seq_info <- seqinfo(genome_info)
    genome_info <- NA
  } else if (is.character(genome_info)) {
    seq_info <- NULL
  } else {
    genome_info <- NA
    seq_info <- NULL
  }

  return(list(genome_info = genome_info, seq_info = seq_info))
}

#' Read a BED or BEDGraph file
#'
#' @param file A path to a file or a connection.
#' @param col_names An optional character vector for including additional
#' columns in `file` that are not part of the BED/narrowPeaks specification.
#' @param genome_info An optional character string or a Ranges object
#' that contains information about the genome build. For example the USSC identifier
#'"hg19" will add build information to the returned GRanges.
#' @param overlap_ranges An optional Ranges object. Only the intervals in the file
#' that overlap the Ranges will be returned.
#'
#' @description This is a lightweight wrapper to the import family
#' of functions defined in \pkg{rtracklayer}.
#'
#' @description Read common interval based formats as GRanges.
#'
#' @details This is a lightweight wrapper to the import family
#' of functions defined in \pkg{rtracklayer}.
#' The `read_narrowpeaks` function parses the ENCODE narrowPeak BED format (see
#' <https://genome.ucsc.edu/FAQ/FAQformat.html#format12> for details.). As
#' such the parser expects four additional columns called (corresponding to
#' the narrowPeaks spec):
#' \itemize{
#'   \item signalValue
#'   \item pValue
#'   \item qValue
#'   \item peak
#' }
#'
#' @return A GRanges object
#'
#' @importFrom rtracklayer import.bed
#' @importFrom Seqinfo seqinfo
#' @importFrom methods is
#' @seealso \code{rtracklayer::\link[rtracklayer:BEDFile-class]{BEDFile()}}
#'
#' @examples
#' test_path <- system.file("tests", package = "rtracklayer")
#' bed_file <- file.path(test_path, "test.bed")
#' gr <- read_bed(bed_file)
#' gr
#' gr <- read_bed(bed_file, genome_info = "hg19")
#' gr
#' olap <-  as_granges(data.frame(seqnames = "chr7", start = 1, end = 127473000))
#' gr <- read_bed(bed_file,
#'               overlap_ranges = olap)
#' # bedGraph
#' bg_file <- file.path(test_path, "test.bedGraph")
#' gr <- read_bed_graph(bg_file)
#' gr
#' # narrowpeaks
#' np_file <- system.file("extdata", "demo.narrowPeak.gz",  package="rtracklayer")
#' gr <- read_narrowpeaks(np_file, genome_info = "hg19")
#' gr
#' @export
#' @rdname io-bed-read
read_bed <- function(file, col_names = NULL, genome_info = NULL,
                     overlap_ranges = NULL) {
  # check genome_info input
  args <- norm_args_reader(genome_info)

  import.bed(file, colnames = col_names,
             genome = args$genome_info,
             seqinfo = args$seq_info,
             which = overlap_ranges,
             trackLine = FALSE)
}


#' Write a BED or BEDGraph file
#'
#' @param x A GRanges object
#' @param file File name, URL or connection specifying a file to write x to.
#'             Compressed files with extensions such as '.gz' are handled
#'             automatically. If you want to index the file with tabix use the
#'             `index` argument.
#' @param index Compress and index the output file
#'              with bgzf and tabix (default = FALSE). Note that tabix indexing will sort the
#'              data by chromosome and start.
#'
#' @description This is a lightweight wrapper to the export family
#' of functions defined in \pkg{rtracklayer}.
#'
#' @importFrom rtracklayer export.bed
#' @seealso \code{rtracklayer::\link[rtracklayer:BEDFile-class]{BEDFile()}}
#' @export
#' @return The write functions return a BED(Graph)File invisibly
#' @rdname io-bed-write
#' @examples 
#' \dontrun{
#'   test_path <- system.file("tests", package = "rtracklayer")
#'   bed_file <- file.path(test_path, "test.bed")
#'   gr <- read_bed(bed_file)
#'   bed_file_out <- file.path(tempdir(), "new.bed")
#'   write_bed(gr, bed_file_out)
#'   read_bed(bed_file_out)
#'   #' bedgraph
#'   bg_file <- file.path(test_path, "test.bedGraph")
#'   gr <- read_bed_graph(bg_file)
#'   bg_file_out <- file.path(tempdir(), "new.bg")
#'   write_bed(gr, bg_file_out)
#'   read_bed(bg_file_out)
#'   # narrowpeaks
#'   np_file <- system.file("extdata", "demo.narrowPeak.gz",package="rtracklayer")
#'   gr <- read_narrowpeaks(np_file, genome_info = "hg19")
#'   np_file_out <- file.path(tempdir(), "new.bg")
#'   write_narrowpeaks(gr, np_file_out)
#'   read_narrowpeaks(np_file_out)
#' }
write_bed <- function(x, file, index = FALSE) {
  export.bed(x, file, index = index)
}

#' @rdname io-bed-read
#' @export
#' @importFrom rtracklayer import.bedGraph
read_bed_graph <- function(file, col_names = NULL, genome_info = NULL,
                           overlap_ranges = NULL) {

  args <- norm_args_reader(genome_info)

  import.bedGraph(file, colnames = col_names,
                  genome = args$genome_info,
                  seqinfo = args$seq_info,
                  which = overlap_ranges,
                  trackLine = FALSE)
}


#' @importFrom rtracklayer export.bedGraph
#' @export
#' @rdname io-bed-write
write_bed_graph <- function(x, file, index = FALSE) {
  export.bedGraph(x, file, index = index)
}

#' @export
#' @rdname io-bed-read
read_narrowpeaks <- function(file,
                             col_names = NULL,
                             genome_info = NULL,
                             overlap_ranges = NULL) {

  args <- norm_args_reader(genome_info)

  extra_cols <- c(signalValue="numeric",
                  pValue="numeric",
                  qValue="numeric",
                  peak="integer")

  import.bed(file,
             colnames = col_names,
             extraCols = extra_cols,
             genome = args$genome_info,
             seqinfo = args$seq_info,
             which = overlap_ranges)
}


#' @export
#' @rdname io-bed-write
write_narrowpeaks <- function(x, file) {
  # at the moment just use write.table
  extra_cols <- c("signalValue", "pValue", "qValue", "peak")
  valid_np <- all(extra_cols %in% names(mcols(x)))
  if (!valid_np) {
    stop(paste("For a valid narrowPeaks there must be columns called:",
               paste(extra_cols, collapse = ","), "in x."),
         call. = FALSE)
  }

  # bed files are 0-based
  rng <- DataFrame(chr = seqnames(x),
                   start = start(x) - 1L,
                   end = end(x),
                   strand = gsub("\\*", "\\.", as.character(strand(x))))

  np_df <- mcols(x)[, extra_cols]

  np_name <- try(mcols(x)[, "name", drop = FALSE], silent = TRUE)
  if (is(np_name, "try-error")) {
    np_name <- DataFrame(name = rep(".", nrow(rng)))
  }

  np_score <- try(mcols(x)[, "score", drop = FALSE], silent = TRUE)
  if (is(np_score, "try-error")) {
    np_score <- DataFrame(score = rep(0L, nrow(rng)))
  }
  np_col_order <- c("chr", "start", "end", "name", "score", "strand",
                    extra_cols)

  np_df <- cbind(rng, np_name, np_score, np_df)[, np_col_order]

  utils::write.table(np_df,
                     file,
                     sep = "\t",
                     row.names = FALSE,
                     col.names = FALSE,
                     na = ".",
                     quote = FALSE)
}

context("reading/writing bedGraph files")
# tests adapted from rtracklayer

createCorrectGR <- function(seqinfo) {
  ir <- shift(IRanges(end = seq(300, 2700, by = 300), width = 300), 59102000)
  score <- seq(-1, 1, by = 0.25)
  space <- factor(c(rep("chr19", 6), "chr17", rep("chr18", 2)),
                  seqlevels(seqinfo))
  correct_gr <- GRanges(space, ir, score = score)
  if (!any(is.na(genome(seqinfo))))
    genome(correct_gr) <- unname(genome(seqinfo)[1])
  seqinfo(correct_gr) <- seqinfo
  correct_gr
}

test_that("read_bed_graph returns correct GRanges", {
  test_path <- system.file("tests", package = "rtracklayer")
  test_bg <- file.path(test_path, "test.bedGraph")
  correct_gr <- createCorrectGR(Seqinfo(c("chr19", "chr17", "chr18")))

  test_gr <- read_bed_graph(test_bg)
  expect_identical(test_gr, correct_gr)

  test_bg_file <- rtracklayer::BEDGraphFile(test_bg)
  test_gr <- read_bed_graph(test_bg)
  expect_identical(test_gr, correct_gr)
  test_bg_con <- file(test_bg)
  test_gr <- read_bed_graph(test_bg_con)
  expect_identical(test_gr, correct_gr)


  # check overlaps
  correct_which <- filter_by_overlaps(correct_gr, correct_gr[3:4])
  test_gr <- read_bed_graph(test_bg, overlap_ranges = correct_gr[3:4])
  expect_identical(correct_which, test_gr)

  # test genome_info
  skip_if_not(
    requireNamespace("BSgenome.Hsapiens.UCSC.hg19", quietly = TRUE),
    message = "'BSgenome.Hsapiens.UCSC.hg19' must be installed to run tests"
  )

  library(BSgenome.Hsapiens.UCSC.hg19)
  hg19_seqinfo <- SeqinfoForBSGenome("hg19")
  correct_genome <- createCorrectGR(hg19_seqinfo)
  test_gr <- read_bed_graph(test_bg, genome_info = "hg19")
  expect_identical(correct_genome, test_gr)

  hg19_gr <- get_genome_info(hg19_seqinfo)
  test_gr <- read_bed_graph(test_bg, genome_info = hg19_gr)
  expect_identical(correct_genome, test_gr)

})

test_that("writing bedGraph files works", {

  correct_gr <- createCorrectGR(Seqinfo(c("chr19", "chr17", "chr18")))
  test_bg_out <- file.path(tempdir(), "test.bedGraph")
  on.exit(unlink(test_bg_out))

  write_bed_graph(correct_gr, test_bg_out)

  test_gr <- read_bed_graph(test_bg_out)
  expect_identical(correct_gr, test_gr)

  test_foo_out <- file.path(tempdir(), "test.foo")
  write_bed_graph(correct_gr, test_foo_out)
  on.exit(unlink(test_foo_out))
  test_gr <- read_bed_graph(test_bg_out)
  expect_identical(correct_gr, test_gr)
  test_bg_out_file <- rtracklayer::BEDGraphFile(test_bg_out)
  write_bed_graph(correct_gr, test_bg_out_file)
  test_gr <- read_bed_graph(test_bg_out_file)
  expect_identical(correct_gr, test_gr)

  # gzipped output
  test_bg_gz <- paste(test_bg_out, ".gz", sep = "")
  on.exit(unlink(test_bg_gz))
  write_bed_graph(correct_gr, test_bg_gz)
  test_gr <- read_bed_graph(test_bg_gz)
  expect_identical(correct_gr, test_gr)

})

context("reading/writing bed files")
# tests adapted from rtracklayer
createCorrectGR <- function(seqinfo) {
  ir <- IRanges(c(127471197, 127472364, 127473531, 127474698, 127475865),
                width = 1167)
  space <- factor(rep(c("chr7", "chr9"), c(3, 2)), seqlevels(seqinfo))
  blocks <- S4Vectors::split(IRanges(c(1, 501, 1068, 1, 668, 1, 1, 1),
                                     c(300, 700, 1167, 250, 1167, 1167, 1167, 1167)),
                             rep(seq_len(5), c(3, 2, 1, 1, 1)))
  names(blocks) <- NULL
  correct_gr <- GRanges(space, ir,
                        strand = strand(c("+", "+", "-", "+", "-")),
                        name = c("Pos1", "Pos2", "Neg1", "Pos3", "Neg2"),
                        score = c(0, 2, 0, 5, 5),
                        itemRgb = c("#FF0000", "#FF0000", "#FF0000",
                                    "#FF0000", "#0000FF"),
                        thick = ir, blocks)
  seqinfo(correct_gr) <- seqinfo
  correct_gr
}



test_that("read_bed returns correct GRanges",{

  test_path <- system.file("tests", package = "rtracklayer")

  test_bed <- file.path(test_path, "test.bed")
  correct_gr <- createCorrectGR(Seqinfo(c("chr7", "chr9")))

  test_gr <- read_bed(test_bed)
  # should always return GRanges
  expect_s4_class(test_gr, "GRanges")
  expect_identical(correct_gr, test_gr)

  test_bed_file <- rtracklayer::BEDFile(test_bed)
  test_gr <- read_bed(test_bed_file)
  expect_identical(correct_gr, test_gr)

  test_bed_con <- file(test_bed)
  test_gr <- read_bed(test_bed_con)
  expect_identical(correct_gr, test_gr)

  # check colnames
  subcols <- c("name", "thick")
  # drops strand info
  test_gr <- read_bed(test_bed, col_names = subcols)

  expect_identical(correct_gr %>%
                     mutate(strand = "*") %>%
                     select(dplyr::all_of(subcols)),
                   test_gr)
  test_gr <- read_bed(test_bed, col_names = c(subcols, "strand"))
  expect_identical(correct_gr %>%
                     select(dplyr::all_of(subcols)),
                   test_gr)

  # check overlaps
  which <- correct_gr[1:2]
  correct_which <- filter_by_overlaps(correct_gr, which)
  test_gr <- read_bed(test_bed, overlap_ranges = which)
  expect_identical(correct_which, test_gr)

  # check that import by genome name works
  skip_if_not(
    requireNamespace("BSgenome.Hsapiens.UCSC.hg19", quietly = TRUE),
    message = "'BSgenome.Hsapiens.UCSC.hg19' must be installed to run tests"
  )

  library(BSgenome.Hsapiens.UCSC.hg19)
  hg19_seqinfo <- SeqinfoForBSGenome(genome = "hg19")
  correct_gr <- createCorrectGR(hg19_seqinfo)
  test_gr <- read_bed(test_bed, genome_info = "hg19")
  expect_identical(correct_gr, test_gr)

  # if genome_info is a ranges object
  hg19_gr <- get_genome_info(hg19_seqinfo)
  test_gr <- read_bed(test_bed, genome_info = hg19_gr)
  expect_identical(correct_gr, test_gr)

})

test_that("write bed returns correct bed files", {

  correct_gr <- createCorrectGR(Seqinfo(c("chr7", "chr9")))
  ## the 'gsub' is to handle Windows paths (for later coercion to URL)
  dir <- tempdir()
  test_bed_out <- gsub("\\\\", "/", file.path(dir, "test.bed"))
  on.exit(unlink(test_bed_out))

  write_bed(correct_gr, test_bed_out)
  test_gr <- read_bed(test_bed_out)
  expect_identical(correct_gr, test_gr)

  # gzipped output
  test_bed_gz <- paste(test_bed_out, ".gz", sep = "")
  write_bed(correct_gr, test_bed_gz)
  test_gr <- read_bed(test_bed_gz)
  expect_identical(correct_gr, test_gr)

  # tabixed output
  test_bed_bgz <- paste(test_bed_out, ".bgz", sep = "")
  on.exit(unlink(paste(test_bed_bgz, ".tbi", sep = "")))
  write_bed(correct_gr, test_bed_out, index = TRUE)
  which <- correct_gr[1:2]
  correct_which <- filter_by_overlaps(correct_gr, which)
  test_gr <- read_bed(test_bed_bgz, overlap_ranges =  which)
  expect_identical(correct_which, test_gr)

  # url input
  # skip on windows for now
  skip_on_os("windows")
  test_bed_url <- paste("file://", test_bed_out, sep = "")
  write_bed(correct_gr, test_bed_url)
  test_gr <- read_bed(test_bed_url)
  expect_identical(correct_gr, test_gr)

  test_bed_gz_url <- paste("file://", test_bed_gz, sep = "")
  write_bed(correct_gr, test_bed_gz_url)
  test_gr <- read_bed(test_bed_gz_url)
  expect_identical(correct_gr, test_gr)

})

test_that("read_narrowpeaks returns correct GRanges", {
  file <- system.file("extdata", "demo.narrowPeak.gz",  package="rtracklayer")
  gr <- read_narrowpeaks(file, genome_info = "hg19")
  expect_equal(length(gr), 6)
  expect_equal(colnames(mcols(gr)),
              c("name","score","signalValue","pValue","qValue","peak"))

  # matches write output
  test_np_out <- file.path(tempdir(), "test.narrowPeak.gz")
  write_narrowpeaks(gr, test_np_out)
  test_gr <- read_narrowpeaks(test_np_out, genome_info = "hg19")
  expect_equal(gr, test_gr)

})

## -----------------------------------------------------------------------------
suppressPackageStartupMessages(library(plyranges))
set.seed(100)
df <- data.frame(start=c(2:-1, 13:15),
                 width=c(0:3, 2:0))

# produces IRanges
rng <- df %>% as_iranges()
rng


## -----------------------------------------------------------------------------
# seqname is required for GRanges, metadata is automatically kept
grng <- df %>%
  transform(seqnames = sample(c("chr1", "chr2"), 7, replace = TRUE),
         strand = sample(c("+", "-"), 7, replace = TRUE),
         gc = runif(7)) %>%
  as_granges()

grng

## ----anchor_fig, echo = FALSE, out.width="400px", fig.align="center"----------
knitr::include_graphics("anchors.png", dpi = 150)

## -----------------------------------------------------------------------------
rng <- as_iranges(data.frame(start=c(1, 2, 3), end=c(5, 2, 8)))
grng <- as_granges(data.frame(start=c(1, 2, 3), end=c(5, 2, 8),
                          seqnames = "seq1",
                          strand = c("+", "*", "-")))
mutate(rng, width = 10)
mutate(anchor_start(rng), width = 10)
mutate(anchor_end(rng), width = 10)
mutate(anchor_center(rng), width = 10)
mutate(anchor_3p(grng), width = 10) # leave negative strand fixed
mutate(anchor_5p(grng), width = 10) # leave positive strand fixed

## -----------------------------------------------------------------------------
rng2 <- stretch(anchor_center(rng), 10)
rng2
stretch(anchor_end(rng2), 10)
stretch(anchor_start(rng2), 10)
stretch(anchor_3p(grng), 10)
stretch(anchor_5p(grng), 10)

## -----------------------------------------------------------------------------
shift_left(rng, 100)
shift_right(rng, 100)
shift_upstream(grng, 100)
shift_downstream(grng, 100)

## -----------------------------------------------------------------------------
grng <- data.frame(seqnames = sample(c("chr1", "chr2"), 7, replace = TRUE),
         strand = sample(c("+", "-"), 7, replace = TRUE),
         gc = runif(7),
         start = 1:7,
         width = 10) %>%
  as_granges()

grng_by_strand <- grng %>%
  group_by(strand)

grng_by_strand

## -----------------------------------------------------------------------------
grng %>% filter(gc < 0.3)
# filtering by group
grng_by_strand %>% filter(gc == max(gc))

## -----------------------------------------------------------------------------
ir0 <- data.frame(start = c(5,10, 15,20), width = 5) %>%
  as_iranges()
ir1 <- data.frame(start = 2:6, width = 3:7) %>%
  as_iranges()
ir0
ir1
ir0 %>% filter_by_overlaps(ir1)
ir0 %>% filter_by_non_overlaps(ir1)

## -----------------------------------------------------------------------------
ir1 <- ir1 %>%
  mutate(gc = runif(length(.)))

ir0 %>%
  group_by_overlaps(ir1) %>%
  summarise(gc = mean(gc))

## -----------------------------------------------------------------------------
query <- data.frame(seqnames = "chr1",
               strand = c("+", "-"),
               start = c(1, 9),
               end =  c(7, 10),
               key.a = letters[1:2]) %>%
  as_granges()

subject <- data.frame(seqnames = "chr1",
               strand = c("-", "+"),
               start = c(2, 6),
               end = c(4, 8),
               key.b = LETTERS[1:2]) %>%
  as_granges()

## ----olap, echo = FALSE, fig.cap = "Query and Subject Ranges", message = FALSE----
library(ggplot2)
query_df <- as.data.frame(query)[, -6]
query_df$key <- "Query"
subject_df <- as.data.frame(subject)[, -6]
subject_df$key <- "Subject"
melted_ranges <- rbind(query_df, subject_df)
ggplot(melted_ranges, aes(xmin = start, xmax = end, ymin = 1, ymax = 3)) +
  geom_rect() +
  facet_grid(key ~ .) +
  scale_x_continuous(breaks = seq(1, 10, by = 1)) +
  xlab("Position") +
  theme(axis.text.y = element_blank(),
        axis.ticks.y = element_blank(),
        axis.title.y = element_blank())

## ----olaps-diagram, echo = FALSE, out.width="600px"---------------------------
knitr::include_graphics("olaps.png")

## -----------------------------------------------------------------------------
intersect_rng <- join_overlap_intersect(query, subject)
intersect_rng

## ----intersect-join, echo = FALSE, fig.cap="Intersect Join"-------------------
intersect_df <- as.data.frame(intersect_rng)[, -c(6,7)]
intersect_df$key <- "Intersect Join"
melted_ranges <- rbind(query_df, subject_df, intersect_df)
melted_ranges$key <- factor(melted_ranges$key,
                              levels = c("Query", "Subject", "Intersect Join"))
ggplot(melted_ranges, aes(xmin = start, xmax = end, ymin = 1, ymax = 3)) +
  geom_rect() +
  facet_grid(key ~ .) +
  scale_x_continuous(breaks = seq(1, 10, by = 1)) +
  xlab("Position") +
  theme(axis.text.y = element_blank(),
        axis.ticks.y = element_blank(),
        axis.title.y = element_blank())

## -----------------------------------------------------------------------------
inner_rng <- join_overlap_inner(query, subject)
inner_rng

## ----inner-join, echo = FALSE, fig.cap = "Inner Join", message = FALSE--------
inner_df <- as.data.frame(inner_rng)[, -c(6,7)]
inner_df$ymin <- c(1,4)
inner_df$ymax <- c(3,6)
inner_df$key <- "Inner Join"
melted_ranges <- rbind(query_df, subject_df)
melted_ranges$ymin <- 1
melted_ranges$ymax <- 3
melted_ranges <- rbind(melted_ranges, inner_df)
melted_ranges$key <- factor(melted_ranges$key,
                              levels = c("Query", "Subject", "Inner Join"))

ggplot(melted_ranges, aes(xmin = start, xmax = end, ymin = ymin, ymax = ymax)) +
  geom_rect() +
  facet_grid(key ~ ., scales = "free_y") +
  scale_x_continuous(breaks = seq(1, 10, by = 1)) +
  xlab("Position") +
  theme(axis.text.y = element_blank(),
        axis.ticks.y = element_blank(),
        axis.title.y = element_blank())

## -----------------------------------------------------------------------------
find_overlaps(query, subject)

## -----------------------------------------------------------------------------
left_rng <- join_overlap_left(query, subject)
left_rng

## ----olap-left, echo = FALSE, fig.cap = "Left Join", message = FALSE----------
left_df <- as.data.frame(left_rng)[, -c(6,7)]
left_df$ymin <- c(1,4, 1)
left_df$ymax <- c(3,6, 3)
left_df$key <- "Left Join"
melted_ranges <- rbind(query_df, subject_df)
melted_ranges$ymin <- 1
melted_ranges$ymax <- 3
melted_ranges <- rbind(melted_ranges, left_df)
melted_ranges$key <- factor(melted_ranges$key,
                              levels = c("Query", "Subject", "Left Join"))

ggplot(melted_ranges,
       aes(xmin = start, xmax = end, ymin = ymin, ymax = ymax)) +
  geom_rect() +
  facet_grid(key ~ ., scales = "free_y") +
  scale_x_continuous(breaks = seq(1, 10, by = 1)) +
  xlab("Position") +
  theme(axis.text.y = element_blank(),
        axis.ticks.y = element_blank(),
        axis.title.y = element_blank())

## ----neigbours, echo = FALSE, out.width="600px"-------------------------------
knitr::include_graphics("neighbours.png")

## -----------------------------------------------------------------------------
join_nearest(ir0, ir1)
join_follow(ir0, ir1)
join_precede(ir0, ir1) # nothing precedes returns empty `Ranges`
join_precede(ir1, ir0)

## ----ex1----------------------------------------------------------------------
intensities <- data.frame(seqnames = "VI",
                          start = c(3320:3321,3330:3331,3341:3342),
                          width = 1) %>%
  as_granges()

intensities

genes <- data.frame(seqnames = "VI",
                    start = c(3322, 3030),
                    end = c(3846, 3338),
                    gene_id=c("YFL064C", "YFL065C")) %>%
  as_granges()

genes

## -----------------------------------------------------------------------------
olap <- join_overlap_inner(intensities, genes) %>%
  select(gene_id)
olap

## -----------------------------------------------------------------------------
olap %>%
  group_by(start) %>%
  summarise(n = n())

## -----------------------------------------------------------------------------
grp_by_olap <- ir0 %>%
  group_by_overlaps(ir1)
grp_by_olap
grp_by_olap %>%
  mutate(n_overlaps = n())

## -----------------------------------------------------------------------------
ir0 %>%
  mutate(n_overlaps = count_overlaps(., ir1))

## -----------------------------------------------------------------------------
sessionInfo()

expand_rng_by_cigar <- function(x, type) {
  # check for cigar column
  stopifnot(any(names(mcols(x)) %in% "cigar"))
  if (type == "gaps") {
    drop.D.ranges <- TRUE
  } else {
    drop.D.ranges <- FALSE
  }
  # extract alignment ranges, returns an IRangesList
  rng <- extractAlignmentRangesOnReference(mcols(x)$cigar, 
                                           pos=start(x), 
                                           drop.D.ranges = drop.D.ranges)
  
  n <- elementNROWS(rng)
  seqnames <- BiocGenerics::rep.int(seqnames(x), n)
  strand <- BiocGenerics::rep.int(strand(x), n)
  grng <- GRanges(seqnames = seqnames, 
                  strand = strand, 
                  ranges = rng@unlistData)
  
  seqinfo(grng) <- seqinfo(x)
  grp <- Rle(BiocGenerics::rep.int(seq_along(x), n))
  mcols(grng) <- mcols(x)[grp,]
  mcols(grng)[[type]] <- grp
  group_by(grng, !!rlang::sym(type))
}

#' Group a GRanges object by introns or gaps
#' 
#' @param x a GenomicRanges object with a cigar string column
#' 
#' @details Creates a grouped Ranges object from a cigar string
#' column, for `chop_by_introns()` will check for the presence of
#' "N" in the cigar string and create a new column called
#' `intron` where TRUE indicates the alignment has a skipped
#' region from the reference. For `chop_by_gaps()` will check
#' for the presence of "N" or "D" in the cigar string and
#' create a new column called "gaps" where TRUE indicates
#' the alignment has a deletion from the reference or has an intron.  
#' 
#' @return a GRanges object
#' @examples 
#' if (require(pasillaBamSubset)) {
#'    bamfile <- untreated1_chr4()
#'    # define a region of interest
#'    roi <- data.frame(seqnames = "chr4", start = 5e5, end = 7e5) %>%
#'             as_granges()
#'    # results in a grouped ranges object
#'    rng <- read_bam(bamfile) %>% 
#'             filter_by_overlaps(roi) %>%
#'             chop_by_gaps()
#'    # to find ranges that have gaps use filter with `n()`
#'    rng %>% filter(n() >= 2)
#'   
#' }
#' 
#' @importFrom BiocGenerics rep.int
#' @importFrom S4Vectors Rle elementNROWS
#' @importFrom GenomicAlignments extractAlignmentRangesOnReference
#' @rdname ranges-chop
#' @export
chop_by_introns <- function(x) UseMethod("chop_by_introns")

#' @export
chop_by_introns.GenomicRanges <- function(x) {
  expand_rng_by_cigar(x, "introns")
}
#' @export
chop_by_introns.DeferredGenomicRanges <- function(x) {
  chop_by_introns(load_delegate(x))
}

#' @rdname ranges-chop
#' @export
chop_by_gaps <- function(x) UseMethod("chop_by_gaps")

#' @export
chop_by_gaps.GenomicRanges <- function(x) {
  expand_rng_by_cigar(x, "gaps")
}

#' @export
chop_by_gaps.DeferredGenomicRanges <- function(x) {
  chop_by_gaps(load_delegate(x))
}

# PPA grouped mutate speedup, 2025
check_colnames <- function(col_names) {
  if (any(col_names %in% "")) {
    stop("mutate must have name-variable pairs as input", call. = FALSE)
  }
  return(invisible(NULL))
}


mutate_mcols <- function(.data, .mutated) {
  all_cols <- names(.mutated)
  only_mcols <- !(all_cols %in%
                    c("start", "end", "width", "seqnames", "strand"))
  .mutated <- .mutated[only_mcols]
  update_cols <- all_cols[only_mcols]

  matches_mcols <- match(update_cols, names(mcols(.data)))
  idx_mcols <- !is.na(matches_mcols)

  if (any(idx_mcols)) {
    mcols(.data)[matches_mcols[idx_mcols]] <- .mutated[idx_mcols]
  }

  if (!all(idx_mcols)) {
    if (is.null(mcols(.data))) {
      mcols(.data) <- S4Vectors::DataFrame(.mutated[!idx_mcols])
    } else {
      mcols(.data) <- S4Vectors::DataFrame(list(mcols(.data), 
                                                .mutated[!idx_mcols]))
    }
  }
  .data
}

# PPA grouped mutate speedup, 2025
mutate_mcols_grp <- function(.data, dots) {
  
  # generate grouped df
  grps <- dplyr::group_vars(.data)
  df <- as.data.frame(dplyr::ungroup(.data))
  df <- dplyr::group_by(df,!!!rlang::syms(grps))
  
  mcols(.data) <- dplyr::mutate(df, !!!dots) %>%
    dplyr::ungroup() %>%
    dplyr::select(-tidyselect::any_of(c("start", "end", "width", "seqnames", "strand"))) %>%
    as("DataFrame")
  
  return(.data)
}



#' @importFrom methods selectMethod
mutate_core <- function(.data, .mutated) {
  all_cols <- names(.mutated)
  core_cols <- all_cols[all_cols %in%
                           c("start", "end", "width", "seqnames", "strand")]
  if (length(core_cols == 0)) {
    .data
  }

  for (col in core_cols) {
    modifier <- match.fun(paste0("set_", col))
    .data <- modifier(.data, .mutated[[col]])
  }
  .data
}


mutate_rng <- function(.data, dots) {
  col_names <- names(dots)
  check_colnames(col_names)

  overscope <- overscope_ranges(.data)
  .mutated <- overscope_eval_update(overscope, dots)
  .data <- mutate_core(.data, .mutated)
  mutate_mcols(.data, .mutated)
}

# idea could simply dispatch to summarise here, and store 
# list columns, if the length is smaller then we can repeat,
# otherwise we try to expand 
mutate_grp <- function(.data, dots) {
  
  inx <- .group_rows(.data)
  rng <- unname(S4Vectors::split(.data@delegate, .data@group_indices))
  rng <- S4Vectors::endoapply(rng, function(x) {
    mutate_rng(x, dots)    
  })
  
  rng <- unlist(rng)[BiocGenerics::order(unlist(inx))]
  new(class(.data),
      delegate = rng, 
      group_keys =  .data@group_keys, 
      group_indices = .data@group_indices,
      n = .data@n )
}
#' Modify a Ranges object
#'
#' @param .data a `Ranges` object
#' @param ... Pairs of name-value expressions. The name-value pairs can either
#' create new metadata columns or modify existing ones.
#'
#' @importFrom dplyr mutate
#' @rdname mutate-ranges
#' @return a Ranges object
#' @method mutate Ranges
#'
#' @examples
#' df <- data.frame(start = 1:10,
#'                  width = 5,
#'                  seqnames = "seq1",
#'                  strand = sample(c("+", "-", "*"), 10, replace = TRUE),
#'                  gc = runif(10))
#' rng <- as_granges(df)
#'
#' # mutate adds new columns
#' rng %>%
#'     mutate(avg_gc = mean(gc), row_id = 1:n())
#' # can also compute on newly created columns
#' rng %>%
#'     mutate(score = gc * width, score2 = score + 1)
#' # group by partitions the data and computes within each group
#' rng %>%
#'     group_by(strand) %>%
#'     mutate(avg_gc = mean(gc), row_id = 1:n())
#'
#' # mutate can be used in conjuction with anchoring to resize ranges
#' rng %>%
#'     mutate(width = 10)
#' # by default width modfication fixes by start
#' rng %>%
#'     anchor_start() %>%
#'     mutate(width = 10)
#' # fix by end or midpoint
#' rng %>%
#'     anchor_end() %>%
#'     mutate(width = width + 1)
#' rng %>%
#'     anchor_center() %>%
#'     mutate(width = width + 1)
#' # anchoring by strand
#' rng %>%
#'     anchor_3p() %>%
#'     mutate(width = width * 2)
#' rng %>%
#'     anchor_5p() %>%
#'     mutate(width = width * 2)
#' @export
mutate.Ranges <- function(.data, ...) {
  dots <- set_dots_named(...)
  mutate_rng(.data, dots)
}

#' @method mutate AnchoredIntegerRanges
#' @export
mutate.AnchoredIntegerRanges <- mutate.Ranges

#' @method mutate AnchoredGenomicRanges
#' @export
mutate.AnchoredGenomicRanges <- mutate.Ranges

#' @method mutate DelegatingGenomicRanges
#' @export
mutate.DelegatingGenomicRanges <- function(.data, ...) {
  dots <- set_dots_named(...)
  delegate <- .data@delegate
  .data@delegate <- mutate_rng(delegate, dots)
  return(.data)
}

#' @method mutate DelegatingIntegerRanges
#' @export
mutate.DelegatingIntegerRanges <- mutate.DelegatingGenomicRanges

#' @method mutate GroupedGenomicRanges
#' @export
mutate.GroupedGenomicRanges <- function(.data, ...) {
  
  dots <- set_dots_named(...)
  check_colnames(names(dots))
  core_cols <- names(dots) %in% c("start", "end", "width", "seqnames", "strand")
  
  # if any S4 columns in mcols use plyranges group mutate
  if (any(sapply(mcols(.data), isS4))) {
    message(
      paste0("metadata contains S4 columns, using lapply group operations.\n",
             "removing S4 columns will speed up group operations.")
    )
    return(mutate_grp(.data, dots))
  }
  # PPA grouped mutate speedup, 2025
  # if mutating core columns, call plyranges group for those
  if(any(core_cols)) {
    .data <- mutate_grp(.data, dots[core_cols])
  }
  # if only core columns in call, return object
  if(all(core_cols)) {
    return(.data)
  } else {
    # use vanilla dplyr group operations on metadata
    mutate_mcols_grp(.data, dots[!core_cols])
  }
}

#' @method mutate GroupedIntegerRanges
#' @export
mutate.GroupedIntegerRanges <- mutate.GroupedGenomicRanges

rng_arrange <- function(.data, dots) {
  rng_os <- overscope_ranges(.data)
  rng_list <- lapply(dots, rlang::eval_tidy, data = rng_os)

  if (length(rng_list) == 1L) {
    idx <- order(rng_list[[1]])
  } else {
    idx <- Reduce(order, rng_list)
  }
  .data[idx, ]
}
#' Sort a Ranges object
#'
#' @param .data A Ranges object.
#' @param ... Comma seperated list of variable names.
#'
#' @importFrom dplyr arrange
#' @examples
#' rng <- as_iranges(data.frame(start = 1:10, width = 10:1))
#' rng <- mutate(rng, score = runif(10))
#' arrange(rng, score)
#' # you can also use dplyr::desc to arrange by descending order
#' @rdname ranges-arrange
#' @return A sorted Ranges object
#' @method arrange Ranges
#' @export
arrange.Ranges <- function(.data, ...) {
  dots <- rlang::enquos(...)
  rng_arrange(.data, dots)
}

#' @method arrange DelegatingGenomicRanges
#' @export
arrange.DelegatingGenomicRanges <- function(.data, ...) {
  dots <- rlang::enquos(...)
  delegate <- .data@delegate
  .data@delegate <- rng_arrange(delegate, dots)
  .data
}

#' @method arrange DelegatingIntegerRanges
#' @export
arrange.DelegatingIntegerRanges <- arrange.DelegatingGenomicRanges
# -- TODO: allow sorting by group?

# ranges-join-precede.R

#' Find preceding Ranges
#'
#' @param x,y Ranges objects, which ranges in x precede those in y.
#' @param suffix A character vector of length two used to identify
#' metadata columns coming from x and y.
#'
#' @details By default `join_precede` will return the ranges
#' in x that come before the ranges in y and ignore any strand information.
#' The function `join_precede_right` will find all ranges in y
#' that are on the right-hand side of the ranges in x ignoring any strand
#' information. Finally, `join_precede_downstream` will find all ranges in y
#' that are that are downstream of the ranges in x. On the positive strand this
#' will result in ranges in y that are right of those in x and on the negative
#' strand it will result in ranges in y that are left of those in x.
#'
#' @return A Ranges object corresponding to the ranges in `y` that are
#' preceded by the ranges in `x`, all metadata is copied over from the
#' right-hand side ranges `y`.
#'
#' @examples
#' subject <- data.frame(start = c(5,10, 15,20), width = 5, gc = runif(4)) %>%
#'              as_iranges()
#' query <- data.frame(start = 2:6, width = 3:7, label = letters[1:5]) %>%
#'              as_iranges()
#'
#' join_precede(query, subject)
#'
#' query  <- data.frame(seqnames = "chr1",
#'                start = c(11,101),
#'                end = c(21, 200),
#'                name = c("a1", "a2"),
#'                strand = c("+", "-"),
#'                score = c(1,2)) %>%
#'            as_granges()
#' subject <- data.frame(seqnames = "chr1",
#'                       strand = c("+", "-", "+", "-"),
#'                       start = c(21,91,101,201),
#'                       end = c(30,101,110,210),
#'                       name = paste0("b", 1:4),
#'                       score = 1:4) %>%
#'                    as_granges()
#'
#' join_precede(query, subject)
#' join_precede_right(query, subject)
#' join_precede_downstream(query, subject)
#' @rdname ranges-precede
#' @importFrom IRanges precede
#' @export
join_precede <- function(x,y, suffix = c(".x", ".y")) { UseMethod("join_precede") }

#' @export
join_precede.IntegerRanges <- function(x,y, suffix = c(".x", ".y")) {
  hits <- precede(x,y)
  expand_by_hits(x,y, suffix, hits)
}

#' @export
join_precede.GenomicRanges <- function(x,y, suffix = c(".x", ".y")) {
  hits <- precede(x,y, ignore.strand = TRUE)
  expand_by_hits(x,y,suffix,hits)
}


#' @rdname ranges-precede
#' @importFrom IRanges precede
#' @export
join_precede_right <- function(x,y, suffix = c(".x", ".y")) { UseMethod("join_precede_right") }

#' @export
join_precede_right.IntegerRanges <- function(x,y, suffix = c(".x", ".y")) {
  hits <- precede(x,y, select = "all")
  expand_by_hits(x,y,suffix,hits)
}

#' @export
join_precede_right.GenomicRanges <- function(x,y, suffix = c(".x", ".y")) {
  hits <- precede(x,y, select = "all", ignore.strand = TRUE)
  expand_by_hits(x,y,suffix,hits)
}


#' @rdname ranges-precede
#' @importFrom IRanges precede
#' @export
join_precede_downstream <- function(x,y, suffix = c(".x", ".y")) {UseMethod("join_precede_downstream")}

#' @export
join_precede_downstream.GenomicRanges <- function(x,y, suffix = c(".x", ".y")) {
  hits <- precede(x,y, select = "all", ignore.strand = FALSE)
  expand_by_hits(x,y,suffix,hits)
}

# Various column wise helpers for ranges

#' Tools for working with named Ranges
#'
#' @param .data a Ranges object
#' @param var Name of column to use for names
#' 
#' @details The function `names_to_column()` and `id_to_column()` always places
#' `var` as the first column in `mcols(.data)`, shifting all other columns 
#' to the left. The `id_to_column()` creates a column with sequential row
#' identifiers starting at 1, it will also remove any existing names. 
#' 
#' @return 
#' Returns a Ranges object with empty names
#' 
#' @examples 
#' ir <- IRanges::IRanges(start = 1:3, width = 4, names = c("a", "b", "c"))
#' remove_names(ir)
#' ir_noname <- names_to_column(ir)
#' ir_noname
#' ir_with_id <- id_to_column(ir)
#' ir_with_id
#' @rdname ranges-names
#' @export
remove_names <- function(.data) {
  stopifnot(is(.data, "Ranges"))
  if (is.null(names(.data))) return(.data)
  names(.data) <- NULL
  .data
}


#' @rdname ranges-names
#' @export
names_to_column <- function(.data, var = "name") {
  stopifnot(is(.data, "Ranges"))
  if (var %in% names(mcols(.data))) {
    stop(paste("Column", var, "already exists in .data."))
  }
  if (is.null(names(.data))) {
    stop("No names present in .data")
  }
  .data <- mutate(.data, `:=`(!!var, names(.data)))
  .data <- select(.data, !!var, tidyselect::everything())
  remove_names(.data)
}

#' @rdname ranges-names
#' @export
id_to_column <- function(.data, var = "id") {
  stopifnot(is(.data, "Ranges"))
  if (var %in% names(mcols(.data))) {
    stop(paste("Column", var, "already exists in .data."))
  }
  
  .data <- mutate(.data, `:=`(!!var, seq_len(length(.data))))
  .data <- select(.data, !!var, tidyselect::everything())
  remove_names(.data)
}

# test-construction.R
context("Ranges construction")

test_that("invalid data.frame call throws an error", {
  expect_error(as_iranges(data.frame()))
  expect_error(as_iranges(data.frame(start = 1)))
})

test_that("valid data.frame returns expected class", {
  df <- data.frame(start = 1:3, width = 2:4)
  expect_s4_class(as_iranges(df), "IRanges")
  df2 <-df
  df2$strand <- "*"
  expect_s4_class(as_iranges(df2), "IRanges")

  # but there's not enough info for GRanges
  expect_error(as_granges(df2))

  df3 <- df2
  df3$seqnames <- "chr1"
  expect_s4_class(as_granges(df3), "GRanges")

  df4 <- df
  df4$seqnames <- "chr1"
  expect_s4_class(as_granges(df4), "GRanges")

})

test_that("non-standard evaluation works as expected", {
  df <- data.frame(st= 1:3, width = 2:4)
  expect_s4_class(as_iranges(df, start = st), "IRanges")
  # adding new column throws an error
  expect_error(as_iranges(df, start = st, gc = 1))
  expect_s4_class(as_granges(df, start = st, seqnames = "chr1"), "GRanges")
})

test_that("out of bounds Ranges throws error", {
  df <- data.frame(start = 1:3, width = 2:4)
  expect_error(as_iranges(df, end = 4))
  df2 <- df
  df2$seqnames <- "1"
  df2$strand <- "negative"
  expect_error(as_granges(df2))
})

test_that("DataFrame input", {
  # tests issue 62 https://github.com/sa-lee/plyranges/issues/62
  df <- DataFrame(st = 1:3, width = 2:4,
                  grps = IRanges::FactorList("a", c("b", "c"), "d", compress = FALSE))

  ir <- as_iranges(df, start = st)
  expect_identical(mcols(ir)$grps, df$grps)

  df$strand <- "+"
  df$chr <- "chr1"
  gr <- as_granges(df, start = st, seqnames = chr)
  expect_identical(mcols(gr)$grps, df$grps)
})

test_that("Working with names", {
  ir <- IRanges::IRanges(start = 1:3, width = 4, names = c("a", "b", "c"))
  expect_identical(unname(ir), remove_names(ir))
  ir_noname <- names_to_column(ir)
  expect_null(names(ir_noname))
  expect_identical(names(ir), mcols(ir_noname)[["name"]])
  ir_id <- id_to_column(ir_noname)
  expect_identical(seq_len(length(ir)), mcols(ir_id)[["id"]])


  gr <- GenomicRanges::GRanges(seqnames = "chr1", ranges = ir)
  expect_identical(unname(gr), remove_names(gr))
  gr_noname <- names_to_column(gr)
  expect_null(names(gr_noname))
  expect_identical(names(gr), mcols(gr_noname)[["name"]])
  gr_id <- id_to_column(gr_noname)
  expect_identical(seq_len(length(gr)), mcols(gr_id)[["id"]])
  expect_identical(mcols(gr_id),
                   S4Vectors::DataFrame(id = seq_len(length(gr)),
                                        name = names(gr)))
})

#' Count the number of overlaps between two Ranges objects
#'
#' @param x,y Objects representing ranges
#' @param maxgap,minoverlap The maximimum gap between intervals as an integer
#' greater than or equal to zero. The minimum amount of overlap between intervals
#' as an integer greater than zero, accounting for the maximum gap.
#'
#' @return An integer vector of same length as x.
#'
#' @examples
#' query <- data.frame(start = c(5,10, 15,20), width = 5, gc = runif(4)) %>%
#'              as_iranges()
#' subject <- data.frame(start = 2:6, width = 3:7, label = letters[1:5]) %>%
#'              as_iranges()
#' query %>% mutate(n_olap = count_overlaps(., subject),
#'                  n_olap_within = count_overlaps_within(., subject))
#'
#' @importFrom IRanges countOverlaps
#' @rdname ranges-count-overlaps
#' @export
count_overlaps <- function(x, y, maxgap, minoverlap) {
  UseMethod("count_overlaps")
}

#' @rdname ranges-count-overlaps
#' @export
count_overlaps.IntegerRanges <- function(x,y, maxgap = -1L, minoverlap = 0L) {
  countOverlaps(x,y, maxgap, minoverlap, type = "any")
}

#' @rdname ranges-count-overlaps
#' @export
count_overlaps.GenomicRanges <- function(x,y, maxgap = -1L, minoverlap = 0L) {
  countOverlaps(x,y, maxgap, minoverlap, type = "any", ignore.strand = TRUE)
}

#' @rdname ranges-count-overlaps
#' @export
count_overlaps_within <- function(x, y, maxgap, minoverlap) {
  UseMethod("count_overlaps_within")
}

#' @rdname ranges-count-overlaps
#' @export
count_overlaps_within.IntegerRanges <- function(x,y, maxgap = 0L, minoverlap = 1L) {
  countOverlaps(x,y, maxgap, minoverlap, type = "within")
}

#' @rdname ranges-count-overlaps
#' @export
count_overlaps_within.GenomicRanges <- function(x,y, maxgap = 0L, minoverlap = 1L) {
  countOverlaps(x,y, maxgap, minoverlap, type = "within", ignore.strand = TRUE)
}

#' @rdname ranges-count-overlaps
#' @export
count_overlaps_directed <- function(x, y, maxgap, minoverlap) {
  UseMethod("count_overlaps_directed")
}

#' @rdname ranges-count-overlaps
#' @export
count_overlaps_directed.GenomicRanges <- function(x,y, maxgap = -1L, minoverlap = 0L) {
  countOverlaps(x,y, maxgap, minoverlap, type = "any", ignore.strand = FALSE)
}

#' @rdname ranges-count-overlaps
#' @export
count_overlaps_within_directed <- function(x, y, maxgap, minoverlap) {
  UseMethod("count_overlaps_directed")
}

#' @rdname ranges-count-overlaps
#' @export
count_overlaps_within_directed.GenomicRanges <- function(x,y, maxgap = -1L, minoverlap = 0L) {
  countOverlaps(x,y, maxgap, minoverlap, type = "within", ignore.strand = FALSE)
}

# abstract class to represent operations 
# performed over a File

#' An abstract class to represent operations performed over a file
#' 
#' @details This class is used internally by DeferredGenomicRanges objects.
#' Currently, this class is only implemented for bam files (as a 
#' BamFileOperator) but will eventually be extended to the other avaialable
#' readers.
#' 
#' @rdname ranges-class
#' @export
setClass("FileOperator", contains = "VIRTUAL")

#' @importFrom Rsamtools BamFile ScanBamParam 
#' @export
#' @rdname ranges-class
setClass("BamFileOperator", 
         slots = c(
           input = "BamFile",
           param = "ScanBamParam",
           paired = "logical"
         ),
         contains = "FileOperator"
)

new_bam_ops <- function(file, index = file, paired = FALSE) {
  if (is.null(index)) {
    index <- character()
  }
  new("BamFileOperator", 
      input = Rsamtools::BamFile(file, index = index),
      param = Rsamtools::ScanBamParam(),
      paired = paired
  )
}

# Generic File loader
#' @importFrom Rsamtools bamFlag bamWhich ScanBamParam
#' @importFrom GenomicAlignments readGAlignments readGAlignmentPairs
#' @importFrom S4Vectors DataFrame
load_genomic_file <- function(ops) UseMethod("load_genomic_file")

#' @export
load_genomic_file.BamFileOperator <- function(ops) {
  if (ops@paired) return(paired_alignments(ops))
  unpaired_alignments(ops)
}

check_which_label <- function(ops) {
  length(bamWhich(ops@param)) > 0
}

galn_to_grng <- function(alignments) {
  grng <- granges(alignments, use.mcols = TRUE)
  mcols(grng)[["cigar"]] <- GenomicAlignments::cigar(alignments)
  mcols(grng)[["qwidth"]] <- GenomicAlignments::qwidth(alignments)
  mcols(grng)[["njunc"]] <- GenomicAlignments::njunc(alignments)
  seqinfo(grng) <- seqinfo(alignments)
  grng
}

paired_alignments <- function(ops) {
  alignments <- readGAlignmentPairs(
    ops@input,
    param = ops@param,
    with.which_label = check_which_label(ops)
  )
  r1 <- S4Vectors::first(alignments)
  r2 <- S4Vectors::second(alignments)
  r1_grng <- galn_to_grng(r1)
  mcols(r1_grng)[["read_pair_id"]] <- seq_along(r1_grng)
  mcols(r1_grng)[["read_pair_group"]] <- 1L
  r2_grng <- galn_to_grng(r2)
  mcols(r2_grng)[["read_pair_id"]] <- seq_along(r1_grng)
  mcols(r2_grng)[["read_pair_group"]] <- 2L
  grng <- c(r1_grng, r2_grng)
  grng <- grng[order(grng$read_pair_id)]
  return(group_by(grng, !!rlang::sym("read_pair_id")))
}

unpaired_alignments <- function(ops) {
  galn_to_grng(
    suppressWarnings(
      readGAlignments(
        file = ops@input,
        param = ops@param,
        with.which_label = check_which_label(ops)
      )
    )
  )
}

#' Compute coverage over a Ranges object
#'
#' @param x a `Ranges` object
#' @param shift shift how much should each range in x be shifted by? (default = 0L)
#' @param width width how long should the returned coverage score be?
#' This must be either a  positive integer or NULL (default = NULL)
#' @param weight weight how much weight should be assigned to each range? Either
#' an integer or numeric vector or a column in x. (default = 1L)
#' @param ... other optional parameters to pass to coverage
#'
#' @return An expanded Ranges object with a score column corresponding to
#' the coverage value over that interval. Note that compute_coverage
#' drops metadata associated with the orginal ranges.
#' @seealso \code{IRanges::\link[IRanges:coverage-methods]{coverage()}}, 
#'          \code{GenomicRanges::\link[GenomicRanges:coverage-methods]{coverage()}}
#' @examples
#' rng <- as_iranges(data.frame(start = 1:10, width = 5))
#' compute_coverage(rng)
#' compute_coverage(rng, shift = 14L)
#' compute_coverage(rng, width = 10L)
#' @importFrom IRanges coverage ranges
#' @importFrom S4Vectors runValue
#' @export
compute_coverage <- function(x, shift, width, weight, ...) {
  UseMethod("compute_coverage")
}

#' @export
compute_coverage.default <- function(x, shift = 0L, width = NULL, weight = 1L, ...) {
  as_ranges(coverage(x, shift, width, weight, ...))
}


setMethod("coverage", "DelegatingGenomicRanges",
          function(x, shift = 0L, width = NULL, weight = 1L, ...) {
            coverage(load_delegate(x), shift, width, weight, ...)
          })

#' Read a WIG file
#'
#' @param file A path to a file or a connection.
#' @param genome_info An optional character string or a Ranges object
#' that contains information about the genome build. For example the USSC identifier
#'"hg19" will add build information to the returned GRanges.
#' @param overlap_ranges An optional Ranges object. Only the intervals in the file
#' that overlap the Ranges will be returned.
#'
#' @description This is a lightweight wrapper to the import family
#' of functions defined in \pkg{rtracklayer}.
#'
#' @return A GRanges object
#'
#' @importFrom rtracklayer import.wig
#' @seealso \code{rtracklayer::\link[rtracklayer:WIGFile-class]{WIGFile()}}
#' @rdname io-wig-read
#' @examples
#' test_path <- system.file("tests", package = "rtracklayer")
#' test_wig <- file.path(test_path, "step.wig")
#' gr <- read_wig(test_wig)
#' gr
#' gr <- read_wig(test_wig, genome_info = "hg19")
#' @return A GRanges object
#' @export
read_wig <- function(file, genome_info = NULL, overlap_ranges = NULL) {

  args <- norm_args_reader(genome_info)

  import.wig(file,
            trackLine = FALSE,
            genome = args$genome_info,
            seqinfo = args$seq_info,
            which = overlap_ranges)
}


#' Write a WIG file
#'
#' @param x A GRanges object
#' @param file File name, URL or connection specifying a file to write x to.
#'             Compressed files with extensions such as '.gz' are handled
#'             automatically.
#' @importFrom rtracklayer export.wig
#' @seealso \code{rtracklayer::\link[rtracklayer:WIGFile-class]{WIGFile()}}
#' @export
#' @return The write function returns a WIGFile invisibly.
#' @rdname io-wig-write
write_wig <- function(x, file) {
  export.wig(x, file)
}

context("reading/writing WIG files")

createCorrectGR <- function(seqinfo) {
  score <- c(seq(10, 20, by = 2.5), seq(17.5, 10, by = -2.5),
             seq(1000, 100, by = -100))
  start <- c(59104701 + cumsum(c(0, 200, 500, 200, 300, 180, 220, 390, 1180)),
             59108021 + cumsum(c(0, rep(300, 9))))
  width <- c(rep(1, 9), rep(200, 10))
  space <- factor(c(rep("chr19", 9), rep("chr18", 10)), seqlevels(seqinfo))
  correct_gr <- GRanges(space, IRanges(start, width = width), score = score)
  if (!any(is.na(genome(seqinfo))))
    genome(correct_gr) <- unname(genome(seqinfo)[1])
  seqinfo(correct_gr) <- seqinfo
  correct_gr
}

test_that("reading WIG files", {
  test_path <- system.file("tests", package = "rtracklayer")
  test_wig <- file.path(test_path, "step.wig")

  correct_gr <- createCorrectGR(Seqinfo(c("chr19", "chr18")))
  test_gr <- read_wig(test_wig)
  expect_identical(correct_gr, test_gr)
  test_wig_file <- WIGFile(test_wig)
  test_gr <- read_wig(test_wig_file)
  expect_identical(correct_gr, test_gr)

  test_wig_con <- file(test_wig)
  test_gr <- read_wig(test_wig_con)
  expect_identical(correct_gr, test_gr)


  ## genome_info checks
  hg19_seqinfo <- SeqinfoForBSGenome("hg19")
  correct_genome <- createCorrectGR(hg19_seqinfo)
  test_gr <- read_wig(test_wig, genome_info = "hg19")
  expect_identical(correct_genome, test_gr)

  hg19_gr <- get_genome_info(hg19_seqinfo)
  test_gr <- read_wig(test_wig, genome_info = hg19_gr)

  expect_identical(correct_genome, test_gr)

  # overlaps
  which <- correct_gr[3:4]
  correct_which <- filter_by_overlaps(correct_gr, which)
  test_gr <- read_wig(test_wig, overlap_ranges = which)
  expect_identical(correct_which, test_gr)

})

test_that("writing wig files", {

  correct_gr <- createCorrectGR(Seqinfo(c("chr19", "chr18")))

  test_wig_out <- file.path(tempdir(), "test.wig")
  on.exit(unlink(test_wig_out))
  write_wig(correct_gr, test_wig_out)
  test_gr <-read_wig(test_wig_out)
  expect_identical(correct_gr, test_gr)

  test_foo_out <- file.path(tempdir(), "test.foo")
  write_wig(correct_gr, test_foo_out)
  on.exit(unlink(test_foo_out))
  test_gr <- read_wig(test_wig_out)
  expect_identical(test_gr, correct_gr)

  test_wig_out_file <- rtracklayer::WIGFile(test_wig_out)
  write_wig(correct_gr, test_wig_out_file)
  test_gr <- read_wig(test_wig_out_file)
  expect_identical(test_gr, correct_gr)

  test_wig_gz <- paste(test_wig_out, ".gz", sep = "")
  on.exit(unlink(test_wig_gz))
  write_wig(correct_gr, test_wig_gz)
  test_gr <- read_wig(test_wig_gz)
  expect_identical(test_gr, correct_gr)

})

partition_rng <- function(rglist) {
  n <- elementNROWS(rglist)
  partitions <- rep.int(seq_along(n), n)
  res <- rglist@unlistData
  mcols(res)[["partition"]] <- partitions
  res
}


#' @rdname ranges-tile
#' @importFrom IRanges tile
#' @export
tile_ranges <- function(x, width) { UseMethod("tile_ranges") }

#' @export
tile_ranges.Ranges <- function(x, width) {
  rng <- tile(x, width = width)
  partition_rng(rng)
}

#' Slide or tile over a Ranges object
#' 
#' @param x a Ranges object
#' @param width the maximum width of each window/tile (integer vector of length 1)
#' @param step the distance between start position of each sliding window (integer vector of length 1)
#' 
#' @details The `tile_ranges()` function paritions a Ranges object `x` by the given the 
#' `width` over all ranges in `x`, truncated by the sequence end. 
#' The `slide_ranges()` function makes sliding windows within each range of `x`
#' of size `width` and sliding by `step`. 
#' Both `slide_ranges()` and `tile_ranges()` return a new Ranges object
#' with a metadata column called "partition" which contains the index of the
#' input range `x` that a parition belongs to.
#' 
#' @return a Ranges object
#' @importFrom IRanges slidingWindows
#' @examples
#'  
#' gr <- data.frame(seqnames = c("chr1", rep("chr2", 3), rep("chr1", 2), rep("chr3", 4)),
#'                  start = 1:10,
#'                  end = 11,
#'                  strand = c("-", rep("+", 2), rep("*", 2), rep("+", 3), rep("-", 2))) %>%
#'       as_granges() %>%
#'       set_genome_info(seqlengths = c(11,12,13))
#' 
#' # partition ranges into subranges of width 2, odd width ranges
#' # will have one subrange of width 1              
#' tile_ranges(gr, width = 2)
#' 
#' # make sliding windows of width 3, moving window with step size of 2
#' slide_ranges(gr, width = 3, step = 2)
#' 
#' @seealso \code{GenomicRanges::\link[GenomicRanges:tile-methods]{tile()}}
#' @rdname ranges-tile
#' @export
slide_ranges <- function(x, width, step) { UseMethod("slide_ranges") }

#' @export
slide_ranges.Ranges <- function(x, width, step) {
  rng <- slidingWindows(x, width = width, step = step)
  partition_rng(rng)
}

# Element wise set operations

#' Row-wise set operations on Ranges objects
#'
#' @param x,y Ranges objects
#'
#'
#' @details Each of these functions acts on the rows between pairs of
#' Ranges object.
#' The function \code{\%union\%()}.
#' will return the entire range between two ranges objects assuming there
#' are no gaps, if you would like to force gaps use [span()] instead.
#' The function \code{\%intersect\%()} will create a new ranges object
#' with a hit column indicating whether or not the two ranges intersect.
#' The function \code{\%setdiff\%()}will return the ranges for each
#' row in x that are not in the corresponding row of y.
#' The function [between()] will return the gaps between
#' two ranges.
#'
#' @return A Ranges object
#' @importFrom IRanges punion pintersect pgap psetdiff
#' @seealso [IRanges::punion()][IRanges::pintersect()][IRanges::pgap()][IRanges::psetdiff()]
#' @examples
#' x <- as_iranges(data.frame(start = 1:10, width = 5))
#' # stretch x by 3 on the right
#' y <- stretch(anchor_start(x), 3)
#' # take the rowwise union
#' x %union% y
#' # take the rowwise intersection
#' x %intersect% y
#' # asymetric difference
#' y %setdiff% x
#' x %setdiff% y
#' # if there are gaps between the rows of each range use span
#' y <- as_iranges(data.frame(start = c(20:15, 2:5),
#' width = c(10:15,1:4)))
#' # fill in the gaps and take the rowwise union
#' span(x,y)
#' # find the gaps
#' between(x,y)
#' @export
#' @rdname element-setops
`%union%` <- function(x, y) {
  punion(x,y, fill.gap = FALSE)
}

#' @export
#' @rdname element-setops
`%intersect%` <- function(x,y) {
  pintersect(x,y)
}
#' @export
#' @rdname element-setops
`%setdiff%` <- function(x, y) {
  psetdiff(x,y)
}

#' @export
#' @rdname element-setops
between <- function(x, y) {
  pgap(x,y)
}

#' @export
#' @rdname element-setops
span <- function(x, y) {
  punion(x, y, fill.gap = TRUE)
}

# ranges-eval-utils.R
# some helpers for 'tidy' evaluation on ranges


#' Create an overscoped environment from a Ranges object
#' 
#' @param x a Ranges object
#' @param envir the environment to place the Ranges in (default = `parent.frame()`)
#' 
#' @details This is the backend for non-standard evaluation in `plyranges`.
#' 
#' @seealso [rlang::new_data_mask()], [rlang::eval_tidy()]
#' @return an environment
#' 
#' @export
overscope_ranges <- function(x, envir = parent.frame()) {
  UseMethod("overscope_ranges")
}

#' @export
overscope_ranges.Ranges <- function(x, envir = parent.frame()) {
  env <- as.env(x, envir)
  
  new_data_mask(env, top = parent.env(env))
}

#' @export
overscope_ranges.DelegatingGenomicRanges <- function(x, envir = parent.frame()) {
  overscope_ranges(x@delegate, envir)
}

#' @export
overscope_ranges.DelegatingIntegerRanges <- overscope_ranges.DelegatingGenomicRanges

#' @export
overscope_ranges.GroupedGenomicRanges <- function(x, envir = parent.frame()) {
  env <- as.env(x@delegate, 
                envir, 
                tform = function(col) unname(S4Vectors::splitAsList(col, x@group_indices)))
  new_data_mask(env, top = parent.env(env))
}


#' @export
overscope_ranges.GroupedIntegerRanges <- overscope_ranges.GroupedGenomicRanges



#' @importFrom rlang env_bind := new_data_mask eval_tidy
overscope_eval_update <- function(overscope, dots, bind_envir = TRUE) {
  update <- vector("list", length(dots))
  names(update) <- names(dots)
  for (i in seq_along(update)) {
    quo <- dots[[i]]
    update[[i]] <- eval_tidy(quo, data = overscope)
    # sometimes we want to compute on previously constructed columns
    # we can do this by binding the evaluated expression to
    # the overscope environment
    if (bind_envir) {
      new_col <- names(dots)[[i]]
      rlang::env_bind(overscope, !!new_col := update[[i]])
    }
  }
  return(update)
}


# Port of dplyrs `n` function
# It works by searching for a vector in the overscope environment
# and calling length on it.

#' Compute the number of ranges in each group.
#'
#' @description This function should only be used
#' within `summarise()`, `mutate()` and `filter()`.
#'
#' @examples
#' ir <- as_iranges(
#'                  data.frame(start = 1:10,
#'                             width = 5,
#'                             name = c(rep("a", 5), rep("b", 3), rep("c", 2))
#'                             )
#'                 )
#' by_names <- group_by(ir, name)
#' summarise(by_names, n = n())
#' mutate(by_names, n = n())
#' filter(by_names, n() >= 3)
#' @return `n()` will only be evaluated inside a function call, where it
#' returns an integer.
#'
#' @importFrom rlang env_get env_parent
#' @export
n <- function() {
  up_env <- parent.frame()
  parent_env <- rlang::env_parent(up_env)
  if (rlang::env_has(parent_env, "start")) {
    .data <- rlang::env_get(parent_env, "start")
    if (is(.data, "IntegerList")) {
      return(lengths(.data))
    } else {
      return(length(.data))
    }
  }
  stop("This function should not be called directly")
}


#' Compute the number of distinct unique values in a vector or List
#' 
#' @param var a vector of values
#' @return an integer vector 
#' 
#' @description This is a wrapper to `length(unique(x))` or 
#' `lengths(unique(x))` if `x` is a List object
#' 
#' @examples 
#' x <- CharacterList(c("a", "b", "c", "a"),  "d")
#' n_distinct(x)
#' n_distinct(unlist(x))
#' @importFrom BiocGenerics unique
#' @export
n_distinct <- function(var) {
  if (inherits(var, "List")) {
    return(lengths(BiocGenerics::unique(var)))
  } else {
    return(length(BiocGenerics::unique(var)))
  }
}

is_empty_quos <- function(quos) {
  length(quos) == 0L
}

context("Column wise selection")

gr0 <- GRanges()
ir0 <- IRanges()

mdata <- DataFrame(score = as.integer(c(0,3,4,5,10,14,0, 0, 4,8)),
                   gc = runif(10),
                   counts = rpois(10, 2))
ir1 <- IRanges(start = 1:10, width = 5)
mcols(ir1) <- mdata
gr1 <- GRanges(seqnames = "seq1",
               strand = c(rep("+", 4), rep("-", 3), rep("*", 3)),
               ranges = ir1)


test_that("no selection returns the same object", {
  expect_identical(select(ir0), ir0)
  expect_identical(select(gr0), gr0)
  expect_identical(select(ir1), ir1)
  expect_identical(select(gr1), gr1)
})

test_that("dropping core parts throws an error...", {
  expect_error(select(ir0, start))
  expect_error(select(gr0, start))
})

test_that("...unless allowing to drop ranges", {
  res <- select(ir0, start, .drop_ranges = TRUE)
  
  expect_s4_class(res, "DataFrame")
  expect_equal(ncol(res), 1)
  expect_equal(length(res$start), 0)
  
  res <- select(gr0, start, .drop_ranges = TRUE)
  expect_s4_class(res, "DataFrame")
  expect_equal(ncol(res), 1)
  expect_equal(length(res$start), 0)
  
  expect_identical(select(ir1, score, gc, counts, .drop_ranges = TRUE),
                   mcols(ir1))
  expect_identical(select(gr1, score, gc, counts, .drop_ranges = TRUE),
                   mcols(gr1))

})

test_that("reordering/dropping works as expected", {
  gr2 <- select(gr1, counts, gc, score)
  expect_equal(names(mcols(gr2)), c("counts", "gc", "score"))
  gr3 <- select(gr1, -1)
  gr4 <- select(gr1, -score)
  expect_identical(gr3, gr4)
  ir2 <- select(ir1, -counts)
  expect_equal(names(mcols(ir2)), c("score", "gc"))
  ir3 <- select(ir1, 1:2)
  expect_identical(ir2, ir3)
})

test_that("dropping everything sets mcols slot to empty", {
  gr5 <- select(gr1, -score:-counts)
  mcols(gr1) <- NULL
  expect_identical(gr5, gr1)
  ir4 <- select(ir1, -score:-counts)
  expect_equivalent(ir4, IRanges(ir1))
})

context("colwise mutate/summarise")

mdata <- DataFrame(score = as.integer(c(0,3,4,5,10,14,0, 0, 4,8)),
                   grp = rep(c("A", "B"), 5))

ir1 <- IRanges(start = 1:10, width = 5)
mcols(ir1) <- mdata
gr1 <- GRanges(seqnames = "seq1",
               strand = c(rep("+", 4), rep("-", 3), rep("*", 3)),
               ranges = ir1)

test_that("summarise evaluates correctly", {
  df <- DataFrame(mean = 4.8,  n = 10L)
  expect_equal(summarise(ir1, mean = mean(score), n = n()), df)
  expect_equal(summarise(gr1, mean = mean(score), n = n()), df)

  test <- ir1 %>% group_by(grp)
  gdf <- DataFrame(grp = c("A", "B"), mean = c(3.6, 6), sum = c(18L, 30L))
  expect_equal(ir1 %>% 
                 group_by(grp) %>% 
                 summarise(mean = mean(score), sum = sum(score)), 
               gdf)
  expect_equal(gr1 %>% 
                 group_by(grp) %>% 
                 summarise(mean = mean(score), sum = sum(score)), 
               gdf)

  gdf <- DataFrame(grp = c("A", "B"), n = c(5L,5L))
  expect_equal(ir1 %>% group_by(grp) %>% summarise(n = n()), gdf)
  expect_equal(gr1 %>% group_by(grp) %>% summarise(n = n()), gdf)
})

test_that("mutate allows out of scope columns", {
  gr2 <- gr1
  mcols(gr2)$score2 <- score(gr2) + 1L
  mcols(gr2)$score3 <- mcols(gr2)$score2 * 2L
  expect_identical(gr1 %>%
                     mutate(score2 = score + 1L,
                            score3 = score2*2L),
                   gr2)

  ir2 <- ir1
  mcols(ir2)$score2 <- mcols(ir2)$score*4L
  mcols(ir2)$score3 <- mcols(ir2)$score2 + 3L
  expect_identical(ir1 %>%
                     mutate(score2 = score * 4L,
                            score3 = score2 + 3L), ir2)

})

test_that("mutating by groups", {
  gr2 <- gr1
  mcols(gr2)$gt_grp_score <- (score(gr1) > 3.6 & mcols(gr1)$grp == "A") |
    (score(gr1) > 6 & mcols(gr1)$grp == "B")
  expect_identical(gr1 %>%
                     group_by(grp) %>%
                     mutate(gt_grp_score = score > mean(score)) %>%
                     ungroup(),
                   gr2)
  ir2 <- ir1
  mcols(ir2)$lt_grp_score <- (mcols(ir1)$score < 3.6 & mcols(ir1)$grp == "A") |
    (mcols(ir1)$score < 6 & mcols(ir1)$grp == "B")
  expect_identical(ir1 %>%
                     group_by(grp) %>%
                     mutate(lt_grp_score = score < mean(score)) %>%
                     ungroup(),
                   ir2)

})




test_that("mutating by groups with S4 columns", {
  S4_column <- IntegerList(
    a = NULL,
    b = c(4, 5),
    c = 3,
    d = c(2, 5),
    e = 1,
    f = c(6, 7, 8),
    g = NULL,
    h = c(5, 6),
    i = c(9, 10, 11),
    j = NULL
  )
  mcols(gr1)$exon_id <- S4_column
  mcols(ir1)$exon_id <- S4_column
  
  gr2 <- gr1
  mcols(gr2)$gt_grp_score <- (score(gr1) > 3.6 &
                                mcols(gr1)$grp == "A") |
    (score(gr1) > 6 & mcols(gr1)$grp == "B")
  expect_identical(gr1 %>%
                     group_by(grp) %>%
                     mutate(gt_grp_score = score > mean(score)) %>%
                     ungroup(),
                   gr2)
  ir2 <- ir1
  mcols(ir2)$lt_grp_score <- (mcols(ir1)$score < 3.6 &
                                mcols(ir1)$grp == "A") |
    (mcols(ir1)$score < 6 & mcols(ir1)$grp == "B")
  expect_identical(ir1 %>%
                     group_by(grp) %>%
                     mutate(lt_grp_score = score < mean(score)) %>%
                     ungroup(),
                   ir2)
  
})

---
title: "Additional examples of plyranges"
author: "Stuart Lee"
package: plyranges
date: "`r Sys.Date()`"
output:
  BiocStyle::html_document:
    toc_float: true
vignette: >
  %\VignetteIndexEntry{Additional examples of plyranges}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
---

# Quick overview

## About `Ranges`

`Ranges` objects can either represent sets of integers as `IRanges` (which have
start, end and width attributes) or represent genomic intervals (which have
additional attributes, sequence name, and strand) as `GRanges`.  In addition,
both types of `Ranges` can store information about their intervals as metadata
columns (for example GC content over a genomic interval).

`Ranges` objects follow the tidy data principle: each row of a `Ranges` object
corresponds to an interval, while each column will represent a variable about
that interval, and generally each object will represent a single unit of
observation (like gene annotations).

We can construct a `IRanges` object from a `data.frame` with a `start` or
`width` using the `as_iranges()` method.

```{r, message=FALSE}
library(plyranges)
df <- data.frame(start = 1:5, width = 5)
as_iranges(df)
# alternatively with end
df <- data.frame(start = 1:5, end = 5:9)
as_iranges(df)
```

We can also construct a `GRanges` object in a similar manner. Note that a
`GRanges` object requires at least a seqnames column to be present in the
data.frame (but not necessarily a strand column).

```{r}
df <- data.frame(seqnames = c("chr1", "chr2", "chr2", "chr1", "chr2"),
                 start = 1:5,
                 width = 5)
as_granges(df)
# strand can be specified with `+`, `*` (mising) and `-`
df$strand <- c("+", "+", "-", "-", "*")
as_granges(df)
```

# Example: finding GWAS hits that overlap known exons
Let's look at a more a realistic example (taken from HelloRanges vignette).

```{r, include=FALSE}
dir <- system.file(package = "HelloRangesData", "extdata/")
genome <- as_granges(read.delim(file.path(dir, "hg19.genome"),
                     header = FALSE),
                     seqnames = V1, start = 1L, width = V2)

gwas <- read_bed(file.path(dir, "gwas.bed"), genome_info = genome)
exons <- read_bed(file.path(dir, "exons.bed"), genome_info = genome)
```

Suppose we have two _GRanges_ objects: one containing coordinates of known
exons and another containing SNPs from a GWAS.

The first and last 5 exons are printed below, there are two additional columns
corresponding to the exon name, and a score.

We could check the number of exons per chromosome using `group_by` and
`summarise`.
```{r}
exons
exons %>%
  group_by(seqnames) %>%
  summarise(n = n())
```

Next we create a column representing the transcript_id with `mutate`:

```{r}
exons <- exons %>%
  mutate(tx_id = sub("_exon.*", "", name))
```

To find all GWAS SNPs that overlap exons, we use `join_overlap_inner`. This
will create a new _GRanges_ with the coordinates of SNPs that overlap exons, as
well as metadata from both objects.

```{r}
olap <- join_overlap_inner(gwas, exons)
olap
```

For each SNP we can count the number of times it overlaps a transcript.

```{r}
olap %>%
  group_by(name.x, tx_id) %>%
  summarise(n = n())
```

We can also generate 2bp splice sites on either side of the exon using
`flank_left` and `flank_right`. We add a column indicating the side of flanking
for illustrative purposes. The `interweave` function pairs the left and right
ranges objects.

```{r}
left_ss <- flank_left(exons, 2L)
right_ss <- flank_right(exons, 2L)
all_ss <- interweave(left_ss, right_ss, .id = "side")
all_ss
```

# Session information

```{r}
sessionInfo()
```

---
title: "Getting started with the plyranges package"
author: "Stuart Lee"
package: plyranges
date: "`r Sys.Date()`"
output:
  BiocStyle::html_document:
    toc_float: true
vignette: >
  %\VignetteIndexEntry{Introduction}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
---

# `Ranges` revisited

In Bioconductor there are two classes, `IRanges` and `GRanges`, that are
standard data structures for representing genomics data.  Throughout this
document I refer to either of these classes as `Ranges` if an operation can be
performed on either class, otherwise I explicitly mention if a function is
appropriate for an `IRanges` or `GRanges`.

`Ranges` objects can either represent sets of integers as `IRanges` (which have
start, end and width attributes) or represent genomic intervals (which have
additional attributes, sequence name, and strand) as `GRanges`. In addition,
both types of `Ranges` can store information about their intervals as metadata
columns (for example GC content over a genomic interval).

`Ranges` objects follow the tidy data principle: each row of a `Ranges` object
corresponds to an interval, while each column will represent a variable about
that interval, and generally each object will represent a single unit of
observation (like gene annotations).

Consequently, `Ranges` objects provide a powerful representation for reasoning
about genomic data. In this vignette, you will learn more about `Ranges`
objects and how via grouping, restriction and summarisation you can perform
common data tasks.


# Constructing `Ranges`

To construct an `IRanges` we require that there are at least two columns that
represent at either a starting coordinate, finishing coordinate or the width of
the interval.

```{r}
suppressPackageStartupMessages(library(plyranges))
set.seed(100)
df <- data.frame(start=c(2:-1, 13:15),
                 width=c(0:3, 2:0))

# produces IRanges
rng <- df %>% as_iranges()
rng

```

To construct a `GRanges` we require a column that represents that sequence name
( contig or chromosome id), and an optional column to represent the
strandedness of an interval.

```{r}
# seqname is required for GRanges, metadata is automatically kept
grng <- df %>%
  transform(seqnames = sample(c("chr1", "chr2"), 7, replace = TRUE),
         strand = sample(c("+", "-"), 7, replace = TRUE),
         gc = runif(7)) %>%
  as_granges()

grng
```

# Arithmetic on Ranges

Sometimes you want to modify a genomic interval by altering the width of the
interval while leaving the start, end or midpoint of the coordinates unaltered.
This is achieved with the `mutate` verb along with `anchor_*` adverbs.

The act of anchoring fixes either the start, end, center coordinates of the
`Range` object, as shown in the figure and code below and anchors are used in
combination with either `mutate` or `stretch`.
By default, the start coordinate will be anchored, so regardless of strand. 
For behavior similar to `GenomicRanges::resize`, use `anchor_5p`.

```{r anchor_fig, echo = FALSE, out.width="400px", fig.align="center"}
knitr::include_graphics("anchors.png", dpi = 150)
```

```{r}
rng <- as_iranges(data.frame(start=c(1, 2, 3), end=c(5, 2, 8)))
grng <- as_granges(data.frame(start=c(1, 2, 3), end=c(5, 2, 8),
                          seqnames = "seq1",
                          strand = c("+", "*", "-")))
mutate(rng, width = 10)
mutate(anchor_start(rng), width = 10)
mutate(anchor_end(rng), width = 10)
mutate(anchor_center(rng), width = 10)
mutate(anchor_3p(grng), width = 10) # leave negative strand fixed
mutate(anchor_5p(grng), width = 10) # leave positive strand fixed
```

Similarly, you can modify the width of an interval using the `stretch` verb.
Without anchoring, this function will extend the interval in either direction
by an integer amount. With anchoring, either the start, end or midpoint are
preserved.

```{r}
rng2 <- stretch(anchor_center(rng), 10)
rng2
stretch(anchor_end(rng2), 10)
stretch(anchor_start(rng2), 10)
stretch(anchor_3p(grng), 10)
stretch(anchor_5p(grng), 10)
```

`Ranges` can be shifted left or right. If strand information is available we
can also shift upstream or downstream.

```{r}
shift_left(rng, 100)
shift_right(rng, 100)
shift_upstream(grng, 100)
shift_downstream(grng, 100)
```

# Grouping `Ranges`

`plyranges` introduces a new class of `Ranges` called `RangesGrouped`, this is
a similar idea to the grouped `data.frame\tibble` in `dplyr`.

Grouping can act on either the core components or the metadata columns of a
`Ranges` object.

It is most effective when combined with other verbs such as `mutate()`,
`summarise()`, `filter()`, `reduce_ranges()` or `disjoin_ranges()`.
```{r}
grng <- data.frame(seqnames = sample(c("chr1", "chr2"), 7, replace = TRUE),
         strand = sample(c("+", "-"), 7, replace = TRUE),
         gc = runif(7),
         start = 1:7,
         width = 10) %>%
  as_granges()

grng_by_strand <- grng %>%
  group_by(strand)

grng_by_strand
```

# Restricting `Ranges`

The verb `filter` can be used to restrict rows in the `Ranges`. Note that
grouping will cause the `filter` to act within each group of the data.
```{r}
grng %>% filter(gc < 0.3)
# filtering by group
grng_by_strand %>% filter(gc == max(gc))
```

We also provide the convenience methods `filter_by_overlaps` and
`filter_by_non_overlaps` for restricting by any overlapping `Ranges`.

```{r}
ir0 <- data.frame(start = c(5,10, 15,20), width = 5) %>%
  as_iranges()
ir1 <- data.frame(start = 2:6, width = 3:7) %>%
  as_iranges()
ir0
ir1
ir0 %>% filter_by_overlaps(ir1)
ir0 %>% filter_by_non_overlaps(ir1)
```

# Summarising `Ranges`

The `summarise` function will return a `DataFrame` because the information
required to return a `Ranges` object is lost. It is often most useful to use
`summarise()` in combination with the `group_by()` family of functions.

```{r}
ir1 <- ir1 %>%
  mutate(gc = runif(length(.)))

ir0 %>%
  group_by_overlaps(ir1) %>%
  summarise(gc = mean(gc))
```

# Joins, or another way at looking at overlaps between `Ranges`

A join acts on two GRanges objects, a query and a subject.

```{r}
query <- data.frame(seqnames = "chr1",
               strand = c("+", "-"),
               start = c(1, 9),
               end =  c(7, 10),
               key.a = letters[1:2]) %>%
  as_granges()

subject <- data.frame(seqnames = "chr1",
               strand = c("-", "+"),
               start = c(2, 6),
               end = c(4, 8),
               key.b = LETTERS[1:2]) %>%
  as_granges()
```

```{r olap, echo = FALSE, fig.cap = "Query and Subject Ranges", message = FALSE}
library(ggplot2)
query_df <- as.data.frame(query)[, -6]
query_df$key <- "Query"
subject_df <- as.data.frame(subject)[, -6]
subject_df$key <- "Subject"
melted_ranges <- rbind(query_df, subject_df)
ggplot(melted_ranges, aes(xmin = start, xmax = end, ymin = 1, ymax = 3)) +
  geom_rect() +
  facet_grid(key ~ .) +
  scale_x_continuous(breaks = seq(1, 10, by = 1)) +
  xlab("Position") +
  theme(axis.text.y = element_blank(),
        axis.ticks.y = element_blank(),
        axis.title.y = element_blank())
```

The join operator is relational in the sense that metadata from the query and
subject ranges is retained in the joined range.  All join operators in the
`plyranges` DSL generate a set of hits based on overlap or proximity of ranges
and use those hits to merge the two datasets in different ways.  There are four
supported matching algorithms: _overlap_, _nearest_, _precede_, and _follow_.
We can further restrict the matching by whether the query is completely
_within_ the subject, and adding the _directed_ suffix ensures that matching
ranges have the same direction (strand).


```{r olaps-diagram, echo = FALSE, out.width="600px"}
knitr::include_graphics("olaps.png")
```

The first function, `join_overlap_intersect()` will return a `Ranges` object
where the start, end, and width coordinates correspond to the amount of any
overlap between the left and right input `Ranges`. It also returns any
metadatain the subject range if the subject overlaps the query.

```{r}
intersect_rng <- join_overlap_intersect(query, subject)
intersect_rng
```

```{r intersect-join, echo = FALSE, fig.cap="Intersect Join"}
intersect_df <- as.data.frame(intersect_rng)[, -c(6,7)]
intersect_df$key <- "Intersect Join"
melted_ranges <- rbind(query_df, subject_df, intersect_df)
melted_ranges$key <- factor(melted_ranges$key,
                              levels = c("Query", "Subject", "Intersect Join"))
ggplot(melted_ranges, aes(xmin = start, xmax = end, ymin = 1, ymax = 3)) +
  geom_rect() +
  facet_grid(key ~ .) +
  scale_x_continuous(breaks = seq(1, 10, by = 1)) +
  xlab("Position") +
  theme(axis.text.y = element_blank(),
        axis.ticks.y = element_blank(),
        axis.title.y = element_blank())
```


The `join_overlap_inner()` function will return the `Ranges` in the query that
overlap any `Ranges` in the subject. Like the `join_overlap_intersect()`
function metadata of the subject `Range` is returned if it overlaps the query.

```{r}
inner_rng <- join_overlap_inner(query, subject)
inner_rng
```


```{r inner-join, echo = FALSE, fig.cap = "Inner Join", message = FALSE}
inner_df <- as.data.frame(inner_rng)[, -c(6,7)]
inner_df$ymin <- c(1,4)
inner_df$ymax <- c(3,6)
inner_df$key <- "Inner Join"
melted_ranges <- rbind(query_df, subject_df)
melted_ranges$ymin <- 1
melted_ranges$ymax <- 3
melted_ranges <- rbind(melted_ranges, inner_df)
melted_ranges$key <- factor(melted_ranges$key,
                              levels = c("Query", "Subject", "Inner Join"))

ggplot(melted_ranges, aes(xmin = start, xmax = end, ymin = ymin, ymax = ymax)) +
  geom_rect() +
  facet_grid(key ~ ., scales = "free_y") +
  scale_x_continuous(breaks = seq(1, 10, by = 1)) +
  xlab("Position") +
  theme(axis.text.y = element_blank(),
        axis.ticks.y = element_blank(),
        axis.title.y = element_blank())
```

We also provide a convenience method called `find_overlaps` that computes the
same result as `join_overlap_inner()`.

```{r}
find_overlaps(query, subject)
```

The `join_overlap_left()` method will perform an outer left join.


First any overlaps that are found will be returned similar to
`join_overlap_inner()`. Then any non-overlapping ranges will be returned, with
missing values on the metadata columns.

```{r}
left_rng <- join_overlap_left(query, subject)
left_rng
```

```{r olap-left, echo = FALSE, fig.cap = "Left Join", message = FALSE}
left_df <- as.data.frame(left_rng)[, -c(6,7)]
left_df$ymin <- c(1,4, 1)
left_df$ymax <- c(3,6, 3)
left_df$key <- "Left Join"
melted_ranges <- rbind(query_df, subject_df)
melted_ranges$ymin <- 1
melted_ranges$ymax <- 3
melted_ranges <- rbind(melted_ranges, left_df)
melted_ranges$key <- factor(melted_ranges$key,
                              levels = c("Query", "Subject", "Left Join"))

ggplot(melted_ranges,
       aes(xmin = start, xmax = end, ymin = ymin, ymax = ymax)) +
  geom_rect() +
  facet_grid(key ~ ., scales = "free_y") +
  scale_x_continuous(breaks = seq(1, 10, by = 1)) +
  xlab("Position") +
  theme(axis.text.y = element_blank(),
        axis.ticks.y = element_blank(),
        axis.title.y = element_blank())
```

Compared with `filter_by_overlaps()` above, the overlap left join expands the
`Ranges` to give information about each interval on the query `Ranges` that
overlap those on the subject `Ranges` as well as the intervals on the left that
do not overlap any range on the right.

## Finding your neighbours
We also provide methods for finding nearest, preceding or following `Ranges`.
Conceptually this is identical to our approach for finding overlaps, except the
semantics of the join are different.

```{r neigbours, echo = FALSE, out.width="600px"}
knitr::include_graphics("neighbours.png")
```

```{r}
join_nearest(ir0, ir1)
join_follow(ir0, ir1)
join_precede(ir0, ir1) # nothing precedes returns empty `Ranges`
join_precede(ir1, ir0)
```


## Example: dealing with multi-mapping
This example is taken from the Bioconductor support
[site](https://support.bioconductor.org/p/100046/).

We have two `Ranges` objects. The first contains single nucleotide positions
corresponding to an intensity measurement such as a ChiP-seq experiment, while
the other contains coordinates for two genes of interest.


We want to identify which positions in the `intensities` `Ranges` overlap the
genes, where each row corresponds to a position that overlaps a single gene.

First we create the two `Ranges` objects
```{r ex1}
intensities <- data.frame(seqnames = "VI",
                          start = c(3320:3321,3330:3331,3341:3342),
                          width = 1) %>%
  as_granges()

intensities

genes <- data.frame(seqnames = "VI",
                    start = c(3322, 3030),
                    end = c(3846, 3338),
                    gene_id=c("YFL064C", "YFL065C")) %>%
  as_granges()

genes
```

Now to find where the positions overlap each gene, we can perform an overlap
join. This will automatically carry over the gene_id information as well as
their coordinates (we can drop those by only selecting the gene_id).
```{r}
olap <- join_overlap_inner(intensities, genes) %>%
  select(gene_id)
olap
```

Several positions match to both genes. We can count them using `summarise` and
grouping by the `start` position:

```{r}
olap %>%
  group_by(start) %>%
  summarise(n = n())
```

## Grouping by overlaps

It's also possible to group by overlaps. Using this approach we can count the
number of overlaps that are greater than 0.

```{r}
grp_by_olap <- ir0 %>%
  group_by_overlaps(ir1)
grp_by_olap
grp_by_olap %>%
  mutate(n_overlaps = n())
```

Of course we can also add overlap counts via the `count_overlaps()` function.

```{r}
ir0 %>%
  mutate(n_overlaps = count_overlaps(., ir1))
```

# Data Import/Output

We provide convenience functions via `rtracklayer` and `GenomicAlignments` for
reading/writing the following data formats to/from `Ranges` objects.

| `plyranges` functions | File Format |
|-----------------------|-------------|
| `read_bam()` | BAM | 
| `read_bed()`/`write_bed()` | BED | 
| `read_bed_graph()`/ `write_bed_graph()` | BEDGraph | 
| `read_narrowpeaks()`/`write_narrowpeaks()` | narrowPeaks | 
| `read_gff()` / `write_gff()` | GFF(1-3)/ GTF | 
| `read_bigwig()` / `write_bigwig()` | BigWig | 
| `read_wig()` /`write_wig()` | Wig |


# Learning more

There are many other resources and workshops available to learn to use
`plyranges` and related Bioconductor packages, especially for more realistic
analyses than the ones covered here:

- The [fluentGenomics](https://bioconductor.org/packages/release/workflows/html/fluentGenomics.html) workflow package is an end-to-end workflow package for integrating differential
expression results with differential accessibility results.
- The [Bioc 2018 Workshop book](https://bioconductor.github.io/BiocWorkshops/fluent-genomic-data-analysis-with-plyranges.html) has worked examples of using `plyranges` to analyse publicly available genomics data.
- The [extended vignette in the plyrangesWorkshops package](https://github.com/sa-lee/plyrangesWorkshops) has a detailed
walk through of using plyranges for coverage analysis.
- The [case study](https://github.com/mikelove/plyrangesTximetaCaseStudy) by Michael Love using plyranges with [tximeta](https://bioconductor.org/packages/release/bioc/html/tximeta.html) to follow
up on interesting hits from a combined RNA-seq and ATAC-seq analysis.
- The [journal article](https://genomebiology.biomedcentral.com/articles/10.1186/s13059-018-1597-8) ([preprint here](https://www.biorxiv.org/content/early/2018/05/23/327841)) has
details about the overall philosophy and design of plyranges.

# Session information

```{r}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-Operator.R
\docType{class}
\name{FileOperator-class}
\alias{FileOperator-class}
\alias{BamFileOperator-class}
\title{An abstract class to represent operations performed over a file}
\description{
An abstract class to represent operations performed over a file
}
\details{
This class is used internally by DeferredGenomicRanges objects.
Currently, this class is only implemented for bam files (as a
BamFileOperator) but will eventually be extended to the other avaialable
readers.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-bind.R
\name{interweave}
\alias{interweave}
\title{Interweave a pair of Ranges objects together}
\usage{
interweave(left, right, .id = NULL)
}
\arguments{
\item{left, right}{Ranges objects.}

\item{.id}{When supplied a new column that represents the origin column
and is linked to each row of the resulting Ranges object.}
}
\value{
a Ranges object
}
\description{
Interweave a pair of Ranges objects together
}
\details{
The output of \code{interweave()} takes pairs of Ranges
objects and combines them into a single  Ranges object. If an .id
argument is supplied, an origin column with name .id is created indicated which side
the resulting Range comes from (eit)
}
\examples{
gr <- as_granges(data.frame(start = 10:15,
                            width = 5,
                            seqnames = "seq1",
                            strand = c("+", "+", "-", "-", "+", "*")))
interweave(flank_left(gr, width = 5L), flank_right(gr, width = 5L))
interweave(flank_left(gr, width = 5L), flank_right(gr, width = 5L), .id = "origin")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-GroupedRanges.R, R/dplyr-groups.R
\docType{class}
\name{GroupedGenomicRanges-class}
\alias{GroupedGenomicRanges-class}
\alias{GroupedIntegerRanges-class}
\alias{group_by-ranges}
\alias{group_by.GenomicRanges}
\alias{ungroup.GroupedGenomicRanges}
\alias{groups.GroupedGenomicRanges}
\alias{groups.GroupedIntegerRanges}
\title{Group a Ranges by one or more variables}
\usage{
\method{group_by}{GenomicRanges}(.data, ..., add = FALSE)

\method{ungroup}{GroupedGenomicRanges}(x, ...)

\method{groups}{GroupedGenomicRanges}(x)

\method{groups}{GroupedIntegerRanges}(x)
}
\arguments{
\item{.data}{a Ranges object.}

\item{...}{Variable names to group by. These can be either metadata columns
or the core variables of a Ranges.}

\item{add}{if \code{.data} is already a GroupedRanges object, when add = FALSE
the (default), \code{group_by()} will override existing groups. If add = TRUE,
additional groups will be added.}

\item{x}{a GroupedRanges object.}
}
\value{
The \code{group_by()} function will return a GroupedRanges object.
These have the same appearance as a regular Ranges object but with an
additional groups slot.
}
\description{
The function \code{group_by} takes a Ranges object and defines
groups by one or more variables. Operations are then performed on the Ranges
by their "group". \code{ungroup()} removes grouping.
}
\details{
\code{group_by()} creates a new object of class \code{GroupedGenomicRanges} if
the input is a \code{GRanges} object or an object of class \code{GroupedIntegerRanges}
if the input is a \code{IRanges} object. Both of these classes contain a slot
called \code{groups} corresponding to the names of grouping variables. They
also inherit from their parent classes, \code{Ranges} and \code{GenomicRanges}
respectively. \code{ungroup()} removes the grouping and will return
either a \code{GRanges} or \code{IRanges} object.
}
\section{Accessors}{

To return grouping variables on a grouped Ranges use either
\describe{
\item{\code{groups(x)}}{Returns a list of symbols}
\item{\code{group_vars(x)}}{Returns a character vector}
}
}

\examples{
set.seed(100)
df <- data.frame(start = 1:10,
                 width = 5,
                 gc = runif(10),
                 cat = sample(letters[1:2], 10, replace = TRUE))
rng <- as_iranges(df)
rng_by_cat <- rng \%>\% group_by(cat)
# grouping does not change appearance or shape of Ranges
rng_by_cat
# a list of symbols
groups(rng_by_cat)
# ungroup removes any grouping
ungroup(rng_by_cat)
# group_by works best with other verbs
grng <- as_granges(df,
                   seqnames = "chr1",
                   strand = sample(c("+", "-"), size = 10, replace = TRUE))

grng_by_strand <- grng \%>\% group_by(strand)
grng_by_strand
# grouping with other verbs
grng_by_strand \%>\% summarise(gc = mean(gc))
grng_by_strand \%>\% filter(gc == min(gc))
grng_by_strand \%>\%
  ungroup() \%>\%
  summarise(gc = mean(gc))


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-bind.R
\name{bind_ranges}
\alias{bind_ranges}
\title{Combine Ranges by concatentating them together}
\usage{
bind_ranges(..., .id = NULL)
}
\arguments{
\item{...}{Ranges objects to combine. Each argument can be a Ranges object,
or a list of Ranges objects.}

\item{.id}{Ranges object identifier. When .id is supplied a new column
is created that links each row to the original Range object. The contents
of the column correspond to the named arguments or the names of the list
supplied.}
}
\value{
a concatenated Ranges object
}
\description{
Combine Ranges by concatentating them together
}
\note{
Currently GRangesList or IRangesList objects are not supported.
}
\examples{
gr <- as_granges(data.frame(start = 10:15,
                            width = 5,
                            seqnames = "seq1"))
gr2 <- as_granges(data.frame(start = 11:14,
                            width = 1:4,
                            seqnames = "seq2"))
bind_ranges(gr, gr2)

bind_ranges(a = gr, b = gr2, .id = "origin")

bind_ranges(gr, list(gr, gr2), gr2)

bind_ranges(list(a = gr, b = gr2), c = gr, .id = "origin")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-eval.R
\name{overscope_ranges}
\alias{overscope_ranges}
\title{Create an overscoped environment from a Ranges object}
\usage{
overscope_ranges(x, envir = parent.frame())
}
\arguments{
\item{x}{a Ranges object}

\item{envir}{the environment to place the Ranges in (default = \code{parent.frame()})}
}
\value{
an environment
}
\description{
Create an overscoped environment from a Ranges object
}
\details{
This is the backend for non-standard evaluation in \code{plyranges}.
}
\seealso{
\code{\link[rlang:as_data_mask]{rlang::new_data_mask()}}, \code{\link[rlang:eval_tidy]{rlang::eval_tidy()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io-bigwig.R
\name{write_bigwig}
\alias{write_bigwig}
\title{Write a BigWig file}
\usage{
write_bigwig(x, file)
}
\arguments{
\item{x}{A GRanges object}

\item{file}{File name, URL or connection specifying a file to write x to.
Compressed files with extensions such as '.gz' are handled
automatically.}
}
\value{
The write functions return a BigWigFile invisibly
}
\description{
This is a lightweight wrapper to the export family
of functions defined in \pkg{rtracklayer}.
}
\examples{
\dontrun{
 if (.Platform$OS.type != "windows") {
  test_path <- system.file("tests", package = "rtracklayer")
  bw_file <- file.path(test_path, "test.bw")
  gr <- read_bigwig(bw_file)
  gr
  bw_out <- file.path(tempdir(), "test_out.bw")
  write_bigwig(gr ,bw_out)
  read_bigwig(bw_out)
 }
}
}
\seealso{
\code{rtracklayer::\link[rtracklayer:BigWigFile-class]{BigWigFile()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-expand.R
\name{expand_ranges}
\alias{expand_ranges}
\title{Expand list-columns in a Ranges object}
\usage{
expand_ranges(
  data,
  ...,
  .drop = FALSE,
  .id = NULL,
  .keep_empty = FALSE,
  .recursive = FALSE
)
}
\arguments{
\item{data}{A Ranges object}

\item{...}{list-column names to expand then unlist}

\item{.drop}{Should additional list columns be dropped (default = FALSE)?
By default \code{expand_ranges()} will keep other list columns even if they are nested.}

\item{.id}{A character vector of length equal to number of list columns.
If supplied will create new column(s) with name \code{.id}
identifying the index of the list column (default = NULL).}

\item{.keep_empty}{If a list-like column contains empty elements, should
those elements be kept? (default = FALSE)}

\item{.recursive}{If there are multiple list-columns, should the columns be
treated as parallel? If FALSE each column will be unnested recursively,
otherwise they are treated as parallel, that is each list column has
identical lengths. (deafualt = FALSE)}
}
\value{
a GRanges object with expanded list columns
}
\description{
Expand list-columns in a Ranges object
}
\examples{
grng <- as_granges(data.frame(seqnames = "chr1", start = 20:23, width = 1000))
grng <- mutate(grng, 
               exon_id = IntegerList(a = 1, b = c(4,5), c = 3, d = c(2,5))
               )
expand_ranges(grng)
expand_ranges(grng, .id = "name")

# empty list elements are not preserved by default
grng <- mutate(grng, 
               exon_id = IntegerList(a = NULL, b = c(4,5), c= 3, d = c(2,5))
               )
expand_ranges(grng)
expand_ranges(grng, .keep_empty = TRUE)
expand_ranges(grng, .id = "name", .keep_empty = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endo-coverage.R
\name{compute_coverage}
\alias{compute_coverage}
\title{Compute coverage over a Ranges object}
\usage{
compute_coverage(x, shift, width, weight, ...)
}
\arguments{
\item{x}{a \code{Ranges} object}

\item{shift}{shift how much should each range in x be shifted by? (default = 0L)}

\item{width}{width how long should the returned coverage score be?
This must be either a  positive integer or NULL (default = NULL)}

\item{weight}{weight how much weight should be assigned to each range? Either
an integer or numeric vector or a column in x. (default = 1L)}

\item{...}{other optional parameters to pass to coverage}
}
\value{
An expanded Ranges object with a score column corresponding to
the coverage value over that interval. Note that compute_coverage
drops metadata associated with the orginal ranges.
}
\description{
Compute coverage over a Ranges object
}
\examples{
rng <- as_iranges(data.frame(start = 1:10, width = 5))
compute_coverage(rng)
compute_coverage(rng, shift = 14L)
compute_coverage(rng, width = 10L)
}
\seealso{
\code{IRanges::\link[IRanges:coverage-methods]{coverage()}},
\code{GenomicRanges::\link[GenomicRanges:coverage-methods]{coverage()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-DeferredGenomicRanges.R
\docType{class}
\name{DeferredGenomicRanges-class}
\alias{DeferredGenomicRanges-class}
\title{DeferredGenomiRanges objects}
\description{
Enables deferred reading of files (currently only BAM files) by
caching results after a plyranges verb is called.
}
\section{Slots}{

\describe{
\item{\code{delegate}}{a GenomicRanges object to be cached}

\item{\code{ops}}{A FileOperator object}
}}

\seealso{
\code{read_bam()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-add-distance.R
\name{add_nearest_distance}
\alias{add_nearest_distance}
\alias{add_nearest_distance_left}
\alias{add_nearest_distance_right}
\alias{add_nearest_distance_upstream}
\alias{add_nearest_distance_downstream}
\title{Add distance to nearest neighbours between two Ranges objects}
\usage{
add_nearest_distance(x, y = x, name = "distance")

add_nearest_distance_left(x, y = x, name = "distance")

add_nearest_distance_right(x, y = x, name = "distance")

add_nearest_distance_upstream(x, y = x, name = "distance")

add_nearest_distance_downstream(x, y = x, name = "distance")
}
\arguments{
\item{x}{The query ranges}

\item{y}{the subject ranges within which the nearest ranges are found.
If missing, query ranges are used as the subject.}

\item{name}{column name to create containing distance values}
}
\value{
ranges in \code{x} with additional column containing the distance to the
nearest range in \code{y}.
}
\description{
Appends distance to nearest subject range to query ranges similar to setting
\code{distance} in \code{join_nearest_}. Distance is set to \code{NA} for features with no
nearest feature by the selected nearest metric.
}
\details{
By default \code{add_nearest_distance} will find arbitrary nearest
neighbours in either direction and ignore any strand information.
The \code{add_nearest_distance_left} and \code{add_nearest_distance_right}  methods
will find arbitrary nearest neighbour ranges on x that are left/right of
those on y and ignore any strand information.

The \code{add_nearest_distance_upstream} method will find arbitrary nearest
neighbour ranges on x that are upstream of those on y. This takes into
account strandedness of the ranges.
On the positive strand nearest upstream will be on the
left and on the negative strand nearest upstream will be on the right.

The \code{add_nearest_distance_downstream} method will find arbitrary nearest
neighbour ranges on x that are upstream of those on y. This takes into
account strandedness of the ranges. On the positive strand nearest downstream
will be on the right and on the negative strand nearest upstream will be on
the left.
}
\examples{
query <- data.frame(start = c(5,10, 15,20),
                   width = 5,
                   gc = runif(4)) \%>\%
             as_iranges()
subject <- data.frame(start = c(2:6, 24),
                      width = 3:8,
                      label = letters[1:6]) \%>\%
             as_iranges()
             
add_nearest_distance(query, subject)
add_nearest_distance_left(query, subject)
add_nearest_distance_left(query)
}
\seealso{
\code{\link{join_nearest}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-rangewise-setops.R
\name{intersect_ranges}
\alias{intersect_ranges}
\alias{intersect_ranges_directed}
\alias{union_ranges}
\alias{union_ranges_directed}
\alias{setdiff_ranges}
\alias{setdiff_ranges_directed}
\alias{complement_ranges}
\alias{complement_ranges_directed}
\title{Vector-wise Range set-operations}
\usage{
intersect_ranges(x, y)

intersect_ranges_directed(x, y)

union_ranges(x, y)

union_ranges_directed(x, y)

setdiff_ranges(x, y)

setdiff_ranges_directed(x, y)

complement_ranges(x)

complement_ranges_directed(x)
}
\arguments{
\item{x, y}{Two Ranges objects to compare.}
}
\value{
A Ranges object
}
\description{
Vector-wise Range set-operations
}
\details{
These are usual set-operations that act on the sets of the
ranges represented in x and y. By default these operations will ignore
any strand information. The directed versions of these functions will
take into account strand for GRanges objects.
}
\examples{
gr1 <- data.frame(seqnames = "chr1",
                  start = c(2,9),
                  end = c(7,9),
                  strand = c("+", "-")) \%>\%
               as_granges()
gr2 <- data.frame(seqnames = "chr1", start = 5, width = 5, strand = "-") \%>\%
         as_granges()

union_ranges(gr1, gr2)
union_ranges_directed(gr1, gr2)

intersect_ranges(gr1, gr2)
intersect_ranges_directed(gr1, gr2)

setdiff_ranges(gr1, gr2)
setdiff_ranges_directed(gr1, gr2)
# taking the complement of a ranges requires annotation information
gr1 <- set_genome_info(gr1, seqlengths = 100)
complement_ranges(gr1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io-bed.R
\name{write_bed}
\alias{write_bed}
\alias{write_bed_graph}
\alias{write_narrowpeaks}
\title{Write a BED or BEDGraph file}
\usage{
write_bed(x, file, index = FALSE)

write_bed_graph(x, file, index = FALSE)

write_narrowpeaks(x, file)
}
\arguments{
\item{x}{A GRanges object}

\item{file}{File name, URL or connection specifying a file to write x to.
Compressed files with extensions such as '.gz' are handled
automatically. If you want to index the file with tabix use the
\code{index} argument.}

\item{index}{Compress and index the output file
with bgzf and tabix (default = FALSE). Note that tabix indexing will sort the
data by chromosome and start.}
}
\value{
The write functions return a BED(Graph)File invisibly
}
\description{
This is a lightweight wrapper to the export family
of functions defined in \pkg{rtracklayer}.
}
\examples{
\dontrun{
  test_path <- system.file("tests", package = "rtracklayer")
  bed_file <- file.path(test_path, "test.bed")
  gr <- read_bed(bed_file)
  bed_file_out <- file.path(tempdir(), "new.bed")
  write_bed(gr, bed_file_out)
  read_bed(bed_file_out)
  #' bedgraph
  bg_file <- file.path(test_path, "test.bedGraph")
  gr <- read_bed_graph(bg_file)
  bg_file_out <- file.path(tempdir(), "new.bg")
  write_bed(gr, bg_file_out)
  read_bed(bg_file_out)
  # narrowpeaks
  np_file <- system.file("extdata", "demo.narrowPeak.gz",package="rtracklayer")
  gr <- read_narrowpeaks(np_file, genome_info = "hg19")
  np_file_out <- file.path(tempdir(), "new.bg")
  write_narrowpeaks(gr, np_file_out)
  read_narrowpeaks(np_file_out)
}
}
\seealso{
\code{rtracklayer::\link[rtracklayer:BEDFile-class]{BEDFile()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-arithmetic-setters.R
\name{set_width}
\alias{set_width}
\alias{set_start}
\alias{set_end}
\alias{set_seqnames}
\alias{set_strand}
\title{Functional setters for Ranges objects}
\usage{
set_width(x, width)

set_start(x, start = 0L)

set_end(x, end = 0L)

set_seqnames(x, seqnames)

set_strand(x, strand)
}
\arguments{
\item{x}{a Ranges object}

\item{width}{integer amount to modify width by}

\item{start}{integer amount to modify start by}

\item{end}{integer amount to modify end by}

\item{seqnames}{update seqnames column}

\item{strand}{update strand column}
}
\value{
a Ranges object
}
\description{
Functional setters for Ranges objects
}
\details{
These methods are used internally in \code{mutate()} to modify
core columns in Ranges objects.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dplyr-arrange.R
\name{arrange.Ranges}
\alias{arrange.Ranges}
\title{Sort a Ranges object}
\usage{
\method{arrange}{Ranges}(.data, ...)
}
\arguments{
\item{.data}{A Ranges object.}

\item{...}{Comma seperated list of variable names.}
}
\value{
A sorted Ranges object
}
\description{
Sort a Ranges object
}
\examples{
rng <- as_iranges(data.frame(start = 1:10, width = 10:1))
rng <- mutate(rng, score = runif(10))
arrange(rng, score)
# you can also use dplyr::desc to arrange by descending order
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-eval.R
\name{n}
\alias{n}
\title{Compute the number of ranges in each group.}
\usage{
n()
}
\value{
\code{n()} will only be evaluated inside a function call, where it
returns an integer.
}
\description{
This function should only be used
within \code{summarise()}, \code{mutate()} and \code{filter()}.
}
\examples{
ir <- as_iranges(
                 data.frame(start = 1:10,
                            width = 5,
                            name = c(rep("a", 5), rep("b", 3), rep("c", 2))
                            )
                )
by_names <- group_by(ir, name)
summarise(by_names, n = n())
mutate(by_names, n = n())
filter(by_names, n() >= 3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-overlap-self-joins.R
\name{join_overlap_self}
\alias{join_overlap_self}
\alias{join_overlap_self_within}
\alias{join_overlap_self_directed}
\alias{join_overlap_self_within_directed}
\title{Find overlaps within a Ranges object}
\usage{
join_overlap_self(x, maxgap, minoverlap)

join_overlap_self_within(x, maxgap, minoverlap)

join_overlap_self_directed(x, maxgap, minoverlap)

join_overlap_self_within_directed(x, maxgap, minoverlap)
}
\arguments{
\item{x}{A Ranges object}

\item{maxgap, minoverlap}{The maximimum gap between intervals as an integer
greater than or equal to zero. The minimum amount of overlap between intervals
as an integer greater than zero, accounting for the maximum gap.}
}
\value{
a Ranges object
}
\description{
Find overlaps within a Ranges object
}
\details{
Self overlaps find any overlaps (or overlaps within or overlaps
directed) between a ranges object and itself.
}
\examples{
query <- data.frame(start = c(5,10, 15,20), width = 5, gc = runif(4)) \%>\%
             as_iranges()

join_overlap_self(query)

# -- GRanges objects, strand is ignored by default
query  <- data.frame(seqnames = "chr1",
               start = c(11,101),
               end = c(21, 200),
               name = c("a1", "a2"),
               strand = c("+", "-"),
               score = c(1,2)) \%>\%
           as_granges()

# ignores strandedness
join_overlap_self(query)
join_overlap_self_within(query)
# adding directed prefix includes strand
join_overlap_self_directed(query)


}
\seealso{
\code{\link[=find_overlaps]{find_overlaps()}}, \code{\link[=join_overlap_inner]{join_overlap_inner()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-join-precede.R
\name{join_precede}
\alias{join_precede}
\alias{join_precede_right}
\alias{join_precede_downstream}
\title{Find preceding Ranges}
\usage{
join_precede(x, y, suffix = c(".x", ".y"))

join_precede_right(x, y, suffix = c(".x", ".y"))

join_precede_downstream(x, y, suffix = c(".x", ".y"))
}
\arguments{
\item{x, y}{Ranges objects, which ranges in x precede those in y.}

\item{suffix}{A character vector of length two used to identify
metadata columns coming from x and y.}
}
\value{
A Ranges object corresponding to the ranges in \code{y} that are
preceded by the ranges in \code{x}, all metadata is copied over from the
right-hand side ranges \code{y}.
}
\description{
Find preceding Ranges
}
\details{
By default \code{join_precede} will return the ranges
in x that come before the ranges in y and ignore any strand information.
The function \code{join_precede_right} will find all ranges in y
that are on the right-hand side of the ranges in x ignoring any strand
information. Finally, \code{join_precede_downstream} will find all ranges in y
that are that are downstream of the ranges in x. On the positive strand this
will result in ranges in y that are right of those in x and on the negative
strand it will result in ranges in y that are left of those in x.
}
\examples{
subject <- data.frame(start = c(5,10, 15,20), width = 5, gc = runif(4)) \%>\%
             as_iranges()
query <- data.frame(start = 2:6, width = 3:7, label = letters[1:5]) \%>\%
             as_iranges()

join_precede(query, subject)

query  <- data.frame(seqnames = "chr1",
               start = c(11,101),
               end = c(21, 200),
               name = c("a1", "a2"),
               strand = c("+", "-"),
               score = c(1,2)) \%>\%
           as_granges()
subject <- data.frame(seqnames = "chr1",
                      strand = c("+", "-", "+", "-"),
                      start = c(21,91,101,201),
                      end = c(30,101,110,210),
                      name = paste0("b", 1:4),
                      score = 1:4) \%>\%
                   as_granges()

join_precede(query, subject)
join_precede_right(query, subject)
join_precede_downstream(query, subject)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dplyr-slice.R
\name{slice.Ranges}
\alias{slice.Ranges}
\alias{slice.GroupedGenomicRanges}
\alias{slice.GroupedIntegerRanges}
\alias{slice_head.Ranges}
\alias{slice_head.GroupedGenomicRanges}
\alias{slice_head.GroupedIntegerRanges}
\alias{slice_tail.Ranges}
\alias{slice_tail.GroupedGenomicRanges}
\alias{slice_tail.GroupedIntegerRanges}
\alias{slice_sample.Ranges}
\alias{slice_sample.GroupedGenomicRanges}
\alias{slice_sample.GroupedIntegerRanges}
\alias{slice_min.Ranges}
\alias{slice_min.GroupedGenomicRanges}
\alias{slice_min.GroupedIntegerRanges}
\alias{slice_max.Ranges}
\alias{slice_max.GroupedGenomicRanges}
\alias{slice_max.GroupedIntegerRanges}
\title{Choose ranges by their position}
\usage{
\method{slice}{Ranges}(.data, ..., .preserve = FALSE)

\method{slice}{GroupedGenomicRanges}(.data, ..., .preserve = FALSE)

\method{slice}{GroupedIntegerRanges}(.data, ..., .preserve = FALSE)

\method{slice_head}{Ranges}(.data, ..., n = NULL, prop = NULL, .preserve = FALSE)

\method{slice_head}{GroupedGenomicRanges}(.data, ..., n = NULL, prop = NULL, .preserve = FALSE)

\method{slice_head}{GroupedIntegerRanges}(.data, ..., n = NULL, prop = NULL, .preserve = FALSE)

\method{slice_tail}{Ranges}(.data, ..., n = NULL, prop = NULL, .preserve = FALSE)

\method{slice_tail}{GroupedGenomicRanges}(.data, ..., n = NULL, prop = NULL, .preserve = FALSE)

\method{slice_tail}{GroupedIntegerRanges}(.data, ..., n = NULL, prop = NULL, .preserve = FALSE)

\method{slice_sample}{Ranges}(
  .data,
  ...,
  n = NULL,
  prop = NULL,
  weight_by = NULL,
  replace = FALSE,
  .preserve = FALSE
)

\method{slice_sample}{GroupedGenomicRanges}(
  .data,
  ...,
  n = NULL,
  prop = NULL,
  weight_by = NULL,
  replace = FALSE,
  .preserve = FALSE
)

\method{slice_sample}{GroupedIntegerRanges}(
  .data,
  ...,
  n = NULL,
  prop = NULL,
  weight_by = NULL,
  replace = FALSE,
  .preserve = FALSE
)

\method{slice_min}{Ranges}(
  .data,
  order_by,
  ...,
  n = NULL,
  prop = NULL,
  with_ties = TRUE,
  .preserve = FALSE
)

\method{slice_min}{GroupedGenomicRanges}(
  .data,
  order_by,
  ...,
  n = NULL,
  prop = NULL,
  with_ties = TRUE,
  .preserve = FALSE
)

\method{slice_min}{GroupedIntegerRanges}(
  .data,
  order_by,
  ...,
  n = NULL,
  prop = NULL,
  with_ties = TRUE,
  .preserve = FALSE
)

\method{slice_max}{Ranges}(
  .data,
  order_by,
  ...,
  n = NULL,
  prop = NULL,
  with_ties = TRUE,
  .preserve = FALSE
)

\method{slice_max}{GroupedGenomicRanges}(
  .data,
  order_by,
  ...,
  n = NULL,
  prop = NULL,
  with_ties = TRUE,
  .preserve = FALSE
)

\method{slice_max}{GroupedIntegerRanges}(
  .data,
  order_by,
  ...,
  n = NULL,
  prop = NULL,
  with_ties = TRUE,
  .preserve = FALSE
)
}
\arguments{
\item{.data}{a \code{Ranges} object.}

\item{...}{<\code{\link[rlang:args_data_masking]{data-masking}}> Integer row values
indicating positions to keep.  Only used by \code{slice()}; ignored elsewhere.}

\item{.preserve}{Ignored; retained for compatibility with the dplyr generic.}

\item{n}{Number of ranges to select.  Mutually exclusive with \code{prop}.
For \code{slice_min()} and \code{slice_max()}, negative values are not supported.}

\item{prop}{Fraction of ranges to select, in \verb{(0, 1]}.  Mutually exclusive
with \code{n}.}

\item{weight_by}{<\code{\link[rlang:args_data_masking]{data-masking}}> Unquoted name
of a non-negative numeric metadata column used as sampling weights.  Used
by \code{slice_sample()}.}

\item{replace}{Should sampling be with replacement?  Default \code{FALSE}.  Used
by \code{slice_sample()}.}

\item{order_by}{<\code{\link[rlang:args_data_masking]{data-masking}}> Unquoted name
of a numeric metadata column to order by.  Used by \code{slice_min()} and
\code{slice_max()}.}

\item{with_ties}{If \code{TRUE} (the default), all ranges tied at the boundary
rank are kept, which may return more than \code{n} ranges.  Set to \code{FALSE} to
return exactly \code{n}.  Used by \code{slice_min()} and \code{slice_max()}.}
}
\value{
A \code{GRanges} (or \code{IRanges}) object of the same class as \code{.data}, with
grouping structure re-applied when \code{.data} was grouped.
}
\description{
The \code{slice} family of functions let you select ranges by position.  They
follow the same conventions as their dplyr counterparts for data frames:
\itemize{
\item \code{slice()} selects ranges by explicit integer position.
\item \code{slice_head()} / \code{slice_tail()} select the first or last ranges.
\item \code{slice_sample()} randomly selects ranges.
\item \code{slice_min()} / \code{slice_max()} select ranges with the smallest or largest
values of a metadata column.
}

All functions preserve grouping: when \code{.data} has been grouped with
\code{\link[dplyr:group_by]{dplyr::group_by()}}, selection is performed within each group.
}
\examples{
df <- data.frame(start = 1:10,
                 width = 5,
                 seqnames = "seq1",
                 strand = sample(c("+", "-", "*"), 10, replace = TRUE),
                 gc = runif(10))
rng <- as_granges(df)
dplyr::slice(rng, 1:2)
dplyr::slice(rng, -n())
dplyr::slice(rng, -5:-n())

by_strand <- group_by(rng, strand)

# slice with group by finds positions within each group
dplyr::slice(by_strand, n())
dplyr::slice(by_strand, which.max(gc))

# if the index is beyond the number of groups slice are ignored
dplyr::slice(by_strand, 1:3)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io-bed.R
\name{read_bed}
\alias{read_bed}
\alias{read_bed_graph}
\alias{read_narrowpeaks}
\title{Read a BED or BEDGraph file}
\usage{
read_bed(file, col_names = NULL, genome_info = NULL, overlap_ranges = NULL)

read_bed_graph(
  file,
  col_names = NULL,
  genome_info = NULL,
  overlap_ranges = NULL
)

read_narrowpeaks(
  file,
  col_names = NULL,
  genome_info = NULL,
  overlap_ranges = NULL
)
}
\arguments{
\item{file}{A path to a file or a connection.}

\item{col_names}{An optional character vector for including additional
columns in \code{file} that are not part of the BED/narrowPeaks specification.}

\item{genome_info}{An optional character string or a Ranges object
that contains information about the genome build. For example the USSC identifier
"hg19" will add build information to the returned GRanges.}

\item{overlap_ranges}{An optional Ranges object. Only the intervals in the file
that overlap the Ranges will be returned.}
}
\value{
A GRanges object
}
\description{
This is a lightweight wrapper to the import family
of functions defined in \pkg{rtracklayer}.

Read common interval based formats as GRanges.
}
\details{
This is a lightweight wrapper to the import family
of functions defined in \pkg{rtracklayer}.
The \code{read_narrowpeaks} function parses the ENCODE narrowPeak BED format (see
\url{https://genome.ucsc.edu/FAQ/FAQformat.html#format12} for details.). As
such the parser expects four additional columns called (corresponding to
the narrowPeaks spec):
\itemize{
\item signalValue
\item pValue
\item qValue
\item peak
}
}
\examples{
test_path <- system.file("tests", package = "rtracklayer")
bed_file <- file.path(test_path, "test.bed")
gr <- read_bed(bed_file)
gr
gr <- read_bed(bed_file, genome_info = "hg19")
gr
olap <-  as_granges(data.frame(seqnames = "chr7", start = 1, end = 127473000))
gr <- read_bed(bed_file,
              overlap_ranges = olap)
# bedGraph
bg_file <- file.path(test_path, "test.bedGraph")
gr <- read_bed_graph(bg_file)
gr
# narrowpeaks
np_file <- system.file("extdata", "demo.narrowPeak.gz",  package="rtracklayer")
gr <- read_narrowpeaks(np_file, genome_info = "hg19")
gr
}
\seealso{
\code{rtracklayer::\link[rtracklayer:BEDFile-class]{BEDFile()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-genomeinfo.R
\name{ranges-info}
\alias{ranges-info}
\alias{genome_info}
\alias{set_genome_info}
\alias{get_genome_info}
\title{Construct annotation information}
\usage{
genome_info(
  genome = NULL,
  seqnames = NULL,
  seqlengths = NULL,
  is_circular = NULL
)

set_genome_info(
  .data,
  genome = NULL,
  seqnames = NULL,
  seqlengths = NULL,
  is_circular = NULL
)

get_genome_info(.data)
}
\arguments{
\item{genome}{A character vector of length one indicating the genome build.}

\item{seqnames}{A character vector containing the name of sequences.}

\item{seqlengths}{An optional integer vector containg the lengths of sequences.}

\item{is_circular}{An optional logical vector indicating whether a sequence is ciruclar.}

\item{.data}{A Ranges object to annotate or retrieve an annotation for.}
}
\value{
a GRanges object containing annotations. To retrieve the annotations
as a Ranges object use \code{get_genome_info}.
}
\description{
To construct annotations by supplying annotation information
use \code{genome_info}. To add
annotations to an existing Ranges object use \code{set_genome_info}. To retrieve
an annotation as a Ranges object use \code{get_genome_info}.
}
\examples{
x <- genome_info(genome = "toy",
                 seqnames = letters[1:4],
                 seqlengths = c(100, 300, 15, 600),
                 is_circular = c(NA, FALSE, FALSE, TRUE))
x

rng <- as_granges(data.frame(seqnames = "a", start = 30:50, width = 10))
rng
rng <- set_genome_info(rng,
                       genome = "toy",
                       seqnames = letters[1:4],
                       seqlengths = c(100, 300, 15, 600),
                       is_circular = c(NA, FALSE, FALSE, TRUE))
get_genome_info(rng)

\dontrun{
if (interactive()) {
 # requires internet connection
 genome_info(genome = "hg38")
}
}

}
\seealso{
\code{\link[Seqinfo:Seqinfo-class]{Seqinfo::Seqinfo()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dplyr-mutate.R
\name{mutate.Ranges}
\alias{mutate.Ranges}
\title{Modify a Ranges object}
\usage{
\method{mutate}{Ranges}(.data, ...)
}
\arguments{
\item{.data}{a \code{Ranges} object}

\item{...}{Pairs of name-value expressions. The name-value pairs can either
create new metadata columns or modify existing ones.}
}
\value{
a Ranges object
}
\description{
Modify a Ranges object
}
\examples{
df <- data.frame(start = 1:10,
                 width = 5,
                 seqnames = "seq1",
                 strand = sample(c("+", "-", "*"), 10, replace = TRUE),
                 gc = runif(10))
rng <- as_granges(df)

# mutate adds new columns
rng \%>\%
    mutate(avg_gc = mean(gc), row_id = 1:n())
# can also compute on newly created columns
rng \%>\%
    mutate(score = gc * width, score2 = score + 1)
# group by partitions the data and computes within each group
rng \%>\%
    group_by(strand) \%>\%
    mutate(avg_gc = mean(gc), row_id = 1:n())

# mutate can be used in conjuction with anchoring to resize ranges
rng \%>\%
    mutate(width = 10)
# by default width modfication fixes by start
rng \%>\%
    anchor_start() \%>\%
    mutate(width = 10)
# fix by end or midpoint
rng \%>\%
    anchor_end() \%>\%
    mutate(width = width + 1)
rng \%>\%
    anchor_center() \%>\%
    mutate(width = width + 1)
# anchoring by strand
rng \%>\%
    anchor_3p() \%>\%
    mutate(width = width * 2)
rng \%>\%
    anchor_5p() \%>\%
    mutate(width = width * 2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io-bam.R
\name{read_bam}
\alias{read_bam}
\title{Read a BAM file}
\usage{
read_bam(file, index = file, paired = FALSE)
}
\arguments{
\item{file}{A connection or path to a BAM file}

\item{index}{The path to the BAM index file}

\item{paired}{Whether to treat alignments as paired end (TRUE)
or single end (FALSE). Default is FALSE.}
}
\value{
A DeferredGenomicRanges object
}
\description{
Read a BAM file
}
\details{
Reading a BAM file is deferred until an action
such as using \code{summarise()} or \code{mutate()} occurs. If paired is set to
TRUE, when alignments are loaded, the GRanges has two additional
columns called read_pair_id and read_pair_group corresponding
to paired reads and is grouped by the read_pair_group.

Certain verbs have different behaviour, after using \code{read_bam()}.

For \code{select()} valid columns are the fields available in the
BAM file. Valid entries are qname (QNAME), flag (FLAG),
rname (RNAME), strand, pos (POS), qwidth (width of query),
mapq (MAPQ), cigar (CIGAR), mrnm (RNEXT), mpos (PNEXT), isize
(TLEN), seq (SEQ), and qual (QUAL). Any two character tags in the BAM file
are also valid.

For \code{filter()} the following fields are valid, to select the FALSE option
place \code{!} in front of the field:
\itemize{
\item \code{is_paired} Select either unpaired (FALSE) or paired (TRUE) reads.
\item \code{is_proper_pair} Select either improperly paired (FALSE) or properly
paired (TRUE) reads. This is dependent on the alignment software used.
\item `is_unmapped_query``	Select unmapped (TRUE) or mapped (FALSE) reads.
\item \code{has_unmapped_mate} Select reads with mapped (FALSE) or unmapped (TRUE) mates.
\item \code{is_minus_strand} 	Select reads aligned to plus (FALSE) or minus (TRUE) strand.
\item \code{is_mate_minus_strand}	Select reads where mate is aligned to plus (FALSE) or
minus (TRUE) strand.
\item \code{is_first_mate_read}	Select reads if they are the first mate (TRUE) or
not (FALSE).
\item \code{is_second_mate_read} Select reads if they are the second mate (TRUE) or
not (FALSE).
\item \code{is_secondary_alignment} Select reads if their alignment status is
secondary (TRUE) or not (FALSE). This might be relevant if there are
multimapping reads.
\item \code{is_not_passing_quality_controls} Select reads that either pass
quality controls (FALSE) or that do not (TRUE).
\item \code{is_duplicate}	Select reads that are unduplicated (FALSE) or
duplicated (TRUE). This may represent reads that are PCR or
optical duplicates.
}
}
\examples{
if (require(pasillaBamSubset)) {
   bamfile <- untreated1_chr4()
   # nothing is read until an action has been performed
   print(read_bam(bamfile))
   # define a region of interest
   roi <- data.frame(seqnames = "chr4", start = 5e5, end = 7e5) \%>\%
            as_granges()
   rng <- read_bam(bamfile) \%>\% 
            select(mapq) \%>\%
            filter_by_overlaps(roi)
}

}
\seealso{
\code{Rsamtools::\link[Rsamtools:BamFile]{BamFile()}},\code{\link[GenomicAlignments:readGAlignments]{GenomicAlignments::readGAlignments()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io-bigwig.R
\name{read_bigwig}
\alias{read_bigwig}
\title{Read a BigWig file}
\usage{
read_bigwig(file, genome_info = NULL, overlap_ranges = NULL)
}
\arguments{
\item{file}{A path to a file or URL.}

\item{genome_info}{An optional character string or a Ranges object
that contains information about the genome build. For example the identifier
"hg19" will add build information to the returned GRanges.}

\item{overlap_ranges}{An optional Ranges object. Only the intervals in the file
that overlap the Ranges will be loaded.}
}
\value{
a GRanges object
}
\description{
Read a BigWig file
}
\examples{
if (.Platform$OS.type != "windows") {
  test_path <- system.file("tests", package = "rtracklayer")
  bw_file <- file.path(test_path, "test.bw")
  gr <- read_bigwig(bw_file)
  gr
}
}
\seealso{
\code{rtracklayer::\link[rtracklayer:BigWigFile-class]{BigWigFile()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-construct.R
\name{as_ranges}
\alias{as_ranges}
\title{Coerce an Rle or RleList object to Ranges}
\usage{
as_ranges(.data)
}
\arguments{
\item{.data}{a \code{\link[S4Vectors:Rle-class]{S4Vectors::Rle()}} or an \code{\link[IRanges:AtomicList-class]{IRanges::RleList()}} object.}
}
\value{
an \code{\link[IRanges:IRanges-constructor]{IRanges::IRanges()}} object if the input is an
\code{\link[S4Vectors:Rle-class]{S4Vectors::Rle()}} object or a \code{\link[GenomicRanges:GRanges-class]{GenomicRanges::GRanges()}} object for
an \code{\link[IRanges:AtomicList-class]{IRanges::RleList()}} object.
}
\description{
Coerce an Rle or RleList object to Ranges
}
\details{
This function is behind \code{\link[=compute_coverage]{compute_coverage()}}.
}
\examples{
x <- S4Vectors::Rle(10:1, 1:10)
as_ranges(x)

# must have names set
y <- IRanges::RleList(chr1 = x)
as_ranges(y)

}
\seealso{
\code{S4Vectors::\link[S4Vectors:Rle-class]{Rle()}},
\code{IRanges::\link[IRanges:AtomicList]{RleList()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-join-mcols.R
\name{join_mcols_inner}
\alias{join_mcols_inner}
\alias{join_mcols_left}
\title{Join data by metadata columns}
\usage{
join_mcols_inner(x, y, by = NULL, ...)

join_mcols_left(x, y, by = NULL, ...)
}
\arguments{
\item{x}{Object representing ranges, with metadata
columns containing variables for matching}

\item{y}{A table of data, DataFrame, data.frame or tibble}

\item{by}{Specifications of the columns used for merging.
Passed to \code{S4Vectors::merge()}}

\item{...}{Additional arguments passed to \code{S4Vectors::merge()}}
}
\value{
Object representing ranges, with new metadata columns
}
\description{
These functions enable joins based on metadata
of \code{x} and data in \code{y}, utilizing
\code{\link[S4Vectors:Vector-merge]{S4Vectors::merge()}}.
}
\details{
The function \code{\link[=join_mcols_inner]{join_mcols_inner()}} returns ranges in
\code{x} only when a match is found in the columns specified with
\code{by} in the table \code{y}.

The function \code{\link[=join_mcols_left]{join_mcols_left()}} returns all ranges in \code{x}
regardless of a match in \code{y}, with duplications possible
from multiple matches.
}
\examples{

x <- as_granges(data.frame(
  seqnames = "chr1", start = 1:5, width=10,
  id=c(1:4,4), foo=letters[1:5]
))
y <- DataFrame(id=c(1:2,2,4), bar=letters[6:9])

# metadata joins:
join_mcols_inner(x, y, by="id")
join_mcols_left(x, y, by="id")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-anchors.R
\name{anchor}
\alias{anchor}
\alias{unanchor}
\alias{anchor_start}
\alias{anchor_end}
\alias{anchor_center}
\alias{anchor_centre}
\alias{anchor_3p}
\alias{anchor_5p}
\title{Anchored Ranges objects}
\usage{
anchor(x)

unanchor(x)

anchor_start(x)

anchor_end(x)

anchor_center(x)

anchor_centre(x)

anchor_3p(x)

anchor_5p(x)
}
\arguments{
\item{x}{a Ranges object}
}
\value{
a RangesAnchored object which has the same appearance as a regular
Ranges object but with an additional slot displaying an anchor.
}
\description{
The \code{GRangesAnchored} class and the \code{IRangesAnchored}
class allow components of a \code{GRanges} or \code{IRanges} (start, end, center)
to be held fixed.
}
\details{
Anchoring will fix a Ranges start, end, or center positions,
so these positions will remain the same when performing arithimetic.
For \code{GRanges} objects, the function
(\code{anchor_3p()}) will fix the start for the negative strand,
while \code{anchor_5p()} will fix the end for the
positive strand. Anchoring modifies how arithmetic is performed, for example
modifying the width of a range with \code{set_width()} or stretching a
range with \code{stretch()}. To remove anchoring use \code{unanchor()}.
}
\section{Constructors}{

Depending on how you want to fix the components of a Ranges, there are
five ways to construct a RangesAnchored class. Here \code{x} is either
an \code{IRanges} or \code{GRanges} object.
\describe{
\item{\code{anchor_start(x)}}{Fix the start coordinates}
\item{\code{anchor_end(x)}}{Fix the end coordinates}
\item{\code{anchor_center(x)}}{Fix the center coordinates}
\item{\code{anchor_3p(x)}}{On the negative strand fix the start coordinates,
and for positive or unstranded ranges fix the end coordinates.}
\item{\code{anchor_5p(x)}}{On the positive or unstranded ranges fix the start coordinates,
coordinates and for negative stranded ranges fix the end coordinates.}
}
}

\section{Accessors}{

To see what has been anchored use the function \code{anchor}.
This will return a character vector containing a valid anchor.
It will be set to one of \code{c("start", "end", "center")} for an
\code{IRanges} object or one of
\code{c("start", "end", "center", "3p", "5p")} for a \code{GRanges} object.
}

\examples{
df <- data.frame(start = 1:10, width = 5)
rng <- as_iranges(df)
rng_by_start <- anchor_start(rng)
rng_by_start
anchor(rng_by_start)
mutate(rng_by_start, width = 3L)
grng <- as_granges(df,
                   seqnames = "chr1",
                   strand = c(rep("-", 5), rep("+", 5)))
rng_by_5p <- anchor_5p(grng)
rng_by_5p
mutate(rng_by_5p, width = 3L)

}
\seealso{
\link{mutate.Ranges}, \link{stretch}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dplyr-summarize.R
\name{summarise.Ranges}
\alias{summarise.Ranges}
\title{Reduce multiple values in a Ranges down to a single value}
\usage{
\method{summarise}{Ranges}(.data, ...)
}
\arguments{
\item{.data}{a Ranges object}

\item{...}{Name-value pairs of summary functions. The name will be the
name of the variable in the result. The value should be an expression that
will return a value that has length one or length equal to the number of
groups.}
}
\value{
A \code{S4Vectors::\link[S4Vectors:DataFrame-class]{DataFrame()}}
}
\description{
Reduce multiple values in a Ranges down to a single value
}
\details{
Creates one or more variables as a \code{S4Vectors::\link[S4Vectors:DataFrame-class]{DataFrame()}}
from the input Ranges object. If the ranges object is grouped, there will
be a row for each group. Because grouping may remove whether a Ranges object
is valid, a DataFrame is always returned.
}
\examples{
df <- data.frame(start = 1:10, width = 5,  seqnames = "seq1",
strand = sample(c("+", "-", "*"), 10, replace = TRUE), gc = runif(10))
rng <- as_granges(df)
rng \%>\% summarise(gc = mean(gc))
rng \%>\% group_by(strand) \%>\% summarise(gc = mean(gc))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dplyr-select.R
\name{select.Ranges}
\alias{select.Ranges}
\title{Select metadata columns of the Ranges object by name or position}
\usage{
\method{select}{Ranges}(.data, ..., .drop_ranges = FALSE)
}
\arguments{
\item{.data}{a \code{Ranges} object}

\item{...}{One or more metadata column names.}

\item{.drop_ranges}{If TRUE select will always return a tibble. In this
case, you may select columns that form the core part of the Ranges object.}
}
\value{
a Ranges object or a tibble
}
\description{
Select metadata columns of the Ranges object by name or position
}
\details{
Note that by default select only acts on the metadata columns (and
will therefore return a Ranges object) if a core component of a Ranges is dropped or selected
without the other required components (this includes the seqnames, strand, start, end,
width names), then select will throw an error unless .drop_ranges is set to TRUE.
}
\examples{
df <- data.frame(start = 1:10, width = 5,  seqnames = "seq1",
strand = sample(c("+", "-", "*"), 10, replace = TRUE), gc = runif(10), counts = rpois(10, 2))
rng <- as_granges(df)
select(rng, -gc)
select(rng, gc)
select(rng, counts, gc)
select(rng, 2:1)
select(rng, seqnames, strand, .drop_ranges = TRUE)
}
\seealso{
\code{\link[dplyr:select]{dplyr::select()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-overlap-count.R
\name{count_overlaps}
\alias{count_overlaps}
\alias{count_overlaps.IntegerRanges}
\alias{count_overlaps.GenomicRanges}
\alias{count_overlaps_within}
\alias{count_overlaps_within.IntegerRanges}
\alias{count_overlaps_within.GenomicRanges}
\alias{count_overlaps_directed}
\alias{count_overlaps_directed.GenomicRanges}
\alias{count_overlaps_within_directed}
\alias{count_overlaps_within_directed.GenomicRanges}
\title{Count the number of overlaps between two Ranges objects}
\usage{
count_overlaps(x, y, maxgap, minoverlap)

\method{count_overlaps}{IntegerRanges}(x, y, maxgap = -1L, minoverlap = 0L)

\method{count_overlaps}{GenomicRanges}(x, y, maxgap = -1L, minoverlap = 0L)

count_overlaps_within(x, y, maxgap, minoverlap)

\method{count_overlaps_within}{IntegerRanges}(x, y, maxgap = 0L, minoverlap = 1L)

\method{count_overlaps_within}{GenomicRanges}(x, y, maxgap = 0L, minoverlap = 1L)

count_overlaps_directed(x, y, maxgap, minoverlap)

\method{count_overlaps_directed}{GenomicRanges}(x, y, maxgap = -1L, minoverlap = 0L)

count_overlaps_within_directed(x, y, maxgap, minoverlap)

\method{count_overlaps_within_directed}{GenomicRanges}(x, y, maxgap = -1L, minoverlap = 0L)
}
\arguments{
\item{x, y}{Objects representing ranges}

\item{maxgap, minoverlap}{The maximimum gap between intervals as an integer
greater than or equal to zero. The minimum amount of overlap between intervals
as an integer greater than zero, accounting for the maximum gap.}
}
\value{
An integer vector of same length as x.
}
\description{
Count the number of overlaps between two Ranges objects
}
\examples{
query <- data.frame(start = c(5,10, 15,20), width = 5, gc = runif(4)) \%>\%
             as_iranges()
subject <- data.frame(start = 2:6, width = 3:7, label = letters[1:5]) \%>\%
             as_iranges()
query \%>\% mutate(n_olap = count_overlaps(., subject),
                 n_olap_within = count_overlaps_within(., subject))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endo-tile.R
\name{tile_ranges}
\alias{tile_ranges}
\alias{slide_ranges}
\title{Slide or tile over a Ranges object}
\usage{
tile_ranges(x, width)

slide_ranges(x, width, step)
}
\arguments{
\item{x}{a Ranges object}

\item{width}{the maximum width of each window/tile (integer vector of length 1)}

\item{step}{the distance between start position of each sliding window (integer vector of length 1)}
}
\value{
a Ranges object
}
\description{
Slide or tile over a Ranges object
}
\details{
The \code{tile_ranges()} function paritions a Ranges object \code{x} by the given the
\code{width} over all ranges in \code{x}, truncated by the sequence end.
The \code{slide_ranges()} function makes sliding windows within each range of \code{x}
of size \code{width} and sliding by \code{step}.
Both \code{slide_ranges()} and \code{tile_ranges()} return a new Ranges object
with a metadata column called "partition" which contains the index of the
input range \code{x} that a parition belongs to.
}
\examples{
 
gr <- data.frame(seqnames = c("chr1", rep("chr2", 3), rep("chr1", 2), rep("chr3", 4)),
                 start = 1:10,
                 end = 11,
                 strand = c("-", rep("+", 2), rep("*", 2), rep("+", 3), rep("-", 2))) \%>\%
      as_granges() \%>\%
      set_genome_info(seqlengths = c(11,12,13))

# partition ranges into subranges of width 2, odd width ranges
# will have one subrange of width 1              
tile_ranges(gr, width = 2)

# make sliding windows of width 3, moving window with step size of 2
slide_ranges(gr, width = 3, step = 2)

}
\seealso{
\code{GenomicRanges::\link[GenomicRanges:tile-methods]{tile()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-eval.R
\name{n_distinct}
\alias{n_distinct}
\title{Compute the number of distinct unique values in a vector or List}
\usage{
n_distinct(var)
}
\arguments{
\item{var}{a vector of values}
}
\value{
an integer vector
}
\description{
This is a wrapper to \code{length(unique(x))} or
\code{lengths(unique(x))} if \code{x} is a List object
}
\examples{
x <- CharacterList(c("a", "b", "c", "a"),  "d")
n_distinct(x)
n_distinct(unlist(x))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-overlap-find.R,
%   R/ranges-overlap-groups.R
\name{find_overlaps}
\alias{find_overlaps}
\alias{find_overlaps.IntegerRanges}
\alias{find_overlaps.GenomicRanges}
\alias{find_overlaps_within}
\alias{find_overlaps_within.IntegerRanges}
\alias{find_overlaps_within.GenomicRanges}
\alias{find_overlaps_directed}
\alias{find_overlaps_directed.GenomicRanges}
\alias{find_overlaps_within_directed}
\alias{find_overlaps_within_directed.GenomicRanges}
\alias{group_by_overlaps}
\alias{group_by_overlaps.IntegerRanges}
\alias{group_by_overlaps.GenomicRanges}
\title{Find overlap between two Ranges}
\usage{
find_overlaps(x, y, maxgap, minoverlap, suffix = c(".x", ".y"))

\method{find_overlaps}{IntegerRanges}(x, y, maxgap = -1L, minoverlap = 0L, suffix = c(".x", ".y"))

\method{find_overlaps}{GenomicRanges}(x, y, maxgap = -1L, minoverlap = 0L, suffix = c(".x", ".y"))

find_overlaps_within(x, y, maxgap, minoverlap, suffix = c(".x", ".y"))

\method{find_overlaps_within}{IntegerRanges}(
  x,
  y,
  maxgap = -1L,
  minoverlap = 0L,
  suffix = c(".x", ".y")
)

\method{find_overlaps_within}{GenomicRanges}(
  x,
  y,
  maxgap = -1L,
  minoverlap = 0L,
  suffix = c(".x", ".y")
)

find_overlaps_directed(x, y, maxgap, minoverlap, suffix = c(".x", ".y"))

\method{find_overlaps_directed}{GenomicRanges}(
  x,
  y,
  maxgap = -1L,
  minoverlap = 0L,
  suffix = c(".x", ".y")
)

find_overlaps_within_directed(x, y, maxgap, minoverlap, suffix = c(".x", ".y"))

\method{find_overlaps_within_directed}{GenomicRanges}(x, y, maxgap, minoverlap, suffix = c(".x", ".y"))

group_by_overlaps(x, y, maxgap, minoverlap)

\method{group_by_overlaps}{IntegerRanges}(x, y, maxgap = -1L, minoverlap = 0L)

\method{group_by_overlaps}{GenomicRanges}(x, y, maxgap = -1L, minoverlap = 0L)
}
\arguments{
\item{x, y}{Objects representing ranges}

\item{maxgap, minoverlap}{The maximimum gap between intervals as an integer
greater than or equal to negative one. The minimum amount of overlap between intervals
as an integer greater than zero, accounting for the maximum gap.}

\item{suffix}{A character vector of length two used to identify metadata columns
coming from x and y.}
}
\value{
A Ranges object with rows corresponding to the
ranges in x that overlap y.  In the case of \code{group_by_overlaps()}, returns
a GroupedRanges object, grouped by the number of overlaps
of ranges in x that overlap y (stored in a column called query).
}
\description{
Find overlap between two Ranges
}
\details{
\code{find_overlaps()} will search for any overlaps between ranges
x and y and return a Ranges object of length equal to the number of times x
overlaps y. This  Ranges object will have additional metadata columns
corresponding to the metadata columns in y. \code{find_overlaps_within()} is
the same but will only search for overlaps within y. For GRanges objects strand is
ignored, unless \code{find_overlaps_directed()} is used. If the Ranges objects have no
metadata, one could use \code{group_by_overlaps()} to be able to
identify the index of the input Range x that overlaps a Range in y.
Alternatively,
\code{pair_overlaps()} could be used to place the x ranges next to the range
in y they overlap.
}
\examples{
query <- data.frame(start = c(5,10, 15,20), width = 5, gc = runif(4)) \%>\%
             as_iranges()
subject <- data.frame(start = 2:6, width = 3:7, label = letters[1:5]) \%>\%
             as_iranges()

find_overlaps(query, subject)
find_overlaps(query, subject, minoverlap = 5)
find_overlaps_within(query, subject) # same result as minoverlap
find_overlaps(query, subject, maxgap = 1)

# -- GRanges objects, strand is ignored by default
query  <- data.frame(seqnames = "chr1",
               start = c(11,101),
               end = c(21, 200),
               name = c("a1", "a2"),
               strand = c("+", "-"),
               score = c(1,2)) \%>\%
           as_granges()
subject <- data.frame(seqnames = "chr1",
                      strand = c("+", "-", "+", "-"),
                      start = c(21,91,101,201),
                      end = c(30,101,110,210),
                      name = paste0("b", 1:4),
                      score = 1:4) \%>\%
                   as_granges()

# ignores strandedness
find_overlaps(query, subject, suffix = c(".query", ".subject"))
find_overlaps(query, subject, suffix = c(".query", ".subject"), minoverlap = 2)
# adding directed prefix includes strand
find_overlaps_directed(query, subject, suffix = c(".query", ".subject"))

}
\seealso{
\code{IRanges::\link[IRanges:findOverlaps-methods]{findOverlaps()}},
\code{GenomicRanges::\link[GenomicRanges:findOverlaps-methods]{findOverlaps()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io-gff.R
\name{write_gff}
\alias{write_gff}
\alias{write_gff1}
\alias{write_gff2}
\alias{write_gff3}
\title{Write a GFF(123) file}
\usage{
write_gff(x, file, index = FALSE)

write_gff1(x, file, index = FALSE)

write_gff2(x, file, index = FALSE)

write_gff3(x, file, index = FALSE)
}
\arguments{
\item{x}{A GRanges object}

\item{file}{Path or connection to write to}

\item{index}{If TRUE the output file will be compressed and indexed using bgzf and
tabix.}
}
\value{
The write function returns a GFFFile object invisibly
}
\description{
This is a lightweight wrapper to the export family
of functions defined in \pkg{rtracklayer}.
}
\examples{
\dontrun{
 test_path <- system.file("tests", package = "rtracklayer")
 test_gff3 <- file.path(test_path, "genes.gff3")
 gr <- read_gff3(test_gff3)
 out_gff3 <- file.path(tempdir(), "test.gff3")
 write_gff3(gr, out_gff3)
 read_gff3(out_gff3)
}
}
\seealso{
\code{rtracklayer::\link[rtracklayer:GFFFile-class]{GFFFile()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-arithmetic-flank.R
\name{flank_left}
\alias{flank_left}
\alias{flank_right}
\alias{flank_upstream}
\alias{flank_downstream}
\title{Generate flanking regions}
\usage{
flank_left(x, width = 0L)

flank_right(x, width = 0L)

flank_upstream(x, width = 0L)

flank_downstream(x, width = 0L)
}
\arguments{
\item{x}{a Ranges object.}

\item{width}{the width of the flanking region relative to the ranges in
\code{x}. Either an integer vector of length 1 or an integer vector
the same length as x. The width can be negative in which case the
flanking region is reversed.}
}
\value{
A Ranges object of same length as \code{x}.
}
\description{
Find flanking regions  to the left or right or
upstream or downstream of a Ranges object.
}
\details{
The function
\code{flank_left} will create the flanking region to the left of starting
coordinates in \code{x}, while \code{flank_right} will create the flanking
region to the right of the starting coordinates in \code{x}. The function
\code{flank_upstream} will \code{flank_left} if the strand of rows in \code{x} is
not negative and will \code{flank_right} if the strand of rows in \code{x} is
negative. The function \code{flank_downstream} will \code{flank_right} if the strand of rows in \code{x} is
not negative and will \code{flank_leftt} if the strand of rows in \code{x} is
negative.

By default \code{flank_left} and \code{flank_right} will
ignore strandedness of any ranges, while \code{flank_upstream} and
\code{flank_downstream} will take into account the strand of \code{x}.
}
\examples{
gr <- as_granges(data.frame(start = 10:15,
                            width = 5,
                            seqnames = "seq1",
                            strand = c("+", "+", "-", "-", "+", "*")))
flank_left(gr, width = 5L)
flank_right(gr, width = 5L)
flank_upstream(gr, width = 5L)
flank_downstream(gr, width = 5L)
}
\seealso{
\code{IRanges::\link[IRanges:intra-range-methods]{flank()}},
\code{GenomicRanges::\link[GenomicRanges:intra-range-methods]{flank()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io-gff.R
\name{read_gff}
\alias{read_gff}
\alias{read_gff1}
\alias{read_gff2}
\alias{read_gff3}
\title{Read a GFF/GTF/GVT file}
\usage{
read_gff(file, col_names = NULL, genome_info = NULL, overlap_ranges = NULL)

read_gff1(file, col_names = NULL, genome_info = NULL, overlap_ranges = NULL)

read_gff2(file, col_names = NULL, genome_info = NULL, overlap_ranges = NULL)

read_gff3(file, col_names = NULL, genome_info = NULL, overlap_ranges = NULL)
}
\arguments{
\item{file}{A path to a file or a connection.}

\item{col_names}{An optional character vector for parsing specific
columns in \code{file} that are part of the GFF specification. These should
name either fixed fields, like source or type, or, for GFF2 and GFF3,
any attribute.}

\item{genome_info}{An optional character string or a Ranges object
that contains information about the genome build. For example the UCSC identifier
"hg19" will add build information to the returned GRanges.}

\item{overlap_ranges}{An optional Ranges object. Only the intervals in the file
that overlap the Ranges will be returned.}
}
\value{
A GRanges object

a GRanges object
}
\description{
This is a lightweight wrapper to the import family
of functions defined in \pkg{rtracklayer}.
}
\examples{
test_path <- system.file("tests", package = "rtracklayer")
# gff3
test_gff3 <- file.path(test_path, "genes.gff3")
gr <- read_gff3(test_gff3)
gr
# alternatively with read_gff
gr <- read_gff(test_gff3, genome_info = "hg19")
gr
}
\seealso{
\code{rtracklayer::\link[rtracklayer:GFFFile-class]{GFFFile()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plyranges.R
\docType{package}
\name{plyranges-package}
\alias{plyranges}
\alias{plyranges-package}
\title{plyranges: a grammar of genomic data manipulation}
\description{
plyranges is a dplyr like API to the Ranges/GenomicRanges infrastructure
in Bioconductor.
}
\details{
plryanges provides a consistent interface for importing and
wrangling genomics data from a variety of sources. The package defines a
grammar of genomic data manipulation through a set of verbs. These verbs
can be used to construct human readable analysis pipelines based on Ranges
objects.
\itemize{
\item Modify genomic regions with the \code{set_width()} and \code{stretch()} functions.
\item Modify genomic regions while fixing the start/end/center coordinates
with the \code{anchors()} family of functions.
\item Sort genomic ranges with \code{arrange()}.
\item Modify, subset, and aggregate genomic data with the \code{mutate()},
\code{filter()}, and \code{summarise()}functions.
\item Any of the above operations can be performed on partitions of the
data with \code{group_by()}.
\item Find nearest neighbour genomic regions with the \code{join_nearest()} family
of functions.
\item Find overlaps between ranges with the \code{join_overlap_inner()} family of functions.
\item Merge all overlapping and adjacent genomic regions with \code{reduce_ranges()}.
\item Merge the end points of all genomic regions with \code{disjoin_ranges()}.
\item Import and write common genomic data formats with the \code{read_/write_} family
of functions.
}

For more details on the features of plryanges, read the vignette:
\code{browseVignettes(package = "plyranges")}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://tidyomics.github.io/plyranges}
  \item Report bugs at \url{https://github.com/tidyomics/plyranges}
}

}
\author{
\strong{Maintainer}: Michael Love \email{michaelisaiahlove@gmail.com} [contributor]

Authors:
\itemize{
  \item Stuart Lee (\href{https://orcid.org/0000-0003-1179-8436}{ORCID})
  \item Michael Lawrence [contributor]
  \item Dianne Cook [contributor]
}

Other contributors:
\itemize{
  \item Spencer Nystrom (\href{https://orcid.org/0000-0003-1000-1579}{ORCID}) [contributor]
  \item Pierre-Paul Axisa [contributor]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-arithmetic-stretch.R
\name{stretch}
\alias{stretch}
\title{Stretch a genomic interval}
\usage{
stretch(x, extend)
}
\arguments{
\item{x}{a Ranges object, to fix by either the start, end or center
of an interval use \code{anchor_start(x)}, \code{anchor_end(x)},
\code{anchor_center(x)}. To fix by strand use \code{anchor_3p(x)} or
\code{anchor_5p(x)}.}

\item{extend}{the amount to alter the width of a Ranges object by. Either an
integer vector of length 1 or an integer vector the same length as x.}
}
\value{
a Ranges object with modified start or end (or both) coordinates
}
\description{
By default, \code{stretch(x)} will anchor by the center of a Ranges
object. This means that half of the value of \code{extend} will be added to
the end of the range and the remaining half subtracted from the start of
the Range. The other anchors will leave the start/end fixed and stretch
the end/start respectively.
}
\examples{
rng <- as_iranges(data.frame(start=c(2:-1, 13:15), width=c(0:3, 2:0)))
rng2 <- stretch(anchor_center(rng), 10)
stretch(anchor_start(rng2), 10)
stretch(anchor_end(rng2), 10)
grng <- as_granges(data.frame(seqnames = "chr1",
                         strand = c("+", "-", "-", "+", "+", "-", "+"),
                         start=c(2:-1, 13:15),
                         width=c(0:3, 2:0)))
stretch(anchor_3p(grng), 10)
stretch(anchor_5p(grng), 10)
}
\seealso{
\code{anchor()}, \code{mutate()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-overlap-joins-intersect.R,
%   R/ranges-overlap-joins-outer.R
\name{join_overlap_intersect}
\alias{join_overlap_intersect}
\alias{join_overlap_intersect_within}
\alias{join_overlap_intersect_directed}
\alias{join_overlap_intersect_within_directed}
\alias{join_overlap_inner}
\alias{join_overlap_inner_within}
\alias{join_overlap_inner_directed}
\alias{join_overlap_inner_within_directed}
\alias{join_overlap_left}
\alias{join_overlap_left_within}
\alias{join_overlap_left_directed}
\alias{join_overlap_left_within_directed}
\title{Join by overlapping Ranges}
\usage{
join_overlap_intersect(x, y, maxgap, minoverlap, suffix = c(".x", ".y"))

join_overlap_intersect_within(x, y, maxgap, minoverlap, suffix = c(".x", ".y"))

join_overlap_intersect_directed(
  x,
  y,
  maxgap,
  minoverlap,
  suffix = c(".x", ".y")
)

join_overlap_intersect_within_directed(
  x,
  y,
  maxgap,
  minoverlap,
  suffix = c(".x", ".y")
)

join_overlap_inner(x, y, maxgap = -1L, minoverlap = 0L, suffix = c(".x", ".y"))

join_overlap_inner_within(
  x,
  y,
  maxgap = -1L,
  minoverlap = 0L,
  suffix = c(".x", ".y")
)

join_overlap_inner_directed(
  x,
  y,
  maxgap = -1L,
  minoverlap = 0L,
  suffix = c(".x", ".y")
)

join_overlap_inner_within_directed(
  x,
  y,
  maxgap = -1L,
  minoverlap = 0L,
  suffix = c(".x", ".y")
)

join_overlap_left(x, y, maxgap, minoverlap, suffix = c(".x", ".y"), distance)

join_overlap_left_within(x, y, maxgap, minoverlap, suffix = c(".x", ".y"))

join_overlap_left_directed(
  x,
  y,
  maxgap,
  minoverlap,
  suffix = c(".x", ".y"),
  distance
)

join_overlap_left_within_directed(
  x,
  y,
  maxgap,
  minoverlap,
  suffix = c(".x", ".y")
)
}
\arguments{
\item{x, y}{Objects representing ranges}

\item{maxgap, minoverlap}{The maximimum gap between intervals as an integer
greater than or equal to zero. The minimum amount of overlap between intervals
as an integer greater than zero, accounting for the maximum gap.}

\item{suffix}{Character to vectors to append to common columns in x and y
(default = \code{c(".x", ".y")}).}

\item{distance}{logical, whether to add a column named "distance"
containing the distance between the ranges, relevant with non-zero
\code{maxgap} values. Available for \code{\link[=join_overlap_left]{join_overlap_left()}}
and the \verb{_directed()} version.}
}
\value{
a GRanges object
}
\description{
Join by overlapping Ranges
}
\details{
The function \code{\link[=join_overlap_intersect]{join_overlap_intersect()}} finds
the genomic intervals that are the overlapping ranges between x and y and
returns a new ranges object with metadata columns from x and y.

The function \code{\link[=join_overlap_inner]{join_overlap_inner()}} is equivalent to \code{\link[=find_overlaps]{find_overlaps()}}.

The function \code{\link[=join_overlap_left]{join_overlap_left()}} performs a left outer join between x
and y. It returns all ranges in x that overlap or do not overlap ranges in y
plus metadata columns common to both. If there is no overlapping range
the metadata column will contain a missing value.

The function \code{\link[=join_overlap_self]{join_overlap_self()}} find all overlaps between a ranges
object x and itself.

All of these functions have two suffixes that modify their behavior.
The \code{within} suffix, returns only ranges in x that are completely
overlapped within in y. The \code{directed} suffix accounts for the strandedness
of the ranges when performing overlaps.
}
\examples{
x <- as_iranges(data.frame(start = c(11, 101), end = c(21, 201)))
y <- as_iranges(data.frame(start = c(10, 20, 50, 100, 1),
                           end = c(19, 21, 105, 202, 5)))

# self
join_overlap_self(y)

# intersect takes common interval
join_overlap_intersect(x,y)

# within
join_overlap_intersect_within(x,y)

# left, and inner join, it's often useful having an id column here
y <- y \%>\% mutate(id = 1:n()) 
x <- x \%>\% mutate(id = 1:n()) 
join_overlap_inner(x,y)
join_overlap_left(y,x, suffix = c(".left", ".right"))

}
\seealso{
\code{\link[=join_overlap_self]{join_overlap_self()}}, \code{\link[=join_overlap_left]{join_overlap_left()}}, \code{\link[=find_overlaps]{find_overlaps()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-arithmetic-shift.R
\name{shift_left}
\alias{shift_left}
\alias{shift_right}
\alias{shift_upstream}
\alias{shift_downstream}
\title{Shift all coordinates in a genomic interval left or right, upstream or downstream}
\usage{
shift_left(x, shift = 0L)

shift_right(x, shift = 0L)

shift_upstream(x, shift = 0L)

shift_downstream(x, shift = 0L)
}
\arguments{
\item{x}{a Ranges object .}

\item{shift}{the amount to move the genomic interval in the Ranges object by.
Either a non-negative integer vector of length 1 or an integer vector
the same length as x.}
}
\value{
a Ranges object with start and end coordinates shifted.
}
\description{
Shift all coordinates in a genomic interval left or right, upstream or downstream
}
\details{
Shifting left or right will ignore any strand information
in the Ranges object, while shifting upstream/downstream will shift coordinates
on the positive strand left/right and the negative strand right/left. By
default, unstranded features are treated as positive. When
using \code{\link[=shift_upstream]{shift_upstream()}} or \code{\link[=shift_downstream]{shift_downstream()}} when the \code{shift} argument is
indexed by the strandedness of the input ranges.
}
\examples{
ir <- as_iranges(data.frame(start = 10:15, width = 5))
shift_left(ir, 5L)
shift_right(ir, 5L)
gr <- as_granges(data.frame(start = 10:15,
                            width = 5,
                            seqnames = "seq1",
                            strand = c("+", "+", "-", "-", "+", "*")))
shift_upstream(gr, 5L)
shift_downstream(gr, 5L)
}
\seealso{
\code{IRanges::\link[IRanges:intra-range-methods]{shift()}},
\code{GenomicRanges::\link[GenomicRanges:intra-range-methods]{shift()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-setops.R
\name{\%union\%}
\alias{\%union\%}
\alias{\%intersect\%}
\alias{\%setdiff\%}
\alias{between}
\alias{span}
\title{Row-wise set operations on Ranges objects}
\usage{
x \%union\% y

x \%intersect\% y

x \%setdiff\% y

between(x, y)

span(x, y)
}
\arguments{
\item{x, y}{Ranges objects}
}
\value{
A Ranges object
}
\description{
Row-wise set operations on Ranges objects
}
\details{
Each of these functions acts on the rows between pairs of
Ranges object.
The function \code{\%union\%()}.
will return the entire range between two ranges objects assuming there
are no gaps, if you would like to force gaps use \code{\link[=span]{span()}} instead.
The function \code{\%intersect\%()} will create a new ranges object
with a hit column indicating whether or not the two ranges intersect.
The function \code{\%setdiff\%()}will return the ranges for each
row in x that are not in the corresponding row of y.
The function \code{\link[=between]{between()}} will return the gaps between
two ranges.
}
\examples{
x <- as_iranges(data.frame(start = 1:10, width = 5))
# stretch x by 3 on the right
y <- stretch(anchor_start(x), 3)
# take the rowwise union
x \%union\% y
# take the rowwise intersection
x \%intersect\% y
# asymetric difference
y \%setdiff\% x
x \%setdiff\% y
# if there are gaps between the rows of each range use span
y <- as_iranges(data.frame(start = c(20:15, 2:5),
width = c(10:15,1:4)))
# fill in the gaps and take the rowwise union
span(x,y)
# find the gaps
between(x,y)
}
\seealso{
[IRanges::punion()][IRanges::pintersect()][IRanges::pgap()][IRanges::psetdiff()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-join-nearest.R
\name{join_nearest}
\alias{join_nearest}
\alias{join_nearest_left}
\alias{join_nearest_right}
\alias{join_nearest_upstream}
\alias{join_nearest_downstream}
\title{Find nearest neighbours between two Ranges objects}
\usage{
join_nearest(x, y, suffix = c(".x", ".y"), distance = FALSE)

join_nearest_left(x, y, suffix = c(".x", ".y"), distance = FALSE)

join_nearest_right(x, y, suffix = c(".x", ".y"), distance = FALSE)

join_nearest_upstream(x, y, suffix = c(".x", ".y"), distance = FALSE)

join_nearest_downstream(x, y, suffix = c(".x", ".y"), distance = FALSE)
}
\arguments{
\item{x, y}{Ranges objects, add the nearest neighbours of ranges in x to
those in y.}

\item{suffix}{A character vector of length two used to identify metadata columns}

\item{distance}{logical vector whether to add a column named "distance"
containing the distance to the nearest region. If set to a character vector
of length 1, will use that as distance column name.}
}
\value{
A Ranges object corresponding to the nearest ranges, all metadata
is copied over from the right-hand side ranges \code{y}.
}
\description{
Find nearest neighbours between two Ranges objects
}
\details{
By default \code{join_nearest} will find arbitrary nearest
neighbours in either direction and ignore any strand information.
The \code{join_nearest_left} and \code{join_nearest_right}  methods
will find arbitrary nearest neighbour ranges on x that are left/right of
those on y and ignore any strand information.

The \code{join_nearest_upstream} method will find arbitrary nearest
neighbour ranges on x that are upstream of those on y. This takes into
account strandedness of the ranges.
On the positive strand nearest upstream will be on the
left and on the negative strand nearest upstream will be on the right.

The \code{join_nearest_downstream} method will find arbitrary nearest
neighbour ranges on x that are upstream of those on y. This takes into
account strandedness of the ranges.On the positive strand nearest downstream
will be on the right and on the negative strand nearest upstream will be on
the left.
}
\examples{
query <- data.frame(start = c(5,10, 15,20),
                   width = 5,
                   gc = runif(4)) \%>\%
             as_iranges()
subject <- data.frame(start = c(2:6, 24),
                      width = 3:8,
                      label = letters[1:6]) \%>\%
             as_iranges()

join_nearest(query, subject)
join_nearest_left(query, subject)
join_nearest_right(query, subject)

subject  <- data.frame(seqnames = "chr1",
               start = c(11,101),
               end = c(21, 200),
               name = c("a1", "a2"),
               strand = c("+", "-"),
               score = c(1,2)) \%>\%
           as_granges()
query <- data.frame(seqnames = "chr1",
                      strand = c("+", "-", "+", "-"),
                      start = c(21,91,101,201),
                      end = c(30,101,110,210),
                      name = paste0("b", 1:4),
                      score = 1:4) \%>\%
                   as_granges()
join_nearest_upstream(query, subject)
join_nearest_downstream(query, subject)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dplyr-filter.R
\name{filter-ranges}
\alias{filter-ranges}
\alias{filter.Ranges}
\title{Subset a \code{Ranges} object}
\usage{
\method{filter}{Ranges}(.data, ..., .preserve = FALSE)
}
\arguments{
\item{.data}{A \code{Ranges} object}

\item{...}{valid logical predictates to subset .data by. These
are determined by variables in \code{.data}. If more than
one condition is supplied, the conditions are combined with \code{&}. Only
rows where the condition evaluates to \code{TRUE} are kept.}

\item{.preserve}{when FALSE (the default) grouping structure is recalculated, TRUE is currently not implemented.}
}
\value{
a Ranges object
}
\description{
Subset a \code{Ranges} object
}
\details{
For any Ranges objects
\code{filter} can act on all core components of the class including start, end,
width (for IRanges) or seqnames and strand (for GRanges) in addition to
metadata columns. If the Ranges object is grouped, \code{filter} will act
seperately on each  parition of the data.
}
\examples{
set.seed(100)
df <- data.frame(start = 1:10,
                 width = 5,
                 seqnames = "seq1",
                 strand = sample(c("+", "-", "*"), 10, replace = TRUE),
                 gc = runif(10))

rng <- as_granges(df)

filter(rng, strand == "+")
filter(rng, gc > 0.5)

# multiple criteria
filter(rng, strand == "+" | start > 5)
filter(rng, strand == "+" & start > 5)

# multiple conditions are the same as and
filter(rng, strand == "+", start > 5)

# grouping acts on each subset of the data
rng \%>\%
  group_by(strand) \%>\%
  filter(gc > 0.5)

}
\seealso{
\code{\link[dplyr:filter]{dplyr::filter()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-chop.R
\name{chop_by_introns}
\alias{chop_by_introns}
\alias{chop_by_gaps}
\title{Group a GRanges object by introns or gaps}
\usage{
chop_by_introns(x)

chop_by_gaps(x)
}
\arguments{
\item{x}{a GenomicRanges object with a cigar string column}
}
\value{
a GRanges object
}
\description{
Group a GRanges object by introns or gaps
}
\details{
Creates a grouped Ranges object from a cigar string
column, for \code{chop_by_introns()} will check for the presence of
"N" in the cigar string and create a new column called
\code{intron} where TRUE indicates the alignment has a skipped
region from the reference. For \code{chop_by_gaps()} will check
for the presence of "N" or "D" in the cigar string and
create a new column called "gaps" where TRUE indicates
the alignment has a deletion from the reference or has an intron.
}
\examples{
if (require(pasillaBamSubset)) {
   bamfile <- untreated1_chr4()
   # define a region of interest
   roi <- data.frame(seqnames = "chr4", start = 5e5, end = 7e5) \%>\%
            as_granges()
   # results in a grouped ranges object
   rng <- read_bam(bamfile) \%>\% 
            filter_by_overlaps(roi) \%>\%
            chop_by_gaps()
   # to find ranges that have gaps use filter with `n()`
   rng \%>\% filter(n() >= 2)
  
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-construct.R
\name{as_iranges}
\alias{as_iranges}
\alias{as_granges}
\title{Construct a I/GRanges object from a tibble or data.frame}
\usage{
as_iranges(.data, ..., keep_mcols = TRUE)

as_granges(.data, ..., keep_mcols = TRUE)
}
\arguments{
\item{.data}{a \code{\link[=data.frame]{data.frame()}} or \code{\link[dplyr:reexports]{dplyr::tibble()}} to
construct a Ranges object from}

\item{...}{optional named arguments specifying which the columns in .data
containin the core components a Ranges object.}

\item{keep_mcols}{place the remaining columns into the metadata columns slot
(default=TRUE)}
}
\value{
a Ranges object.
}
\description{
The as_i(g)ranges function looks for column names in .data called start,
end, width, seqnames and strand in order to construct an IRanges or GRanges
object. By default other columns in .data are placed into the mcols (
metadata columns) slot of the returned object.
}
\examples{
df <- data.frame(start=c(2:-1, 13:15), width=c(0:3, 2:0))
as_iranges(df)

df <- data.frame(start=c(2:-1, 13:15), width=c(0:3, 2:0), strand = "+")
# will return an IRanges object
as_iranges(df)

df <- data.frame(start=c(2:-1, 13:15), width=c(0:3, 2:0),
strand = "+", seqnames = "chr1")
as_granges(df)

# as_g/iranges understand alternate name specification
df <- data.frame(start=c(2:-1, 13:15), width=c(0:3, 2:0),
strand = "+", chr = "chr1")
as_granges(df, seqnames = chr)

# can also handle DFrame input
df <- methods::as(df, "DFrame")
df$y <- IRanges::IntegerList(c(1,2,3), NA, 5, 6, 8, 9, 10:12)
as_iranges(df)
as_granges(df, seqnames = chr)

}
\seealso{
\code{IRanges::\link[IRanges:IRanges-constructor]{IRanges()}},
\code{GenomicRanges::\link[GenomicRanges:GRanges-class]{GRanges()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-colwise.R
\name{remove_names}
\alias{remove_names}
\alias{names_to_column}
\alias{id_to_column}
\title{Tools for working with named Ranges}
\usage{
remove_names(.data)

names_to_column(.data, var = "name")

id_to_column(.data, var = "id")
}
\arguments{
\item{.data}{a Ranges object}

\item{var}{Name of column to use for names}
}
\value{
Returns a Ranges object with empty names
}
\description{
Tools for working with named Ranges
}
\details{
The function \code{names_to_column()} and \code{id_to_column()} always places
\code{var} as the first column in \code{mcols(.data)}, shifting all other columns
to the left. The \code{id_to_column()} creates a column with sequential row
identifiers starting at 1, it will also remove any existing names.
}
\examples{
ir <- IRanges::IRanges(start = 1:3, width = 4, names = c("a", "b", "c"))
remove_names(ir)
ir_noname <- names_to_column(ir)
ir_noname
ir_with_id <- id_to_column(ir)
ir_with_id
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-pairs.R
\name{pair_overlaps}
\alias{pair_overlaps}
\alias{pair_nearest}
\alias{pair_precede}
\alias{pair_follow}
\title{Pair together two ranges objects}
\usage{
pair_overlaps(x, y, maxgap, minoverlap, suffix)

pair_nearest(x, y, suffix)

pair_precede(x, y, suffix)

pair_follow(x, y, suffix)
}
\arguments{
\item{x, y}{Ranges objects to pair together.}

\item{maxgap, minoverlap}{The maximimum gap between intervals as an integer
greater than or equal to negative one. The minimum amount of overlap between intervals
as an integer greater than zero, accounting for the maximum gap.}

\item{suffix}{A character vector of length two used to identify metadata columns
coming from x and y.}
}
\value{
a DataFrame with two ranges columns and the corresponding metadata
columns.
}
\description{
Pair together two ranges objects
}
\details{
These functions return a DataFrame object, and is one way of
representing paired alignments with plyranges.
}
\examples{
query <- data.frame(start = c(5,10, 15,20), width = 5, gc = runif(4)) \%>\%
             as_iranges()
subject <- data.frame(start = 2:6, width = 3:7, label = letters[1:5]) \%>\%
             as_iranges()

pair_overlaps(query, subject)
pair_overlaps(query, subject, minoverlap = 5)
pair_nearest(query, subject)


query  <- data.frame(seqnames = "chr1",
               start = c(11,101),
               end = c(21, 200),
               name = c("a1", "a2"),
               strand = c("+", "-"),
               score = c(1,2)) \%>\%
           as_granges()
subject <- data.frame(seqnames = "chr1",
                      strand = c("+", "-", "+", "-"),
                      start = c(21,91,101,201),
                      end = c(30,101,110,210),
                      name = paste0("b", 1:4),
                      score = 1:4) \%>\%
                   as_granges()

# ignores strandedness
pair_overlaps(query, subject, suffix = c(".query", ".subject"))
pair_follow(query, subject, suffix = c(".query", ".subject"))
pair_precede(query, subject, suffix = c(".query", ".subject"))
pair_precede(query, subject, suffix = c(".query", ".subject"))
}
\seealso{
[join_nearest()][join_overlap_inner()][join_precede()][join_follow()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-overlap-filter.R
\name{filter_by_overlaps}
\alias{filter_by_overlaps}
\alias{filter_by_non_overlaps}
\alias{filter_by_overlaps_directed}
\alias{filter_by_non_overlaps_directed}
\title{Filter by overlapping/non-overlapping ranges}
\usage{
filter_by_overlaps(x, y, maxgap = -1L, minoverlap = 0L)

filter_by_non_overlaps(x, y, maxgap, minoverlap)

filter_by_overlaps_directed(x, y, maxgap = -1L, minoverlap = 0L)

filter_by_non_overlaps_directed(x, y, maxgap, minoverlap)
}
\arguments{
\item{x, y}{Objects representing ranges}

\item{maxgap}{The maximimum gap between intervals as a single
integer greater than or equal to -1. If you modify this argument,
\code{minoverlap} must be held fixed.}

\item{minoverlap}{The minimum amount of overlap between intervals
as a single integer greater than 0. If you modify this argument,
\code{maxgap} must be held fixed.}
}
\value{
a Ranges object
}
\description{
Filter by overlapping/non-overlapping ranges
}
\details{
By default, \code{filter_by_overlaps} and
\code{filter_by_non_overlaps} ignore strandedness for \code{\link[GenomicRanges:GRanges-class]{GenomicRanges::GRanges()}}
objects. To perform stranded operations use \code{filter_by_overlaps_directed} and \code{filter_by_non_overlaps_directed}. The argument \code{maxgap} is the maximum number of positions
between two ranges for them to be considered overlapping. Here the default
is set to be -1 as that is the the gap between two ranges that
has its start or end strictly inside the other. The argugment
\code{minoverlap} refers to the minimum number of positions
overlapping between ranges, to consider there to be overlap.
}
\examples{
df <- data.frame(seqnames = c("chr1", rep("chr2", 2),
                              rep("chr3", 3), rep("chr4", 4)),
                 start = 1:10,
                 width = 10:1,
                 strand = c("-", "+", "+", "*", "*", "+", "+", "+", "-", "-"),
                 name = letters[1:10])
query <- as_granges(df)

df2 <- data.frame(seqnames = c(rep("chr2", 2), rep("chr1", 3), "chr2"),
                  start = c(4,3,7,13,1,4),
                  width = c(6,6,3,3,3,9),
                  strand = c(rep("+", 3), rep("-", 3)))
subject <- as_granges(df2)

filter_by_overlaps(query, subject)

filter_by_overlaps_directed(query, subject)

filter_by_non_overlaps(query, subject)

filter_by_non_overlaps_directed(query, subject)

}
\seealso{
\code{IRanges::\link[IRanges:findOverlaps-methods]{subsetByOverlaps()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-reduce.R
\name{reduce_ranges}
\alias{reduce_ranges}
\alias{reduce_ranges_directed}
\title{Reduce then aggregate a Ranges object}
\usage{
reduce_ranges(.data, min.gapwidth = 1L, ...)

reduce_ranges_directed(.data, min.gapwidth = 1L, ...)
}
\arguments{
\item{.data}{a Ranges object to reduce}

\item{min.gapwidth}{Ranges separated by a gap of at least min.gapwidth positions are not merged.}

\item{...}{Name-value pairs of summary functions.}
}
\value{
a Ranges object with the
}
\description{
Reduce then aggregate a Ranges object
}
\examples{
set.seed(10)
df <- data.frame(start = sample(1:10), 
                 width = 5,  
                 seqnames = "seq1",
                 strand = sample(c("+", "-", "*"), 10, replace = TRUE), 
                 gc = runif(10))
                 
rng <- as_granges(df)
rng \%>\% reduce_ranges()
rng \%>\% reduce_ranges(gc = mean(gc))
rng \%>\% reduce_ranges_directed(gc = mean(gc))
rng \%>\% reduce_ranges_directed(gc = mean(gc), min.gapwidth = 10)

x <- data.frame(start = c(11:13, 2, 7:6), 
               width=3, 
               id=sample(letters[1:3], 6, replace = TRUE),
               score= sample(1:6))
x <- as_iranges(x)
x \%>\% reduce_ranges()
x \%>\% reduce_ranges(score = sum(score))
x \%>\% group_by(id) \%>\% reduce_ranges(score = sum(score))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io-wig.R
\name{write_wig}
\alias{write_wig}
\title{Write a WIG file}
\usage{
write_wig(x, file)
}
\arguments{
\item{x}{A GRanges object}

\item{file}{File name, URL or connection specifying a file to write x to.
Compressed files with extensions such as '.gz' are handled
automatically.}
}
\value{
The write function returns a WIGFile invisibly.
}
\description{
Write a WIG file
}
\seealso{
\code{rtracklayer::\link[rtracklayer:WIGFile-class]{WIGFile()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io-wig.R
\name{read_wig}
\alias{read_wig}
\title{Read a WIG file}
\usage{
read_wig(file, genome_info = NULL, overlap_ranges = NULL)
}
\arguments{
\item{file}{A path to a file or a connection.}

\item{genome_info}{An optional character string or a Ranges object
that contains information about the genome build. For example the USSC identifier
"hg19" will add build information to the returned GRanges.}

\item{overlap_ranges}{An optional Ranges object. Only the intervals in the file
that overlap the Ranges will be returned.}
}
\value{
A GRanges object

A GRanges object
}
\description{
This is a lightweight wrapper to the import family
of functions defined in \pkg{rtracklayer}.
}
\examples{
test_path <- system.file("tests", package = "rtracklayer")
test_wig <- file.path(test_path, "step.wig")
gr <- read_wig(test_wig)
gr
gr <- read_wig(test_wig, genome_info = "hg19")
}
\seealso{
\code{rtracklayer::\link[rtracklayer:WIGFile-class]{WIGFile()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dplyr-pull.R
\name{pull-ranges}
\alias{pull-ranges}
\alias{pull.Ranges}
\title{Extract a single column from a Ranges object as a vector}
\usage{
\method{pull}{Ranges}(.data, var = -1, name = NULL, ...)
}
\arguments{
\item{.data}{a \code{Ranges} object}

\item{var}{A variable specified as:
\itemize{
\item a literal variable name
\item a positive integer, giving the position counting from the left. In this
case order is start, end, width, (strand, seqnames), gc and score.
\item a negative integer, giving the position counting from the right.
The default returns the last column (on the assumption that's the column
you've created most recently). This argument is taken by expression and
supports quasiquotation (you can unquote column names and column locations).
}}

\item{name}{An optional parameter that specifies the column to be used as
names for a named vector. Specified in a similar manner as \code{var}.}

\item{...}{For use by methods.}
}
\description{
Extract a single column from a Ranges object as a vector
}
\examples{
df <- data.frame(start = 1:10,
                 width = 5,
                 seqnames = "seq1",
                 strand = sample(c("+", "-", "*"), 10, replace = TRUE),
                 gc = runif(10),
                 score = rpois(10, 2))
rng <- as_granges(df)

# Pull parts of the range
pull(rng, start)
# equivalent to start(rng)

# Pull by column name
pull(rng, gc)
pull(rng, score)

# Pull by position (positive from left, negative from right)
pull(rng, 1)    # First metadata column
pull(rng, -1)   # Last metadata column (default)
pull(rng, -2)   # Second to last metadata column

# Pull with names
pull(rng, score, name = gc)

}
\seealso{
\code{\link[dplyr:pull]{dplyr::pull()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ranges-disjoin.R
\name{disjoin_ranges}
\alias{disjoin_ranges}
\alias{disjoin_ranges_directed}
\title{Disjoin then aggregate a Ranges object}
\usage{
disjoin_ranges(.data, ...)

disjoin_ranges_directed(.data, ...)
}
\arguments{
\item{.data}{a Ranges object to disjoin}

\item{...}{Name-value pairs of summary functions.}
}
\value{
a Ranges object that is now disjoint (no bases overlap).
}
\description{
Disjoin then aggregate a Ranges object
}
\examples{
df <- data.frame(start = 1:10, width = 5,  seqnames = "seq1",
strand = sample(c("+", "-", "*"), 10, replace = TRUE), gc = runif(10))
rng <- as_granges(df)
rng \%>\% disjoin_ranges()
rng \%>\% disjoin_ranges(gc = mean(gc))
rng \%>\% disjoin_ranges_directed(gc = mean(gc))
}
